#!/usr/bin/env python
"""
_AlcaSkimInjectorComponent_

Poll T0AST for RECO files and generate Alca workflows and jobspecs for them.
Inject the jobs into a job queue while they wait for skim configs.
"""

__revision__ = "$Id: AlcaSkimInjectorComponent.py,v 1.65 2010/07/26 13:04:56 hufnagel Exp $"
__version__ = "$Revision: 1.65 $"

import logging
import os
import sys
import shutil

from JobQueue import JobQueueDB

from T0 import Globals
from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListJobs
from T0.State.Database.Writer import InsertJob

from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig

from WMCore.WMBS.Job import Job

from WMCore.JobSplitting.SplitterFactory import SplitterFactory

from T0.GenericTier0.GenericTier0Component import GenericTier0Component
from T0.JobFactory.AlcaSkimJob import AlcaSkimJob
from T0.WorkflowFactory.AlcaSkimWorkflow import AlcaSkimWorkflow

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec


class AlcaSkimInjectorComponent(GenericTier0Component):
    """
    _AlcaSkimInjectorComponent_

    Poll T0AST for reco files and generate workflows and jobspecs for them.
    Inject the jobs into a job queue while they wait for alca configs.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)

        # Data structure to hold WorkflowSpecs.  This is a two level
        # dictionary where the key to the first level is the run number,
        # and the second level is dataset id.
        self.workflows = {}

        # cache datasetpath information to avoid queries
        self.datasetpaths = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2
        
        # Lookup some parameters from the config file:
        #   scramArch - SCRAM Architecture
        #   cmsPath - Path to CMS software
        #   eventsPerJobs - Max events to process per AlcaSkim job
        #   recoTimeout - The amount of time to hold AlkcaSkim jobs for
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/afs/cern.ch/cms/sw")
        self.pollInterval = self.lookupParameter("PollInterval", "00:05:00")

        # place for all workflows, jobspecs, configs, etc
        self.jobCacheDir = self.lookupParameter("JobCacheDir",
                                                self.args["ComponentDir"])

        # WMCore splitter factory for creating AlcaSkim jobs
        self.splitterFactory = SplitterFactory(package = "T0.JobSplitting")

        self.alcaSkimSubscription = None

        # See if lazy downloads are enabled
        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        logging.info("AlcaSkim Injector Component Started")


    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if message == "AlcaSkimInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "AlcaSkimInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "AlcaSkimInjector:PollForJobs":
            self.pollForJobs()
        elif message == "AlcaSkimInjector:CleanUp":
            self.cleanupRun(payload)
        else:
            logging.error("Unknown message: %s, payload: %s" % (message, payload))
        
        return


    def retrieveWorkflow(self, run, datasetpath):
        """
        _retrieveWorkflow_

        Retrieve a WorkflowSpec from the cache.  If a workflow is not in the
        cache it will be created.  Returns None if an error occured while
        trying to create the workflow.
        """
        if run not in self.workflows.keys():
            self.workflows[run] = {}
        if datasetpath not in self.workflows[run].keys():
            self.workflows[run][datasetpath] = None

        existingWorkflow = self.workflows[run][datasetpath]

        if existingWorkflow != None:
            return existingWorkflow

        newWorkflow = self.createWorkflow(run, datasetpath)
        self.workflows[run][datasetpath] = newWorkflow

        return newWorkflow


    def createWorkflow(self, run, datasetpath):
        """
        _createWorkflow_

        Create a workflow for a particular run number and datasetpath id
        """
        logging.debug("createWorkflow(): Running run %s, datasetpath %s" % \
                      (run, datasetpath))

        # no check needed, done before
        runConfig = self.getRunConfig(run)

        primds_id = self.datasetpaths[datasetpath]['primds_id']
        primds = self.datasetpaths[datasetpath]['primds']

        scenario = runConfig.getScenario(primds_id)

        recoConfig = runConfig.getRecoConfigurationByPrimaryID(primds_id)
        if recoConfig == None:
            logging.error("No reco config for run %s, datasetpath %s" % (run,
                                                                         datasetpath))
            return None
    
        alcaConfig = runConfig.getAlcaConfigurationByPrimaryID(primds_id)
        if alcaConfig == None:
            logging.error("No alca config for run %s, datasetpath %s" % (run,
                                                                         datasetpath))
            return None

        if alcaConfig.ConfigURL == None:
            configFile = None
        else:
            configFile = runConfig.retrieveConfigFromURL(alcaConfig.ConfigURL)

        outModuleInfoList = [ runConfig.getOutputModuleInfo("AlcaSkim", "ALCARECO",
                                                            datasetID = primds_id) ]

        workflowName = "AlcaSkim-Run%s-%s" % (run, primds)

        workflowDir = os.path.join(self.jobCacheDir, "workflows",
                                   "Run%s" % run)
        workflowSpecFile = os.path.join(workflowDir,
                                        "%s-workflow.xml" % workflowName)

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No Workflow for run, dataset: %s, %s" % (run, primds))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowSpecFile)

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return workflowSpec

        workflow = AlcaSkimWorkflow(run,
                                    alcaConfig.CMSSWVersion,
                                    self.cmsPath,
                                    self.scramArch,
                                    *outModuleInfoList)

        workflowSpec = workflow.makeWorkflowSpec(workflowName,
                                                 scenario = scenario,
                                                 configFile = configFile,
                                                 alcaProducers = recoConfig.AlcaProducers,
                                                 globalTag = recoConfig.GlobalTag,
                                                 enableLazyDownload = self.useLazyDownload,
                                                 production = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)
        
        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)

        workflowSpec.save(workflowSpecFile)

        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)


    def createJobSpecs(self, run, datasetpath, jobs):
        """
        _createJobSpecs_

        Loop over all jobs and process them (they are all for same run/datasetpath)

        Generate job specs for every job

        """
        logging.debug("createJobSpecs(): Running...")

        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %s" % run)
            return {}

        if not self.datasetpaths.has_key(datasetpath):
            self.datasetpaths[datasetpath] = ListDatasets.listDatasetPathInfo(self.t0astDBConn,
                                                                              datasetpath)

        primds = self.datasetpaths[datasetpath]['primds']

        workflowSpec = self.retrieveWorkflow(run, datasetpath)
        if workflowSpec == None:
            logging.error("No workflow for run %s, datasetpath %s" % (run,
                                                                      datasetpath))
            return {}

        jobSpecBaseDir = os.path.join(self.jobCacheDir, "jobspecs", "Run%d" % run)

        jobFactory = AlcaSkimJob(workflowSpec)

        jobSpecDict = {}
        for job in jobs:

            maxEvents = -1
            skipEvents = 0

            jobSpecDict[job['id']] = jobFactory(job['id'],
                                                primds,
                                                job['files'],
                                                maxEvents, skipEvents,
                                                jobSpecBaseDir)

        return jobSpecDict


    def pollForJobs(self):
        """
        _pollForJobs_

        Poll T0AST for new AlcaSkim jobs

        Call job creation code, then queue them in JobQueue
        and update name and status in WMBS

        """
        logging.debug("pollForJobs(): Running...")

        self.t0astDBConn.beginTransaction()

        wmbsJobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                              subscription = self.alcaSkimSubscription)

        wmbsJobFactory()

        self.t0astDBConn.commit()

        jobs = ListJobs.listNewWMBSJobs(self.t0astDBConn, self.alcaSkimSubscription['id'])

        # sort jobs by run and datasetpath
        jobDict = {}
        for job in jobs:
            run = job['run']
            datasetpath = job['datasetpath']
            if jobDict.has_key(run):
                if jobDict[run].has_key(datasetpath):
                    jobDict[run][datasetpath].append(job)
                else:
                    jobDict[run][datasetpath] = [job]
            else:
                jobDict[run] = { datasetpath : [job] }

        # process jobs by run and datasetpath
        jobSpecDict = {}
        for run,tempDict in jobDict.items():
            for datasetpath, jobs in tempDict.items():
                jobSpecDict.update( self.createJobSpecs(run,datasetpath,jobs) )

        if len(jobSpecDict) > 0:

            # aggregate information for efficient database operations
            wmbsJobs = []
            jobNames = {}
            for jobId, jobSpec in jobSpecDict.items():
                wmbsJobs.append(Job(id = jobId))
                jobNames[jobId] = jobSpec.parameters["JobName"]

            # now do actual PA and T0AST DB modifications
            try:
                self.paDBConn.beginTransaction()

                for jobSpec in jobSpecDict.values():

                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters["JobSpecFile"],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority,
                                             siteName = None,
                                             status = "new")

                self.t0astDBConn.beginTransaction()

                InsertJob.updateWMBSJobNames(self.t0astDBConn, jobNames)
                self.stateChanger.propagate(wmbsJobs, 'created', 'new')

                self.paDBConn.commit()
                self.t0astDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(jobSpecDict))
            except StandardError, ex:
                self.t0astDBConn.rollback()
                self.paDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        self.ms.publishUnique("AlcaSkimInjector:PollForJobs", "", self.pollInterval)
        self.ms.commit()
        return

    def cleanupRun(self, payload):
        """
        _cleanupRun_
    
        Once a run has completed push out any files that have not yet
        been merged and perform the following cleanup:
          - Clear out the workflow template for the run
          - Clear all workflows and jobspec files off disk
        """
        try:
            run = int(payload)
        except ValueError:
            logging.error("Run number is not valid: %s" % payload)
            return
    
        logging.debug("Cleaning up run %s" % run)
    
        if run in self.workflows.keys():
            del self.workflows[run]
    
        workflowDir = self.jobCacheDir + "/workflows/Run%s" % run
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % run
    
        shutil.rmtree(workflowDir, ignore_errors = True)
        shutil.rmtree(jobSpecDir, ignore_errors = True)
        return

    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("AlcaSkimInjector:StartDebug")
        self.ms.subscribeTo("AlcaSkimInjector:EndDebug")
        self.ms.subscribeTo("AlcaSkimInjector:PollForJobs")

        self.ms.subscribeTo("CleanUp")

        self.connectT0AST()
        self.connectPADB()

        # Load up the alcaskim subscription for easy access
        self.alcaSkimSubscription = Globals.retrieveAlcaSubscription()

        self.ms.publishUnique("AlcaSkimInjector:PollForJobs", "", self.pollInterval)
        self.ms.commit()

        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)

        # wait for messages
        self.messageLoop()
