#!/usr/bin/env python
"""
_BlockAuditorComponent_

"""

__revision__ = "$Id: BlockAuditorComponent.py,v 1.55 2009/09/14 04:58:57 sryu Exp $"
__version__ = "$Revision: 1.55 $"

import logging

from DBSAPI.dbsApi import DbsApi
from DBSAPI.dbsApiException import DbsBadResponse
from ProdAgentCore.Configuration import loadProdAgentConfiguration

from T0.GenericTier0.GenericTier0Component import GenericTier0Component
from T0.State.Database.Reader.ListPhEDExSubscription import listStorageNodesByBlock
from T0.State.Database.Reader.ListBlock import listBlocksByStatus
from T0.State.Database.Reader.ListBlock import isParentBlockMigrated
from T0.ExportSystemAPI.PhEDExDataServiceAPI import isBlockTransfered
from T0.ExportSystemAPI.PhEDExDataServiceAPI import getNodeList
from T0.ExportSystemAPI.ExportSystemExceptions import PhEDExServiceError
from T0.ExportSystemAPI.ExportSystemExceptions import DBSServiceError

from T0.BlockAuditor.BlockAuditorAPI import updateBlockFileStatusToExported
from T0.BlockAuditor.BlockAuditorAPI import updateBlockMigrateStatus
   
class BlockAuditorComponent(GenericTier0Component):
    """
    _BlockAuditorComponent_

    Listen for the call back from PhEDEx about file (block) transfer complete,
    Post on the global DBS, clean up files and tables
    
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component to setup logging.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("BlockAuditor Component Started")
        self.pollInterval = self.lookupParameter("PollInterval", "00:03:00")
        self.migrateAfterTransfer = self.lookupParameter(
                                            "MigrateAfterTransfer", "Off")

        self.localDBSUrl = None
        self.globalDBSUrl = None
        self.dbsApi = None
        self.alert = None

        
    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "BlockAuditor:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "BlockAuditor:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "BlockAuditor:Poll":
            self.poll()
        else:
            logging.error("Unrecognized event: %s" % event)

        return
    
    def loadDBSConf(self):
        
        paConfig = loadProdAgentConfiguration()
        localDBSConfig = paConfig.getConfig("LocalDBS")
        globalDBSConfig = paConfig.getConfig("GlobalDBSDLS")
        phedexConfig = paConfig.getConfig("PhEDExConfig")
        
        self.localDBSUrl = localDBSConfig.get("DBSURL", None)
        self.CAF_DBS_URL = localDBSConfig.get("CAFDBSURL", None)
        self.globalDBSUrl = globalDBSConfig.get("DBSURL", None)
        
        self.CAF_NODE = localDBSConfig.get("CAFNode", None)
        self.phedexDSUrl = phedexConfig.get("DataServiceURL", None)
        # DbsApi gives an error if url is not set.
        # url here is irrelavent 
        self.dbsApi = DbsApi({'url':localDBSConfig.get("DBSURL", None)})
        
    
    def poll(self):
        """
        1. get the call back from PhEDEx about file transfer complete
        2. determine whether block is completed or not
        3. for the completed block, update to global DBS
        4. clean up files 
        5. clean up database 
        
        """
        if self.dbsApi == None:
            self.loadDBSConf()
            
        #self.handleTestBlock()
        injectedBlockList = listBlocksByStatus(self.t0astDBConn, "InFlight")
        injectedBlocks = self.sortBlockListByStreamType(injectedBlockList)
        
        self.updateExpressTransferStatus(injectedBlocks["Express"])        
        self.updateNormalTransferStatus(injectedBlocks["Normal"])
        
        blockStatusList = []
        if self.migrateAfterTransfer == "On":
            logging.info("MigrateAfterTransfer flag is On")
            blockStatusList = ["Exported"]
        else:
            blockStatusList = ["InFlight", "Exported"]    
        
        blockList = listBlocksByStatus(self.t0astDBConn, blockStatusList,
                                       "NotMigrated")
        blockDict = self.sortBlockListByStreamType(blockList)
        
        self.migrateExpressBlocks(blockDict["Express"])
        self.migrateNormalBlocks(blockDict["Normal"])                    
                
        self.ms.publishUnique('BlockAuditor:Poll', '',
                              self.pollInterval)
        self.ms.commit()
    
    def updateExpressTransferStatus(self, blockList):
        """
        _updateExpressTransferStatus_
        
        update the list of blocks' transfer status
        Express stream should have been stage out already. 
        Update all the express blocks "Exported" state as well as files 
        belong to the block 
        """
        # get all the closed block for export checking    
        #blockList = listBlocksByStatus(self.t0astDBConn, "Closed")
        logging.info("""Express block transfer status update: 
                        Everything should be staged out already""")
        # set the expressFlag so it won't get deleted later
        for block in blockList:
            updateBlockFileStatusToExported(self.t0astDBConn, 
                                            block, self.CAF_NODE,
                                            expressFlag=True)
            
                
    def updateNormalTransferStatus(self, blockList):
        """
        _updateNormalTransferStatus_
        
        check whether given block is transfered to the subscribed location
        then mark all the parent file as deletable and update the block status
        to 'Exported'
        """
        # get all the closed block for export checking    
        #blockList = listBlocksByStatus(self.t0astDBConn, "Closed")
        logging.info("Block transfer check")
        for block in blockList:
            nodeList = self.blockTransferCheck(block)
            if nodeList != False:
                logging.info(" block %s has transfered to %s " % 
                             (block["BLOCK_NAME"], nodeList))
                
    def migrateExpressBlocks(self, blockList):
        """
        _migrateExpressBlocks_
        
        If the expressed block is closed, it means, 
        the block is already transfered to CAF (stage out directly to CAF lfn.
        So it just needed to be migrated to global dbs.
        
        Also the node info needs to be updated in block table. 
        """
        logging.debug("migrating express block list")
        
        for block in blockList:
            self.migrateBlockToGlobalDBS(self.CAF_DBS_URL, block)
    
    def migrateNormalBlocks(self, blockList):
        """
        _migrateNormalBlocks_
        
        Migrate to Global DBS 
        Also the node info needs to be updated in block table. 
        """
        logging.debug("migrating normal block list")
        for block in blockList:
            self.migrateBlockToGlobalDBS(self.localDBSUrl, block)
            
    def blockTransferCheck(self, block):
        """
        """
        nodeList = listStorageNodesByBlock(self.t0astDBConn, block)
                
        logging.debug("checking %s subscription node list: %s" % 
                              (block.getPrimaryDatasetName(), nodeList))
                
        try:
            if isBlockTransfered(self.phedexDSUrl, block, nodeList):
            # change block block to exported if it is repack block
            # reco block to completed if no other processing is planned for reco
            
            # update exportable subscrition to completed for all the files belong to that block
                # to do: get se list from node list
                #TODO: update block table. exported time,  node id 
                seList = nodeList
                updateBlockFileStatusToExported(self.t0astDBConn, block, seList)
                
                return nodeList
            else:
                return False
        except PhEDExServiceError, ex:
            self.publishAlert("Severe", str(ex))
            return False
                
    def migrateBlockToGlobalDBS(self, localDBSUrl, block):
        """
        """
        if not isParentBlockMigrated(self.t0astDBConn, block["BLOCK_ID"]):
            logging.debug("""Not all the parent block is migrated for 
                               \nBlcok Name:%s \nBlock ID:%s Will try later""" 
                                % (block["BLOCK_NAME"], block["BLOCK_ID"]))
            return
        
        logging.debug("""start to migrate to global DBS \nBlock Path:%s
                               \nBlcok Name:%s \nBlock ID:%s""" 
                                % (block.getDatasetPath(), block["BLOCK_NAME"],
                                   block["BLOCK_ID"]))
        try:
            
            self.dbsApi.migrateDatasetContents(localDBSUrl, 
                                       self.globalDBSUrl,
                                       block.getDatasetPath(), 
                                       block["BLOCK_NAME"]
                                       )
            updateBlockMigrateStatus(self.t0astDBConn, block, "Migrated")
        
        except DbsBadResponse, e:
            logging.error(""" Some DBS Error for data path path %s block %s ID:%s again later.
                               \n local dbs: %s
                               \n global dbs: %s
                               \nError from DBS: %s
                            """ % (block.getDatasetPath(), block["BLOCK_NAME"], 
                                   block["BLOCK_ID"], localDBSUrl, self.globalDBSUrl, e))
            raise DbsBadResponse, e
        return
    
    def handleTestBlock(self):
        """
        _handleTestBlock_
        """
        # get exported but not completed block list
        # only repacked block can be exported not completed.
        # if all the files in the block is exported but not yet reconstructed.
        # this should/may be in separate polling cycle.        
        injectedBlockList = listBlocksByStatus(self.t0astDBConn, "InFlight")
        
        for block in injectedBlockList:
            updateBlockFileStatusToExported(self.t0astDBConn, 
                                            block, self.CAF_NODE)
            # mark every block as deletable since ExpressInjector processed 
            # all tiers
            markFilesInBlockToDelete(self.t0astDBConn, block)        
    
    
    def sortBlockListByStreamType(self, blockList):
        """
        _sortBlockListByStreamType_
        
        sort block list by stream type:
        Currently only 2 different stream type, Express and Non Express
        """
        blockDict  = {}
        blockDict["Express"] = []
        blockDict["Normal"]= []
        for block in blockList:
            if self.isBlockExpressStream(block):
                blockDict["Express"].append(block)
            else:
                blockDict["Normal"].append(block)
        return blockDict
                
    def isBlockExpressStream(self, blockInfo):
        """
        _isBlockExpressStream_
        
        Check whether given block is express stream,
        If it is, return True other wise return False
        """
        runConfig = self.getRunConfig(blockInfo["RUN_ID"])
        return runConfig.isDatasetExpress(blockInfo["DATASET_ID"])
        
    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("BlockAuditor:StartDebug")
        self.ms.subscribeTo("BlockAuditor:EndDebug")
        self.ms.subscribeTo("BlockAuditor:Poll")

        self.ms.publishUnique("BlockAuditor:Poll", "", self.pollInterval)
        self.ms.commit()

        self.connectT0AST()

        # wait for messages
        self.messageLoop()
