"""
API interface to T0AST DB specific to BlockInjector.
The main purpose of the module is combine T0AST DB queries 
in a way it should happen in one transaction

"""

__revision__ = "$ID: $"
__version__ = "$Revision: 1.70 $"
import os
import logging
import datetime
from sets import Set

from T0 import Globals
from T0.DataStructs.T0ASTFile import T0ASTFile

from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListBlock
from T0.State.Database.Reader.ListBlock import listActiveBlock
from T0.State.Database.Reader.ListBlock import listAllActiveBlocks
from T0.State.Database.Reader.ListBlock import listActiveBlocksByRunStatus
from T0.State.Database.Reader.ListBlock import getBlockInfoByID
from T0.State.Database.Reader.ListBlock import getExpressBlockCreationTime
from T0.State.Database.Reader.ListBlock import listBlocksByStatus

from T0.State.Database.Writer.InsertBlock import insertBlock
from T0.State.Database.Writer.InsertBlock import insertBlockParentage
from T0.State.Database.Writer.InsertBlock import closeBlockByBlockID
from T0.State.Database.Writer.InsertBlock import insertFileBlockAssoc
from T0.State.Database.Writer.InsertBlock import updateBlockSizeAndFileCount
from T0.State.Database.Writer.InsertBlock import updateExportStartTime
from T0.State.Database.Writer.InsertBlock import updateBlockStatusByID
from T0.State.Database.Writer.InsertPhEDExSubscription import insertPhEDExSubscriptionMade


def listFilesBySubscriptionStatus(dbConn, wmbsSub, status, runID=None):
    """
    _listFilesBySubscriptionStatus_

    Get a list of the T0AST file objects in T0AST with a particular export status.  
    The status must be a string that corresponds to a row in the export_status
    table.  Data is returned is list of T0ASTFile.
    """
    logging.debug("Subscription id : %s type: %s" % (wmbsSub['id'], 
                                                     wmbsSub['type']))
    if runID == None:
        wmbsFiles = wmbsSub.filesOfStatus(status)        
    else:
        logging.debug("Getting % status of Run id %s" % (status, runID))
        wmbsFiles = wmbsSub.filesOfStatusByRun(status, runID)        
        
    filesWithBlock = []
    filesWithoutBlock  =[]
    
    for wmbsFile in wmbsFiles:
        t0astFile = T0ASTFile(wmbsFile) 
        t0astFile.datasetPathID = \
            ListDatasets.listDatasetIDForWMBSFile(dbConn, wmbsFile["id"])
        
        #To do: not sure this is needed (might not be needed)
        datasetIDs = ListDatasets.listDatasetIDsFromID(dbConn, 
                                              t0astFile.datasetPathID)
        t0astFile["DATASET_ID"] = datasetIDs["PRIMARY"]
            
        datasetNames = \
              ListDatasets.listDatasetNamesForWMBSFile(dbConn, wmbsFile["id"])
        t0astFile["PRIMARY_DATASET"] = datasetNames["PRIMARY"]
        t0astFile["PROCESSED_DATASET"] = datasetNames["PROCESSED"]
        t0astFile["DATA_TIER"] = datasetNames["TIER"]
             
        t0astFile["BLOCK_ID"] = ListBlock.getBlockIDByFileID(dbConn, wmbsFile["id"])
        
        if t0astFile["BLOCK_ID"] != None:
            # this can only happen some thing goes wrong with export
            # this will return the files to recover
            filesWithBlock.append(t0astFile)
        else:
            filesWithoutBlock.append(t0astFile)
        
    return {"BLOCK":filesWithBlock, "NO_BLOCK":filesWithoutBlock}

def _orderFilesByTier(t0astFileList):
    """
    _orderFilesByTier_
    
    order t0ast files by give TIER_ORDER defined in GlobalValues.
    if there are missing tiers, they will be going to at the end of list
    TODO: not very optimal solution - need to be improved
    """
    orderedFiles = []
            
    for dataTier in Globals.GlobalValues.TIER_ORDER:
        for t0astFile in t0astFileList:
            if t0astFile["DATA_TIER"] == dataTier:
                orderedFiles.append(t0astFile)
    
    # in case there is tier not defined in TIER_ORDER
    for t0astFile in t0astFileList:
        if t0astFile["DATA_TIER"] not in Globals.GlobalValues.TIER_ORDER:
            orderedFiles.append(t0astFile)
    
    return orderedFiles
        
def _addFileToBlockSet(blockSet, blockInfo, t0astFile, isNewBlock = False):
    """
        _addFileToBlockSet_  
        add file to a blockset structure
        block set is dict format of 
        {blockID:{'block':blockInfo, 'fileList':[file1, file2, ...], 
                  'isNewBlock':True}} 
    """
        
    if not blockSet.has_key(blockInfo["BLOCK_ID"]):
        blockSet[blockInfo["BLOCK_ID"]] = {}
        blockSet[blockInfo["BLOCK_ID"]]['fileList'] = []
            
    blockSet[blockInfo["BLOCK_ID"]]['block'] = blockInfo
    if t0astFile != None: 
        blockSet[blockInfo["BLOCK_ID"]]['fileList'].append(t0astFile)
    blockSet[blockInfo["BLOCK_ID"]]['isNewBlock'] = isNewBlock
        
    return    

def _addFileToDatasetPathSet(datasetPathSet, blockInfo, t0astFile):
    """
        _addFileToBlockSet_  
        add file to a DatasetPathSet structure
        DatasetPathSet is dict format of 
        {datasetPath: {blockName:{'fileList': [file1, file2, file], 
                'status':'Closed' },  
        .....}
    """
    datasetPath = blockInfo.getDatasetPath()
        
    if not datasetPathSet.has_key(datasetPath) :
        datasetPathSet[datasetPath] = {}

    if not datasetPathSet[datasetPath].has_key(
                                    blockInfo["BLOCK_NAME"]):
        datasetPathSet[datasetPath][blockInfo["BLOCK_NAME"]] = {'fileList': [], 
                                                'status': blockInfo["STATUS"]}
    
    if t0astFile != None:     
        datasetPathSet[datasetPath][
                       blockInfo["BLOCK_NAME"]]['fileList'].append(t0astFile)
    return

def assignFilesToBlock(dbConn, dbsSubscription, conditionFunc, expressFlag=False):
    """
        1. Find all exportable repacked files.
        2. Use existing active blocks or open new the active blocks
           (update repacked on block_id, (insert into active blocks))
           inject the file to PhEDEx, update local DBS 
        3. close the block when closing condition is met and send message for 
           closeout to BlockAuditor 
    """

    # block set is nested dict format of 
    # {blockID:{'block':blockInfo, 'fileList':[file1, file2, ...], 
    #              'isNewBlock':True}, ....} 
    blockSet = {} 
    isNewBlock = False
    
    t0astFiles = listFilesBySubscriptionStatus(dbConn, dbsSubscription,
                                              "Available")
    
    # get the list of files block is already assigned but not in DBS
    # this will be normally empty list but if DBS upload is failed for some 
    # reason, this will pick up the the not luploaded files 
    # when BlockInjector restarts
    for t0astFile in t0astFiles["BLOCK"]:
        blockInfo = getBlockInfoByID(dbConn, t0astFile["BLOCK_ID"])
        # update run_id info (in t0ast block can't contain more than one run
        # due to the structure of active block so this can be done by modifying
        # sql query in getBlockInfoByID
        
        blockInfo["RUN_ID"] = t0astFile.getRunID()
        _addFileToBlockSet(blockSet, blockInfo, t0astFile, isNewBlock)
        
    orderedT0ASTFiles = _orderFilesByTier(t0astFiles["NO_BLOCK"])    
    
    for t0astFile in orderedT0ASTFiles:
        blockInfo = listActiveBlock(dbConn, t0astFile)
        dbConn.beginTransaction()
        if blockInfo == None:
            # insertBlock will insert into ActiveBlock as well
            # also insert into block_run_assoc table
            blockInfo = insertBlock(dbConn, t0astFile)
            isNewBlock = True
        else:
            
            blockInfo["BLOCKSIZE"] += t0astFile["FILESIZE"]
            blockInfo["FILECOUNT"] += 1
            updateBlockSizeAndFileCount(dbConn, blockInfo)
            
        # update the t0astFile Export status and blockID    
        insertFileBlockAssoc(dbConn, t0astFile["ID"], blockInfo["BLOCK_ID"])

        # add block parentage if block is not set already
        # this doesn't prevent tangled block
        if blockInfo.getDataTier() != "RAW":
            logging.debug("""Block: File ID : %s, Block ID: %s 
                             DataTier: %s""" % (t0astFile["ID"], 
                                                blockInfo["BLOCK_ID"],
                                                blockInfo.getDataTier()))
            
            # TODO:Express parentage needs to be handled differently
            if not expressFlag:
                fileIDs = t0astFile.getParentList(dataType="id")
                logging.debug("Parent file ids: %s" % fileIDs)
                insertBlockParentage(dbConn, fileIDs, blockInfo["BLOCK_ID"])
                
         # it is not guaranteed that RAW file is assigned to a block before
         # corresponding RECO file did. check whether this gives an error
#        elif blockInfo.getDataTier() == "RAW":
#            fileIDs = wmbsFile.getDescendants(level=2, type="id")
            
            
        # check the block closing condition
        # if condition is met close the block
        if conditionFunc(blockInfo):
            closeBlockByBlockID(dbConn, blockInfo)
            logging.debug("BlockInjectorAPI: block %s (block id: %s) closed" 
                          % (blockInfo["BLOCK_NAME"], blockInfo["BLOCK_ID"]))
            
        #one transaction
        dbConn.commit()
        
        _addFileToBlockSet(blockSet, blockInfo, t0astFile, isNewBlock)
        
    # This returned blockSet will be used
    # call local DBS update
    # create the xml file for PhEDEx injection
        
    return blockSet

def getTransferableFiles(dbConn, transferSubscription):
    """
    _getTransferableFiles_
    
    get transferable files (updated in local dbs but not yet injected in PhEDEx
    and organize files convenient for PhEDEx injection)
    {datasetPath: {blockName:[file1, file2, file]},  .....}
    and also organize files to BlockSet
    blockset is nested dict format of 
    {blockID:{'block':blockInfo, 'fileList':[file1, file2, ...], 
                  'isNewBlock':True}, ....} 
    """
    
    t0astFiles = listFilesBySubscriptionStatus(dbConn, transferSubscription,
                                              "Available")
    if len(t0astFiles["NO_BLOCK"]) != 0:
        msg = """ Error: There are files which are not assigned to block: %s
              """ % t0astFiles["NO_BLOCK"]  
        raise Exception, msg
    
    # block set is nested dict format of 
    # {blockID:{'block':blockInfo, 'fileList':[file1, file2, ...], 
    #              'isNewBlock':True}, ....} 
    blockSet = {}
    blocksPerDatasetPath = {}
    for t0astFile in t0astFiles["BLOCK"]:
        blockInfo = getBlockInfoByID(dbConn, t0astFile["BLOCK_ID"])
        # update run_id info (in t0ast block can't contain more than one run
        # due to the structure of active block so this can be done by modifying
        # sql query in getBlockInfoByID
        
        blockInfo["RUN_ID"] = t0astFile.getRunID()
        
        _addFileToBlockSet(blockSet, blockInfo, t0astFile)
        _addFileToDatasetPathSet(blocksPerDatasetPath, blockInfo, t0astFile)
        
    return {'DatasetPathSet': blocksPerDatasetPath, 'BlockSet':blockSet}
        
def convertToDatasetPathSet(blockSet):
    """
    _convertToDatasetPathSet_
    
    convert blockset structure to datasetPathSet structure.
    normally it is used on empty blockSet which doesn't have any files listed.
    
    """
    datasetPathSet = {}
    for blockDict in blockSet.values():
        if (len(blockDict['fileList']) == 0):
            _addFileToDatasetPathSet(datasetPathSet, blockDict['block'], None)
        else:
            for file in blockDict['fileList']:
                _addFileToDatasetPathSet(datasetPathSet, blockDict['block'],
                                         file)
        
    return datasetPathSet
            
def updateFilesToInFlightStatus(dbConn, blockSet, transferSubscription):
    """
    _updateFilesToInFlightStatus_
    
    update file status to InFlight after DBS and PhEDEx are updated.
    Also update the time stamp in Block export_start_time
    """
    logging.debug("Uplating files to InFlight status")
    for blockDict in blockSet.values():
        dbConn.beginTransaction()
        if ( len(blockDict['fileList']) > 0 ):
            wmbsFileList = []
            for file in blockDict['fileList']:
                wmbsFileList.append(file.convertToWMBSFile())
            transferSubscription.acquireFiles(wmbsFileList)
            updateExportStartTime(dbConn, blockDict['block']["BLOCK_ID"])
        
        # check block status if it is DBSuploaded update to InFlight status.
        # InFlight status means all the files in the block are updated 
        # in DBS and injected in PhEDEx.
        if blockDict['block']["STATUS"] == "DBSUploaded":
            updateBlockStatusByID(dbConn, blockDict['block'], "InFlight")
        
        dbConn.commit()
    return

def updateFilesToTransferableStatus(dbConn, blockSet, dbsSubscription,
                                   transferFileset):
    """
    _updateFilesToTrasferableStatus_
    
    update file status to Transferable after Loacal dbs is updated
    """
    logging.debug("Uplating files to Transferable status")
    for blockDict in blockSet.values():
        dbConn.beginTransaction()
        if ( len(blockDict['fileList']) > 0 ):
            wmbsFileList = []
            for file in blockDict['fileList']:
                wmbsFile = file.convertToWMBSFile()
                wmbsFileList.append(wmbsFile)
                transferFileset.addFile(wmbsFile)
            transferFileset.commit()
            # TODO: doesn't handle the dbs failed case, should be successful all the 
            # time: comment out acquireFiles function. - directly go to complete status
            # dbsSubscription.acquireFiles(fileList)
    
            dbsSubscription.completeFiles(wmbsFileList)
            
        # check block status if it it is closed update to DBSUploaded status.
        # InFlight status means all the files in the block are updated 
        # in DBS and injected in PhEDEx.
        if blockDict['block']["STATUS"] == "Closed":
            updateBlockStatusByID(dbConn, blockDict['block'], "DBSUploaded")
        
        dbConn.commit()
    
    return

def getBlocksByStatus(dbConn, status, getRunConfig, expressFlag = False):
    """
    _getBlocksByStatus_
    
    """
    logging.debug("getting blocks of %s status" % status)
    allBlockList = listBlocksByStatus(dbConn, status)
    blockList = []
    runConfigCache = {}
    for block in allBlockList:
        runConfig = runConfigCache.setdefault(block["RUN_ID"], 
                                    getRunConfig(block["RUN_ID"]))
        
        if expressFlag:
            if runConfig.isDatasetExpress(block["DATASET_ID"]):
                blockList.append(block)
        else:
            if not runConfig.isDatasetExpress(block["DATASET_ID"]):
                blockList.append(block)
    
    if status == "Closed":
        subscription = Globals.retrieveDBSUploadSubscription()
    elif status == "DBSUploaded":
        subscription = Globals.retrieveTransferSubscription()
    return _convertToBlockSet(dbConn, blockList, subscription)
        
        
    
def closeBlocksByRun(dbConn, stageMap, getRunConfig):
    """
    _closeBlocksByRun_
    
    close regular blocks by its proper run states
    """ 
    # performance may improved if runID is directly used in 
    # closeBlockByBlockID
    # needed in block table. Possible to make one sql statement.
    allBlockList = listActiveBlocksByRunStatus(dbConn, stageMap["Stage"])
    blockList = []
    runConfigCache = {}    
    for block in allBlockList:
        if block.getDataTier() in stageMap["TierList"]:
            runConfig = runConfigCache.setdefault(block["RUN_ID"], 
                                    getRunConfig(block["RUN_ID"]))
            if not runConfig.isDatasetExpress(block["DATASET_ID"]):
                blockList.append(block)
    
    #if block is empty do nothing
    if len(blockList) == 0:
        return
    
    dbConn.beginTransaction()
    closeBlockByBlockID(dbConn, blockList)
    logging.debug("BlockInjectorAPI: Close blocks by Run: %s closed" 
                          % blockList)
    dbConn.commit()
    
    return


def closeExpressBlocksByTimeout(dbConn, getRunConfig, timeout):
    """
    closeExpressBlockByTimeout
    
    close express blocks by time out
    """
    allBlockList = listAllActiveBlocks(dbConn)
    
    blockList = []
    runConfigCache = {}
    for block in allBlockList:
        runConfig = runConfigCache.setdefault(block["RUN_ID"], 
                                    getRunConfig(block["RUN_ID"]))
        if runConfig.isDatasetExpress(block["DATASET_ID"]):
            creationTime = getExpressBlockCreationTime(dbConn, 
                                                   block["BLOCK_ID"])
            if timeout <= \
                (datetime.datetime.now() - datetime.datetime.fromtimestamp(creationTime)):
                logging.debug("Using the time out - no file arrived")
                blockList.append(block)
    
    #if block is empty do nothing
    if len(blockList) == 0:
        return
    
    dbConn.beginTransaction()
    closeBlockByBlockID(dbConn, blockList)
    logging.debug("BlockInjectorAPI: Close Express blocks by Timeout: %s closed" 
                          % blockList)
    dbConn.commit()
            
    return 

def _convertToBlockSet(dbConn, blockList, subscription):
    """
    _convertToBlockSet
    
    convert block list to block set.
    
    This is used for non file related close out condition.
    ie run boundary or timeout condition so the FileList should be
    always empty.
    """
    blockSet = {}
                
    # when run closed is called, normally, there shouldn't be 
    # any files available for export, but if DBS upload fails and 
    # run close signal picked up first, there will be files 
    # available 
    
    availableFilesByRun = {}
    for blockInfo in blockList:
        if not availableFilesByRun.has_key(blockInfo["RUN_ID"]):
            t0astFiles = listFilesBySubscriptionStatus(dbConn, subscription,
                                              "Available", blockInfo['RUN_ID'])   
            availableFilesByRun[blockInfo["RUN_ID"]] = t0astFiles["BLOCK"]
            logging.debug("-----Block %s: \nfile length: %s" % (blockInfo, 
                                                                len(t0astFiles["BLOCK"])))    
    
    # block set is dict format of 
    # {blockID:{'block':blockInfo, 'fileList':[file1, file2, ...] }
    # in this case file list will be always empty
     
    for blockInfo in blockList:
        updateFlag = False
        for file in availableFilesByRun[blockInfo["RUN_ID"]]:
            if file["BLOCK_ID"] == blockInfo["BLOCK_ID"]:
                logging.debug("adding file %s" % file)
                _addFileToBlockSet(blockSet, blockInfo, file)        
                updateFlag = True
        if not updateFlag:
            _addFileToBlockSet(blockSet, blockInfo, None)
            
    return blockSet
 
def insertPhEDExSubscription(dbConn, subscription):
    dbConn.beginTransaction()
    insertPhEDExSubscriptionMade(dbConn, subscription)
    dbConn.commit()
    return
