#!/usr/bin/env python
"""
_BlockInjectorComponent_

Create the logical block for exportable RAW, RECO.
Manage the block open and closing
 
"""

__revision__ = "$Id: BlockInjectorComponent.py,v 1.129 2010/08/10 16:37:15 sryu Exp $"
__version__ = "$Revision: 1.129 $"

import logging
import datetime

from ProdAgentCore.Configuration import loadProdAgentConfiguration
from DBSAPI.dbsApi import DbsApi
from WMCore.WMBS.Fileset import Fileset as WMBSFileset

from T0 import Globals

from T0.GenericTier0.GenericTier0Component import GenericTier0Component
from T0.State.Database.Reader.ListPhEDExSubscription import getSubscriptionPolicy

from T0.State.Database.Reader import ListDatasets

from T0.ExportSystemAPI.PhEDExDataServiceAPI import subscribeToPhEDEx
from T0.ExportSystemAPI.PhEDExDataServiceAPI import injectBlocksToPhEDEx 
from T0.ExportSystemAPI.DBSUploadAPI import insertDBSBlocks
from T0.ExportSystemAPI.DBSUploadAPI import updateDBSRunStatus
from T0.ExportSystemAPI.ExportSystemExceptions import PhEDExServiceError
from T0.ExportSystemAPI.ExportSystemExceptions import DBSServiceError

from T0.BlockInjector.BlockInjectorAPI import assignFilesToBlock
from T0.BlockInjector.BlockInjectorAPI import closeBlocksByRun
from T0.BlockInjector.BlockInjectorAPI import updateFilesToInFlightStatus
from T0.BlockInjector.BlockInjectorAPI import updateFilesToTransferableStatus
from T0.BlockInjector.BlockInjectorAPI import insertPhEDExSubscription
from T0.BlockInjector.BlockInjectorAPI import convertToDatasetPathSet
from T0.BlockInjector.BlockInjectorAPI import getTransferableFiles
from T0.BlockInjector.BlockInjectorAPI import closeExpressBlocksByTimeout
from T0.BlockInjector.BlockInjectorAPI import getBlocksByStatus

class BlockInjectorComponent(GenericTier0Component):
    """
    _BlockInjectorComponent_

    create the logical block for exportable RAW, RECO, HLTBug files to transfer to T1
    using PhEDEx.
    Also clean up files after transfer is done
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component to setup logging.
        """
        GenericTier0Component.__init__(self, **args)
        
        # Lookup maximun block size that can be exported
        # through the PhedEx.  Default is 200 gigabytes.
        self.disableThreshold  = self.convertToBoolean(
                                          self.lookupParameter(
                                                "DisableThreshold", "False"))
        
        self.disableExpressThreshold  = self.convertToBoolean(
                                          self.lookupParameter(
                                             "DisableExpressThreshold", "False"))
        
        self.blockSizeThreshold = int(self.lookupParameter(
                                           "blockSizeThreshold",
                                           "200000000000"))
        self.numOfFilesThreshold = int(self.lookupParameter(
                                            "numOfFilesThreshold", 100))
        self.numOfExpressFilesThreshold = int(self.lookupParameter(
                                            "numOfExpressFilesThreshold", 1))
        timeStr = self.lookupParameter("ExpressTimeThreshold", "01:00:00")
        self.expressTimeThreshold = self.convertToTimeDelta(timeStr)
        
        self.pollInterval = self.lookupParameter("PollInterval", "00:03:00")
        
        self.runPollNumber = int(self.lookupParameter("RunPollNumber", 5))
        
        self.T0_BUFFER_NODE = self.lookupParameter(
                                         "T0BufferNode", "T0_CH_CERN_Export")
        
        self.localDBSUrl = None
        self.globalDBSUrl = None
        self.phedexDSUrl = None
        self.endedRunNumber = None
        # get it to do all the polling right after startup
        self.pollingCount = self.runPollNumber

        self.expressDBSSubscription = None
        self.expressTransferSubscription = None
        self.dbsSubscription = None
        self.transferSubscription = None
        
        self.expressTransferFileset = None
        self.transferFileset = None
        # get local storage element info
        self.localSEList = ["srm-cms.cern.ch"]
        #TODO: verify  whether caf use the same name 
        self.CAF_SE_LIST = self.localSEList 
        # t0 buffer node as source for PhEDEx
        # the node name should be registered in PhEDEx
        # real one 
        #self.T0_BUFFER_NODE = "T0_CH_CERN_Export"
        # the fake one for the test bed
        #self.T0_BUFFER_NODE = "TX_Test1_Buffer"
      
        # contains run numbers to check whether
        # new run close status is updated in DBS or not.
        #TODO: Set might not be a best container for this.
        # maybe limit the cache size to prevent from growing infinitely 
        self.dbsRunCloseCache = set()
        
        logging.info("BlockInjector Component Started")
        
    
    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component receiving a message.
        """
        # always have a debug on/off switch
        
        if event == "BlockInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "BlockInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "BlockInjector:Poll":
            self.poll()
        elif event == "BlockInjector:SubscriptionPoll":
            self.subscriptionPoll()
        else:
            logging.error("Unrecognized event: %s" % event)

        return

    def loadExportSystemConf(self):
        
        paConfig = loadProdAgentConfiguration()
        localDBSConfig = paConfig.getConfig("LocalDBS")
        globalDBSConfig = paConfig.getConfig("GlobalDBSDLS")
        phedexConfig = paConfig.getConfig("PhEDExConfig")
        self.localDBSUrl = localDBSConfig.get("DBSURL", None)
        self.CAF_DBS_URL = localDBSConfig.get("CAFDBSURL", None)
        self.globalDBSUrl = globalDBSConfig.get("DBSURL", None)
        
        self.CAF_NODE = localDBSConfig.get("CAFNode", None)
        # the configuration should be File:Section format
        self.phedexDBargs = phedexConfig.get("DBPARAM", None)
        self.xmlDir = phedexConfig.get("PhEDExDropBox", None)
        self.phedexDSUrl = phedexConfig.get("DataServiceURL", None)
        
    def poll(self,type = None):
        """
        _poll_

        Single cycle invocation of finding all exportable files and create the
        them by primary dataset, data tier(RAW, RECO)
        
        runNumber is only used as block closing condition parameter
        
        This could create the small blocks at the end of the run.
        """
        
        
        logging.info("loading WMBS Subscription")
        
        # make sure runConfigCache is initialized
        # runConfigCache should be initialized 
        # (set the T0AST database connection) 
        # before block is created by T0ASTFile
        
        self.initRunConfigCache()
        # set subscriptions
        self.expressDBSSubscription = Globals.retrieveExpressDBSSubscription()
        self.expressTransferSubscription = \
                           Globals.retrieveExpressTransferSubscription()
        self.dbsSubscription = Globals.retrieveDBSUploadSubscription()
        self.transferSubscription = Globals.retrieveTransferSubscription()
        
        if self.localDBSUrl == None or self.phedexDSUrl == None:
            self.loadExportSystemConf()
            
        if self.expressTransferFileset == None: 
            self.expressTransferFileset = WMBSFileset(
                                             name="ExpressTransferable")
        if self.transferFileset == None: 
            self.transferFileset = WMBSFileset(name = "Transferable")
                
        logging.info("Searching Exportable Poll Invoked")
                
        self.exportExpressBlocks()
        self.exportExpressTimedOutBlocks()
        
        self.exportRegularBlocks()
        
        # checking run status is close out state to close the block and 
        # export it
        # polling cycle is 10 times longer than normal polling cycle
        # also update the run status when export is finished
        self.pollingCount += 1
            
        if self.pollingCount >= self.runPollNumber:

            logging.debug("----- run Polling starts %s" % self.pollingCount)
            
            
            # close all the blocks in order. (not sure it is necessary)
            # close all repack blocks if the run stage is in right stage 
            # close all reco blocks if the run stage is in right stage 
            # close all alcaskim blocks if the run stage is in right stage 
            
            for tier in Globals.GlobalValues.TIER_ORDER:
                self.exportRegularBlocksForRunClose(tier)
            
            self.closeDBSRunByPrimaryDataset()
               
            self.pollingCount = 0

        self.ms.publishUnique('BlockInjector:Poll', '',
                              self.pollInterval)
        self.ms.commit()
        return
    
    def exportRegularBlocksForRunClose(self, dataTier):
        """
        _exportRegularBlocksForRunClose_
        
        close out all the regular blocks according to the proper run state 
        """
        stageMap = Globals.closeOutStageMap(dataTier)
        
        logging.info("Searching Run Closed Block to Export for regular block")
        
        closeBlocksByRun(self.t0astDBConn, stageMap, 
                         self.getRunConfig)
        
        fileBlocks = getBlocksByStatus(self.t0astDBConn, "Closed", 
                                       self.getRunConfig)
        
        self.localDBSUpload(fileBlocks, 
                            self.localDBSUrl, self.localSEList,
                            self.dbsSubscription, self.transferFileset)
    
        fileBlocks = getBlocksByStatus(self.t0astDBConn, "DBSUploaded",
                                       self.getRunConfig)
        
        datasetPathSet = convertToDatasetPathSet(fileBlocks)
    
        self.phedexTransfer(datasetPathSet, fileBlocks, self.globalDBSUrl, 
                            self.T0_BUFFER_NODE, self.transferSubscription,
                            phedexSubFlag=True)
        return
    
    def exportExpressTimedOutBlocks(self):
        """
        _exportExpressTimedOutBlocks_
        
        close all the express blocks which were timed out. and export those blocks
        """
        logging.info("Searching Timed out express Block to Export ")
            
        closeExpressBlocksByTimeout(self.t0astDBConn, 
                                    self.getRunConfig, self.expressTimeThreshold)
        
        fileBlocks = getBlocksByStatus(self.t0astDBConn, "Closed",
                                       self.getRunConfig, expressFlag=True)
                               
        self.localDBSUpload(fileBlocks, 
                            self.CAF_DBS_URL, self.CAF_SE_LIST,  
                            self.expressDBSSubscription, 
                            self.expressTransferFileset, expressFlag=True)
    
        fileBlocks = getBlocksByStatus(self.t0astDBConn, "DBSUploaded",
                                       self.getRunConfig, expressFlag=True)
        
        datasetPathSet = convertToDatasetPathSet(fileBlocks)
    
        self.phedexTransfer(datasetPathSet, fileBlocks, self.globalDBSUrl, 
                            self.CAF_NODE, self.expressTransferSubscription,
                            phedexSubFlag=False)
        return
    
    def exportRegularBlocks(self):
        """
        _exportRegularBlocks_
        just wrapper function of exportBlocks funcion with argument set for
        regular block export
        """
        logging.info("Searching Regular files to assign to blocks")
        
        fileBlocks = assignFilesToBlock(self.t0astDBConn, 
                                              self.dbsSubscription,
                                              self.closeCondition)
        
        self.localDBSUpload(fileBlocks, 
                            self.localDBSUrl, self.localSEList,
                            self.dbsSubscription, self.transferFileset)
        
        datasetPathSet = getTransferableFiles(self.t0astDBConn, 
                                              self.transferSubscription)
        
        self.phedexTransfer(datasetPathSet['DatasetPathSet'], 
                            datasetPathSet['BlockSet'], self.globalDBSUrl, 
                            self.T0_BUFFER_NODE, self.transferSubscription,
                            phedexSubFlag=True)
        
        return 
    
    def exportExpressBlocks(self):
        """
        _exportExpressBlocks_
        just wrapper function of exportBlocks funcion with argument set for
        express block export
        """
        logging.info("Searching Express files to assign to blocks")
        
        expressFileBlocks = assignFilesToBlock(self.t0astDBConn, 
                                              self.expressDBSSubscription,
                                              self.expressCloseCondition, 
                                              expressFlag=True)
        
        self.localDBSUpload(expressFileBlocks, self.CAF_DBS_URL, 
                            self.CAF_SE_LIST,  self.expressDBSSubscription, 
                            self.expressTransferFileset, expressFlag=True)
        
        expressDPSet = getTransferableFiles(self.t0astDBConn, 
                                            self.expressTransferSubscription)
        
        self.phedexTransfer(expressDPSet['DatasetPathSet'], 
                            expressDPSet['BlockSet'], self.globalDBSUrl, 
                            self.CAF_NODE, self.expressTransferSubscription,
                            phedexSubFlag=False)
        
        return
        
    def phedexSubscriptionHandler(self, fileBlockSet):
        
        if len(fileBlockSet) == 0:
            logging.info("Empty block set: Nothing to make PheEDEx Subscription")
            return
        #reorganize fileBlockSet for subscription
        blockList = []
        blockSetByRun = set() 
        for blockDict in fileBlockSet.values():
            blockTuple = (blockDict["block"]["DATASET_PATH_ID"], 
                          blockDict["block"]["RUN_ID"])
                
            if not (blockTuple in blockSetByRun):
                blockSetByRun.add(blockTuple)
                blockList.append(blockDict["block"])
            
        # check the new subscription for blockList: will only create the 
        # subscription per dataset path not per block
        subPolicy = getSubscriptionPolicy(self.t0astDBConn, blockList)
        # if there is no new subscription do nothing
        if subPolicy == None:
            logging.debug("No new subscription found")
            return

# commented out: this function should call only after corresponding block is called        
#        dataPathSet = Set()
#        for subscription in subPolicy.getSubscriptionList():
#            for dataPath in subscription.getDatasetPaths():
#                dataPathSet.add(dataPath)
#        
#        # might need to keep cache for datapath list for preventing injecting duplication
#        # need to inject empty dataset before subscription is made
#        injectDatasetsToPhEDEx(self.phedexDSUrl, self.globalDBSUrl, 
#                               dataPathSet, self.T0_BUFFER_NODE)
#
        for subscription in subPolicy.getSubscriptionList():
            subscribeToPhEDEx(self.phedexDSUrl, subscription, self.globalDBSUrl)        
            # update subscription table
            insertPhEDExSubscription(self.t0astDBConn, subscription)
            logging.debug("PhEDEx Subscription is made for %s" % subscription)
        
        return
    
    def localDBSUpload(self, blockSet, localDBSUrl, localSEList, 
                       dbsSubscription, transferFileset, expressFlag=False):
        
        logging.info("Using local DBS: %s" % localDBSUrl)
        if not blockSet:
            logging.info("No blocks to be uploaded")
            return
        
        # insert blocks to local dbs 
        logging.debug("Inserting blocks to LocalDBS") 
        localDBSApi = DbsApi({'url':localDBSUrl})
        
        try:
            insertDBSBlocks(localDBSApi, blockSet, localSEList, expressFlag)
            if len(blockSet) != 0:
                updateFilesToTransferableStatus(self.t0astDBConn, blockSet, 
                                           dbsSubscription, transferFileset)
        except DBSServiceError, e:
            logging.error(str(e))
        
        return
    
    def phedexTransfer(self, datasetPathSet, blockSet, globalDBSUrl, bufferNode,
                       transferSubscription, phedexSubFlag = False):
            
        if not blockSet:
            logging.info("No blocks to be transfered")
            return
        
        logging.debug("Injecting blocks to PhEDEx")
        
        # try to inject blocks to PhEDEx, if it fails, 
        try:
            logging.debug("Use PhEDEx data service: %s" % self.phedexDSUrl)
            logging.debug("Pointing Global DBS: %s" % globalDBSUrl)
            
            injectBlocksToPhEDEx(self.phedexDSUrl, globalDBSUrl, 
                                 datasetPathSet, bufferNode)
        except PhEDExServiceError, e:
            msg = """ PhEDEx data service error, Contact PhEDEx service: %s""" % str(e)
            #TODO check duplicate file update error when PhEDEx supports the error message for
            # non strict option
            
            # if request entry is too large make small piece and try again
            if str(e).find("Request Entity Too Large") != -1:
                logging.warning("""Error:%s \nBreak into small pieces and inject multiple times 
                                    :one injection per block """ % str(e))
                
                # commented out: one injection per dataset - still too large
                # Use this if PhEDEx increase the limit
#                for (k, v) in datasetPathSet.iteritems():
#                    oneDatasetPathSet = {}
#                    oneDatasetPathSet[k] = v
#                    injectBlocksToPhEDEx(self.phedexDSUrl, globalDBSUrl, 
#                                         oneDatasetPathSet, bufferNode)
                # inject one block at a time
                for (k, v) in blockSet.iteritems():
                    oneBlockSet = {}
                    oneBlockSet[k] = v
                    dsPathSet = convertToDatasetPathSet(oneBlockSet)
                    injectBlocksToPhEDEx(self.phedexDSUrl, globalDBSUrl, 
                                         dsPathSet, bufferNode)
                    self._subscribeAndUpdateStatus(oneBlockSet, 
                                        transferSubscription, phedexSubFlag)
            else:
                msg += "\nWill try later" 
                msg = Globals.stackTraceLog(msg)
                self.publishAlert("Severe", msg)
                # subscription will not be made either
                return
        else:
            self._subscribeAndUpdateStatus(blockSet, transferSubscription, phedexSubFlag)
        
        return 

    def _subscribeAndUpdateStatus(self, blockSet, transferSubscription, phedexSubFlag):
        # subscription is made after at least one block is injected 
        # so we don't have to make the explicit subscription for the datsetpath
        if phedexSubFlag:
            logging.debug("Trying to Subscribe to PhEDEx for num of blockset %s" % 
                          len(blockSet))
            try:
                self.phedexSubscriptionHandler(blockSet)
            except PhEDExServiceError, ex:
                self.publishAlert("Severe", str(ex))
        else:
            logging.debug("Subscription flag is set to False") 
        # update file status to InFlight if DBS and PhEDEx injection is
        # successful
        updateFilesToInFlightStatus(self.t0astDBConn, blockSet, transferSubscription)
        
        
    def closeDBSRunByPrimaryDataset(self):
        """
        _closeDBSRunByPrimaryDataset_

        close runs for datasets in DBS once the runs are
        either in CloseoutT1Skimming or Complete

        only handle bulk datasets !
        
        FIXME: what about datasets from express streams ?

        """
        logging.info("Closing run per dataset in DBS")
        
        datasetpathsByRun = ListDatasets.listDatasetPathsForCompleteRuns(self.t0astDBConn)

        for run, datasetpaths in datasetpathsByRun.items():

            if run in self.dbsRunCloseCache:
                continue

            try:
                updateDBSRunStatus(DbsApi({'url':self.localDBSUrl}),
                                   list(datasetpaths), run)
                self.dbsRunCloseCache.add(run)
            except DBSServiceError:
                self.publishAlert("Severe", "DBS run close update failed for run %s: Will try again" % run)


    def closeCondition(self, blockInfo):
        """
        _closeCondition_
        
        Closing conditions for regular blocks
        Warning: This can be extended but only file related condition. 
        Which means when the new file is arrived, conditions will be check
        Periodical time related condition is checked separately, 
        i.e. run based closing condition, since the data is pulled by the
        different condition.
        To do: might be able to combine those two categories and simplify 
        the codes  
        """
        
        ## if self.endedRunNumber != None:
        ##     if blockInfo["RUN_ID"] == self.endedRunNumber:
        ##         return True
        if self.disableThreshold:
            return False
        if self.blockSizeThreshold < blockInfo["BLOCKSIZE"]:
            return True
        elif self.numOfFilesThreshold <= blockInfo["FILECOUNT"]:
            return True
        else:
            return False
        
    def expressCloseCondition(self, blockInfo):
        """
        _expressCloseCondition_
        
        Closing conditions for express blocks
        Warning: This can be extended but only file related condition. 
        Which means when the new file is arrived, conditions will be check
        Periodical time related condition is checked separately, 
        i.e. timeout based closing condition, since the data is pulled by the
        different condition.
        To do: might be able to combine those two categories and simplify 
        the codes  
        """
        ## if self.endedRunNumber != None:
        ##     if blockInfo["RUN_ID"] == self.endedRunNumber:
        ##         return True
        if self.disableExpressThreshold:
            return False
        elif self.numOfExpressFilesThreshold <= blockInfo["FILECOUNT"]:
            logging.debug("Useing the file count threshold")
            return True
        else:
            logging.debug("Block is not ready to close")
            return False
    
    def convertToBoolean(self, value):
        if str(value).lower() in ("true", "yes"):
            return True
        else:
            return False 
        
    def convertToTimeDelta(self, timeStr):
        """
        time format should be '01:00:02'
        """
        splittedTime = timeStr.split(':')
        return datetime.timedelta(hours=int(splittedTime[0]), 
                                  minutes=int(splittedTime[1]),
                                  seconds=int(splittedTime[2]))
                                          
    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and enter the
        message loop.
        """
        self.ms.subscribeTo("BlockInjector:StartDebug")
        self.ms.subscribeTo("BlockInjector:EndDebug")
        self.ms.subscribeTo("BlockInjector:Poll")

        # do first poll quickly
        self.ms.publishUnique("BlockInjector:Poll", "", "00:00:20")
        self.ms.commit()

        self.connectT0AST()

        # wait for messages
        self.messageLoop()
