#!/usr/bin/env python
"""
_BlockAuditorComponent_

"""

__revision__ = "$Id: CleanUpManagerComponent.py,v 1.35 2010/11/05 11:30:16 hufnagel Exp $"
__version__ = "$Revision: 1.35 $"

import logging
import os
import re
import subprocess
import threading

from T0 import Globals
from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0.State.Database.Writer import InsertCleanup
from T0.State.Database.Reader import ListCleanup

from ProdCommon.FwkJobRep.SiteLocalConfig import loadSiteLocalConfig


class CleanUpManagerComponent(GenericTier0Component):
    """
    _CleanUpManagerComponent_

    
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component to setup logging.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("CleanUpManager Component Started")
        self.pollInterval = self.lookupParameter("PollInterval", "00:01:00")

        self.streamerPool = self.lookupParameter("StreamerPool", "t0export")
        self.tempPool = self.lookupParameter("TempPool", "t0input")
        self.mergedPool = self.lookupParameter("MergedPool", "t0input")

        self.maxBulkDelete = int(self.lookupParameter("MaxBulkDelete", 1))
        if self.maxBulkDelete > 100:
            self.maxBulkDelete = 100

        # switch between streamer deletions and just logging the call
        if self.lookupParameter("DeleteStreamers", "Off") == "On":
            self.deleteStreamers = True
        else:
            self.deleteStreamers = False

        # switch between unmerged file deletions and just logging the call
        if self.lookupParameter("DeleteTempFiles", "Off") == "On":
            self.deleteTempFiles = True
        else:
            self.deleteTempFiles = False

        # switch between merged file deletions and just logging the call
        if self.lookupParameter("DeleteMergedFiles", "Off") == "On":
            self.deleteMergedFiles = True
        else:
            self.deleteMergedFiles = False

        # if logging the stager_rm calls, put the logs here
        self.commandLogDir = self.lookupParameter("CommandLogDir", self.args['ComponentDir'])
        
        # TFC needed for LFN -> PFN conversion
        self.tfcInstance = loadSiteLocalConfig().trivialFileCatalog()


    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if message == "CleanUpManager:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "CleanUpManager:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "CleanUpManager:Poll":
            self.poll()
        else:
            logging.error("Unrecognized message: %s" % message)

        return


    def removeFilesWithStagerRM(self, castorPool, lfnToIdDict,
                                executeFlag, fileLocation):
        """
        _removeFilesWithStagerRM_
        
        remove list of files using stager_rm  command
        
        """
        if len(lfnToIdDict) == 0:
            return

        command = "stager_rm -S %s" %  castorPool

        pfnToLfnDict = {}
        for lfn in lfnToIdDict.keys():
            pfn = self.tfcInstance.matchLFN("direct", lfn)
            if pfn == None:
                msg = "Unable to map LFN to PFN:\n"
                msg += "LFN: %s\n" % lfn
                logging.error(msg)
                return []
            else:
                pfnToLfnDict[pfn] = lfn
                command += " -M %s" % pfn

        if executeFlag:

            proc = subprocess.Popen(command, shell = True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            proc.wait()
            output = proc.stdout.read()

            path = os.path.join(fileLocation, "StagerRM_Log.txt")
            fileHandle = open(path, "a")
            fileHandle.writelines("command = %s\n" % command)
            fileHandle.write(output)
            fileHandle.close()

            if proc.returncode == 0:
                return lfnToIdDict.values()
            else:
                idList = []

                regExpParser = re.compile('^([^ ]+)[ ]+(SUBREQUEST_READY|SUBREQUEST_FAILED 2 File not found on disk cache|SUBREQUEST_FAILED 2 File not found on this service class|SUBREQUEST_FAILED 2 No such file or directory)')

                outputLines = output.split('\n')
                for outputLine in outputLines:
                    match = regExpParser.match(outputLine)
                    if match != None:
                        idList.append(lfnToIdDict[pfnToLfnDict[match.group(1)]])

                return idList

        else:

            path = os.path.join(fileLocation, "StagerRM_Call.txt")
            fileHandle = open(path, "a")
            fileHandle.writelines("%s\n" % command)
            fileHandle.close()

            return lfnToIdDict.values()


    def poll(self):
        """
        _poll_

        mark deletable express streamers as such, updating deleted flag to 1
        mark deletable repack streamers as such, updating deleted flag to 1
        delete all streamers with deleted = 1, taking note of command return code
        for all streamers that were deleted update deleted flag to 2

        put new deletable WMBS files into delete fileset
        delete all delatable WMBS files, taking note of command return code
        for all files that were deleted complete them for the delete subscription

        """
        logging.debug("poll(): Running...")

        # updating status for new deletable streamers
        InsertCleanup.updateDeletableExpressStreamers(self.t0astDBConn)
        InsertCleanup.updateDeletableRepackStreamers(self.t0astDBConn)

        # getting all deletable streamers
        streamers = ListCleanup.getDeletableStreamers(self.t0astDBConn)

        for lfn in streamers.keys():
            logging.info("deleting streamer %s" % lfn)

        # split into self.maxBulkDelete big chunks and bulk stager_rm
        deletedStreamerIDs = []
        lfnToIdDict = {}
        for lfn in streamers.keys():
            lfnToIdDict[lfn] = streamers[lfn]
            del streamers[lfn]
            if len(lfnToIdDict) >= self.maxBulkDelete:
                deletedStreamerIDs.extend( self.removeFilesWithStagerRM(self.streamerPool, lfnToIdDict,
                                                                        self.deleteStreamers, self.commandLogDir) )
                lfnToIdDict = {}
        if len(lfnToIdDict) > 0:
            deletedStreamerIDs.extend( self.removeFilesWithStagerRM(self.streamerPool, lfnToIdDict,
                                                                        self.deleteStreamers, self.commandLogDir) )
            lfnToIdDict = {}

        # updating status for all streamers deleted in this cycle
        InsertCleanup.updateDeletedStreamers(self.t0astDBConn, deletedStreamerIDs)

        # putting new deletable WMBS files into delete fileset
        InsertCleanup.updateDeletableFiles(self.t0astDBConn)

        # getting all deletable WMBS files
        wmbsFiles = ListCleanup.getDeletableFiles(self.t0astDBConn)

        for lfn in wmbsFiles['TEMP'].keys():
            logging.info("deleting temp file %s" % lfn)
        for lfn in wmbsFiles['MERGED'].keys():
            logging.info("deleting merged file %s" % lfn)

        # split into self.maxBulkDelete big chunks and bulk stager_rm
        deletedFileIDs = []
        lfnToIdDict = {}
        for lfn in wmbsFiles['TEMP'].keys():
            lfnToIdDict[lfn] = wmbsFiles['TEMP'][lfn]
            del wmbsFiles['TEMP'][lfn]
            if len(lfnToIdDict) >= self.maxBulkDelete:
                deletedFileIDs.extend( self.removeFilesWithStagerRM(self.tempPool, lfnToIdDict,
                                                                    self.deleteTempFiles,
                                                                    self.commandLogDir) )
                lfnToIdDict = {}
        if len(lfnToIdDict) > 0:
            deletedFileIDs.extend( self.removeFilesWithStagerRM(self.tempPool, lfnToIdDict,
                                                                self.deleteTempFiles,
                                                                self.commandLogDir) )
            lfnToIdDict = {}

        lfnToIdDict = {}
        for lfn in wmbsFiles['MERGED'].keys():
            lfnToIdDict[lfn] = wmbsFiles['MERGED'][lfn]
            del wmbsFiles['MERGED'][lfn]
            if len(lfnToIdDict) >= self.maxBulkDelete:
                deletedFileIDs.extend( self.removeFilesWithStagerRM(self.mergedPool, lfnToIdDict,
                                                                    self.deleteMergedFiles,
                                                                    self.commandLogDir) )
                lfnToIdDict = {}
        if len(lfnToIdDict) > 0:
            deletedFileIDs.extend( self.removeFilesWithStagerRM(self.mergedPool, lfnToIdDict,
                                                                self.deleteMergedFiles,
                                                                self.commandLogDir) )
            lfnToIdDict = {}


        # marking all WMBS files deleted in this cycle as deletion complete
        InsertCleanup.updateDeletedFiles(self.t0astDBConn, deletedFileIDs)

        self.ms.publishUnique("CleanUpManager:Poll", "", self.pollInterval)
        self.ms.commit()
        return


    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("CleanUpManager:StartDebug")
        self.ms.subscribeTo("CleanUpManager:EndDebug")
        self.ms.subscribeTo("CleanUpManager:Poll")
        
        self.ms.publishUnique("CleanUpManager:Poll", "", self.pollInterval)
        self.ms.commit()

        self.connectT0AST()

        # wait for messages
        self.messageLoop()
