#!/usr/bin/env python
"""
_ListFilesOverInterval_

Retrieve meta data about files belonging to a particular dataset and produced
over a certain interval.
"""

__revision__ = "$Id: ListFilesOverInterval.py,v 1.1 2010/03/17 15:22:45 sryu Exp $"
__version__ = "$Revision: 1.1 $"

import time

from WMCore.Database.DBFormatter import DBFormatter

class ListFilesOverInterval(DBFormatter):
    sql = """SELECT wmbs_file_details.lfn AS lfn,
                    wmbs_file_details.filesize AS file_size,
                    wmbs_file_details.events AS events,
                    wmbs_file_runlumi_map.run AS run,
                    wmbs_file_runlumi_map.lumi as lumi
                    FROM wmbs_file_details
               INNER JOIN wmbs_file_runlumi_map ON
                 wmbs_file_details.id = wmbs_file_runlumi_map.fileid
               INNER JOIN wmbs_fileset_files ON
                 wmbs_file_details.id = wmbs_fileset_files.fileid
               INNER JOIN wmbs_fileset ON
                 wmbs_fileset_files.fileset = wmbs_fileset.id
               INNER JOIN wmbs_file_dataset_path_assoc ON
                 wmbs_file_details.id = wmbs_file_dataset_path_assoc.file_id
               INNER JOIN dataset_path ON
                 wmbs_file_dataset_path_assoc.dataset_path_id = dataset_path.id
             WHERE wmbs_fileset.name = :fileset_name AND
                   wmbs_fileset_files.insert_time >= :start_time AND
                   wmbs_fileset_files.insert_time <= :end_time AND
                   dataset_path.primary_dataset =
                     (SELECT id FROM primary_dataset WHERE name = :prids) AND
                   dataset_path.processed_dataset =
                     (SELECT id FROM processed_dataset WHERE name = :procds) AND
                   dataset_path.data_tier =
                     (SELECT id FROM data_tier WHERE name = :datatier)"""

    # Unmerged express files use a different dataset mapping scheme because Dirk didn't
    # like the one we use for bulk files.  This means we need a slightly different query
    # to access express unmerged files.
    unSql = """SELECT wmbs_file_details.lfn AS lfn,
                      wmbs_file_details.filesize AS file_size,
                      wmbs_file_details.events AS events,
                      wmbs_file_runlumi_map.run AS run,
                      wmbs_file_runlumi_map.lumi as lumi
                      FROM wmbs_file_details
                 INNER JOIN wmbs_file_runlumi_map ON
                   wmbs_file_details.id = wmbs_file_runlumi_map.fileid
                 INNER JOIN wmbs_fileset_files ON
                   wmbs_file_details.id = wmbs_fileset_files.fileid
                 INNER JOIN wmbs_fileset ON
                   wmbs_fileset_files.fileset = wmbs_fileset.id
                 INNER JOIN express_file_info ON
                   wmbs_file_details.id = express_file_info.file_id
               WHERE wmbs_fileset.name = :fileset_name AND
                     wmbs_fileset_files.insert_time >= :start_time AND
                     wmbs_fileset_files.insert_time <= :end_time AND
                     express_file_info.primary_dataset =
                       (SELECT id FROM primary_dataset WHERE name = :prids) AND
                     express_file_info.processed_dataset =
                       (SELECT id FROM processed_dataset WHERE name = :procds) AND
                     express_file_info.data_tier =
                       (SELECT id FROM data_tier WHERE name = :datatier)"""    

    def format(self, results):
        """
        _format_

        Format the result of the query into a dictionary for each file and
        return a list of dictionaries.  We get a row back for each run/lumi
        combination. 
        """
        results = self.formatDict(results)

        resultsDict = {}
        for result in results:
            if resultsDict.has_key(result["lfn"]):
                runDict = resultsDict[result["lfn"]]["runs"]
                if runDict.has_key(str(result["run"])):
                    runDict[str(result["run"])].append(int(result["lumi"]))
                else:
                    runDict[str(result["run"])] = [int(result["lumi"])]
            else:
                run = {str(result["run"]): [int(result["lumi"])]}
                resultsDict[result["lfn"]] = {"lfn": result["lfn"],
                                              "file_size": int(result["file_size"]),
                                              "events": int(result["events"]),
                                              "runs": run}
        resultsList = []        
        for lfn in resultsDict.keys():
            resultsList.append(resultsDict[lfn])

        return resultsList
    
    def execute(self, fileType, startTime, primaryDataset, processedDataset, dataTier,
                conn = None, transaction = False):
        """
        _execute_

        Figure out the end time and run the query.  We take a second off the
        end time to assure ourselves that we don't miss any files that may get
        added later this second.

        Create a dictionary that contains the files, start_time, end_time and
        dataset that was queried.
        """
        if fileType == "express":
            filesetName = "ExpressDBSUploadable"
            query = self.sql
        elif fileType == "expressunmerged":
            filesetName = "ExpressMergeable"
            query = self.unSql
        else:
            filesetName = "DBSUploadable"
            query = self.sql
            
        endTime = int(time.time()) - 1
        bindVars = {"start_time": startTime, "end_time": endTime,
                    "prids": primaryDataset, "procds": processedDataset,
                    "datatier": dataTier, "fileset_name": filesetName}

        result = self.dbi.processData(query, bindVars, conn = conn,
                                      transaction = transaction)

        resultDict = {}
        resultDict["files"] = self.format(result)
        resultDict["dataset_path"] = "/%s/%s/%s" % (primaryDataset,
                                                    processedDataset,
                                                    dataTier)
        resultDict["start_time"] = startTime
        resultDict["end_time"] = endTime
        
        return resultDict
