#!/usr/bin/env python
"""
_JobsByStatus_

T0Mon query: base class for all the other JobsByStatus calls
"""

from T0.DAS.Database.Oracle.LongResultQuery import LongResultQuery
from WMCore.Database.DBFormatter import DBFormatter

class JobsByStatus(LongResultQuery):
    """
    retrieve job stats
    
    """
##     subscription = None
    fileset = "YouNeedToOverrideThis"
    datatier = None

    sql = """SELECT wmbs_job.id AS job_id,
                    wmbs_job.state_time AS state_time,
                    wmbs_file_runlumi_map.run AS run_id,
                    primary_dataset.name AS dataset,
                    COUNT(*) AS inputfiles
             FROM wmbs_job
             INNER JOIN wmbs_jobgroup ON
               wmbs_jobgroup.id = wmbs_job.jobgroup AND
               wmbs_jobgroup.subscription = (SELECT id FROM wmbs_subscription
                                             WHERE fileset = (SELECT id FROM wmbs_fileset
                                                              WHERE name = :FILESET))
             INNER JOIN wmbs_job_assoc ON
               wmbs_job_assoc.job = wmbs_job.id
             INNER JOIN wmbs_file_dataset_path_assoc ON
               wmbs_file_dataset_path_assoc.file_id = wmbs_job_assoc.fileid
             INNER JOIN dataset_path ON
               dataset_path.id = wmbs_file_dataset_path_assoc.dataset_path_id
             INNER JOIN primary_dataset ON
               primary_dataset.id = dataset_path.primary_dataset
             INNER JOIN wmbs_file_runlumi_map ON
               wmbs_file_runlumi_map.fileid = wmbs_job_assoc.fileid %s
             %s
             %s
             GROUP BY wmbs_job.id,
                      wmbs_job.state_time,
                      wmbs_file_runlumi_map.run,
                      primary_dataset.name
             ORDER BY wmbs_job.id DESC
             """

    sqlRunCheck = """AND wmbs_file_runlumi_map.run = :RUN"""

    def execute(self, run, status, uplimit, downlimit, conn = None, transaction = False):
        """
        create correct sql query and execute it
         
        """
        try:
            run = int(run)
        except ValueError:
            run = 0

        allowedStates = [ "acquired", "complete", "failed" ]
        try:
            status = str(status)
            status = status.lower()
            if status not in allowedStates:
                raise ValueError
        except ValueError:
            return { "Error" : "wrong status name", "hint" : allowedStates }
        
        bindVars = { 'uplimit' : uplimit,
                     'downlimit' : downlimit,
                     'FILESET' : self.fileset }

        if run > 0:
            sqlRunCheck = self.sqlRunCheck
            bindVars.update( { 'RUN' : run } )
        else:
            sqlRunCheck = ""

        if self.datatier != None:
            sqlDatatierCheck = """INNER JOIN data_tier ON
                                    data_tier.id = dataset_path.data_tier AND
                                    data_tier.name = :DATATIER
                                  """
            bindVars.update( { 'DATATIER' : self.datatier } )
        else:
            sqlDatatierCheck = ""

        if status == allowedStates[0]:
            sqlStatusCheck = """WHERE wmbs_job.state != (SELECT id FROM wmbs_job_state WHERE name = 'cleanout')"""
        elif status == allowedStates[1]:
            sqlStatusCheck = """WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = 'cleanout')
                                AND wmbs_job.outcome = 1"""
        elif status == allowedStates[2]:
            sqlStatusCheck = """WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = 'cleanout')
                                AND wmbs_job.outcome = 0"""

        sql = self.sql % (sqlRunCheck,
                          sqlDatatierCheck,
                          sqlStatusCheck)

        result = self.dbi.processData(sql, bindVars, conn = conn,
                                      transaction = transaction)
        
        return self.formatDict(result)


class JobsByStatus_count(DBFormatter):
    """
    retrieve job count
    
    """
##     subscription = None
    fileset = "YouNeedToOverrideThis"
    datatier = None

    sqlRunCheck = """INNER JOIN wmbs_file_runlumi_map ON
                       wmbs_file_runlumi_map.fileid = wmbs_job_assoc.fileid AND
                       wmbs_file_runlumi_map.run = :RUN
                     """

    def execute(self, run, status, conn = None, transaction = False):
        """
        create correct sql query and execute it
         
        """
        try:
            run = int(run)
        except ValueError:
            run = 0

        allowedStates = [ "acquired", "complete", "failed" ]
        try:
            status = str(status)
            status = status.lower()
            if status not in allowedStates:
                raise ValueError
        except ValueError:
            return { "Error" : "wrong status name", "hint" : allowedStates }
        
        bindVars = { 'FILESET' : self.fileset }

        if run > 0 or self.datatier != None:
            sqlAssocJoin = """INNER JOIN wmbs_job_assoc ON
                                wmbs_job_assoc.job = wmbs_job.id
                              """
        else:
            sqlAssocJoin = ""

        if run > 0:
            sqlRunCheck = self.sqlRunCheck
            bindVars.update( { 'RUN' : run } )
        else:
            sqlRunCheck = ""

        if self.datatier != None:
            sqlDatatierCheck = """INNER JOIN wmbs_file_dataset_path_assoc ON
                                    wmbs_file_dataset_path_assoc.file_id = wmbs_job_assoc.fileid
                                  INNER JOIN dataset_path ON
                                    dataset_path.id = wmbs_file_dataset_path_assoc.dataset_path_id
                                  INNER JOIN data_tier ON
                                    data_tier.id = dataset_path.data_tier AND
                                    data_tier.name = :DATATIER
                                  """
            bindVars.update( { 'DATATIER' : self.datatier } )
        else:
            sqlDatatierCheck = ""

        if status == allowedStates[0]:
            sqlStatusCheck = """WHERE wmbs_job.state != (SELECT id FROM wmbs_job_state WHERE name = 'cleanout')"""
        elif status == allowedStates[1]:
            sqlStatusCheck = """WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = 'cleanout')
                                AND wmbs_job.outcome = 1"""
        elif status == allowedStates[2]:
            sqlStatusCheck = """WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = 'cleanout')
                                AND wmbs_job.outcome = 0"""

        sql = """SELECT COUNT(DISTINCT wmbs_job.id) 
                 FROM wmbs_job
                 INNER JOIN wmbs_jobgroup ON
                   wmbs_jobgroup.id = wmbs_job.jobgroup AND
                   wmbs_jobgroup.subscription = (SELECT id FROM wmbs_subscription
                                                 WHERE fileset = (SELECT id FROM wmbs_fileset
                                                                  WHERE name = :FILESET))
                 %s
                 %s
                 %s
                 %s
                 """ % (sqlAssocJoin,
                        sqlRunCheck,
                        sqlDatatierCheck,
                        sqlStatusCheck)

        result = self.dbi.processData(sql, bindVars, conn = conn,
                                      transaction = transaction)
        
        return self.format(result)

