#!/usr/bin/env python
"""
_StatsJob_

get stats for jobs
if the run parameter is not provided or 0, whole system stats are returned

T0Mon: base class for other classes
"""

from WMCore.Database.DBFormatter import DBFormatter

class StatsJob(DBFormatter):

    fileset = "YouNeedToOverrideThis"
    datatier = None

    sqlRunCheck = """INNER JOIN wmbs_file_runlumi_map ON
                       wmbs_file_runlumi_map.fileid = wmbs_job_assoc.fileid AND
                       wmbs_file_runlumi_map.run = :RUN
                     """

    def execute(self, run, conn = None, transaction = False):
        """
        create correct sql query and execute it

        """
        try:
            run = int(run)
        except ValueError:
            run = 0

        bindVars = { 'FILESET' : self.fileset }

        if run > 0 or self.datatier != None:
            sqlAssocJoin = """INNER JOIN wmbs_job_assoc ON
                                wmbs_job_assoc.job = wmbs_job.id
                              """
            sqlGroupBy = """GROUP BY wmbs_job.id,
                              CASE
                                WHEN wmbs_job_state.name = 'cleanout' AND wmbs_job.outcome = 1 THEN 'complete'
                                WHEN wmbs_job_state.name = 'cleanout' AND wmbs_job.outcome = 0 THEN 'failed'
                                ELSE 'acquired'
                              END
                            """
        else:
            sqlAssocJoin = ""
            sqlGroupBy = ""

        if run > 0:
            sqlRunCheck = self.sqlRunCheck
            bindVars.update( { 'RUN' : run } )
        else:
            sqlRunCheck = ""

        if self.datatier != None:
            sqlDatatierCheck = """INNER JOIN wmbs_file_dataset_path_assoc ON
                                    wmbs_file_dataset_path_assoc.file_id = wmbs_job_assoc.fileid
                                  INNER JOIN dataset_path ON
                                    dataset_path.id = wmbs_file_dataset_path_assoc.dataset_path_id
                                  INNER JOIN data_tier ON
                                    data_tier.id = dataset_path.data_tier AND
                                    data_tier.name = :DATATIER
                                  """
            bindVars.update( { 'DATATIER' : self.datatier } )
        else:
            sqlDatatierCheck = ""

        sql = """SELECT job_state_map.status AS status,
                        COUNT(*) AS count
                 FROM (       
                   SELECT wmbs_job.id AS id,
                          CASE
                            WHEN wmbs_job_state.name = 'cleanout' AND wmbs_job.outcome = 1 THEN 'complete'
                            WHEN wmbs_job_state.name = 'cleanout' AND wmbs_job.outcome = 0 THEN 'failed'
                            ELSE 'acquired'
                          END AS status
                   FROM wmbs_job
                   INNER JOIN wmbs_jobgroup ON
                     wmbs_jobgroup.id = wmbs_job.jobgroup AND
                     wmbs_jobgroup.subscription = (SELECT id FROM wmbs_subscription
                                                   WHERE fileset = (SELECT id FROM wmbs_fileset
                                                                    WHERE name = :FILESET))
                   INNER JOIN wmbs_job_state ON
                     wmbs_job_state.id = wmbs_job.state
                   %s
                   %s
                   %s
                   %s
                 ) job_state_map
                 GROUP BY job_state_map.status
                 """ % (sqlAssocJoin,
                        sqlRunCheck,
                        sqlDatatierCheck,
                        sqlGroupBy)

        result = self.dbi.processData(sql, bindVars, conn = conn,
                                      transaction = transaction)

        resultDicts = self.formatDict(result)

        haveComplete = False
        haveFailed = False
        haveAcquired = False
        for resultDict in resultDicts:
            status = resultDict['status']
            if status == "complete":
                haveComplete = True
            elif status == "failed":
                haveFailed = True
            elif status == "acquired":
                haveAcquired = True

        if not haveComplete:
            resultDicts.append( { 'status' : 'complete',
                                  'count' : 0 } )
        if not haveFailed:
            resultDicts.append( { 'status' : 'failed',
                                  'count' : 0 } )
        if not haveAcquired:
            resultDicts.append( { 'status' : 'acquired',
                                  'count' : 0 } )

        return resultDicts
