#!/usr/bin/env python
"""
_DQMSchedulerComponent_

Poll T0AST for RECO files in the DQM subscription.  
"""

__revision__ = "$Id: DQMSchedulerComponent.py,v 1.41 2010/10/07 15:39:56 hufnagel Exp $"
__version__ = "$Revision: 1.41 $"

import logging
import threading
import sys
import os

from T0.State.Database.Reader import ListFiles
from T0.State.Database.Reader import ListRuns

from T0 import Globals

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow

from WMCore.DAOFactory import DAOFactory

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0.WorkflowFactory.HarvestWorkflow import HarvestWorkflow


class DQMSchedulerComponent(GenericTier0Component):
    """
    _DQMSchedulerComponent_

    Poll T0AST for RECO files in the DQM subscription.  Add them to the CAFWMBS
    and create workflows for DQM as needed.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("DQMScheduler Component Started")

        self.pollInterval = self.lookupParameter("PollInterval", "00:05:00")

        # Scram arch and path to cmssw needed to generate workflows.
        self.scramArch = self.lookupParameter("ScramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("CmsPath", "/uscmst1/prod/sw/cms/")

        # The job cache directory is where all workflows, jobspecs and configs
        # are stored.
        self.jobCacheDir = self.lookupParameter("JobCacheDir", self.args["ComponentDir"])

        # Parameters used to create the DQM workflows.
        self.site = self.lookupParameter("Site", "srm-cms.cern.ch")
        self.DQMCopyToCERN = self.lookupParameter("DQMCopyToCERN", False)
        self.DQMServer = self.lookupParameter("DQMServer", "https://cmsweb.cern.ch/dqm/tier-0")
        self.proxyLocation = self.lookupParameter("proxyLocation", "/afs/cern.ch/user/c/cmsprod/.globus/x509up_u7178")

        # DQM Job period
        self.jobPeriod = int(self.lookupParameter("JobPeriod", 3600))

        # configure dropbox and output directory for conditions
        self.sqliteTargetDir = self.lookupParameter("SqliteTargetDir", None)
        self.dropBoxHost = self.lookupParameter("DropBoxHost", None)
        self.dropBoxValidation = self.lookupParameter("DropBoxValidation", "True")

        # DAO factory that is pointed at T0AST
        self.daoFactory = None

        # Dictionary used to keep track of what fileset are used
        self.filesets = {}

        # IDs of the DQM and MergePack subscription
        self.dqmSubscription = None
        self.alcaHarvestSubscription = None
        self.mergePackSubscription = None

        return

    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "DQMScheduler:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "DQMScheduler:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "DQMScheduler:poll":
            filesets = self.pollForClosableFilesets()
            self.pollForNewFiles("DQM")
            self.pollForNewFiles("ALCAPROMPT")
            self.closeFilesets(filesets)
        else:
            logging.debug("Unknown event: %s" % event)
            
        return

    def pollForNewFiles(self, harvestType):
        """
        _pollForNewFiles_

        poll for new files to be harvested
        add them to the harvest filesets, creating these filesets and
        the corresponding workflows and subscriptions as needed

        FIXME: avoid workflow creation inside the transaction !!!

        """
        myThread = threading.currentThread()

        if harvestType == "DQM":
            subscription = self.dqmSubscription
        elif harvestType == "ALCAPROMPT":
            subscription = self.alcaHarvestSubscription
        
        newFiles = ListFiles.getIncompleteWMBSFiles(self.t0astDBConn, subscription['id'])

        filesetFiles = {}

        for file in newFiles:
            runNumber = file['run']
            runConfig = self.getRunConfig(runNumber)
            primaryDatasetID = file['primds_id']
            primaryDataset = runConfig.getPrimaryDatasetName(primaryDatasetID)
            processedDataset = file['procds']
            dataTier = file['tier']
            streamID = runConfig.getStreamIDfromDatasetID(primaryDatasetID)
            processingStyle = runConfig.getProcessingStyle(streamID)

            # get harvest CMSSW version and GlobalTag
            if processingStyle == "Express":
                procConfig = runConfig.getExpressConfiguration(streamID)
                cmsswVersion = procConfig.CMSSWVersion
                globalTag = procConfig.GlobalTag
            else:
                procConfig = runConfig.getRecoConfigurationByPrimaryID(primaryDatasetID)
                cmsswVersion = procConfig.CMSSWVersion
                globalTag = procConfig.GlobalTag

            # different setup if we also do periodic harvesting (only for express DQM)
            if processingStyle == "Express" and harvestType == "DQM":

                periodicFilesetName = "T0-Run%s-%s-%s-%s-Periodic" % (runNumber, primaryDataset,
                                                                      processedDataset, dataTier)
                endOfRunFilesetName = "EORT0-Run%s-%s-%s-%s-EndOfRun" % (runNumber, primaryDataset,
                                                                         processedDataset, dataTier)

                if not self.filesets.has_key(periodicFilesetName):

                    myThread.transaction.begin()

                    inputFileset = Fileset(name = periodicFilesetName, is_open = True)
                    inputFileset.create()

                    outputFileset = Fileset(name = endOfRunFilesetName, is_open = True)
                    outputFileset.create()

                    myThread.transaction.commit()

                    self.filesets[periodicFilesetName] = [ inputFileset.id, outputFileset.id ]

                    periodicWorkflowName = "DQMHarvest-Periodic-Run%s-%s-%s" % (runNumber, primaryDataset, dataTier)
                    endOfRunWorkflowName = "DQMHarvest-EndOfRun-Run%s-%s-%s" % (runNumber, primaryDataset, dataTier)

                    self.createWorkflow(runNumber, periodicWorkflowName, primaryDataset, processedDataset, dataTier,
                                        cmsswVersion, globalTag, False, inputFileset, outputFileset, harvestType)

                    self.createWorkflow(runNumber, endOfRunWorkflowName, primaryDataset, processedDataset, dataTier,
                                        cmsswVersion, globalTag, True, outputFileset, None, harvestType)

                for fileset in self.filesets[periodicFilesetName]:
                    if not filesetFiles.has_key(fileset):
                        filesetFiles[fileset] = []
                    filesetFiles[fileset].append(file['id'])

            # simpler for just endOfRun
            else:

                endOfRunFilesetName = "T0-Run%s-%s-%s-%s-EndOfRun" % (runNumber, primaryDataset,
                                                                      processedDataset, dataTier)

                if not self.filesets.has_key(endOfRunFilesetName):

                    myThread.transaction.begin()
                
                    inputFileset = Fileset(name = endOfRunFilesetName, is_open = True)
                    inputFileset.create()

                    myThread.transaction.commit()

                    self.filesets[endOfRunFilesetName] = [ inputFileset.id ]

                    endOfRunWorkflowName = "DQMHarvest-EndOfRun-Run%s-%s-%s" % (runNumber, primaryDataset, dataTier)

                    self.createWorkflow(runNumber, endOfRunWorkflowName, primaryDataset, processedDataset, dataTier,
                                        cmsswVersion, globalTag, True, inputFileset, None, harvestType)

                for fileset in self.filesets[endOfRunFilesetName]:
                    if not filesetFiles.has_key(fileset):
                        filesetFiles[fileset] = []
                    filesetFiles[fileset].append(file['id'])

        # prepare for transaction
        addFilesToFileset = self.daoFactory(classname = "Fileset.BulkAdd")
        filesComplete = self.daoFactory(classname = "Subscriptions.CompleteFiles")

        fileList = set([])
        filesetFileList = []
        for fileset, fileids in filesetFiles.items():
            for fileid in fileids:
                fileList.add(fileid)
                filesetFileList.append ( { 'fileset' : fileset,
                                           'fileid' : fileid } )
        fileList = list(fileList)

        if len(fileList) > 0:

            # actual transaction
            myThread.transaction.begin()

            addFilesToFileset.execute(binds = filesetFileList,
                                      conn = myThread.transaction.conn,
                                      transaction = True)

            for fileid in fileList:
                filesComplete.execute(subscription = subscription['id'],
                                      file = fileid,
                                      conn = myThread.transaction.conn,
                                      transaction = True)

            myThread.transaction.commit()

        return

    def createWorkflow(self, runNumber, workflowName, primaryDataset, processedDataset,
                       dataTier, cmsswVersion, globalTag, endOfRun,
                       inputFileset, outputFileset, harvestType):
        """
        _createWorkflowForDataset_

        Create a DQM workflow for the given run and dataset.
        Create a WMBS subscription, workflow and input dataset as well.

        """
        myThread = threading.currentThread()

        runConfig = self.getRunConfig(runNumber)
        primaryDatasetID = runConfig.getPrimaryDatasetID(primaryDataset)

        workflowDir = os.path.join(self.jobCacheDir, "workflows", "Run%s" % runNumber)
        workflowSpecFile = os.path.join(workflowDir, "%s-workflow.xml" % workflowName)

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No Workflow for run, dataset: %s, %s" % (runNumber, primaryDataset))
                return

            myThread.transaction.begin()

            wmbsWorkflow = Workflow(spec = workflowSpecFile,
                                    name = workflowName,
                                    owner = "Tier0", task = "BOGUS")
            wmbsWorkflow.create()

            if outputFileset != None:
                wmbsWorkflow.addOutput("Anything", outputFileset)

            if endOfRun:
                splitAlgo = "EndOfRun"
            else:
                splitAlgo = "Periodic"

            wmbsSubscription = Subscription(fileset = inputFileset,
                                            workflow = wmbsWorkflow,
                                            split_algo = splitAlgo,
                                            type = "Harvesting")
            wmbsSubscription.create()

            myThread.transaction.commit()

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return

        workflow = HarvestWorkflow(runNumber,
                                   cmsswVersion,
                                   self.cmsPath,
                                   self.scramArch)

        workflowSpec = workflow.makeWorkflowSpec(workflowName,
                                                 scenario = runConfig.getScenario(primaryDatasetID),
                                                 harvestType = harvestType,
                                                 globalTag = globalTag,
                                                 inputDataset = "/%s/%s/%s" % (primaryDataset, processedDataset, dataTier),
                                                 endOfRun = endOfRun,
                                                 enableLazyDownload = True,
                                                 production = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)

        workflowSpec.parameters["WorkflowClass"] = "%sHarvest" % harvestType
        workflowSpec.parameters["WorkflowDescription"] = "%s-%s" % (primaryDataset, dataTier)
        workflowSpec.parameters['OnlySites'] = self.site

        if harvestType == "DQM":
            workflowSpec.parameters["job_period"] = self.jobPeriod
            workflowSpec.parameters['DQMServer'] = self.DQMServer
            workflowSpec.parameters['proxyLocation'] = self.proxyLocation
            workflowSpec.parameters['DQMCopyToCERN'] = self.DQMCopyToCERN

        if harvestType == "ALCAPROMPT":
            workflowSpec.parameters['SqliteTargetDir'] = self.sqliteTargetDir
            workflowSpec.parameters['DropBoxHost'] = self.dropBoxHost
            workflowSpec.parameters['DropBoxValidation'] = self.dropBoxValidation

        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)

        workflowSpec.save(workflowSpecFile)

        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)

    def pollForClosableFilesets(self):
        """
        _pollForClosableFilesets_

        find open filesets matching the naming conventions (name starts with T0)

        for any runs with open filesets check if bulk or express processing is done
        (depending on the filesets check either bulk or express or both)

        return a list of filesets that can be closed

        """
        myThread = threading.currentThread()
        myThread.transaction.begin()

        openFilesetDAO = self.daoFactory(classname = "Fileset.ListOpenByName")
        openFilesetNames = openFilesetDAO.execute(name = "T0%")

        openExpressFilesets = {}
        openBulkFilesets = {}
        closableFilesets = []

        for openFilesetName in openFilesetNames:
            (type, runElement, primaryDataset, junk) = openFilesetName.split("-", 3)
            if type != "T0":
                continue

            run = int(runElement[3:])

            runConfig = self.getRunConfig(run)
            primaryDatasetID = runConfig.getPrimaryDatasetID(primaryDataset)
            streamID = runConfig.getStreamIDfromDatasetID(primaryDatasetID)

            if runConfig.getProcessingStyle(streamID) == "Express":
                if openExpressFilesets.has_key(run):
                    openExpressFilesets[run].append(openFilesetName)
                else:
                    openExpressFilesets[run] = [openFilesetName]
            else:
                if openBulkFilesets.has_key(run):
                    openBulkFilesets[run].append(openFilesetName)
                else:
                    openBulkFilesets[run] = [openFilesetName]

        subComplete = self.daoFactory(classname = "Subscriptions.IsCompleteOnRun")

        for run in openExpressFilesets.keys():
            if ListRuns.listExpressProcCompleteForRun(self.t0astDBConn, run):
                pendingFiles = subComplete.execute(self.mergePackSubscription["id"], run)
                if pendingFiles == 0:
                    closableFilesets.extend(openExpressFilesets[run])

        for run in openBulkFilesets.keys():
            runStatus = ListRuns.listRunState(self.t0astDBConn, run)
            if runStatus[run] in ["CloseOutExport", "CloseOutDQMHarvest",
                                  "CloseOutT1Skimming", "Complete"]:
                closableFilesets.extend(openBulkFilesets[run])

        myThread.transaction.commit()

        return closableFilesets

    def closeFilesets(self, filesets):
        """
        _closeFilesets_

        close all filesets that are specified
        remove them from the fileset cache

        """
        if len(filesets) > 0:

            myThread = threading.currentThread()
            myThread.transaction.begin()

            for fileset in filesets:
                logging.debug("Closing fileset: %s" % fileset)
                wmbsFileset = Fileset(name = fileset)
                wmbsFileset.markOpen(False)
                if fileset in self.filesets.keys():
                    del self.filesets[fileset]

            myThread.transaction.commit()

        self.ms.publishUnique("DQMScheduler:poll", "", self.pollInterval)
        self.ms.commit()
        return
    
    def startComponent(self):
        """
        _startComponent_

        Fire up the message service and define the list of
        messages that this component subscribes to
        """
        self.ms.subscribeTo("DQMScheduler:StartDebug")
        self.ms.subscribeTo("DQMScheduler:EndDebug")
        self.ms.subscribeTo("DQMScheduler:poll")
        self.ms.commit()

        self.connectT0AST(manageGlobal = True)
        self.t0astDBConn.beginTransaction()

        self.daoFactory = DAOFactory(package = "WMCore.WMBS", logger = logging,
                                     dbinterface = self.t0astDBConn.dbInterface)

        self.dqmSubscription = Globals.retrieveDQMSubscription()
        self.alcaHarvestSubscription = Globals.retrieveAlcaHarvestSubscription()
        self.mergePackSubscription = Globals.retrieveExpressMergeSubscription()

        self.t0astDBConn.commit()

        self.ms.publishUnique("DQMScheduler:poll", "", "00:01:00")
        self.ms.commit()

        # Wait for messages.
        self.messageLoop()
