#!/usr/bin/env python
"""
_DropMaker_

Generate the XML file for injecting data into PhEDEx
Modified from ProdCommon.DataMgmt.PhEDEx.DropMaker.py

"""

import logging
from IMProv.IMProvNode import IMProvNode
from IMProv.IMProvDoc import IMProvDoc

class XMLFileblock(list):
    """
    _XMLFileblock_

    Object representing a fileblock for conversion to XML

    """
    def __init__(self, fileblockName, isOpen = "y"):
        list.__init__(self)
        self.fileblockName = fileblockName
        self.isOpen = isOpen

    def addFile(self, lfn, checksums, size):
        """
        _addFile_

        Add a file to this fileblock

        """
        self.append(
            (lfn, checksums, size)
            )
        return

    def save(self):
        """
        _save_

        Serialise this to XML compatible with PhEDEx injection

        """
        result = IMProvNode("block")
        result.attrs['name'] = self.fileblockName
        result.attrs['is-open'] = self.isOpen
        for lfn, checksums, size in self:
            # checksums is a comma separated list of key:value pair
            formattedChecksums = ",".join(["%s:%s" % (x.lower(), y) for x, y \
                                           in checksums.items() \
                                           if y not in (None, '')])
            file = IMProvNode("file")
            file.attrs['name'] = lfn
            file.attrs['checksum'] = formattedChecksums
            file.attrs['bytes'] = size
            result.addNode(file)
        return result
    
class XMLDataset(list):
    """
    <dataset name='DatasetNameHere' is-open='boolean' is-transient='boolean'>
    <block name='fileblockname' is-open='boolean'>
    <file lfn='lfn1Here' checksum='cksum:0' size ='fileSize1Here'/>
    <file lfn='lfn2Here' checksum='cksum:0' size ='fileSize2Here'/> 
    </block>
    </dataset>
    """
    
    def __init__(self, datasetName, datasetOpen = "y", 
                 datasetTransient = "n" ):
        
        self.datasetName = datasetName    
        self.datasetIsOpen = datasetOpen
        self.datasetIsTransient = datasetTransient

        #  //
        # // Fileblocks
        #//
        self.fileblocks = {}


    def getFileblock(self, fileblockName, isOpen = "y"):
        """
        _getFileblock_

        Add a new fileblock with name provided if not present, if it exists,
        return it

        """
        if self.fileblocks.has_key(fileblockName):
            return self.fileblocks[fileblockName]
        
        newFileblock = XMLFileblock(fileblockName, isOpen)
        self.fileblocks[fileblockName] = newFileblock
        return newFileblock

    def save(self):
        """
        _save_

        serialise object into PhEDEx injection XML format

        """
        
        dataset = IMProvNode("dataset")
        dataset.attrs['name'] = self.datasetName
        dataset.attrs['is-open'] = self.datasetIsOpen
        dataset.attrs['is-transient'] = self.datasetIsTransient
        
        for block in self.fileblocks.values():
            dataset.addNode(block.save())
        
        return dataset

class XMLInjectionSpec:
    """
    _XMLInjectionSpec_
    <data version='2'>
    <dbs name='DBSNameHere'>

    <dataset name='DatasetNameHere' is-open='boolean' is-transient='boolean'>
    <block name='fileblockname' is-open='boolean'>
    <file lfn='lfn1Here' checksum='cksum:0' size ='fileSize1Here'/>
    <file lfn='lfn2Here' checksum='cksum:0' size ='fileSize2Here'/> 
    </block>
    </dataset>
    <dataset name='DatasetNameHere' is-open='boolean' is-transient='boolean'>
    <block name='fileblockname' is-open='boolean'>
    <file lfn='lfn1Here' checksum='cksum:0' size ='fileSize1Here'/>
    <file lfn='lfn2Here' checksum='cksum:0' size ='fileSize2Here'/> </block>
    </dataset>
    
    </dbs>
    </data> 
    """
    def __init__(self, dbs):
        self.dbs = dbs
        #  //
        # // dataset attributes
        #//

        #  //
        # // Fileblocks
        #//
        self.datasetPaths = {}


    def getDataset(self, datasetName, isOpen = "y", isTransient = "n" ):
        """
        _getFileblock_

        Add a new fileblock with name provided if not present, if it exists,
        return it

        """
        if self.datasetPaths.has_key(datasetName):
            return self.datasetPaths[datasetName]
        
        newDataset = XMLDataset(datasetName, isOpen, isTransient)
        self.datasetPaths[datasetName] = newDataset
        return newDataset

    def save(self):
        """
        _save_

        serialise object into PhEDEx injection XML format

        """
        result = IMProvNode("data")
        # hard coded as version 2. might need to change
        result.attrs['version'] = '2'
        dbs = IMProvNode("dbs")
        dbs.attrs['name'] = self.dbs
        dbs.attrs['dls'] = 'dbs'
        result.addNode(dbs)
        
        for dataset in self.datasetPaths.values():
            dbs.addNode(dataset.save())
                
        return result

    def write(self, filename):
        """
        _write_

        Write to file using name provided

        """
        handle = open(filename, 'w')
        improv = self.save()
        handle.write(improv.makeDOMElement().toprettyxml())
        handle.close()
        return
        
def makePhEDExDataXML(datasetPathSet, globalDbsUrl):
    """
    _makePhEDExDropForTier0_

    Given a datasetPathSet structure, generate an XML structure for injection.
    DatasetPathSet is dict format of 
        {datasetPath: {blockName:{'fileList': [file1, file2, file], 
                'status':'Closed' },  
        .....}
    """
            
    spec = XMLInjectionSpec(globalDbsUrl)

    for datasetPath in datasetPathSet.keys():
        xmlDataset = spec.getDataset(datasetPath)
        for blockName in datasetPathSet[datasetPath].keys():
            if datasetPathSet[datasetPath][blockName]['status'] in ["Closed", "DBSUploaded", "InFlight"]:
                isOpen = 'n'
            else:
                isOpen = 'y'
            xmlBlock = xmlDataset.getFileblock(blockName, isOpen)
            
            for file in datasetPathSet[datasetPath][blockName]['fileList']:
                filteredCheckSums = {}
                for type, value in file["CHECKSUMS"].items():
                    if type.lower() in ('cksum', 'adler32'):
                        filteredCheckSums[type] = value
                xmlBlock.addFile(file["LFN"], filteredCheckSums, file["FILESIZE"])
        
    improv = spec.save()
    xmlString = improv.makeDOMElement().toprettyxml()
    logging.debug("From DatasetPathSet")
    logging.debug(xmlString)
    return xmlString


def makePhEDExXMLForDatasets(globalDbsUrl, datasetPaths):        
    
    """
    _makePhEDExDropForDataset_

    Given a DBS Url, list of dataset path name, 
    generate an XML structure for injection
   
    """
    spec = XMLInjectionSpec(globalDbsUrl)
    for datasetPath in datasetPaths:
        xmlDataset = spec.getDataset(datasetPath)
      
    improv = spec.save()
    xmlString = improv.makeDOMElement().toprettyxml()
    logging.debug("From DastsetPath")
    logging.debug(xmlString)
    return xmlString

    
    
    

