#!/usr/bin/env python
"""
_ExpressInjectorComponent_

Component for generating Express jobs
"""

__version__ = "$Revision: 1.25 $"
__revision__ = "$Id: ExpressInjectorComponent.py,v 1.25 2011/01/06 00:08:00 hufnagel Exp $"

import os
import sys
import shutil
import logging

from JobQueue import JobQueueDB
from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0.WorkflowFactory.ExpressWorkflow import ExpressWorkflow

from T0.JobFactory.StreamerJob import StreamerJob

from T0 import Globals

from T0.State.Database.Reader import ListJobs
from T0.State.Database.Writer import InsertJob
from T0.State.Database.Writer import InsertRun

from WMCore.JobSplitting.SplitterFactory import SplitterFactory

from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig

from WMCore.WMBS.Job import Job

from ProdAgentCore.Configuration import loadProdAgentConfiguration
from T0.GenericTier0 import Tier0DB


class ExpressInjectorComponent(GenericTier0Component):
    """
    _ExpressInjectorComponent_

    Component for generating Express jobspecs.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.

        """
        GenericTier0Component.__init__(self, **args)

        # Data structure to hold WorkflowSpecs.  This is a two level
        # dictionary with run -> stream_id
        self.workflows = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2

        # Check the ProdAgent configuration for the following parameters:
        #
        #   MaxInputEvents - max number of input events per job
        #
        #   PollInterval - How often T0AST is checked for new files
        #   scramArch - SCRAM Architecture
        #   cmsPath - Path to CMS software
        #   JobCacheDir - Where jobspecs and workflows are stored
        #
        self.maxInputEvents = int(self.lookupParameter("MaxInputEvents", "100"))

        self.pollInterval = self.lookupParameter("PollInterval", "00:01:00")
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")
        self.jobCacheDir = self.lookupParameter("JobCacheDir", self.args["ComponentDir"])

        # WMCore splitter factory for creating Express jobs
        self.splitterFactory = SplitterFactory(package = "T0.JobSplitting")

        self.expressSubscription = None

        # See if lazy downloads are enabled. Whether or not this
        # switch does anything depends on the framework version
        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        self.popconlogDBConn = None

        logging.info("ExpressInjector Component Started")


    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component receiving a message
        """
        # always have a debug on/off switch
        if message == "ExpressInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "ExpressInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "ExpressInjector:Poll":
            self.poll()
        elif message == "CleanUp":
            self.cleanupRun(payload)
        else:
            logging.error("Unknown message: %s, payload: %s" % (message, payload))

        return


    def cleanupRun(self, payload):
        """
        _cleanupRun_

        Remove all workflows and jobspecs for a particular run from disk.
        Remove any workflows for the run from memory.

        """
        try:
            run = int(payload)
        except ValueError:
            logging.error("Run number is not valid: %s" % payload)
            return

        logging.debug("Cleaning up run %s" % run)

        if run in self.workflows.keys():
            del self.workflows[run]

        workflowDir = self.jobCacheDir + "/workflows/Run%s" % run
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % run
        
        shutil.rmtree(workflowDir, ignore_errors=True)
        shutil.rmtree(jobSpecDir, ignore_errors=True)

        return


    def retrieveWorkflow(self, run, stream_id):
        """
        _retrieveWorkflow_

        Retrieve a workflow from the cache.
        If the workflow does not exist a new one will be created.
        If workflow creation fails None will be returned.

        """
        if self.workflows.has_key(run):
            if self.workflows[run].has_key(stream_id):
                return self.workflows[run][stream_id]
        else:
            self.workflows[run] = {}
                    
        # that we made it here means we have to create a new workflow
        workflow = self.createWorkflow(run, stream_id)
        if workflow != None:
            self.workflows[run][stream_id] = workflow

        return workflow


    def createWorkflow(self, run, stream_id):
        """
        _createWorkflow_

        Creates a new express workflow.
        If workflow creation fails None is returned.

        A 'NewWorkflow' message is generated for each new workflow
        with the path to the workflow as the payload.

        """
        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %d" % run)
            return None

        expressConfig = runConfig.getExpressConfiguration(stream_id)
        if expressConfig == None:
            logging.error("No express config for run %d and stream %s" % (run, stream))
            return None

        stream = runConfig.getStreamName(stream_id)

        logging.debug("createWorkflow for run %d, stream %s" % (run, stream))

        scenario = expressConfig.Scenario

        if expressConfig.ProcessingConfigURL == None:
            configFile = None
        else:
            configFile = runConfig.retrieveConfigFromURL(expressConfig.ProcessingConfigURL)

        # filter out ALCARECO from data tiers
        enableAlcaReco = False
        if "ALCARECO" in expressConfig.DataTiers:
            expressConfig.DataTiers.remove("ALCARECO")
            enableAlcaReco = True

        # filter out DQM from data tiers
        enableDQM = False
        if "DQM" in expressConfig.DataTiers:
            expressConfig.DataTiers.remove("DQM")
            enableDQM = True

        outModuleInfoList = []
        if expressConfig.SplitInProcessing:
            outModuleInfoList = runConfig.getOutputModuleInfoList("Express", stream_id,
                                                                  expressConfig.DataTiers,
                                                                  splitIntoDatasets = True)
        else:
            outModuleInfoList = runConfig.getOutputModuleInfoList("Express", stream_id,
                                                                  expressConfig.DataTiers)

        if outModuleInfoList == None:
            logging.error("No output module for run %s and stream %s" % (run, stream))
            return None

        # now add the single output module for combined ALCARECO
        if enableAlcaReco:
            outModuleInfo = runConfig.getOutputModuleInfo("Express","ALCARECO", streamID = stream_id)
            if outModuleInfo == None:
                logging.error("Could not add ALCARECO output for run %s and stream %s" % (run, stream))
                return None

            outModuleInfoList.append(outModuleInfo)

        # now add the single output module for combined DQM
        if enableDQM:
            outModuleInfo = runConfig.getOutputModuleInfo("Express","DQM", streamID = stream_id)
            if outModuleInfo == None:
                logging.error("Could not add DQM output for run %s and stream %s" % (run, stream))
                return None

            outModuleInfoList.append(outModuleInfo)

        workflowSpecName = "Express-Run%s-Stream%s" % (run, stream)
        workflowDir = os.path.join(self.jobCacheDir, "workflows", "Run%s" % run)
        workflowSpecFile = os.path.join(workflowDir, "%s-workflow.xml" % workflowSpecName)

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No workflow for run, stream: %s, %s" % (run, stream))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowSpecFile)

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return workflowSpec

        workflow = ExpressWorkflow(run,
                                   expressConfig.CMSSWVersion,
                                   self.cmsPath,
                                   self.scramArch,
                                   *outModuleInfoList)

        workflowSpec = workflow.makeWorkflowSpec(workflowSpecName,
                                                 scenario = scenario,
                                                 configFile = configFile,
                                                 alcaProducers = expressConfig.AlcaProducers,
                                                 globalTag = expressConfig.GlobalTag,
                                                 enableLazyDownload = self.useLazyDownload,
                                                 productionMode = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)
        
        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)
        
        workflowSpec.save(workflowSpecFile)
            
        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)


    def createJobSpecs(self, run, stream_id, jobDict):
        """
        _createJobSpecs_

        Loop over all jobs and process them (same run and stream_id = same workflow)

        Generate job specs for every job

        """
        logging.debug("createJobSpecs(): Running...")

        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %s" % run)
            return {}

        stream = runConfig.getPrimaryDatasetName(stream_id)
        if stream == None:
            logging.error("Could not get stream name for id %d" % stream_id)
            return {}

        workflowSpec = self.retrieveWorkflow(run, stream_id)
        if workflowSpec == None:
            logging.error("No workflow for run %s, stream %s" % (run, stream))
            return {}

        jobSpecBaseDir = os.path.join(self.jobCacheDir, "jobspecs", "Run%d" % run)

        jobFactory = StreamerJob(workflowSpec)

        jobSpecDict = {}
        for id, files in jobDict.items():

            jobSpecDict[id] = jobFactory("Express", id, files, jobSpecBaseDir, True)

        return jobSpecDict


    def poll(self):
        """
        _poll_

        Poll T0AST for new Express jobs

        Call job creation code, then queue them in JobQueue
        and update name and status in WMBS

        """
        logging.info("poll(): Running...")

        # first check for new express released runs
        try:
            InsertRun.releaseRunsForExpress(self.t0astDBConn, self.popconlogDBConn)
        except StandardError, ex:
            logging.error("ERROR: Failed to release runs for express : %s\n" % str(ex))

        self.t0astDBConn.beginTransaction()

        wmbsJobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                              subscription = self.expressSubscription)

        wmbsJobFactory(maxInputEvents = self.maxInputEvents)

        self.t0astDBConn.commit()

        jobDict = ListJobs.listNewStreamerJobs(self.t0astDBConn, self.expressSubscription['id'])

        # process jobs by run and stream_id
        jobSpecDict = {}
        for run in jobDict.keys():
            for stream_id in jobDict[run].keys():

                normalJobDict = jobDict[run][stream_id]['normal']
                jobSpecDict.update( self.createJobSpecs(run, stream_id, normalJobDict) )

                splitJobDict = jobDict[run][stream_id]['split']
                jobSpecDict.update( self.createJobSpecs(run, stream_id, splitJobDict) )

        if len(jobSpecDict) > 0:

            # aggregate information for efficient database operations
            wmbsJobs = []
            jobNames = {}
            for jobId, jobSpec in jobSpecDict.items():
                wmbsJobs.append(Job(id = jobId))
                jobNames[jobId] = jobSpec.parameters["JobName"]

            # now do actual PA and T0AST DB modifications
            try:
                self.paDBConn.beginTransaction()

                for jobSpec in jobSpecDict.values():

                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters["JobSpecFile"],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority,
                                             siteName = None,
                                             status = "new")

                self.t0astDBConn.beginTransaction()

                InsertJob.updateWMBSJobNames(self.t0astDBConn, jobNames)
                self.stateChanger.propagate(wmbsJobs, 'created', 'new')

                self.paDBConn.commit()
                self.t0astDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(jobSpecDict))
            except StandardError, ex:
                self.t0astDBConn.rollback()
                self.paDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        self.ms.publishUnique("ExpressInjector:Poll", "", self.pollInterval)
        self.ms.commit()
        return


    def startComponent(self):
        """
        _startComponent_

        Fire up the message service and define the list of
        messages that this component subscribes to

        """
        self.ms.subscribeTo("ExpressInjector:StartDebug")
        self.ms.subscribeTo("ExpressInjector:EndDebug")
        self.ms.subscribeTo("ExpressInjector:Poll")
        self.ms.subscribeTo("CleanUp")
        
        self.connectT0AST()
        self.connectPADB()

        # Load up the express subscription for easy access.
        self.expressSubscription = Globals.retrieveExpressSubscription()

        paConfig = loadProdAgentConfiguration()
        popconlogDBConfig = paConfig.getConfig("PopConLogDB")
        if popconlogDBConfig == None:
            self.popconlogDBConn = None
        else:
            self.popconlogDBConn = Tier0DB.Tier0DB(popconlogDBConfig,
                                                   manageGlobal = False)
            self.popconlogDBConn.connect()

        self.ms.publishUnique("ExpressInjector:Poll", "", "00:01:00")
        self.ms.commit()

        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)

        # wait for messages
        self.messageLoop()
