#!/usr/bin/env python
"""
_ExpressMergerComponent_

Create mergepack jobs and send them to job queue
"""

__revision__ = "$Id: ExpressMergerComponent.py,v 1.36 2011/01/13 11:02:33 hufnagel Exp $"
__version__ = "$Revision: 1.36 $"

import os
import sys
import shutil
import logging

from JobQueue import JobQueueDB
from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0 import Globals

from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListJobs
from T0.State.Database.Writer import InsertJob

from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig
from WMCore.WMBS.Job import Job

from T0.WorkflowFactory.MergePackWorkflow import MergePackWorkflow
from T0.JobFactory.MergePackJob import MergePackJob


class ExpressMergerComponent(GenericTier0Component):
    """
    _ExpressMergerComponent_

    Search for mergable files in T0AST that can be gathered together
    and merged into larger files.  Create and submit merge jobs to
    the ProdAgent JobQueue.

    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.

        """
        GenericTier0Component.__init__(self, **args)

        # Data structure to hold WorkflowSpecs.  This is a three level
        # dictionary with run -> primds_id -> datatier
        self.workflows = {}

        # cache datasetpath information to avoid queries
        self.datasetpaths = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2

        # Check the ProdAgent configuration for the following parameters:
        #   MaxAlcaLumiAge - max age of lumi sections for ALCARECO
        #   MaxDqmLumiAge - max age of lumi sections for DQM
        #   MaxLumiAge - max age of lumi sections for other tiers
        #
        #   We accumulate lumi section until the oldest reaches a maximum age,
        #   then trigger a merge on everything we have.
        #   Setting the maximum age to 0 forces mergepack by lumi section.
        #
        #   MaxInputFiles - max number of input files for a job
        #   MaxMergeSize = max size of input for a job
        #
        #   PollInterval - How often T0AST is checked for new files
        #   scramArch - SCRAM Architecture
        #   cmsPath - Path to CMS software
        #   JobCacheDir - Where jobspecs and workflows are stored
        #
        self.maxAlcaLumiAge = int(self.lookupParameter("MaxAlcaLumiAge", "300"))
        self.maxDqmLumiAge = int(self.lookupParameter("MaxDqmLumiAge", "300"))
        self.maxLumiAge = int(self.lookupParameter("MaxLumiAge", "100"))

        self.maxInputFiles = int(self.lookupParameter("MaxInputFiles", "200"))
        self.maxInputSize = int(self.lookupParameter("MaxInputSize", 4*1024*1024*1024))

        self.pollInterval = self.lookupParameter("PollInterval", "00:01:00")
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")
        self.jobCacheDir = self.lookupParameter("JobCacheDir", self.args["ComponentDir"])

        # WMCore splitter factory for creating Merge jobs
        self.splitterFactory = SplitterFactory(package = "T0.JobSplitting")

        self.expressMergeSubscription = None

        # See if lazy downloads are enabled. Whether or not this
        # switch does anything depends on the framework version
        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        logging.info("ExpressMerger Component Started")


    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if message == "ExpressMerger:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "ExpressMerger:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "ExpressMerger:Poll":
            self.poll()
        elif message == "CleanUp":
            self.cleanupRun(payload)
        
        return
    

    def cleanupRun(self, payload):
        """
        _cleanupRun_

        Remove all workflows and jobspecs for a particular run from disk.
        Remove any workflows for the run from memory.

        """
        try:
            run = int(payload)
        except ValueError:
            logging.error("Run number is not valid: %s" % payload)
            return

        logging.debug("Cleaning up run %s" % run)

        if run in self.workflows.keys():
            del self.workflows[run]

        workflowDir = self.jobCacheDir + "/workflows/Run%s" % run
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % run

        shutil.rmtree(workflowDir, ignore_errors=True)
        shutil.rmtree(jobSpecDir, ignore_errors=True)


    def retrieveWorkflow(self, run, datasetpath,
                         primds, procds, datatier):
        """
        _retrieveWorkflow_

        Retrieve a workflow from the cache that the ExpressMerger maintains.
        If the workflow does not exist a new one will be created. If workflow
        creation fails for some reason None will be returned.

        """
        if self.workflows.has_key(run):
            if self.workflows[run].has_key(datasetpath):
                return self.workflows[run][datasetpath]
        else:
            self.workflows[run] = { datasetpath : {} }

        # that we made it here means we have to create a new workflow
        workflow = self.createWorkflow(run, datasetpath,
                                       primds, procds, datatier)
        if workflow != None:
            self.workflows[run][datasetpath] = workflow

        return workflow


    def createWorkflow(self, run, datasetpath,
                       primds, procds, datatier):
        """
        _createWorkflow_

        Creates an express merge workflow for a specific run and dataset 
        combination.  If workflow creation fails None is returned.

        A NewWorkflow message is generated for each new workflow
        with the path to the workflow being passed as the payload.

        """
        logging.debug("createWorkflow(): Running run %s, datasetpath %s" % \
                      (run, datasetpath))

        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %d" % run)
            return None

        primds_id = runConfig.getPrimaryDatasetID(primds)
        stream_id = runConfig.getStreamIDfromDatasetID(primds_id)
        stream = runConfig.getStreamName(stream_id)

        expressConfig = runConfig.getExpressConfiguration(stream_id)
        if expressConfig == None:
            logging.error("No express config for run %d and stream %s" % (run, stream))
            return None

        scenario = expressConfig.Scenario

        if datatier == 'ALCARECO':

            if expressConfig.AlcaMergeConfigURL == None:
                configFile = None
            else:
                configFile = runConfig.retrieveConfigFromURL(expressConfig.AlcaMergeConfigURL)

            outModuleInfo = runConfig.getOutputModuleInfo("Express", datatier,
                                                          streamID = stream_id)

            if outModuleInfo == None:
                logging.error("No output module for run %s, stream %s, datatier %s"
                              % (run, stream, datatier))
                return None

            workflow = MergePackWorkflow(run,
                                         expressConfig.CMSSWVersion,
                                         self.cmsPath,
                                         self.scramArch,
                                         *[outModuleInfo])

            workflowSpecName = "MergePack-Run%s-%s-%s" % (run, stream, datatier)

        elif datatier == 'DQM':

            outModuleInfo = runConfig.getOutputModuleInfo("Express", datatier,
                                                          datasetID = primds_id)
            if outModuleInfo == None:
                logging.error("No output module for run %s, stream %s, primds %s, datatier %s"
                              % (run, stream, primds, datatier))
                return None

            workflow = MergePackWorkflow(run,
                                         expressConfig.CMSSWVersion,
                                         self.cmsPath,
                                         self.scramArch,
                                         *[outModuleInfo])

            workflowSpecName = "MergePack-Run%s-%s-%s" % (run, stream, datatier)

        else:

            if expressConfig.SplitInProcessing:
                outModuleInfo = runConfig.getOutputModuleInfo("Express", datatier,
                                                              datasetID = primds_id)
                if outModuleInfo == None:
                    logging.error("No output module for run %s, stream %s, primds %s, datatier %s"
                                  % (run, stream, primds, datatier))
                    return None

                workflow = MergePackWorkflow(run,
                                             expressConfig.CMSSWVersion,
                                             self.cmsPath,
                                             self.scramArch,
                                             *[outModuleInfo])

                workflowSpecName = "MergePack-Run%s-%s-%s-%s" % (run, stream, datatier, primds)

            else:
                outModuleInfoList = runConfig.getOutputModuleInfoList("Express", stream_id, datatier,
                                                                      splitIntoDatasets = True)
                if outModuleInfoList == None:
                    logging.error("No output modules for run %s, stream %s, datatier %s"
                                  % (run, stream, datatier))
                    return None

                workflow = MergePackWorkflow(run,
                                             expressConfig.CMSSWVersion,
                                             self.cmsPath,
                                             self.scramArch,
                                             *outModuleInfoList)

                workflowSpecName = "MergePack-Run%s-%s-%s" % (run, stream, datatier)


        workflowDir = os.path.join(self.jobCacheDir, "workflows", "Run%d" % run)
        workflowSpecFile = os.path.join(workflowDir, "%s-workflow.xml" % workflowSpecName)

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No workflow for run, stream: %s, %s" % (run, stream))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowSpecFile)

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return workflowSpec

        if datatier == 'ALCARECO':

            if expressConfig.AlcaMergeConfigURL == None:
                configFile = None
            else:
                configFile = runConfig.retrieveConfigFromURL(expressConfig.AlcaMergeConfigURL)

            workflowSpec = workflow.makeWorkflowSpec(workflowSpecName,
                                                     scenario = scenario,
                                                     configFile = configFile,
                                                     alcaProducers = expressConfig.AlcaProducers,
                                                     globalTag = expressConfig.GlobalTag,
                                                     enableLazyDownload = self.useLazyDownload,
                                                     production = self.productionMode)

        elif datatier == 'DQM':

            workflowSpec = workflow.makeWorkflowSpec(workflowSpecName,
                                                     tempOutput = True,
                                                     enableLazyDownload = self.useLazyDownload,
                                                     production = self.productionMode)

        else:

            workflowSpec = workflow.makeWorkflowSpec(workflowSpecName,
                                                     enableLazyDownload = self.useLazyDownload,
                                                     production = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)

        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)
        
        workflowSpec.save(workflowSpecFile)

        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)


    def createJobSpecs(self, run, datasetpath, jobs):
        """
        _createJobSpecs_

        Loop over all jobs and process them (same run and primds_id = same workflow)

        Generate job specs for every job

        """
        logging.debug("createJobSpecs(): Running...")

        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %s" % run)
            return {}

        if not self.datasetpaths.has_key(datasetpath):
            self.datasetpaths[datasetpath] = ListDatasets.listDatasetPathInfo(self.t0astDBConn,
                                                                              datasetpath)
        primds = self.datasetpaths[datasetpath]['primds']
        procds = self.datasetpaths[datasetpath]['procds']
        datatier = self.datasetpaths[datasetpath]['datatier']

        workflowSpec = self.retrieveWorkflow(run, datasetpath,
                                             primds, procds, datatier)
        if workflowSpec == None:
            logging.error("No workflow for run %s, datasetpath %s" % (run,
                                                                      datasetpath))
            return {}

        jobSpecBaseDir = os.path.join(self.jobCacheDir, "jobspecs", "Run%d" % run)

        jobFactory = MergePackJob(workflowSpec)

        jobSpecDict = {}
        for job in jobs:

            jobSpecDict[job['id']] = jobFactory(job['id'],
                                                datatier,
                                                job['files'],
                                                jobSpecBaseDir)

        return jobSpecDict

                        
    def poll(self):
        """
        _poll_

        Poll T0AST for new ExpressMerge jobs

        Call job creation code, then queue them in JobQueue
        and update name and status in WMBS

        """
        logging.debug("poll(): Running...")

        self.t0astDBConn.beginTransaction()

        wmbsJobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                              subscription = self.expressMergeSubscription)

        wmbsJobFactory(maxAlcaLumiAge = self.maxAlcaLumiAge,
                       maxDqmLumiAge = self.maxDqmLumiAge,
                       maxLumiAge = self.maxLumiAge,
                       maxInputFiles = self.maxInputFiles,
                       maxInputSize = self.maxInputSize)

        self.t0astDBConn.commit()

        jobs = ListJobs.listNewWMBSJobs(self.t0astDBConn, self.expressMergeSubscription['id'])

        # sort jobs by run and datasetpath
        jobDict = {}
        for job in jobs:
            run = job['run']
            datasetpath = job['datasetpath']

            # intialize dictionary tree
            if not jobDict.has_key(run):
                jobDict[run] = {}
            if not jobDict[run].has_key(datasetpath):
                jobDict[run][datasetpath] = []

            jobDict[run][datasetpath].append(job)

        # process jobs by run and datasetpath
        jobSpecDict = {}
        for run in jobDict.keys():
            for datasetpath, jobs in jobDict[run].items():
                jobSpecDict.update( self.createJobSpecs(run, datasetpath, jobs) )

        if len(jobSpecDict) > 0:

            # aggregate information for efficient database operations
            wmbsJobs = []
            jobNames = {}
            for jobId, jobSpec in jobSpecDict.items():
                wmbsJobs.append(Job(id = jobId))
                jobNames[jobId] = jobSpec.parameters["JobName"]

            # now do actual PA and T0AST DB modifications
            try:
                self.paDBConn.beginTransaction()

                for jobSpec in jobSpecDict.values():

                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters["JobSpecFile"],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority,
                                             siteName = None,
                                             status = "new")

                self.t0astDBConn.beginTransaction()

                InsertJob.updateWMBSJobNames(self.t0astDBConn, jobNames)
                self.stateChanger.propagate(wmbsJobs, 'created', 'new')

                self.paDBConn.commit()
                self.t0astDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(jobSpecDict))
            except StandardError, ex:
                self.t0astDBConn.rollback()
                self.paDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        self.ms.publishUnique("ExpressMerger:Poll", "", self.pollInterval)
        self.ms.commit()
        return


    def startComponent(self):
        """
        _startComponent_

        Fire up the message service and define the list of
        messages that this component subscribes to

        """
        self.ms.subscribeTo("ExpressMerger:StartDebug")
        self.ms.subscribeTo("ExpressMerger:EndDebug")
        self.ms.subscribeTo("ExpressMerger:Poll")
        self.ms.subscribeTo("CleanUp")
        self.ms.commit()

        self.connectT0AST()
        self.connectPADB()

        # Load up the express merge subscription for easy access.
        self.expressMergeSubscription = Globals.retrieveExpressMergeSubscription()

        self.ms.publishUnique("ExpressMerger:Poll", "", self.pollInterval)
        self.ms.commit()
        
        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)

        # Wait for messages.
        self.messageLoop()

