"""
save all the global variables used by Tier0
Currently just WMBS subscription

"""
import traceback
import sys
import logging
from time import strftime, localtime

from WMCore.WMBS.Workflow import Workflow
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.Services.Twitter import twitter

class GlobalValues:
    """
    Namespace like Container 

    """
    _repackSubscription = None
    _expressSubscription = None
    _expressMergeSubscription = None
    _mergeSubscription = None
    _alcaSubscription = None
    _expressDBSSubscription = None
    _expressTransferSubscription = None
    _dbsUploadSubscription = None
    _transferSubscription = None
    _recoSubscription = None
    _deleteSubscription = None
    _dqmSubscription  = None
    _wmbsPublishSubscription = None
    _promptCalibSubscription = None
    _alcaHarvestSubscription = None
    TIER_ORDER = ["RAW", "RECO", "AOD", "DQM", "ALCARECO", "ALCAPROMPT"]
    # express doesn't really have order (not used currently)
    EXPRESS_TIER_ORDER = ["RAW", "RECO", "ALCARECO", "ALCAPROMPT", "FEVT", "FEVTHLTALL"]
    
    RUN_STATE_ORDER = ["CloseOutRepack",
                       "CloseOutRepackMerge",
                       "CloseOutPromptReco",
                       "CloseOutRecoMerge",
                       "CloseOutAlcaSkim", 
                       "CloseOutAlcaSkimMerge",
                       "CloseOutExport", 
                       "CloseOutT1Skimming", 
                       "Complete"]
    
    CAF_NODE = "T2_CH_CAF"

    #  _tier0Twitter is following format {'username': '', 'passwd': ''}
    _tier0Twitter = None

def _getFollowingList(element, eList):
    """
    verify the element whether it is exist against the eList,
    and return the list of after elements including itself
    if not raise an error
    """
    for index in range(len(eList)):
        if eList[index].lower() == element.lower():
            return eList[index:]
    raise Exception,"no element is mapped to %s in %s" % (
                                                element, eList)
  
def _getPreviousList(element, eList):
    """
    verify the element whether it is exist against the eList, 
    and return the list of previous elements including itself
    if not, raise an error
    """
    for index in range(len(eList)):
        if eList[index].lower() == element.lower():
            return eList[:index+1]
    raise Exception,"no element is mapped to %s in %s" % (
                                                element, eList)

def _getElement(element, eList):
    
    """
    verify the element whether it is exist against the eList, 
    and return itself
    if not, raise an error
    
    """
    for index in range(len(eList)):
        if eList[index].lower() == element.lower():
            return eList[index]
    raise Exception,"no element is mapped to %s in %s" % (
                                                element, eList)


def getFollowingTiers(tier):
    """
    verify the tier whether it is exist against the TIER_ORDER, 
    and return the list of following elements including itself
    if not, raise an error
    """
    return _getFollowingList(tier, GlobalValues.TIER_ORDER)

def getPreviousTiers(tier):
    """
    verify the tier whether it is exist against the TIER_ORDER, 
    and return the list of previous elements including itself
    if not, raise an error
    """
    return _getPreviousList(tier, GlobalValues.TIER_ORDER)

def getTier(tier):
    """
    verify the tier whether it is exist against the TIER_ORDER, 
    and return itself
    if not, raise an error
    """
    return _getElement(tier, GlobalValues.TIER_ORDER)

def getFollowingStates(state):
    """
    """
    return _getFollowingList(state, GlobalValues.RUN_STATE_ORDER)            

def getPreviousStates(state):
    """
    """
    return _getPreviousList(state, GlobalValues.RUN_STATE_ORDER)

def getState(state):
    """
    verify the state whether it is exist, and return the same state
    if not raise an error
    """
    return _getElement(state, GlobalValues.RUN_STATE_ORDER)
    

def closeOutStageMap(dataTier):
    """
    _closeOutStageMap_
    
    """
    map = {}
    
    map[getTier("RAW")] = {
                "Stage": getFollowingStates("CloseOutPromptReco"),
                "TierList": getPreviousTiers("RAW")}
    map[getTier("RECO")] = {
                "Stage": getFollowingStates("CloseOutAlcaSkim"),
                "TierList": getPreviousTiers("RECO")}
    map[getTier("AOD")] = {
                "Stage": getFollowingStates("CloseOutAlcaSkim"),
                "TierList": getPreviousTiers("AOD")}
    map[getTier("DQM")] = {
                "Stage": getFollowingStates("CloseOutAlcaSkim"),
                "TierList": getPreviousTiers("DQM")}
    map[getTier("ALCARECO")] = {
                "Stage": getFollowingStates("CloseOutExport"),
                "TierList": getPreviousTiers("ALCARECO")}
    map[getTier("ALCAPROMPT")] = {
                "Stage": getFollowingStates("CloseOutExport"),
                "TierList": getPreviousTiers("ALCAPROMPT")}

    return map[dataTier]        

def getMaxSubscriptionID():
    """
    _getMaxSubscriptionID_

    Given all the Tier0 subscriptions figure out the highest subscription ID.
    This is used by the Tier1Injector so that it can ignore any Tier0 specific
    subscriptions.
    """
    repackSub = retrieveRepackSubscription()
    expressSub = retrieveExpressSubscription()
    mergePackSub = retrieveExpressMergeSubscription()
    mergeSub = retrieveMergeSubscription()
    expressDBSSub = retrieveExpressDBSSubscription()
    expressTransferSub = retrieveExpressTransferSubscription() 
    dbsUploadSub = retrieveDBSUploadSubscription()
    transferSub = retrieveTransferSubscription()
    recoSub = retrieveRecoSubscription()
    alcaSub = retrieveAlcaSubscription()
    deleteSub = retrieveDeleteSubscription()
    wmbsPublishSub = retrieveWMBSPublishSubscription()
    dqmSub = retrieveDQMSubscription()
    promptCalibSub = retrievePromptCalibSubscription()
    alcaHarvestSub = retrieveAlcaHarvestSubscription()

    return max(repackSub["id"], expressSub["id"], mergePackSub["id"], mergeSub["id"],
               expressDBSSub["id"], expressTransferSub["id"], dbsUploadSub["id"],
               transferSub["id"], recoSub["id"], alcaSub["id"], deleteSub["id"],
               wmbsPublishSub["id"], dqmSub["id"], promptCalibSub["id"],
               alcaHarvestSub["id"])

def _getSubscription(workflow, fileset):
    workflow.load()
    fileset.load()
    subscription = Subscription(fileset = fileset, workflow = workflow)
    subscription.load()
    return subscription
    
def tier0TwitterPost(msg, start_time = None):
    """
    Posting a message and time stamp on Twitter. if it fails just log error 
    message and keep going 
    """
    
    def configParameterCheck(name):
        if not twitterConfig.has_key(name):
            logging.error("Twitter Config '%s' is not defined" % name)
            return False
        return True
    
    
    if GlobalValues._tier0Twitter == None:

        from ProdAgentCore.Configuration import loadProdAgentConfiguration

        paConfig = loadProdAgentConfiguration()
        twitterConfig = paConfig.getConfig( "Tier0Twitter" )
        if twitterConfig == None:
            logging.error("Tier0Twitter config is not defined in RepackerConfig.xml")
            return
        
        parameters = ["Post", "User", "Passwd"]
        
        for param in parameters:
            if not configParameterCheck(param):
                return
        
        if twitterConfig["Post"].lower() != "on":
            logging.info("Twitter posting is not turned on: %s" % msg )
            return
        else: 
            GlobalValues._tier0Twitter = {'username': twitterConfig["User"],
                                          'passwd': twitterConfig["Passwd"]}
    try:
        logging.debug("Updating Twitter")
        api = twitter.Api(username=GlobalValues._tier0Twitter['username'], 
                          password=GlobalValues._tier0Twitter['passwd'])
        if start_time == None:
            start_time = localtime()
        else:
            start_time=localtime(start_time)
        msg = strftime("%a, %d %b %Y %H:%M:%S +0000", start_time) + ": " + msg
        api.PostUpdates(msg)
        logging.debug("Twitter update succeed")
    except Exception, ex:
        logging.error("Twitter update failed: %s" % ex )

def retrieveRepackSubscription():
    """
    _retrieveRepackSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._repackSubscription == None:
        GlobalValues._repackSubscription = \
           _getSubscription(Workflow(spec = "Repack", 
                                  owner = "CMSTier0",
                                  name = "Repack"),
                            Fileset( name = "Repack")
                           )
    return GlobalValues._repackSubscription

def retrieveExpressSubscription():
    """
    _retrieveExpressSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._expressSubscription == None:
        GlobalValues._expressSubscription = \
           _getSubscription(Workflow(spec = "Express", 
                                  owner = "CMSTier0",
                                  name = "Express"),
                            Fileset( name = "Express")
                           )
    return GlobalValues._expressSubscription

def retrieveExpressMergeSubscription():
    """
    _retrieveExpressMergeSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._expressMergeSubscription == None:
        GlobalValues._expressMergeSubscription = \
           _getSubscription(Workflow(spec = "ExpressFileMerge", 
                                  owner = "CMSTier0",
                                  name = "ExpressFileMerge"),
                            Fileset( name = "ExpressMergeable")
                           ) 
        
    return GlobalValues._expressMergeSubscription

def retrieveMergeSubscription():
    """
    _retrieveMergeSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._mergeSubscription == None:
        GlobalValues._mergeSubscription = \
           _getSubscription(Workflow(spec = "FileMerge", 
                                  owner = "CMSTier0",
                                  name = "FileMerge"),
                            Fileset( name = "Mergeable")
                           ) 
        
    return GlobalValues._mergeSubscription

def retrieveAlcaSubscription():
    """
    _retrieveAlcaSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._alcaSubscription == None:
        GlobalValues._alcaSubscription = \
           _getSubscription(Workflow(spec = "FileAlcaSkim", 
                                  owner = "CMSTier0",
                                  name = "FileAlcaSkim"),
                            Fileset( name = "AlcaSkimmable" )
                           )

    return GlobalValues._alcaSubscription

def retrievePromptCalibSubscription():
    """
    _retrievePromptCalibSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be
    needed to loadData from caller for further processing 
    """
    if GlobalValues._promptCalibSubscription == None:
        GlobalValues._promptCalibSubscription = \
           _getSubscription(Workflow(spec = "FilePromptCalib", 
                                     owner = "CMSTier0",
                                     name = "FilePromptCalib"),
                            Fileset( name = "PromptCalibrationable" )
                            ) 
        
    return GlobalValues._promptCalibSubscription

def retrieveExpressDBSSubscription():
    """
    _retrieveExpressDBSSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._expressDBSSubscription == None:
        GlobalValues._expressDBSSubscription = \
            _getSubscription(Workflow(spec = "ExpressDBSUpload", 
                                      owner = "CMSTier0",
                                      name = "ExpressDBSUpload"),
                             Fileset(name = "ExpressDBSUploadable")
                             )
             
    return GlobalValues._expressDBSSubscription

def retrieveExpressTransferSubscription():
    """
    _retrieveExpressTransferSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._expressTransferSubscription == None:
        GlobalValues._expressTransferSubscription = \
            _getSubscription(Workflow(spec = "ExpressTransfer", 
                                      owner = "CMSTier0",
                                      name = "ExpressTransfer"),
                             Fileset(name = "ExpressTransferable")
                             )

    return GlobalValues._expressTransferSubscription

def retrieveDBSUploadSubscription():
    """
    _retrieveDBSUploadSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._dbsUploadSubscription == None:
        GlobalValues._dbsUploadSubscription = \
            _getSubscription(Workflow(spec = "FileDBSUpload", 
                                      owner = "CMSTier0",
                                      name = "FileDBSUpload"),
                             Fileset(name = "DBSUploadable")
                             )
             
    return GlobalValues._dbsUploadSubscription

def retrieveTransferSubscription():
    """
    _retrieveTransferSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._transferSubscription == None:
        GlobalValues._transferSubscription = \
            _getSubscription(Workflow(spec = "FileTransfer", 
                                      owner = "CMSTier0",
                                      name = "FileTransfer"),
                             Fileset(name = "Transferable")
                             )
            
    return GlobalValues._transferSubscription

def retrieveRecoSubscription():
    """
    _retrieveRecoSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._recoSubscription == None:
        GlobalValues._recoSubscription = \
            _getSubscription(Workflow(spec = "FileReconstruction", 
                                      owner = "CMSTier0",
                                      name = "FileReconstruction"),
                             Fileset(name = "Reconstructable")
                             )
            
    return GlobalValues._recoSubscription
 
def retrieveDeleteSubscription():
    """
    _retrieveDeleteSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._deleteSubscription == None:
        GlobalValues._deleteSubscription =  \
            _getSubscription(Workflow(spec = "FileDelete", 
                                      owner = "CMSTier0",
                                      name = "FileDelete"),
                             Fileset(name = "Deletable")
                             )
            
    return GlobalValues._deleteSubscription

def retrieveDQMSubscription():
    """
    _retrieveDQMSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._dqmSubscription == None:
        GlobalValues._dqmSubscription =  \
            _getSubscription(Workflow(spec = "DQM", 
                                      owner = "CMSTier0",
                                      name = "DQM"),
                             Fileset(name = "DQMenable")
                             )
            
    return GlobalValues._dqmSubscription

def retrieveWMBSPublishSubscription():
    """
    _retrieveWMBSPublishSubscription_

    Retrieve the wmbs publish subscription from the cache if possible.  Load it
    from the database if it is not in the cache.
    """
    if GlobalValues._wmbsPublishSubscription == None:
        GlobalValues._wmbsPublishSubscription =  \
            _getSubscription(Workflow(spec = "FileWMBSPublish", 
                                      owner = "CMSTier0",
                                      name = "FileWMBSPublish"),
                             Fileset(name = "WMBSPublishable")
                             )
            
    return GlobalValues._wmbsPublishSubscription   

def retrieveAlcaHarvestSubscription():
    """
    _retrieveAlcaHarvestSubscription_
    
    retrieve wmbs Subscription from the cache or database
    Core information should be constant (i.e. name, id) but might be needed to 
    loadData from caller for further processing 
    """
    if GlobalValues._alcaHarvestSubscription == None:
        GlobalValues._alcaHarvestSubscription =  \
            _getSubscription(Workflow(spec = "FileAlcaHarvest", 
                                      owner = "CMSTier0",
                                      name = "FileAlcaHarvest"),
                             Fileset(name = "AlcaHarvestable")
                             )
            
    return GlobalValues._alcaHarvestSubscription

def stackTraceLog(errorMessage):
    crashMessage = ""
    stackTrace = traceback.format_tb(sys.exc_info()[2], None)
    for stackFrame in stackTrace:
        crashMessage += stackFrame
    crashMessage = str(errorMessage) + crashMessage
    logging.error("""Error, Stack trace: \n\t%s""" % crashMessage)        
    
    return crashMessage
