#!/usr/bin/env python
"""
_Express_

Splitting algorithm for express processing.
"""

__revision__ = "$Id: Express.py,v 1.1 2011/01/05 23:57:25 hufnagel Exp $"
__version__  = "$Revision: 1.1 $"

import threading

from WMCore.WMBS.File import File

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUID import makeUUID
from WMCore.DAOFactory import DAOFactory


class Express(JobFactory):
    """
    Split jobs by set of files

    """
    def algorithm(self, groupInstance = None, jobInstance = None, *args,
                  **kwargs):
        """
        _algorithm_

        A file based splitting algorithm

        """
        # extract some global scheduling parameters
        self. maxInputEvents = kwargs['maxInputEvents']

        # flag to know whether a group has been created
        self.createdGroup = False

        # remember what streamers are used in jobs
        self.usedStreamers = []

        myThread = threading.currentThread()
        t0astDaoFactory = DAOFactory(package = "T0.State.Database",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)
        wmbsDaoFactory = DAOFactory(package = "WMCore.WMBS",
                                    logger = myThread.logger,
                                    dbinterface = myThread.dbi)

        getFilesDAO = t0astDaoFactory(classname = "ListFilesForExpress")
        filesDict = getFilesDAO.execute()

        #
        # What follows is a bit hacky due to the way we get express to work
        # with WMBS. Data discovery runs on streamer table, so we need to
        # update the status info there and actually populate the input
        # fileset for the express subscription
        #
        for run in sorted(filesDict.keys()):
            for stream_id, streamersByLumi in filesDict[run].items():
                self.defineJobs(streamersByLumi)

        if len(self.usedStreamers) > 0:

            bindsFilesetAdd = []
            bindsStreamerUpdate = []
            for id in self.usedStreamers:
                bindsStreamerUpdate.append({ 'id' : id })
                bindsFilesetAdd.append({ 'fileset' : self.subscription['fileset'].id,
                                         'fileid' :  id })

            addFilesToFilesetDAO = wmbsDaoFactory(classname = "Fileset.BulkAdd")
            addFilesToFilesetDAO.execute(binds = bindsFilesetAdd)

            markStreamersUsedDAO = t0astDaoFactory(classname = "MarkStreamersUsed")
            markStreamersUsedDAO.execute( binds = bindsStreamerUpdate )

        return


    def defineJobs(self, streamersByLumi):
        """
        _defineJobs_

        schedule jobs

        """
        myThread = threading.currentThread()
        myThread.logger.debug("defineJobs(): Running...")

        splitLumis = []

        for lumi in sorted(streamersByLumi.keys()):

            lumiStreamerList = streamersByLumi[lumi]

            splitLumis.append( { 'run_id' : lumiStreamerList[0]['run'],
                                 'lumi_id' : lumiStreamerList[0]['lumi'],
                                 'stream_id' : lumiStreamerList[0]['stream_id'] } )

            while len(lumiStreamerList) > 0:

                eventsTotal = 0
                streamerList = []

                for streamer in lumiStreamerList:

                    # if first streamer, always use it
                    if len(streamerList) == 0:
                        eventsTotal = streamer['events']
                        streamerList.append(streamer)
                    # otherwise calculate new totals and check if to use streamer
                    else:
                        newEventsTotal = eventsTotal + streamer['events']

                        if newEventsTotal <= self.maxInputEvents:
                            eventsTotal = newEventsTotal
                            streamerList.append(streamer)

                self.createJob(streamerList)

                for streamer in streamerList:
                    lumiStreamerList.remove(streamer)

        if len(splitLumis) > 0:

            myThread = threading.currentThread()
            t0astDaoFactory = DAOFactory(package = "T0.State.Database",
                                         logger = myThread.logger,
                                         dbinterface = myThread.dbi)

            insertSplitLumisDAO = t0astDaoFactory(classname = "InsertSplitLumis")
            insertSplitLumisDAO.execute( binds = splitLumis )

        return


    def createJob(self, streamerList):
        """
        _createJob_

        create a job processing streamer list

        """
        if not self.createdGroup:
            self.newGroup()
            self.createdGroup = True

        self.newJob(name = makeUUID())
        self.currentJob['mask'].setMaxAndSkipEvents(-1, 0)

        for streamer in streamerList:
            self.currentJob.addFile( File(id = streamer['id']) )
            self.usedStreamers.append(streamer['id'])
