#!/usr/bin/env python
"""
_Merge_

Splitting algorithm for merges.
Observes merge group (lumi section and parent file) boundaries.
"""

__revision__ = "$Id: Merge.py,v 1.9 2010/08/11 13:25:55 hufnagel Exp $"
__version__  = "$Revision: 1.9 $"

import threading

from WMCore.WMBS.File import File

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUID import makeUUID
from WMCore.DAOFactory import DAOFactory

# FIXEM: only needed for strict merge testing
from T0.State.Database.Reader import ListFiles


class Merge(JobFactory):
    """
    Split jobs by set of files

    """
    def algorithm(self, groupInstance = None, jobInstance = None, *args,
                  **kwargs):
        """
        _algorithm_

        A file based splitting algorithm, observing merge group
        (lumi section or parent file) boundaries.

        """
        # FIXME: only needed for strict merge testing
        self.t0astDBConn = kwargs['t0astDBConn']

        # extract some global scheduling parameters
        self.minMergeSize = kwargs['minMergeSize']
        self.maxOverMergeSize = kwargs['maxOverMergeSize']
        self.maxInputFiles = kwargs['maxInputFiles']
        self.maxEdmSize = kwargs['maxEdmSize']
        self.maxMergeEvents = kwargs['maxMergeEvents']

        # flag to know whether a group has been created
        self.createdGroup = False

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package = "T0.State.Database",
                                logger = myThread.logger,
                                dbinterface = myThread.dbi)

        getFilesDAO = daoFactory(classname = "ListFilesForMerge")
        filesDict = getFilesDAO.execute(self.subscription['id'])

        for run in sorted(filesDict.keys()):
            for datatier in filesDict[run].keys():

                forceMergeCloseout = filesDict[run][datatier]['closeout']

                for datasetpath in filesDict[run][datatier]['datasetpaths'].keys():

                    if kwargs['strictLumiOrder']:
                        self.defineStrictMergeJobs(run, datatier, datasetpath, forceMergeCloseout,
                                                   filesDict[run][datatier]['datasetpaths'][datasetpath].values(),
                                                   maxMergeSize = kwargs['maxMergeSize'])
                                               
                    else:
                        self.defineMergeJobs(run, datatier, datasetpath, forceMergeCloseout,
                                             filesDict[run][datatier]['datasetpaths'][datasetpath].values(),
                                             maxMergeSize = kwargs['maxMergeSize'])

        return


    def defineMergeJobs(self, run, datatier, datasetpath, forceMergeCloseout, mergeUnits, **params):
        """
        _defineMergeJobs_

        Given a set of mergeUnits that belong to the same run/dataset try to
        create merge jobs according to the scheduling parameters specified.

        """
        myThread = threading.currentThread()
        myThread.logger.debug("defineMergeJobs(): Running...")

        # DQM files ignore any size/event thresholds
        ignoreSize = False
        if datatier == "DQM" or datatier == "ALCAPROMPT":
            ignoreSize = True

        # merge scheduling parameters
        maxMergeSize = params['maxMergeSize']
        sendToErrorDataset = params.get('sendToErrorDataset', False)

        # only use maxEdmSize check for RAW files
        useMaxEdmSize = params.get('useMaxEdmSize', True) and ( datatier == 'RAW')

        mergeJobSize = 0
        mergeJobEvents = 0
        mergeJobInputFiles = 0
        mergeJobUnits = []

        for mergeUnit in mergeUnits:

            if not ignoreSize and useMaxEdmSize and mergeUnit['SIZE'] > self.maxEdmSize:

                # trigger merge on what we have to preserve order
                if len(mergeJobUnits) > 0:
                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = 0
                    mergeJobEvents = 0
                    mergeJobInputFiles = 0
                    mergeJobUnits = []

                # mergeUnit => new single file mergeUnits
                # then make recursive call
                newMergeUnits = []
                for mergeableFile in mergeUnit['FILES']:
                    newMergeUnits.append( { 'RUN' : run,
                                            'EVENTS' : mergeableFile['events'],
                                            'SIZE' : mergeableFile['filesize'],
                                            'FIRSTLUMI' : mergeableFile['first_lumi'],
                                            'LASTLUMI' : mergeableFile['last_lumi'],
                                            'FILES' : [mergeableFile] } )

                self.defineMergeJobs(run, datatier, datasetpath, True,
                                     newMergeUnits,
                                     maxMergeSize = self.maxEdmSize,
                                     useMaxEdmSize = False,
                                     sendToErrorDataset = True)

            elif ( not ignoreSize and mergeUnit['SIZE'] >= maxMergeSize ) or \
                     ( not ignoreSize and mergeUnit['EVENTS'] >= self.maxMergeEvents ) or \
                     len(mergeUnit['FILES']) >= self.maxInputFiles:

                # trigger merge on what we have to preserve order
                if len(mergeJobUnits) > 0:
                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = 0
                    mergeJobEvents = 0
                    mergeJobInputFiles = 0
                    mergeJobUnits = []

                # then issue merge on new mergeUnit
                self.createMergeJob([mergeUnit], sendToErrorDataset)

            else:

                newMergeJobSize = mergeJobSize + mergeUnit['SIZE']
                newMergeJobEvents = mergeJobEvents + mergeUnit['EVENTS']
                newMergeJobInputFiles = mergeJobInputFiles + len(mergeUnit['FILES'])

                # still safe with new file, just add it
                if ( ignoreSize or newMergeJobSize <= maxMergeSize ) and \
                       ( ignoreSize or newMergeJobEvents <= self.maxMergeEvents ) and \
                       newMergeJobInputFiles <= self.maxInputFiles:

                    mergeJobSize = newMergeJobSize
                    mergeJobEvents = newMergeJobEvents
                    mergeJobInputFiles = newMergeJobInputFiles
                    mergeJobUnits.append(mergeUnit)

                # over limits with new file, over minimum without it
                # issue merge job (regular)
                elif not ignoreSize and mergeJobSize > self.minMergeSize:

                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = mergeUnit['SIZE']
                    mergeJobEvents = mergeUnit['EVENTS']
                    mergeJobInputFiles = len(mergeUnit['FILES'])
                    mergeJobUnits = [mergeUnit]
                    

                # over limits with new file, below minimum without it
                # still below override limits (and below event limit)
                # add file, issue merge job (too large)
                elif not ignoreSize and \
                         newMergeJobSize <= self.maxOverMergeSize and \
                         newMergeJobEvents <= self.maxMergeEvents:

                    mergeJobUnits.append(mergeUnit)
                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = 0
                    mergeJobEvents = 0
                    mergeJobInputFiles = 0
                    mergeJobUnits = []

                # over limits with new file, below minimum without it
                # over override limit or event limit with new file
                # issue merge job (too small)
                else:

                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = mergeUnit['SIZE']
                    mergeJobEvents = mergeUnit['EVENTS']
                    mergeJobInputFiles = len(mergeUnit['FILES'])
                    mergeJobUnits = [mergeUnit]

        # finish out leftovers if we are in closeout
        if len(mergeJobUnits) > 0 and forceMergeCloseout:
            self.createMergeJob(mergeJobUnits, sendToErrorDataset)

        return


    def defineStrictMergeJobs(self, run, datasetpath, forceMergeCloseout, mergeUnits, **params):
        """
        _defineStrictMergeJobs_

        Given a set of mergeUnits that belong to the same run/dataset try to
        create merge jobs according to the scheduling parameters specified.

        nforce strict lumi ordering.

        """
        myThread = threading.currentThread()
        myThread.logger.debug("defineStrictMergeJobs(): Running...")

        # merge scheduling parameters
        maxMergeSize = params['maxMergeSize']
        sendToErrorDataset = params.get('sendToErrorDataset', False)

        # only use maxEdmSize check for RAW files
        useMaxEdmSize = params.get('useMaxEdmSize', True) and ( datatier == 'RAW')

        mergeJobSize = 0
        mergeJobEvents = 0
        mergeJobInputFiles = 0
        mergeJobUnits = []

        # get largest lumi with merge jobs for run/datasetpath
        lastLumi = ListFiles.getMaxMergeLumi(self.t0astDBConn, run, datasetpath)

        for mergeUnit in mergeUnits:

##             myThread.logger.info("DEBUG mergeUnit lastLumi = %d" % lastLumi)
##             myThread.logger.info("DEBUG mergeUnit firstLumi = %d" % firstLumi)

            # if we miss a lumi => bail unless we are in forceMergeCloseout
            if mergeUnit['FIRSTLUMI'] != lastLumi + 1:
                if forceMergeCloseout:
                    if len(mergeJobUnits) > 0:
                        self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                        mergeJobSize = 0
                        mergeJobEvents = 0
                        mergeJobInputFiles = 0
                        mergeJobUnits = []
                else:
                    break

            lastLumi = mergeUnit['LASTLUMI']

            #
            # merge unit is above single lumi threshold
            #
            # => handle it individually, split and route to error dataset
            #
            if useMaxEdmSize and mergeUnit['SIZE'] > self.maxEdmSize:

                # trigger merge on what we have to preserve order
                if len(mergeJobUnits) > 0:
                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = 0
                    mergeJobEvents = 0
                    mergeJobInputFiles = 0
                    mergeJobUnits = []

                # mergeUnit => new single file mergeUnits
                # then make recursive call
                newMergeUnits = []
                for mergeableFile in mergeUnit['FILES']:
                    newMergeUnits.append( { 'RUN' : run,
                                            'EVENTS' : mergeableFile['events'],
                                            'SIZE' : mergeableFile['size'],
                                            'FIRSTLUMI' : mergeableFile['first_lumi'],
                                            'LASTLUMI' : mergeableFile['last_lumi'],
                                            'FILES' : [mergeableFile] } )

                self.defineMergeJobs(run, datasetpath,
                                     newMergeUnits,
                                     maxMergeSize = self.maxEdmSize,
                                     useMaxEdmSize = False,
                                     sendToErrorDataset = True,
                                     forceMergeCloseout = True)

##             elif mergeUnit['SIZE'] >= maxMergeSize or \
##                    mergeUnit['EVENTS'] >= self.maxMergeEvents or \
##                    len(mergeUnit['FILES']) >= self.maxInputFiles:

##                 # trigger merge on what we have to preserve order
##                 if len(mergeJobUnits) > 0:
##                     self.createMergeJob(mergeJobUnits, sendToErrorDataset)

##                 # then issue merge on new mergeUnit
##                 self.createMergeJob([mergeUnit], sendToErrorDataset)

##                 mergeJobSize = 0
##                 mergeJobEvents = 0
##                 mergeJobInputFiles = 0
##                 mergeJobUnits = []

            #
            # merge unit is below single lumi threshold, try to combine
            #
            else:

                newMergeJobSize = mergeJobSize + mergeUnit['SIZE']
                newMergeJobEvents = mergeJobEvents + mergeUnit['EVENTS']
                newMergeJobInputFiles = mergeJobInputFiles + len(mergeUnit['FILES'])

                # still safe with new merge unit, just add it
                if newMergeJobSize <= maxMergeSize and \
                       newMergeJobEvents <= self.maxMergeEvents and \
                       newMergeJobInputFiles <= self.maxInputFiles:

                    mergeJobSize = newMergeJobSize
                    mergeJobEvents = newMergeJobEvents
                    mergeJobInputFiles = newMergeJobInputFiles
                    mergeJobUnits.append(mergeUnit)

                # over limits with new merge unit, over minimum without it
                # issue merge job (regular)
                elif mergeJobSize > self.minMergeSize:

                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = mergeUnit['SIZE']
                    mergeJobEvents = mergeUnit['EVENTS']
                    mergeJobInputFiles = len(mergeUnit['FILES'])
                    mergeJobUnits = [mergeUnit]

                # over limits with new file, below minimum without it
                # still below override limits (and below event limit)
                # add file, issue merge job (too large)
                elif newMergeJobSize <= self.maxOverMergeSize and \
                         newMergeJobEvents <= self.maxMergeEvents:

                    mergeJobUnits.append(mergeUnit)
                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = 0
                    mergeJobEvents = 0
                    mergeJobInputFiles = 0
                    mergeJobUnits = []

                # over limits with new file, below minimum without it
                # over override limit or event limit with new file
                # issue merge job (too small)
                else:

                    self.createMergeJob(mergeJobUnits, sendToErrorDataset)
                    mergeJobSize = mergeUnit['SIZE']
                    mergeJobEvents = mergeUnit['EVENTS']
                    mergeJobInputFiles = len(mergeUnit['FILES'])
                    mergeJobUnits = [mergeUnit]

        # finish out leftovers if we are in closeout
        if len(mergeJobUnits) > 0 and forceMergeCloseout:
            self.createMergeJob(mergeJobUnits, sendToErrorDataset)

        return


    def createMergeJob(self, mergeUnits, sendToErrorDataset):
        """
        _createMergeJob_

        Create a merge job from list of mergeUnits.

        """
        if not self.createdGroup:
            self.newGroup()
            self.createdGroup = True

        self.newJob(name = makeUUID())

        # dirty hack to forward sendToErrorDataset to the
        # job creation logic in Tier0Merger
        #
        # FirstEvent is -1 => job outputs to error dataset
        #
        if sendToErrorDataset:
            self.currentJob['mask'].setMaxAndSkipEvents(-1, -1)
        else:
            self.currentJob['mask'].setMaxAndSkipEvents(-1, 0)

        for mergeUnit in mergeUnits:
            for f in mergeUnit['FILES']:
                self.currentJob.addFile( File(id = f['id']) )
