#!/usr/bin/env python
"""
_PromptCalib_

Splitting algorithm for PromptCalib jobs that observes the available files
for the different prompt calib workflows and builds job groups.
The split algorithm plus parameter is read from the calib config table.
"""

__revision__ = "$Id: PromptCalib.py,v 1.3 2010/04/14 10:58:57 naderer Exp $"
__version__  = "$Revision: 1.3 $"

from sets import Set
import threading
import logging

from WMCore.WMBS.File import File

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUID import makeUUID
from WMCore.DAOFactory import DAOFactory

class PromptCalib(JobFactory):
    """
    Split jobs by set of files
    
    """
    def algorithm(self, groupInstance = None, jobInstance = None, *args,
                  **kwargs):
        """
        _algorithm_

        The splitting algorithm reads the filesets and starts the 
        corresponding split algorithm. There one jobgroup is created and,
        depending on the split parameter, one or more jobs.
        In each split algorithm the available files for the jobs are added
        with ascending lumi section.

        """
        myThread = threading.currentThread()
        daoFactory = DAOFactory(package = "T0.State.Database",
                                logger = myThread.logger,
                                dbinterface = myThread.dbi)

        getFilesDAO = daoFactory(classname = "ListFilesForPromptCalib")
        fileset = getFilesDAO.execute(self.subscription["id"])

        jobfiles = []
        i = 0
        while i < len(fileset):
            curr_f = fileset[i]
            if i == len(fileset)-1:
                next_f = {}
                next_f['primds_id'] = next_f['procds_id'] = None
                # next_f['config_url'] = None
                # next_f['split_algo'] = next_f['split_param'] = None
                
            else:
                next_f = fileset[i+1]
            jobfiles.append(curr_f['file_id'])
            something_changed = (curr_f['primds_id'] != next_f['primds_id'] or
                                 curr_f['procds_id'] != next_f['procds_id'])

            #                  or curr_f['split_algo'] != next_f['split_algo'] or
            #                     curr_f['split_param'] != next_f['split_param'] or 
            #                     curr_f['config_url'] != next_f['config_url'])
            
            # If only split_algo, split_param or the config_url changes, we would
            # create different PromptCalib workflows that will use the same input file.
            # This cannot be handled within the context of a single subscription
            # -> would need a subscription for every PromptCalib workflow.
            
            if something_changed:
                if curr_f['split_algo'] == 'FileBased':
                    self.FileBasedJobSplitting(int(curr_f['split_param']), jobfiles)
                elif curr_f['split_algo'] == 'EventBased':
                    self.EventBasedJobSplitting(int(curr_f['split_param']), jobfiles)
                elif curr_f['split_algo'] == 'LumiBased':
                    self.LumiBasedJobSplitting(int(curr_f['split_param']), jobfiles)
                else:
                    logging.error("I'm not prepared for %s as split_algo." %
                                  curr_f['split_algo'])
                    logging.error("Should be either FileBased, EventBased or LumiBased!")
                jobfiles = []
                
            i+=1
           
        return


    def FileBasedJobSplitting(self, filesPerJob, jobfile_ids):
        """
        _FileBasedJobSplitting_

        Split up the files such that each job
        will process a maximum of 'filesPerJob'.
        """

        self.newGroup()
        baseName = makeUUID()
        filesInJob = 0
        totalJobs = 1
        
        for file_id in jobfile_ids:
            if filesInJob == 0 or filesInJob == filesPerJob:
                self.newJob(name = '%s-%i' % (baseName, totalJobs))
                filesInJob = 0
                totalJobs += 1

            filesInJob += 1
            loadedFile = File(id=file_id)
            loadedFile.load()
            self.currentJob.addFile(loadedFile)    
        return


    def EventBasedJobSplitting(self, eventsPerJob, jobfile_ids):
        """
        _EventBasedJobSplitting_

        Split up the files such that each job
        will process a maximum of 'eventsPerJob'.
        """

        self.newGroup()
        self.currentJob = None
        baseName = makeUUID()
        totalJobs = 1
        
        for file_id in jobfile_ids:
            f = File(id = file_id)
            f.load()
            eventsInFile = f['events']
            currentEvent = 0

            if eventsInFile >= eventsPerJob:
                    currentEvent   = 0
                    while currentEvent < eventsInFile:
                        self.newJob(name = '%s-%i' % (baseName, totalJobs))
                        totalJobs += 1
                        self.currentJob.addFile(f)
                        self.currentJob["mask"].setMaxAndSkipEvents(eventsPerJob,
                                                                    currentEvent)
                        currentEvent += eventsPerJob
                    currentEvent = 0
            else:
                if currentEvent + eventsInFile > eventsPerJob:
                    #Create new jobs, because we are out of room
                    self.currentJob["mask"].setMaxAndSkipEvents(eventsPerJob, 0)
                    self.newJob(name = '%s-%i' % (baseName, totalJobs))
                    totalJobs += 1
                    currentEvent = 0

                if currentEvent + eventsInFile <= eventsPerJob:
                    #Add if it will be smaller
                    if self.currentJob:
                        self.currentJob.addFile(f)
                    else:
                        self.newJob(name = '%s-%i' % (baseName, totalJobs))
                        totalJobs += 1
                        self.currentJob.addFile(f)
                    currentEvent += eventsInFile

        if currentEvent > 0:
            self.currentJob["mask"].setMaxAndSkipEvents(eventsPerJob, 0)     
        return
    

    def LumiBasedJobSplitting(self, lumisPerJob, jobfile_ids):
        """
        _LumiBasedJobSplitting_
        
        Split up the files such that each job
        will process a maximum of 'lumisPerJob'.
        """
        
        self.newGroup()
        self.currentJob = None
        baseName = makeUUID()
        totalJobs = 1
        currentLumis = 0
        
        for file_id in jobfile_ids:
            f = File(id = file_id)
            f.loadData()
            lumisInFile = list(f["runs"])[0].lumis

            if len(lumisInFile) >= lumisPerJob:
                    currentLumis   = 0
                    while currentLumis < len(lumisInFile):
                        self.newJob(name = '%s-%i' % (baseName, totalJobs))
                        totalJobs += 1
                        self.currentJob.addFile(f)
                        self.currentJob["mask"].setMaxAndSkipLumis(lumisPerJob,
                                                                   currentLumis)
                        currentLumis += lumisPerJob
                    currentLumis = 0
            else:
                if currentLumis + len(lumisInFile) > lumisPerJob:
                    #Create new jobs, because we are out of room
                    self.currentJob["mask"].setMaxAndSkipLumis(lumisPerJob, 0)
                    self.newJob(name = '%s-%i' % (baseName, totalJobs))
                    totalJobs += 1
                    currentLumis = 0

                if currentLumis + len(lumisInFile) <= lumisPerJob:
                    #Add if it will be smaller
                    if self.currentJob:
                        self.currentJob.addFile(f)
                    else:
                        self.newJob(name = '%s-%i' % (baseName, totalJobs))
                        totalJobs += 1
                        self.currentJob.addFile(f)
                    currentLumis += len(lumisInFile)

        if currentLumis > 0:
            self.currentJob["mask"].setMaxAndSkipLumis(lumisPerJob, 0)
        return
