#!/usr/bin/env python
"""
_Repack_

Splitting algorithm for repacking.
"""

__revision__ = "$Id: Repack.py,v 1.2 2011/01/05 23:56:53 hufnagel Exp $"
__version__  = "$Revision: 1.2 $"

import threading

from WMCore.WMBS.File import File

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUID import makeUUID
from WMCore.DAOFactory import DAOFactory


class Repack(JobFactory):
    """
    Split jobs by set of files

    """
    def algorithm(self, groupInstance = None, jobInstance = None, *args,
                  **kwargs):
        """
        _algorithm_

        A file based splitting algorithm

        """
        # extract some global scheduling parameters
        self.maxStreamerSizeSingleLumi = kwargs['maxStreamerSizeSingleLumi']
        self.maxStreamerEventsSingleLumi = kwargs['maxStreamerEventsSingleLumi']
        self.maxStreamerSizeMultiLumi = kwargs['maxStreamerSizeMultiLumi']
        self.maxStreamerEventsMultiLumi = kwargs['maxStreamerEventsMultiLumi']
        self.maxStreamerCount = kwargs['maxStreamerCount']

        # flag to know whether a group has been created
        self.createdGroup = False

        # remember what streamers are used in jobs
        self.usedStreamers = []

        myThread = threading.currentThread()
        t0astDaoFactory = DAOFactory(package = "T0.State.Database",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)
        wmbsDaoFactory = DAOFactory(package = "WMCore.WMBS",
                                    logger = myThread.logger,
                                    dbinterface = myThread.dbi)

        getFilesDAO = t0astDaoFactory(classname = "ListFilesForRepack")
        filesDict = getFilesDAO.execute()

        #
        # What follows is a bit hacky due to the way we get repacking to work
        # with WMBS. Data discovery runs on streamer table, so we need to
        # update the status info there and actually populate the input
        # fileset for the repacking subscription
        #
        for run in sorted(filesDict.keys()):
            forceRepackCloseout = filesDict[run]['closeout']
            for stream_id, streamersByLumi in filesDict[run]['streams'].items():
                self.defineJobs(streamersByLumi, forceRepackCloseout)

        if len(self.usedStreamers) > 0:

            bindsFilesetAdd = []
            bindsStreamerUpdate = []
            for id in self.usedStreamers:
                bindsStreamerUpdate.append({ 'id' : id })
                bindsFilesetAdd.append({ 'fileset' : self.subscription['fileset'].id,
                                         'fileid' :  id })

            addFilesToFilesetDAO = wmbsDaoFactory(classname = "Fileset.BulkAdd")
            addFilesToFilesetDAO.execute(binds = bindsFilesetAdd)

            markStreamersUsedDAO = t0astDaoFactory(classname = "MarkStreamersUsed")
            markStreamersUsedDAO.execute( binds = bindsStreamerUpdate )

        return


    def defineJobs(self, streamersByLumi, forceRepackCloseout):
        """
        _defineJobs_

        schedule jobs

        """
        myThread = threading.currentThread()
        myThread.logger.debug("defineJobs(): Running...")

        jobSizeTotal = 0
        jobEventsTotal = 0
        jobStreamerList = []

        splitLumis = []

        for lumi in sorted(streamersByLumi.keys()):

            lumiStreamerList = streamersByLumi[lumi]

            # calculate new total size and event
            lumiSizeTotal = 0
            lumiEventsTotal = 0
            for streamer in lumiStreamerList:
                lumiEventsTotal += streamer['events']
                lumiSizeTotal += streamer['filesize']
                        
            # lumi is larger than split limits
            #
            # => handle lumi individually and split
            #
            if lumiSizeTotal > self.maxStreamerSizeSingleLumi or \
                   lumiEventsTotal > self.maxStreamerEventsSingleLumi:

                splitLumis.append( { 'run_id' : lumiStreamerList[0]['run'],
                                     'lumi_id' : lumiStreamerList[0]['lumi'],
                                     'stream_id' : lumiStreamerList[0]['stream_id'] } )

                while len(lumiStreamerList) > 0:

                    eventsTotal = 0
                    sizeTotal = 0
                    streamerList = []
                    for streamer in lumiStreamerList:
                    
                        # if first streamer, always use it
                        if len(streamerList) == 0:
                            eventsTotal = streamer['events']
                            sizeTotal = streamer['filesize']
                            streamerList.append(streamer)
                        # otherwise calculate new totals and check if to use streamer
                        else:
                            newEventsTotal = eventsTotal + streamer['events']
                            newSizeTotal = sizeTotal + streamer['filesize']

                            if newSizeTotal <= self.maxStreamerSizeSingleLumi and \
                                   newEventsTotal <= self.maxStreamerEventsSingleLumi:
                                eventsTotal = newEventsTotal
                                sizeTotal = newSizeTotal
                                streamerList.append(streamer)
                        
                    self.createJob(streamerList)

                    for streamer in streamerList:
                        lumiStreamerList.remove(streamer)

            # lumi is smaller than split limits
            # check if it can be combined with previous lumi(s)
            #
            # yes => just add lumi to job (with an additional order check)
            #
            # no => issue job for previous lumi(s), save current for next job
            #
            else:

                newSizeTotal = jobSizeTotal + lumiSizeTotal
                newEventsTotal = jobEventsTotal + lumiEventsTotal
                newStreamerCount = len(jobStreamerList) + len(lumiStreamerList)

                # always take the first one
                if len(jobStreamerList) == 0:

                    jobSizeTotal = newSizeTotal
                    jobEventsTotal = newEventsTotal
                    jobStreamerList.extend(lumiStreamerList)

                elif newSizeTotal <= self.maxStreamerSizeMultiLumi and \
                         newEventsTotal <= self.maxStreamerEventsMultiLumi and \
                         newStreamerCount <= self.maxStreamerCount:

                    jobSizeTotal = newSizeTotal
                    jobEventsTotal = newEventsTotal
                    jobStreamerList.extend(lumiStreamerList)

                else:

                    self.createJob(jobStreamerList)

                    jobSizeTotal = lumiSizeTotal
                    jobEventsTotal = lumiEventsTotal
                    jobStreamerList = lumiStreamerList

        # if we are in closeout issue repack job for leftovers
        if len(jobStreamerList) > 0 and forceRepackCloseout:
            self.createJob(jobStreamerList)

        if len(splitLumis) > 0:

            myThread = threading.currentThread()
            t0astDaoFactory = DAOFactory(package = "T0.State.Database",
                                         logger = myThread.logger,
                                         dbinterface = myThread.dbi)

            insertSplitLumisDAO = t0astDaoFactory(classname = "InsertSplitLumis")
            insertSplitLumisDAO.execute( binds = splitLumis )

        return


    def defineStrictJobs(self, streamersByLumi, forceRepackCloseout):
        """
        _defineStrictJobs_

        schedule jobs

        """
        myThread = threading.currentThread()
        myThread.logger.debug("defineStrictJobs(): Running...")

        jobSizeTotal = 0
        jobEventsTotal = 0
        jobStreamerList = []

        # get largest lumi with repack jobs for run/stream FIXME !!!
        #lastLumi = getMaxRepackLumi(dbConn, run, stream)
        lastLumi = 0

        for lumi in sorted(streamersByLumi.keys()):

            lumiStreamerList = streamersByLumi[lumi]

            # calculate lumi size and event count
            lumiSizeTotal = 0
            lumiEventsTotal = 0
            for streamer in lumiStreamerList:
                lumiEventsTotal += streamer['events']
                lumiSizeTotal += streamer['filesize']

            # if we miss a lumi => bail unless we are in forceMergeCloseout
            if lumi != lastLumi + 1:
                if forceRepackCloseout:
                    if len(jobStreamerList) > 0:
                        self.createJob(jobStreamerList)
                        jobSizeTotal = 0
                        jobEventsTotal = 0
                        jobStreamerList = []
                else:
                    break

            lastLumi = lumi

            # lumi is larger than split limits
            #
            # => handle lumi individually and split
            #
            if lumiSizeTotal > self.maxStreamerSizeSingleLumi or \
                   lumiEventsTotal > self.maxStreamerEventsSingleLumi:

                # trigger merge on what we have to preserve order
                if len(jobStreamerList) > 0:
                    self.createJob(streamerList)
                    jobSizeTotal = 0
                    jobEventsTotal = 0
                    jobStreamerList = []

                while len(lumiStreamerList) > 0:

                    eventsTotal = 0
                    sizeTotal = 0
                    streamerList = []
                    for streamer in lumiStreamerList:

                        # if first streamer, always use it
                        if len(streamerList) == 0:
                            eventsTotal = streamer['events']
                            sizeTotal = streamer['filesize']
                            streamerList.append(streamer)
                        # otherwise calculate new totals and check if to use streamer
                        else:
                            newEventsTotal = eventsTotal + streamer['events']
                            newSizeTotal = sizeTotal + streamer['filesize']                        

                            if newSizeTotal <= self.maxStreamerSizeSingleLumi and \
                                   newEventsTotal <= self.maxStreamerEventsSingleLumi:
                                eventsTotal = newEventsTotal
                                sizeTotal = newSizeTotal
                                streamerList.append(streamer)

                    self.createJob(streamerList)

                    for streamer in streamerList:
                        lumiStreamerList.remove(streamer)

            # lumi is smaller than split limits
            # check if it can be combined with previous lumi(s)
            #
            # yes => just add lumi to job (with an additional order check)
            #
            # no => issue job for previous lumi(s), save current for next job
            #
            else:

                newSizeTotal = jobSizeTotal + lumiSizeTotal
                newEventsTotal = jobEventsTotal + lumiEventsTotal
                newStreamerCount = len(jobStreamerList) + len(lumiStreamerList)

                # still safe with new lumi, just add it
                if newSizeTotal <= self.maxStreamerSizeMultiLumi and \
                       newEventsTotal <= self.maxStreamerEventsMultiLumi and \
                       newStreamerCount <= self.maxStreamerCount:

                    jobSizeTotal = newSizeTotal
                    jobEventsTotal = newEventsTotal
                    jobStreamerList.extend(lumiStreamerList)

                # over limits with new lumi, issue repack job
                else:

                    self.createJob(jobStreamerList)

                    jobSizeTotal = lumiSizeTotal
                    jobEventsTotal = lumiEventsTotal
                    jobStreamerList = lumiStreamerList

        # if we are in closeout issue repack job for leftovers
        if len(jobStreamerList) > 0 and forceRepackCloseout:
            self.createJob(jobStreamerList)

        return


    def createJob(self, streamerList):
        """
        _createJob_

        create a job processing streamer list

        """
        if not self.createdGroup:
            self.newGroup()
            self.createdGroup = True

        self.newJob(name = makeUUID())
        self.currentJob['mask'].setMaxAndSkipEvents(-1, 0)

        for streamer in streamerList:
            self.currentJob.addFile( File(id = streamer['id']) )
            self.usedStreamers.append(streamer['id'])
