#!/usr/bin/env python
"""
_PromptCalibInjectorComponent_

Poll T0AST for ALCARECO files and generate PromptCalib workflows and jobspecs for them.
Inject the jobs into a job queue while they wait for calib configs.
"""

__revision__ = "$Id: PromptCalibInjectorComponent.py,v 1.13 2010/04/14 10:50:39 naderer Exp $"
__version__ = "$Revision: 1.13 $"


from sets import Set
import logging
import os
import shutil
import random
import time
import sys

from JobQueue import JobQueueDB

from T0.State.Database.Reader import ListFiles
from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListLumiSections
from T0.State.Database.Reader import ListRuns
from T0.State.Database.Writer import InsertJob
from T0.State.Database.Writer import InsertRun

from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig

from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup

from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUID import makeUUID

from T0.GenericTier0.GenericTier0Component import GenericTier0Component
from T0.JobFactory.PromptCalibJob import PromptCalibJob
from T0.WorkflowFactory.PromptCalibWorkflow import PromptCalibWorkflow

from T0.Globals import retrievePromptCalibSubscription

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec


class PromptCalibInjectorComponent(GenericTier0Component):
    """
    _PromptCalibInjectorComponent_

    Poll T0AST for alcareco files and generate workflows and jobspecs for them.
    Inject the jobs into a job queue while they wait for Calib configs.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)

        # Data structure to hold WorkflowSpecs.  This is a two level
        # dictionary where the key to the first level is the run number,
        # to the second level is the dataset path id.
        self.workflows = {}
        self.files = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2
        
        # Lookup some parameters from the config file:
        #   scramArch - SCRAM Architecture
        #   cmsPath - Path to CMS software
        #   calibTimeout - The amount of time to hold PromptCalib jobs for
        #   lumiThreshold - Lumi percentage for jobs to be created immediately
        self.scramArch = self.lookupParameter("scramArch", "slc5_ia32_gcc434")
        self.cmsPath = self.lookupParameter("cmsPath", "/afs/cern.ch/cms/sw")
        self.calibTimeout = int(self.lookupParameter("calibTimeout", 300))
        self.lumiThreshold = int(self.lookupParameter("lumiThreshold", 100))
        
        # The job cache directory is where all workflows, jobspecs and configs
        # are stored.
        self.jobCacheDir = self.lookupParameter("JobCacheDir",
                                                self.args["ComponentDir"])

        # WMCore splitter factory for splitting up PromptCalib jobs.
        self.splitterFactory = SplitterFactory(package='T0.JobSplitting')

        # Load up the calib subscription for easy access.
        self.promptCalibSubscription = None

        # See if lazy downloads are enabled.  
        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        self.jobGroupCache = {}
        self.jobGroupToRunCache = {}
        self.jobGroupToDatasetCache = {}
        
        logging.info("PromptCalib Injector Component Started")
        
    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "PromptCalibInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "PromptCalibInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "PromptCalibInjector:pollForFiles":
            self.pollForFiles()
        else:
            logging.error("Unknown event: %s, payload: %s" % (event, payload))
        
        return

    def retrieveWorkflow(self, runNumber, datasetpathID):
        """
        _retrieveWorkflow_

        Retrieve a WorkflowSpec from the cache.  If a workflow is not in the
        cache it will be created.  Returns None if an error occured while
        trying to create the workflow.
        """
        if runNumber not in self.workflows.keys():
            self.workflows[runNumber] = {}
        if datasetpathID not in self.workflows[runNumber].keys():
            self.workflows[runNumber][datasetpathID] = None

        existingWorkflow = self.workflows[runNumber][datasetpathID]

        if existingWorkflow != None:
            return existingWorkflow

        newWorkflow = self.createWorkflow(runNumber, datasetpathID)
        self.workflows[runNumber][datasetpathID] = newWorkflow

        return newWorkflow

    def createWorkflow(self, runNumber, datasetpathID):
        """
        _createWorkflow_

        Create a workflow for a particular run number and datasetpathID
        """
        logging.debug("createWorkflow(): Running run %s, datasetpath %s" % \
                      (runNumber, datasetpathID))

        # no check needed, done before
        runConfig = self.getRunConfig(runNumber)

        datasetIDs = ListDatasets.listDatasetIDsFromID(self.t0astDBConn,
                                                       datasetpathID)
        primds_id = datasetIDs["PRIMARY"]
        primds = runConfig.getPrimaryDatasetName(primds_id)

        procds = ListDatasets.listProcessedDatasetNameFromID(self.t0astDBConn,
                                                             datasetIDs["PROCESSED"])
        alcaStream = procds.split("-")[1]
        
        calibConfig = runConfig.getCalibConfiguration(primds_id, procds)
        if calibConfig == None:
            logging.error("No calib config for run %s, dataset %s, alca stream %s" % \
                          (runNumber, primds_id, alcaStream))
            return None

        configFile = runConfig.retrieveConfigFromURL(calibConfig.ConfigURL)
        if configFile == None:
            logging.error("No config file for run %s, dataset %s, alca stream %s" % \
                          (runNumber, primds_id, alcaStream))
            return None
        
        # extract the name of the calib task from the config file url
        #    up to now: calib task name = config file name
        #    last part looks like '.../BeamFit_long_worflow.py?revision=1.3'
        # and get revision info, used in configuration metadata
        # should look like: $Revision=1.3$
        temp = calibConfig.ConfigURL.split("/")[-1]
        try:
            calibTask, configFile_rev = temp.split("?")
            configFile_rev = '$'+configFile_rev.capitalize()+'$'
        except ValueError:
            calibTask = temp
            configFile_rev = '$unknown$'

        calibTask = calibTask.replace('.py','')
        calibTask = calibTask.replace('_cfg','')
        
        outModuleInfo={ "process":"PromptCalib",
                        "dataTier":"ALCARECO",
                        "calibTask":calibTask,
                        "runNumber":runNumber,
                        "dataset":primds,
                        "alcaStream":alcaStream,
                        "globalTag":calibConfig.GlobalTag,
                        "configFile_rev":configFile_rev
                        }

        workflowName = "PromptCalib%s-Run%s-%s" % (calibTask,runNumber, alcaStream)
        workflowDir = os.path.join(self.jobCacheDir, "workflows", "Run%d" % runNumber)
        workflowSpecFile = os.path.join(workflowDir, "%s-workflow.xml" % workflowName)

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No workflow for run %s, dataset %s, alca stream %s" %
                                           (runNumber, primds, alcaStream))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowSpecFile)

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return workflowSpec


        workflow = PromptCalibWorkflow( runNumber,
                                        calibConfig.CMSSWVersion,
                                        calibConfig.GlobalTag,
                                        self.cmsPath,
                                        self.scramArch,
                                        outModuleInfo)
        
        workflowSpec = workflow.makeWorkflowSpec(workflowName,
                                                 configFile,
                                                 self.useLazyDownload,
                                                 production = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)
        
        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)

        workflowSpec.save(workflowSpecFile)

        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)
         
    def createAndSubmitJobs(self, runNumber, datasetpathID, wmbsJobGroups):
        """
        _createAndSubmitJob_

        Loop over all job groups and process them
        (they are all for the same run)

        Each job group can consist of one or more jobs

        Generate job specs for every job and queue them in JobQueue
        Update job name in WMBS
        """
        if len(wmbsJobGroups) == 0:
            logging.debug("No job groups to process, nothing to do")
            return

        runConfig = self.getRunConfig(runNumber)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %s" % \
                          runNumber)
            return None
        
        datasetNames = ListDatasets.listDatasetNamesFromID(self.t0astDBConn,
                                                           datasetpathID)
        procds = datasetNames["PROCESSED"]
        alcaStream = datasetNames["PROCESSED"].split('-')[1]
        
        # loop over all job groups and process them
        for wmbsJobGroup in wmbsJobGroups:

            if len(wmbsJobGroup.jobs) == 0:
                logging.error("WMBSJobGroup %s has no jobs" % wmbsJobGroup)
                continue

            newWorkflow = self.retrieveWorkflow(runNumber, datasetpathID)
            if newWorkflow == None:
                logging.error("No workflow for run %s, datasetpath %s" % (runNumber,
                                                                          datasetpathID))
                continue

            jobSpecBaseDir = os.path.join(self.jobCacheDir, "jobspecs",
                                          "Run%d" % runNumber)

            jobFactory = PromptCalibJob(newWorkflow)

            # create job specs (split this out to avoid
            # file handling inside database transaction)
            createdJobSpecs = {}
            for wmbsJob in wmbsJobGroup.jobs:

                wmbsJob.getMask()
                try:
                    maxEvents = wmbsJob["mask"]["LastEvent"] - wmbsJob["mask"]["FirstEvent"]
                    skipEvents = wmbsJob["mask"]["FirstEvent"]
                except TypeError:
                    maxEvents = -1
                    skipEvents = 0
                try:
                    maxLumis = wmbsJob["mask"]["LastLumi"] - wmbsJob["mask"]["FirstLumi"]
                    skipLumis = wmbsJob["mask"]["FirstLumi"]
                except TypeError:
                    maxLumis = None
                    skipLumis = None
                
                jobSpec = jobFactory(wmbsJob["id"],
                                     alcaStream,
                                     wmbsJob.getFiles(),
                                     maxEvents, skipEvents,
                                     maxLumis, skipLumis,
                                     jobSpecBaseDir)

                createdJobSpecs[wmbsJob["id"]] = jobSpec

            try:
                self.t0astDBConn.beginTransaction()
                self.paDBConn.beginTransaction()
				
                for wmbsJob in wmbsJobGroup.jobs:

                    jobSpec = createdJobSpecs[wmbsJob["id"]]

                    wmbsJob["name"] = jobSpec.parameters["JobName"]

                    wmbsJob.save()

                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters["JobSpecFile"],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority, siteName = None,
                                             status = "new")
                                             
                    self.stateChanger.propagate(wmbsJob, 'created', 'new')
                    
                # cleanup memory cache of job groups and their run/dataset
                del self.jobGroupCache[wmbsJobGroup.id]
                del self.jobGroupToRunCache[wmbsJobGroup.id]
                del self.jobGroupToDatasetCache[wmbsJobGroup.id]
                logging.info("DEBUG Removed JobGroup %s from cache" %  wmbsJobGroup.id)

                self.t0astDBConn.commit()
                self.paDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(createdJobSpecs.keys()))
            except StandardError, ex:
                self.paDBConn.rollback()
                self.t0astDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        return


    def pollForFiles(self):
        """
        _pollForFiles_

        Poll T0AST for files with prompt calib subscription, start timer,
        collect the lumis from the file, call PromptCalib job splitter,
        cache the job groups and pass them off so jobspecs can be created.
        """
        
        logging.debug("pollAlcaRecoFiles(): Running...")
        self.t0astDBConn.beginTransaction()     
        availableFiles = self.promptCalibSubscription.availableFiles()

        for availableFile in availableFiles:
            availableFile.loadData()

            run = list(availableFile["runs"])[0].run            
            runConfig = self.getRunConfig(run)
            if runConfig == None:
                logging.error("Could not retrieve run config for run %s" % run)
                continue
            
            datasetpathID = ListDatasets.listDatasetIDForWMBSFile(self.t0astDBConn,
                                                              availableFile["id"])
            alldatasetIDs = ListDatasets.listDatasetIDsFromID(self.t0astDBConn,
                                                              datasetpathID)
            primds_id = alldatasetIDs["PRIMARY"]
            procds = ListDatasets.listProcessedDatasetNameFromID(self.t0astDBConn,
                                                                 alldatasetIDs["PROCESSED"])
            
            # we only play with those files, where we have a config entry in the database
            calibConfig = runConfig.getCalibConfiguration(primds_id, procds)
            if calibConfig == None:
                logging.error("No calib config for run %s, dataset %s, proc. dataset %s" % \
                              (run, primds_id, procds))
                self.promptCalibSubscription.completeFiles(availableFile)
                continue
            else:
                # the first file per run/datasetpathID starts a timer
                # after calibTimeout, job splitter will be startet at the latest
                if run not in self.files.keys():
                    self.files[run]={}
                if datasetpathID not in self.files[run].keys():
                    self.files[run][datasetpathID]={}
                    self.files[run][datasetpathID]['start'] = int(time.time())
                    self.files[run][datasetpathID]['lumis'] = set()
                for lumi in list(availableFile["runs"])[0].lumis:
                    self.files[run][datasetpathID]['lumis'].add(lumi)
                logging.info("availableFile: %s, procds: %s, lumis %s" %
                        (availableFile["id"], procds, list(availableFile["runs"])[0].lumis))
        
        for run in self.files.keys():
            for datasetpathID in self.files[run].keys():
                waited = int(time.time()) - self.files[run][datasetpathID]['start']

                lumis = self.files[run][datasetpathID]['lumis']
                lumisInRun = ListLumiSections.countLumisByRun(self.t0astDBConn, run)
                lumipercentage = float(len(lumis))/float(lumisInRun)*100
                
                logging.info("Lumis collected: %s" % lumis)
                logging.info("waited: %s, collected lumis %s%%" % (waited, lumipercentage))

                # either the collected lumi sections or timeout can start the job splitter
                if waited >= self.calibTimeout or lumipercentage >= self.lumiThreshold:
                    wmbsJobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                        subscription = self.promptCalibSubscription)
                    wmbsJobFactory()
                    self.t0astDBConn.commit()

        wmbsJobGroupIds = self.promptCalibSubscription.getJobGroups()
        
        logging.info("DEBUG JobGroups for promptCalib = %s" % wmbsJobGroupIds)

        jobGroupsDict = {}
        for wmbsJobGroupId in wmbsJobGroupIds:
            wmbsJobGroup = self.jobGroupCache.get(wmbsJobGroupId, None)
            if wmbsJobGroup == None:
                wmbsJobGroup = JobGroup(id = wmbsJobGroupId,
                                        subscription = self.promptCalibSubscription)
                jobIDs = wmbsJobGroup.listJobIDs()
                for jobID in jobIDs:
                    newJob = Job(id = jobID)
                    wmbsJobGroup.jobs.append(newJob)
                if len(jobIDs) == 0:
                    logging.error("WMBSJobGroup %s has no jobs, ignore it" % wmbsJobGroup)
                    continue

            runNumber = self.jobGroupToRunCache.get(wmbsJobGroup.id, None)
            datasetpathID = self.jobGroupToDatasetCache.get(wmbsJobGroup.id, None)
            if runNumber == None or datasetpathID == None:

                jobGroupInfoList = ListDatasets.listJobGroupInfo(self.t0astDBConn,
                                                                 wmbsJobGroup.id)
                if len(jobGroupInfoList) > 1:
                    logging.error("WMBSJobGroup %s has more than one run/dataset" %
                                  wmbsJobGroup.id)
                    continue
                elif len(jobGroupInfoList) == 0:
                    logging.error("WMBSJobGroup %s has no run/dataset" %
                                  wmbsJobGroup.id)
                    continue

                runNumber = jobGroupInfoList[0]['RUN_ID']
                datasetpathID = jobGroupInfoList[0]['DATASETPATH_ID']
                self.jobGroupToRunCache[wmbsJobGroup.id] = runNumber
                self.jobGroupToDatasetCache[wmbsJobGroup.id] = datasetpathID
                logging.info("DEBUG Added JobGroup %s to cache" %  wmbsJobGroup.id)
            if not jobGroupsDict.has_key(runNumber):
                jobGroupsDict[runNumber] = {}
            if not jobGroupsDict[runNumber].has_key(datasetpathID):
                jobGroupsDict[runNumber][datasetpathID] = Set()
            jobGroupsDict[runNumber][datasetpathID].add(wmbsJobGroup)
            self.jobGroupCache[wmbsJobGroupId] = wmbsJobGroup

        # process job groups by run and dataset
        for runNumber,tempDict in jobGroupsDict.items():
            for datasetpathID, jobGroups in tempDict.items():
                wmbsJobGroupIds = []
                for jobGroup in jobGroups:
                    wmbsJobGroupIds.append(jobGroup.id)
                logging.info("DEBUG Creating jobs for run %d, datasetpathId %d, jobgroups %s" %
                             (runNumber, datasetpathID, wmbsJobGroupIds))
                self.createAndSubmitJobs(runNumber,datasetpathID,jobGroups)
                del self.files[runNumber][datasetpathID]
                logging.info("DEBUG Created jobs for run %d, datasetpathId %d, jobgroups %s" %
                             (runNumber, datasetpathID, wmbsJobGroupIds))

        self.ms.publishUnique("PromptCalibInjector:pollForFiles",
                              "", "00:01:00")
        self.ms.commit()

    def cleanupRun(self, payload):
        """
        _cleanupRun_
    
        Once a run has completed push out any files that have not yet
        been merged and perform the following cleanup:
          - Clear out the workflow template for the run
          - Clear all workflows and jobspec files off disk
        """
        try:
            runNumber = int(payload)
        except ValueError:
            logging.error("Run number is not valid: %s" % payload)
            return
    
        logging.debug("Cleaning up run %s" % runNumber)
    
        if runNumber in self.workflows.keys():
            del self.workflows[runNumber]
    
        workflowDir = self.jobCacheDir + "/workflows/Run%s" % runNumber
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % runNumber
    
        shutil.rmtree(workflowDir, ignore_errors = True)
        shutil.rmtree(jobSpecDir, ignore_errors = True)

    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("PromptCalibInjector:StartDebug")
        self.ms.subscribeTo("PromptCalibInjector:EndDebug")
        self.ms.subscribeTo("PromptCalibInjector:pollForFiles")

        self.connectT0AST()
        self.connectPADB()

        self.promptCalibSubscription = retrievePromptCalibSubscription()

        self.ms.publishUnique("PromptCalibInjector:pollForFiles", 
                              "", "00:00:15")
        self.ms.commit()

        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)

        # Poll for messages...
        self.messageLoop()
