#!/usr/bin/env python
"""
_RuntimePromptReco_

Update the global tag in the jobspec because it is not available
at the time that the jobspec is created.
"""

__revision__ = "$Id: RuntimePromptReco.py,v 1.15 2010/02/13 13:42:47 hufnagel Exp $"
__version__ = "$Revision: 1.15 $"

import os
import pickle

from ProdCommon.MCPayloads.JobSpec import JobSpec
from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec


class JobSetup:
    """
    _JobSetup_

    """
    def __init__(self, workflowSpec, jobSpec):

        self.workflowSpec = WorkflowSpec()
        self.workflowSpec.load(workflowSpec)

        self.jobSpec = JobSpec()
        self.jobSpec.load(jobSpec)

        self.cmssw = {}
        self.cmssw['version1'] = 0
        self.cmssw['version2'] = 0
        self.cmssw['version3'] = 0
        version = self.workflowSpec.parameters.get("CMSSWVersion")
        cmsswVersionList = version.split("_")
        try:
            self.cmssw['version1'] = int(cmsswVersionList[1])
            self.cmssw['version2'] = int(cmsswVersionList[2])
            self.cmssw['version3'] = int(cmsswVersionList[3])
        except ValueError:
            print "Warning: non standard CMSSW version: %s" % version

    def backupPSet(self,filename,process):
        """
        _backupPSet_
        
        Write a backup copy of the current PSet to disk.
        """
        print "Wrote current configurations as %s" % filename
        handle = open(filename, 'w')
        handle.write("import pickle\n")
        handle.write("pickledCfg=\"\"\"%s\"\"\"\n" % pickle.dumps(process))
        handle.write("process = pickle.loads(pickledCfg)\n")
        handle.close()

        return

    def importAndBackupProcess(self):
        """
        _importAndBackupProcess_
        
        Try to import the process object for the job, which is contained in
        PSet.py and save a backup copy of it.
        """
        try:
            from PSet import process
        except ImportError, ex:
            msg = "Failed to import PSet module containing cmsRun Config\n"
            msg += str(ex)
            raise RuntimeError, msg

        print "PSet.py imported"

        self.backupPSet("PSetPreRuntime.log",process)

        return process

    def dumpConfiguration(self,process):
        """
        _dumpConfiguration_
        
        Write out the final configuration used for the job.
        """
        if ( ( self.cmssw['version1'] > 2 ) or \
             ( self.cmssw['version1'] == 2 and self.cmssw['version2'] >= 1 ) ):
            pycfgDump = open("PyCfgFileDump.log", 'w')
            try:
                print "Writing out PyCfgFileDump.log"
                pycfgDump.write(process.dumpPython())
            except Exception, ex:
                msg = "Error writing Python format config dump:\n"
                msg += "%s\n" % str(ex)
                msg += "This needs to be reported to the framework team"
                pycfgDump.write(msg)
            pycfgDump.close()
        else:
            cfgDump = open("CfgFileDump.log", 'w')
            try:
                print "Writing out CfgFileDump.log"
                cfgDump.write(process.dumpConfig())
            except Exception, ex:
                msg = "Error writing CfgFile format config dump\n"
                msg += "%s\n" % str(ex)
                msg += "This needs to be reported to the framework team"
                cfgDump.write(msg)
            cfgDump.close()

        return

    def importPSet(self):
        """
        _importPSet_

        Import the process object for cmsRun

        """
        process = self.importAndBackupProcess()

        import FWCore.ParameterSet.Config as cms

        # override maxEvents and skipEvents
        maxEvents = int(self.jobSpec.parameters.get("MaxEvents", -1))
        skipEvents = int(self.jobSpec.parameters.get("SkipEvents", 0))
        process.maxEvents = cms.untracked.PSet(input = cms.untracked.int32(maxEvents))
        process.source.skipEvents = cms.untracked.uint32(skipEvents)

        # dump and save final configuration
        self.dumpConfiguration(process)
        self.backupPSet("PSet.py",process)

        return

def main():
    """
    _main_

    Verify that the environment variable that points at the workflow spec
    has been set correctly and then make modification to the process object

    """
    print "=========PromptReco Job Setup================="

    workflowSpec = os.environ.get("PRODAGENT_WORKFLOW_SPEC", None)
    if workflowSpec == None:
        msg = "Unable to find WorkflowSpec from "
        msg += "PRODAGENT_WORKFLOW_SPEC variable\n"
        msg += "Unable to proceed\n"
        raise RuntimeError, msg

    if not os.path.exists(workflowSpec):
        msg = "Cannot find WorkflowSpec file:\n %s\n" % workflowSpec
        msg += "Unable to proceed\n"
        raise RuntimeError, msg

    jobSpec = os.environ.get("PRODAGENT_JOBSPEC", None)
    if jobSpec == None:
        msg = "Unable to find JobSpec from "
        msg += "PRODAGENT_JOBSPEC variable\n"
        msg += "Unable to proceed\n"
        raise RuntimeError, msg

    if not os.path.exists(jobSpec):
        msg = "Cannot find JobSpec file:\n %s\n" % jobSpec
        msg += "Unable to proceed\n"
        raise RuntimeError, msg

    instance = JobSetup(workflowSpec,jobSpec)
    instance.importPSet()

    print "=========PromptReco Job Setup Done==========="

if __name__ == '__main__':
    main()
