#!/usr/bin/env python
"""
_RepackerAuditorComponent_

Scan T0AST and trigger various clean up operations.
"""

__revision__ = "$Id: RepackerAuditorComponent.py,v 1.59 2011/01/13 12:00:10 hufnagel Exp $"
__version__ = "$Revision: 1.59 $"

import logging
import time

from T0 import Globals
from T0.State.Database.Writer import InsertRun
from T0.State.Database.Writer import RunLumiCloseout

from T0.State.Database.Reader import ListBlock
from T0.State.Database.Reader import ListFiles
from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListJobs
from T0.State.Database.Reader import ListRuns
from T0.State.Database.Reader import ListStreamers

from ProdAgentCore.Configuration import loadProdAgentConfiguration
from T0.GenericTier0 import Tier0DB

from T0.GenericTier0.GenericTier0Component import GenericTier0Component


class RepackerAuditorComponent(GenericTier0Component):
    """
    _RepackerAuditorComponent_

    Scan T0AST and trigger various clean up operations.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("Repacker Auditor Component Started")
        self.pollInterval = self.lookupParameter("PollInterval", "00:05:00")

        # lumi sections are closed every 23 seconds
        # but run clsoing and the closeout should run less frequently
        self.cycleCount = 0
        intervals = self.pollInterval.split(':')
        self.maxCycle = ( int(intervals[0])*3600 + int(intervals[1])*60 + int(intervals[2]) ) / 23
        self.pollInterval = "00:00:23"

        # connection to StorageManager database, used for run and lumi closing
        self.smDBConn = None

        # switches between old and new mode to close runs and lumis
        # old mode closes runs on timeout and lumis on filecount match
        # new mode used EoR and EoLS records from the SM database
        self.oldCloseoutMode = self.lookupParameter('OldCloseoutMode', "On")

        # Amount of time in seconds to wait after the last streamer is
        # injected for the auditor to close a run (or lumi)
        #
        # Interpreted differently whether in old or new closeout mode
        #
        # Old: direct timeout for closing runs and lumis
        #
        # New:                    lumis are never closed by timeout,always
        #                         wait for EoLS and completeness in T0AST
        # New and runTimeout = 0: run is closed when all lumis are closed,
        #                         however long that may take
        # New and runTimeout > 0: run is closed by timeout same as in old
        #                         closeout mode
        #
        self.runTimeout = int(self.lookupParameter("RunTimeout", "0"))

    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "RepackerAuditor:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "RepackerAuditor:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "RepackerAuditor:closeOutPoll":
            self.closeOutPoll()
        
        return
            
    def stateCloseOutRepack(self, run):
        """
        _stateCloseOutRepack_

        handle the CloseOutRepack state

        all bulk streamers need to have at least one EDM RAW child

        """
        logging.debug("stateCloseOutRepack(): Running %s" % run)

        count = ListStreamers.countRepackStreamersWithoutChild(self.t0astDBConn, run)

        if count != 0:
            logging.debug("Run %s has %s unrepacked bulk streamers." % (run, count))
            return

        logging.debug("Run %s moving to closeOutRepackMerge." % run)

        InsertRun.updateRunStatus(self.t0astDBConn, run, "CloseOutRepackMerge")
        
        Globals.tier0TwitterPost("Run %s : CloseOutRepackMerge" % run)
        return
    
    def stateCloseOutRepackMerge(self, run):
        """
        _stateCloseOutRepackMerge_

        handle the CloseOutRepackMerge state

        all RAW files need to be complete or failed in merge subscription

        ignore files created by incomplete processing groups (split repacked lumis)

        """
        logging.debug("stateCloseOutRepackMerge(): Running %s" % run)

        count = ListFiles.countPendingFilesForMerge(self.t0astDBConn, run, "RAW")

        if count != 0:
            logging.debug("Run %s has %d RAW files to merge." % (run, count))
            return

        logging.debug("Run %s moving to CloseOutPromptReco." % run)

        InsertRun.updateRunStatus(self.t0astDBConn, run, "CloseOutPromptReco")
        
        Globals.tier0TwitterPost("Run %s : CloseOutPromptReco" % run)
        return

    def stateCloseOutPromptReco(self, run):
        """
        _stateCloseOutPromptReco_

        handle the CloseOutPromptReco state

        all files in the reco subscription need to be complete or failed

        """
        logging.debug("stateCloseOutPromptReco(): Running %s" % run)

        sub = Globals.retrieveRecoSubscription()

        count = ListFiles.countPendingFilesForSub(self.t0astDBConn, run, sub['id'])

        if count != 0:
            logging.debug("Run %s has %d files to reco." % (run, count))
            return

        logging.debug("Run %s moving to CloseOutRecoMerge." % run)

        InsertRun.updateRunStatus(self.t0astDBConn, run, "CloseOutRecoMerge")
        
        Globals.tier0TwitterPost("Run %s : CloseOutRecoMerge" % run)
        return
                
    def stateCloseOutRecoMerge(self, run):
        """
        _stateCloseOutRecoMerge_

        handle the CloseOutRecoMerge state

        all RECO/DQM/AOD files need to be complete or failed in merge subscription

        ignore files created by incomplete processing groups (split PromptReco)

        """
        logging.debug("stateCloseOutRecoMerge(): Running %s" % run)

        count = ListFiles.countPendingFilesForMerge(self.t0astDBConn, run, "RECO")

        if count != 0:
            logging.debug("Run %s has %d RECO files to merge." % (run, count))
            return

        count = ListFiles.countPendingFilesForMerge(self.t0astDBConn, run, "DQM")

        if count != 0:
            logging.debug("Run %s has %d DQM files to merge." % (run, count))
            return

        count = ListFiles.countPendingFilesForMerge(self.t0astDBConn, run, "AOD")

        if count != 0:
            logging.debug("Run %s has %d AOD files to merge." % (run, count))
            return

        logging.debug("Run %s moving to CloseOutAlcaSkim." % run)

        InsertRun.updateRunStatus(self.t0astDBConn, run, "CloseOutAlcaSkim")
        
        Globals.tier0TwitterPost("Run %s : CloseOutAlcaSkim" % run)
        return

    def stateCloseOutAlcaSkim(self, run):
        """
        _stateCloseOutAlcaSkim_

        handle the CloseOutAlcaSkim state

        all files in the alcaSkim subscription need to be complete or failed

        ignore files created by incomplete processing groups (split PromptReco)
        
        """
        logging.debug("stateCloseOutAlcaSkim(): Running %s" % run)

        sub = Globals.retrieveAlcaSubscription()

        count = ListFiles.countPendingFilesForSub(self.t0astDBConn, run, sub['id'],
                                                  inputFromSplitProc = True)

        if count != 0:
            logging.debug("Run %s has %d files to alcaSkim." % (run, count))
            return

        logging.debug("Run %s moving to CloseOutAlcaSkimMerge." % run)

        InsertRun.updateRunStatus(self.t0astDBConn, run, "CloseOutAlcaSkimMerge")
        
        Globals.tier0TwitterPost("Run %s : CloseOutAlcaSkimMerge" % run)
        return
                
    def stateCloseOutAlcaSkimMerge(self, run):
        """
        _stateCloseOutAlcaSkimMerge_

        handle the CloseOutAlcaSkimMerge state

        all ALCARECO files need to be complete or failed in merge subscription

        ignore files created by incomplete processing groups (should not apply here)

        """
        logging.debug("stateCloseOutAlcaSkimMerge(): Running %s" % run)

        count = ListFiles.countPendingFilesForMerge(self.t0astDBConn, run, "ALCARECO")

        if count != 0:
            logging.debug("Run %s has %s ALCARECO files to merge." % (run, count))
            return

        count = ListFiles.countPendingFilesForMerge(self.t0astDBConn, run, "ALCAPROMPT")

        if count != 0:
            logging.debug("Run %s has %s ALCAPROMPT files to merge." % (run, count))
            return

        logging.debug("Run %s moving to CloseOutExport." % run)

        InsertRun.updateRunStatus(self.t0astDBConn, run, "CloseOutExport")
        
        Globals.tier0TwitterPost("Run %s : CloseOutExport" % run)
        return
    
    def stateCloseOutExport(self, run):
        """
        _stateCloseOutExport_

        handle CloseOutExport state

        all files in transfer subscription need to be complete or failed

        FIXME: race condition, as files are entered into transfer subscription
        not by the Tier0Accounter but by the export system itself (do some
        combined DBSUpload and transfer subscription here ?)

        """
        logging.debug("stateCloseOutExport(): Running %s" % run)
         
        sub = Globals.retrieveTransferSubscription()

        count = ListFiles.countPendingFilesForSub(self.t0astDBConn, run, sub['id'])

        if count != 0:
            logging.debug("Run %s has %d files to export." % (run, count))
            return

        logging.debug("Run %s moving to CloseOutT1Skimming." % run )

        InsertRun.updateRunStatus(self.t0astDBConn, run, "CloseOutT1Skimming")
        
        Globals.tier0TwitterPost("Run %s : CloseOutT1Skimming" % run)
        return
        
    def stateCloseOutT1Skimming(self, run):
        """
        _stateCloseOutT1Skimming_

        handle CloseOutT1Skimming state

        all blocks need to have status Skimmed

        """
        logging.debug("stateCloseOutT1Skimming(): Running %s" % run)        

        count = ListBlock.countUnskimmedBlocks(self.t0astDBConn, run)

        if count != 0:
            logging.debug("Run %s has %d blocks to skim." % (run, count))
            return

        logging.debug("Run %s moving to Complete." % run)

        InsertRun.updateRunStatus(self.t0astDBConn, run, "Complete")
        
        Globals.tier0TwitterPost("Run %s : Complete" % run)
        return
        
    def closeOutStateMachine(self):
        """
        _closeOutStateMachine_

        Periodically poll the run table looking for runs being closed out
        and pass off their run number to the appropriate methods.

        """
        runsInCloseout = ListRuns.listRunsInCloseout(self.t0astDBConn)

        for run, status in runsInCloseout.items():
            if status == "CloseOutRepack":
                self.stateCloseOutRepack(run)
            elif status == "CloseOutRepackMerge":
                self.stateCloseOutRepackMerge(run)
            elif status == "CloseOutPromptReco":
                self.stateCloseOutPromptReco(run)
            elif status == "CloseOutRecoMerge":
                self.stateCloseOutRecoMerge(run)
            elif status == "CloseOutAlcaSkim":
                self.stateCloseOutAlcaSkim(run)
            elif status == "CloseOutAlcaSkimMerge":
                self.stateCloseOutAlcaSkimMerge(run)
            elif status == "CloseOutExport":
                self.stateCloseOutExport(run)
            elif status == "CloseOutT1Skimming":
                self.stateCloseOutT1Skimming(run)
            else:
                logging.debug("Unknown status: %s for run %s." % (status, run))

        return

    def closeOutPoll(self):
        """
        _closeOutPoll_

        find runs with data and no RunConfig record, trigger
        RunConfig population for them (every polling cycle)

        close lumi sections (every polling cycle)

        close runs and lumis (every nth polling cycle)

        go through all runs in closeout and advance them
        in their closeout state if possible (every
        nth polling cycle)

        """
        logging.debug("closeOutPoll(): Running...")

        # populate RunConfig for new runs
        runs = ListStreamers.getRunsWithoutRunConfig(self.t0astDBConn)
        for run in runs:
            runConfig = self.getRunConfig(run)

        # check for completion of split processed streamer lumis
        RunLumiCloseout.checkSplitStreamerLumisForCompletion(self.t0astDBConn)

        self.cycleCount += 1

        if self.oldCloseoutMode == "Off":

            if self.cycleCount >= self.maxCycle:

                RunLumiCloseout.closeRunsAndLumis(self.t0astDBConn,
                                                  self.smDBConn,
                                                  runTimeout = self.runTimeout,
                                                  closeRuns = True)
                self.closeOutStateMachine()
                self.cycleCount = 0

            else:
            
                RunLumiCloseout.closeRunsAndLumis(self.t0astDBConn,
                                                  self.smDBConn)

        else:

            RunLumiCloseout.closeLumiSections(self.t0astDBConn,
                                              self.smDBConn,
                                              self.runTimeout)

            if self.cycleCount >= self.maxCycle:

                RunLumiCloseout.closeRuns(self.t0astDBConn, self.runTimeout)
                self.closeOutStateMachine()
                self.cycleCount = 0

        self.ms.publishUnique("RepackerAuditor:closeOutPoll", "", self.pollInterval)
        self.ms.commit()
        return

    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("RepackerAuditor:StartDebug")
        self.ms.subscribeTo("RepackerAuditor:EndDebug")
        self.ms.subscribeTo("RepackerAuditor:closeOutPoll")
        self.ms.commit()

        self.connectT0AST()

        paConfig = loadProdAgentConfiguration()
        smDBConfig = paConfig.getConfig("StorageManagerDB")
        self.smDBConn = Tier0DB.Tier0DB(smDBConfig,
                                        manageGlobal = False)
        self.smDBConn.connect()

        # Schedule an initial scan.
        self.ms.publishUnique("RepackerAuditor:closeOutPoll", "", self.pollInterval)

        # Wait for messages...
        self.messageLoop()
