#!/usr/bin/env python
"""
_RepackerInjectorComponent_

Component for generating Repacker jobsspecs
"""

__version__ = "$Revision: 1.94 $"
__revision__ = "$Id: RepackerInjectorComponent.py,v 1.94 2011/01/06 00:06:17 hufnagel Exp $"

import os
import sys
import shutil
import logging

from JobQueue import JobQueueDB
from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0.WorkflowFactory.RepackWorkflow import RepackWorkflow
from T0.WorkflowFactory.ConversionWorkflow import ConversionWorkflow

from T0.JobFactory.StreamerJob import StreamerJob

from T0 import Globals

from T0.State.Database.Reader import ListJobs
from T0.State.Database.Writer import InsertJob

from WMCore.JobSplitting.SplitterFactory import SplitterFactory

from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig

from WMCore.WMBS.Job import Job


class RepackerInjectorComponent(GenericTier0Component):
    """
    _RepackerInjectorComponent_

    Search for bulk streamer files in T0AST that can be gathered together
    and repacked into EDM RAW files.  Create and submit repack jobs to
    the ProdAgent JobQueue.

    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.

        """
        GenericTier0Component.__init__(self, **args)

        # Data structure to hold WorkflowSpecs.  This is a two level
        # dictionary with run -> stream_id
        self.workflows = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2

        # Check the ProdAgent configuration for the following parameters:
        #
        #   MaxSize - max file size for EDM RAW, used for output splitting
        #
        #   MaxStreamerSizeSingleLumi - maximum combined streamer size for
        #                               single lumi (split lumi if over)
        #   MaxStreamerEventsSingleLumi - maximum combined streamer events for
        #                                 single lumi (split lumi if over)
        #   MaxStreamerSizeMultiLumi - maximum combined streamer size for
        #                              multi lumi repack jobs
        #   MaxStreamerEventsMultiLumi - maximum combined streamer events for
        #                                multi lumi repack jobs
        #   MaxStreamerCount - maximum number of streamers for
        #                      multi lumi repack jobs
        #
        #   PollInterval - How often T0AST is checked for new files
        #   scramArch - SCRAM Architecture
        #   cmsPath - Path to CMS software
        #   JobCacheDir - Where jobspecs and workflows are stored
        #
        self.maxSize = int(self.lookupParameter("MaxFileSize", 10 * 1024 * 1024))
        self.maxStreamerSizeSingleLumi = int(self.lookupParameter('MaxStreamerSizeSingleLumi', 20*1024*1024*1024))
        self.maxStreamerEventsSingleLumi = int(self.lookupParameter('MaxStreamerEventsSingleLumi', 500000))
        self.maxStreamerSizeMultiLumi = int(self.lookupParameter('MaxStreamerSizeMultiLumi', 6*1024*1024*1024))
        self.maxStreamerEventsMultiLumi = int(self.lookupParameter('MaxStreamerEventsMultiLumi', 500000))
        self.maxStreamerCount = int(self.lookupParameter('MaxStreamerCount', 1000))
        
        self.pollInterval = self.lookupParameter("PollInterval", "00:01:00")
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")
        self.jobCacheDir = self.lookupParameter("JobCacheDir", self.args["ComponentDir"])

        # WMCore splitter factory for creating Repack jobs
        self.splitterFactory = SplitterFactory(package = "T0.JobSplitting")

        self.repackSubscription = None

        # See if lazy downloads are enabled. Whether or not this
        # switch does anything depends on the framework version
        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        # Check for forcing strict lumi order.
        # Feature is not really fully functional at the moment.
        if self.lookupParameter("StrictLumiOrder", "Off") == "On":
            self.strictLumiOrder = True
        else:
            self.strictLumiOrder = False

        # Check if we support streamer file conversions.
        # Feature is not really fully functional at the moment.
        if self.lookupParameter("SupportConversion", "Off") == "On":
            self.supportConversion = True
        else:
            self.supportConversion = False

        logging.info("RepackerInjector Component Started")


    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component receiving a message

        """
        # always have a debug on/off switch
        if message == "RepackerInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "RepackerInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "RepackerInjector:Poll":
            self.poll()
        elif message == "CleanUp":
            self.cleanupRun(payload)
        else:
            logging.error("Unknown message: %s, payload: %s" % (message, payload))

        return


    def cleanupRun(self, payload):
        """
        _cleanupRun_

        Remove all workflows and jobspecs for a particular run from disk.
        Remove any workflows for the run from memory.

        """
        try:
            run = int(payload)
        except ValueError:
            logging.error("Run number is not valid: %s" % payload)
            return

        logging.debug("Cleaning up run %s" % run)

        if run in self.workflows.keys():
            del self.workflows[run]

        workflowDir = self.jobCacheDir + "/workflows/Run%s" % run
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % run
        
        shutil.rmtree(workflowDir, ignore_errors=True)
        shutil.rmtree(jobSpecDir, ignore_errors=True)

        return


    def retrieveWorkflow(self, run, stream_id):
        """
        _retrieveWorkflow_

        Retrieve a workflow from the cache.
        If the workflow does not exist a new one will be created.
        If workflow creation fails None will be returned.

        """
        if self.workflows.has_key(run):
            if self.workflows[run].has_key(stream_id):
                return self.workflows[run][stream_id]
        else:
            self.workflows[run] = {}
                    
        # that we made it here means we have to create a new workflow
        workflow = self.createWorkflow(run, stream_id)
        if workflow != None:
            self.workflows[run][stream_id] = workflow

        return workflow


    def createWorkflow(self, run, stream_id):
        """
        _createWorkflow_

        Creates a new repack workflow.
        If workflow creation fails None is returned.

        A 'NewWorkflow' message is generated for each new workflow
        with the path to the workflow as the payload.

        """
        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %d" % run)
            return None

        repackConfig = runConfig.getRepackConfiguration(stream_id)
        if repackConfig == None:
            logging.error("No repack configuration for run %s and stream %s" % (run, stream))
            return None

        stream = runConfig.getStreamName(stream_id)
        processingStyle = runConfig.getProcessingStyle(stream_id)

        logging.debug("createWorkflow for run %d, stream %s" % (run, stream))

        if processingStyle == 'Bulk':

            cmsswVer = repackConfig.CMSSWVersion

            outModuleInfoList = runConfig.getOutputModuleInfoList("Repacker", stream_id, "RAW",
                                                                  splitIntoDatasets = True)
            if outModuleInfoList == None:
                logging.error("No output module list for run %s and stream %s" % (run, stream))
                return None

            workflow = RepackWorkflow(run,
                                      cmsswVer,
                                      self.cmsPath,
                                      self.scramArch,
                                      *outModuleInfoList)

            workflowSpecName = "Repack-Run%s-Stream%s" % (run, stream)

        elif processingStyle == 'Convert' or processingStyle == 'RegisterAndConvert':

            cmsswVer = repackConfig.CMSSWVersion

            outModuleInfo = runConfig.getConversionOutputModuleInfo(stream_id)

            if outModuleInfo == None:
                logging.error("No output module for run %s and stream %s" % (run, stream))
                return None

            workflow = ConversionWorkflow(run,
                                          cmsswVer,
                                          self.cmsPath,
                                          self.scramArch,
                                          *[outModuleInfo])

            workflowSpecName = "Repack-Run%s-Stream%s" % (run, stream)

        else:

            logging.error("Unknown processing style %s for run %d and stream %s" %
                          (processingStyle, run, stream))
            return None

        workflowDir = os.path.join(self.jobCacheDir, "workflows", "Run%s" % run)
        workflowSpecFile = os.path.join(workflowDir, "%s-workflow.xml" % workflowSpecName)

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No workflow for run, stream: %s, %s" % (run, stream))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowSpecFile)

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return workflowSpec

        workflowSpec = workflow.makeWorkflow(workflowSpecName,
                                             self.useLazyDownload, self.maxSize,
                                             production = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)

        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)
        
        workflowSpec.save(workflowSpecFile)

        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)


    def createJobSpecs(self, run, stream_id, jobDict, forceUnmergedOutput = False):
        """
        _createJobSpecs_

        Loop over all jobs and process them (same run and stream_id = same workflow)

        Generate job specs for every job

        """
        logging.debug("createJobSpecs(): Running...")

        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %s" % run)
            return {}

        stream = runConfig.getPrimaryDatasetName(stream_id)
        if stream == None:
            logging.error("Could not get stream name for id %d" % stream_id)
            return {}

        workflowSpec = self.retrieveWorkflow(run, stream_id)
        if workflowSpec == None:
            logging.error("No workflow for run %s, stream %s" % (run, stream))
            return {}

        jobSpecBaseDir = os.path.join(self.jobCacheDir, "jobspecs", "Run%d" % run)

        jobFactory = StreamerJob(workflowSpec)

        jobSpecDict = {}
        for id, files in jobDict.items():

            jobSpecDict[id] = jobFactory("Repack", id, files, jobSpecBaseDir, forceUnmergedOutput)

        return jobSpecDict


    def poll(self):
        """
        _poll_

        Poll T0AST for new Repack jobs

        Call job creation code, then queue them in JobQueue
        and update name and status in WMBS

        """
        logging.info("poll(): Running...")

        self.t0astDBConn.beginTransaction()

        wmbsJobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                              subscription = self.repackSubscription)

        wmbsJobFactory(maxStreamerSizeSingleLumi = self.maxStreamerSizeSingleLumi,
                       maxStreamerEventsSingleLumi = self.maxStreamerEventsSingleLumi,
                       maxStreamerSizeMultiLumi = self.maxStreamerSizeMultiLumi,
                       maxStreamerEventsMultiLumi = self.maxStreamerEventsMultiLumi,
                       maxStreamerCount = self.maxStreamerCount)

        self.t0astDBConn.commit()

        jobDict = ListJobs.listNewStreamerJobs(self.t0astDBConn, self.repackSubscription['id'])

        # process jobs by run and stream_id
        jobSpecDict = {}
        for run in jobDict.keys():
            for stream_id in jobDict[run].keys():

                normalJobDict = jobDict[run][stream_id]['normal']
                jobSpecDict.update( self.createJobSpecs(run, stream_id, normalJobDict) )

                splitJobDict = jobDict[run][stream_id]['split']
                jobSpecDict.update( self.createJobSpecs(run, stream_id, splitJobDict,
                                                        forceUnmergedOutput = True) )

        if len(jobSpecDict) > 0:

            # aggregate information for efficient database operations
            wmbsJobs = []
            jobNames = {}
            for jobId, jobSpec in jobSpecDict.items():
                wmbsJobs.append(Job(id = jobId))
                jobNames[jobId] = jobSpec.parameters["JobName"]

            # now do actual PA and T0AST DB modifications
            try:
                self.paDBConn.beginTransaction()

                for jobSpec in jobSpecDict.values():

                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters["JobSpecFile"],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority,
                                             siteName = None,
                                             status = "new")

                self.t0astDBConn.beginTransaction()

                InsertJob.updateWMBSJobNames(self.t0astDBConn, jobNames)
                self.stateChanger.propagate(wmbsJobs, 'created', 'new')

                self.paDBConn.commit()
                self.t0astDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(jobSpecDict))
            except StandardError, ex:
                self.t0astDBConn.rollback()
                self.paDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        self.ms.publishUnique("RepackerInjector:Poll", "", self.pollInterval)
        self.ms.commit()
        return


    def startComponent(self):
        """
        _startComponent_

        Fire up the message service and define the list of
        messages that this component subscribes to

        """
        self.ms.subscribeTo("RepackerInjector:StartDebug")
        self.ms.subscribeTo("RepackerInjector:EndDebug")
        self.ms.subscribeTo("RepackerInjector:Poll")
        self.ms.subscribeTo("CleanUp")
        self.ms.commit()

        self.connectT0AST()
        self.connectPADB()

        # Load up the repack subscription for easy access.
        self.repackSubscription = Globals.retrieveRepackSubscription()

        self.ms.publishUnique("RepackerInjector:Poll", "", "00:01:00")
        self.ms.commit()

        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)
        
        # wait for messages.
        self.messageLoop()
