#!/usr/bin/env python
"""
_FullLumiWithEventLimit_

Algorithm that sorts the trigger segments by run, lumi and stream
and creates jobs for each combinations seperately.

Optionally one can specify a maximum number of events per output
file, a LS would then potentially split up into multiple jobs
if it has too many events for a dataset

"""

import logging
from T0.RepackerScheduler.AlgorithmInterface import AlgorithmInterface
from T0.RepackerScheduler.Registry import registerAlgorithm


class FullLumiWithEventLimit(AlgorithmInterface):
    """
    _FullLumiWithEventLimit_

    """
    
    def __call__(self, *segments, **args):
        """
        _operator()_

        """
        self.maxEvents = args.get('maxEventsPerFile', 1000000)

        # sort segments by run, lumi and stream
        segmentDict = {}
        for segment in segments:

            run = segment['RunID']
            stream = segment['StreamID']
            lumi = segment['LumiID']

            if segmentDict.has_key(run):
                if segmentDict[run].has_key(stream):
                    if segmentDict[run][stream].has_key(lumi):
                        segmentDict[run][stream][lumi].append(segment)
                    else:
                        segmentDict[run][stream][lumi] = [ segment ]
                else:
                    segmentDict[run][stream] = { lumi : [ segment ] }
            else:
                segmentDict[run] = { stream : { lumi : [ segment ] } }

        result = []

        # now schedule jobs
        for run in sorted(segmentDict.keys()):
            for stream in sorted(segmentDict[run].keys()):
                for lumi in sorted(segmentDict[run][stream].keys()):

                    result.extend( self.scheduleJobsPerRunLumiStream(run,stream,lumi,
                                                                     segmentDict[run][stream][lumi]) )

        return result

    def scheduleJobsPerRunLumiStream(self, run, stream, lumi, segmentsByRunLumiStream):
        """
        Takes dict of segments and makes jobs out of them

        These segments belong to the same run, stream and lumi
        so in principle can all go into the same single repack jobs.

        Except for the event limit that is.

        """
        # make dictionary with streamer as key, list of segments as value
        segmentsByStreamer = {}
        for segment in segmentsByRunLumiStream:
            if segmentsByStreamer.has_key(segment['StreamerID']):
                segmentsByStreamer[segment['StreamerID']].append(segment)
            else:
                segmentsByStreamer[segment['StreamerID']] = [ segment ]

        jobEntityList = []

        # start scheduling loop, scheduling jobs and
        # removing the corresponding segments until
        # all segments are assigned to job
        while ( len(segmentsByStreamer.keys()) > 0 ):

            # find streamers that repack into the same group of datasets
            # recursive to catch the case where datasets in the group
            # come from streamers that are found late in the loop
            outputDatasets = []
            scanForStreamers = True
            segmentsByStreamerAndDataset = {}
            while ( scanForStreamers == True ):
                for streamer, segmentsForStreamer in segmentsByStreamer.items():
                    numberOfDatasets = len(outputDatasets)
                    if not segmentsByStreamerAndDataset.has_key(streamer):
                        if numberOfDatasets == 0:
                            for segment in segmentsForStreamer:
                                outputDatasets.append(segment['DatasetID'])
                            segmentsByStreamerAndDataset[streamer] = segmentsForStreamer
                        else:
                            addStreamer = False
                            for segment in segmentsForStreamer:
                                if segment['DatasetID'] in outputDatasets:
                                    addStreamer = True
                                    break
                            if addStreamer:
                                for segment in segmentsForStreamer:
                                    if segment['DatasetID'] not in outputDatasets:
                                        outputDatasets.append(segment['DatasetID'])
                                segmentsByStreamerAndDataset[streamer] = segmentsForStreamer

                if len(outputDatasets) == numberOfDatasets:
                    scanForStreamers = False

            # now schedule jobs per group of datasets
            while ( len(segmentsByStreamerAndDataset) > 0 ):
                segmentsToRepack = self.getSegmentsToRepack(segmentsByStreamerAndDataset)

                for segment in segmentsToRepack:
                    if segmentsByStreamer.has_key(segment['StreamerID']):
                        del segmentsByStreamerAndDataset[segment['StreamerID']]
                        del segmentsByStreamer[segment['StreamerID']]

                jobEntityList.append( self.newJobEntity(*segmentsToRepack) )

        return jobEntityList


    def getSegmentsToRepack(self, segmentsByStreamerAndDataset):
        """
        Goes through segmentsByStreamerAndDataset and creates
        a repack job that observes the maxEvents limit for
        all datasets

        """
        segmentsToBeRepacked = []
        totalEventsPerDataset = {}
        for streamer, segments in segmentsByStreamerAndDataset.items():

            # calculating events from streamer for each dataset
            eventsPerDataset = {}
            for segment in segments:
                if eventsPerDataset.has_key(segment['DatasetID']):
                    eventsPerDataset[segment['DatasetID']] += segment['SegmentSize']
                else:
                    eventsPerDataset[segment['DatasetID']] = segment['SegmentSize']

            # if first streamer always use it
            if len(segmentsToBeRepacked) == 0:
                totalEventsPerDataset = eventsPerDataset
                segmentsToBeRepacked.extend(segments)
                continue

            # combine old total event counts with this streamer
            for dataset, events in totalEventsPerDataset.items():
                if eventsPerDataset.has_key(dataset):
                    eventsPerDataset[dataset] += events
                else:
                    eventsPerDataset[dataset] = events

            # check the new number of events per dataset against maxEvents
            addStreamerToJob = True
            for events in eventsPerDataset.values():
                if events > self.maxEvents:
                    logging.debug("Too many events in run %d, stream_id %d, lumi section %d, split up !"
                                  % (segments[0]['RunID'], segments[0]['StreamID'], segments[0]['LumiID']))
                    addStreamerToJob = False
                    break

            if addStreamerToJob:
                totalEventsPerDataset = eventsPerDataset
                segmentsToBeRepacked.extend(segments)

        return segmentsToBeRepacked


registerAlgorithm(FullLumiWithEventLimit.__name__, FullLumiWithEventLimit)

