#!/usr/bin/env python
# pylint: disable-msg = W0142,W0104
"""
_RepackerSchedulerComponent_

Component for polling T0AST and grouping trigger segments into
repacker job definitions.
"""

__revision__ = "$Id: RepackerSchedulerComponent.py,v 1.30 2009/09/07 22:04:15 hufnagel Exp $"
__version__ = "$Revision: 1.30 $"


import logging

import T0.RepackerScheduler.Algorithms
from T0.RepackerScheduler.Registry import retrieveAlgorithm

from T0.State.Database.Reader.ListTriggerSegments import getNewTriggerSegments

from T0.RepackerScheduler.RepackerSchedulerAPI import registerJobs

from T0.GenericTier0.GenericTier0Component import GenericTier0Component


class RepackerSchedulerComponent(GenericTier0Component):
    """
    _RepackerSchedulerComponent_

    Component that creates new repack and conversion jobs

    Repack jobs are created from new trigger segments
    Conversion jobs are created from 'conversion' streamers

    """
    def __init__(self, **args):
        GenericTier0Component.__init__(self, **args)
        self.algorithm = self.lookupParameter("AlgorithmName", "FullLumi")
        self.pollInterval = self.lookupParameter("PollInterval", "00:02:00")
        self.maxEventsPerFile = int(self.lookupParameter("MaxEventsPerFile", 1000000))
        self.lumiPerRepackJob = int(self.lookupParameter("LumiPerRepackJob", 1))

        msg = "RepackerScheduler Component Started\n"
        msg += "Default Scheduling Algorithm: %s\n" % (self.algorithm)
        msg += "Poll Cycle: %s\n" % self.pollInterval
        logging.info(msg)


    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.

        """
        logging.debug("Event: %s" % event)
        
        if event == "RepackerScheduler:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
            return
        elif event == "RepackerScheduler:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
            return

        elif event == "RepackerScheduler:Poll":
            self.poll()
            return

        elif event == "CloseOutScheduler":
            self.poll()
            return
        
        return


    def defineRepackJobs(self, segments):
        """
        _defineRepackJobs_

        Load algorithm
        Pass list of available trigger segments to algorithm
        Algortithm returns list of job entities

        """
        algo = retrieveAlgorithm(self.algorithm)
        args = { 'maxEventsPerFile' : self.maxEventsPerFile,
                 'lumiPerRepackJob' : self.lumiPerRepackJob }
        jobEntityList = algo(*segments, **args)

        logging.info("Number of scheduled repack jobs = %d" % len(jobEntityList))
        logging.debug("Scheduled repack jobs = %s" % jobEntityList)

        return jobEntityList


    def poll(self):
        """
        _poll_

        Single cycle invocation of all DatasetWorkers to search for jobs

        """
        logging.info("Trigger Segment Poll Invoked")

        #
        # create repack jobs
        #  - get available segments for repacking
        #  - define jobs from them
        #  - store job definitions in T0AST
        #
        triggerSegmentList = getNewTriggerSegments(self.t0astDBConn)
        jobEntityList = self.defineRepackJobs(triggerSegmentList)
        registerJobs(self.t0astDBConn, jobEntityList)

        self.t0astDBConn.commit()
        
        #  //
        # // Lastly, delay the poll cycle every time we poll, since
        #//  we only want it as a catch all/timeout, so we remove the existing
        #  //poll and replace it with another full interval
        # //
        #//
        self.ms.publishUnique('RepackerScheduler:Poll', '', self.pollInterval)
        self.ms.commit()
        return


    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("RepackerScheduler:StartDebug")
        self.ms.subscribeTo("RepackerScheduler:EndDebug")
        self.ms.subscribeTo("CloseOutScheduler")
        self.ms.subscribeTo("RepackerScheduler:Poll")
        
        # generate first polling cycle (delayed, not immediately)
        self.ms.remove("RepackerScheduler:Poll")
        self.ms.publishUnique("RepackerScheduler:Poll", "", self.pollInterval)
        self.ms.commit()
        
        self.connectT0AST()

        # wait for messages
        self.messageLoop()
