#!/usr/bin/env python
"""
_Cache_

In memory cache for a set of RunConfig objects.
Maximum number of RunConfigs in memory configured via constructor.

In the event a RunConfig is not in memory, it is created from information in
T0AST.

If T0AST does not know about this Run, the information is looked up in
RunSummary and ConfDB and written into T0AST before loading it.
"""

__revision__ = "$Id: Cache.py,v 1.68 2011/01/12 21:11:56 hufnagel Exp $"
__version__ = "$Revision: 1.68 $"

_DEFAULT_CACHE_DEPTH = 100

import time
import operator
import logging

from T0.State.Database.Reader import ListRunConfig
from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Writer import InsertDataset
from T0.State.Database.Writer import InsertRunConfig

import T0.ConfDB.RetrieveConfiguration as ConfDB
from T0.RunConfigCache.RunConfig import RunConfig
from T0.RunConfigCache.Tier0Config import retrieveDatasetConfig
from T0.RunConfigCache.Tier0Config import addRepackConfig
from T0.RunConfigCache.Tier0Config import deleteStreamConfig

from ProdAgentCore.Configuration import loadProdAgentConfiguration
from WMCore.Configuration import loadConfigurationFile
from WMCore.Database.CMSCouch import CouchServer

class Cache:
    """
    _Cache_

    Access to Run configuration with in mem cache, T0AST lookup and
    ConfDB callouts
    """
    def __init__(self, cacheDepth = _DEFAULT_CACHE_DEPTH):
        self.depth = cacheDepth
        self.fileBackend = None
        self.t0astDBConn = None
        self.accessTimes = {}
        self.runConfigs = {}

        # The path to the Offline ConfDB file and the directory to cache
        # framework configs is stored in the RunConfig section of the ProdAgent
        # config file.
        paConfig = loadProdAgentConfiguration().getConfig("RunConfig")
        self.offlineConfDB = paConfig["OfflineConfDB"]
        self.configCache = paConfig["ConfigCacheDir"]
        self.couchURL = paConfig.get("CouchURL", None)
        
        logging.debug("RunConfigCache instantiated with depth=%s" % self.depth)

    def getRunConfig(self, runNumber):
        """
        _getRunConfig_

        Retrieve the configuration for a given run.
        """
        if self.runConfigs.has_key(runNumber):
            self.accessTimes[runNumber] = time.time()
            logging.debug("RunConfig %s accessed in memory at %s" %
                          (runNumber, self.accessTimes[runNumber]))
            return self.runConfigs[runNumber]

        #
        # Check if run exists in T0AST, if not pull
        # from RunSummary/ConfDB or Filebackend
        # and write it to T0AST
        #
        if not ListRunConfig.runExists(self.t0astDBConn, runNumber):

            logging.debug("Run %s does not exist in T0AST" % runNumber)

            try:
                self.t0astDBConn.beginTransaction()
                self.insertRunConfig(runNumber)
                self.t0astDBConn.commit()
            except Exception, ex:
                # Three things could have happened here:
                #   - Something bad happened.  We have a database connection
                #     problem, a config problem, etc.
                #   - RunConfig insertion succeeded but Couch insertion failed.
                #   - We have a race between two components trying to insert the
                #     same run config.
                #
                # We'll assume that the problem is the latter.  Wait for a
                # second and then check to see if the RunConfig exists in T0AST.
                # If it does we're golden and we can move on.  If it doesn't
                # we have a real problem and we'll bail out.
                logging.error("Failed to insert run config: %s" % str(ex))
                self.t0astDBConn.rollback()
                time.sleep(1)

                if not ListRunConfig.runExists(self.t0astDBConn, runNumber):
                    return None

            try:
                self.insertConfigIntoCouch(runNumber)
            except Exception, ex:
                logging.error("Error inserting into Couch: %s" % str(ex))
                
        newRunConfig = RunConfig(self.t0astDBConn, runNumber, self.configCache)
        self.addRunConfig(newRunConfig)
        return self.runConfigs[runNumber]

    def insertRunConfig(self, runNumber):
        """
        _insertRunConfig_

        Insert a run config into T0AST.  Pull down the offline configuration as
        well as the online configuration and insert everything into T0AST.
        """
        logging.debug("insertRunConfig(): Running %s" % runNumber)

        #
        # treat central DAQ runs and other runs differently
        #   (central DAQ runs have HLTKey set)
        #
        # FIXME: this will totally break if for some reason we ever share
        #        the same run number for central DAQ and other runs
        #
        versionAndHLTKey = ListRunConfig.retrieveVersionAndHLTKey(self.t0astDBConn,
                                                                  runNumber)

        onlineVersion = versionAndHLTKey[0][0]
        hltkey = versionAndHLTKey[0][1]

        logging.debug( "onlineVersion: %s hltkey: %s" % ( onlineVersion, hltkey ) )

        if hltkey != 'none':

            logging.debug("Pulling HLT config from ConfDB")

            configuration = ConfDB.getConfiguration(runNumber, hltkey)

            if configuration == None:
                raise RuntimeError, "Could not retrieve HLT config for run %s" % runNumber

            tier0Config = loadConfigurationFile(self.offlineConfDB)

            # all HLT defined streams not explicitely configured are repacked
            for streamName in configuration[1].keys():
                if streamName not in tier0Config.Streams.dictionary_().keys():
                    addRepackConfig(tier0Config, streamName)

            # filter out repack/express streams that are not defined in the HLT
            for streamName, streamConfig in tier0Config.Streams.dictionary_().items():
                if streamName not in configuration[1].keys():
                    if streamConfig.ProcessingStyle == "Express" or streamConfig.ProcessingStyle == "Bulk":
                        deleteStreamConfig(tier0Config, streamName)

            InsertRunConfig.insertStreamDatasetTriggerMapping(self.t0astDBConn, runNumber, configuration[1])

            InsertRunConfig.insertT0Config(self.t0astDBConn, runNumber,
                                           tier0Config.Global.Version)

            # write stream configurations to T0AST
            for streamName,streamConfig in tier0Config.Streams.dictionary_().items():
                InsertRunConfig.insertStreamConfig(self.t0astDBConn, runNumber,
                                                   streamName, streamConfig)

            # write dataset configurations to T0AST
            for streamName in configuration[1].keys():
                for datasetName in configuration[1][streamName]:
                    datasetConfig = retrieveDatasetConfig(tier0Config, datasetName)
                    InsertRunConfig.insertDatasetConfig(self.t0astDBConn, runNumber,
                                                        datasetConfig)

            InsertRunConfig.updateRunTable(self.t0astDBConn, runNumber, configuration[0],
                                           tier0Config.Global.AcquisitionEra,
                                           tier0Config.Global.RecoTimeout,
                                           tier0Config.Global.RecoLockTimeout)

        else:

            tier0Config = loadConfigurationFile(self.offlineConfDB)

            InsertRunConfig.insertStreamDatasetTriggerMapping(self.t0astDBConn, runNumber, {})

            InsertRunConfig.insertT0Config(self.t0astDBConn, runNumber,
                                           tier0Config.Global.Version)

            # now that I have all streams, insert everything
            for streamName,streamConfig in tier0Config.Streams.dictionary_().items():
                InsertRunConfig.insertStreamConfig(self.t0astDBConn, runNumber,
                                                   streamName, streamConfig)

            InsertRunConfig.updateRunTable(self.t0astDBConn, runNumber,
                                           "FakeProcessName", "FakeAcquisitionEra")

        return

    def insertConfigIntoCouch(self, runNumber):
        """
        _insertConfigIntoCouch_

        Query T0AST for the configuration for the run and try to insert it into
        Couch.
        """
        couchConfig = {}
        couchConfig["acquisitionEra"] = "Undefined"
        couchConfig["streams"] = {}

        streamsInfo = ListRunConfig.listStreamInfoForRun(self.t0astDBConn, runNumber)
        for streamInfo in streamsInfo:
            couchConfig["acquisitionEra"] = streamInfo[2]
            couchConfig["streams"][streamInfo[0]] = {}

            streamDict = couchConfig["streams"][streamInfo[0]]
            streamDict["processingStyle"] = streamInfo[1]
            streamDict["frameworkVersion"] = streamInfo[3]

            if streamInfo[1] == "Bulk":
                streamDict["primaryDatasets"] = {}

                datasetsInfo = ListRunConfig.listDatasetInfoForRun(self.t0astDBConn, runNumber, streamInfo[0])
                for datasetInfo in datasetsInfo:
                    streamDict["primaryDatasets"][datasetInfo[0]] = {}
                    datasetDict = streamDict["primaryDatasets"][datasetInfo[0]]
                    datasetDict["scenario"] = datasetInfo[1]
                    datasetDict["globalTag"] = datasetInfo[2]
                    datasetDict["repack"] = {"processingVersion": datasetInfo[3],
                                             "frameworkVersion": datasetInfo[4]}
                    datasetDict["promptReco"] = {"enabled": datasetInfo[5],
                                                 "processingVersion": datasetInfo[6],
                                                 "frameworkVersion": datasetInfo[7],
                                                 "config": datasetInfo[8]}
                    datasetDict["alca"] = {"enabled": datasetInfo[9],
                                           "processingVersion": datasetInfo[10],
                                           "frameworkVersion": datasetInfo[11],
                                           "config": datasetInfo[12]}
                    datasetDict["dqm"] = {"enabled": datasetInfo[13]}
                    datasetDict["promptSkim"] = {}

                    skimsInfo = ListRunConfig.listSkimInfoForRun(self.t0astDBConn, runNumber, datasetInfo[0])
                    for skimInfo in skimsInfo:
                        datasetDict["promptSkim"][skimInfo[0]] = {}
                        skimsDict = datasetDict["promptSkim"][skimInfo[0]]
                        skimsDict["dataTier"] = skimInfo[1]
                        skimsDict["frameworkVersion"] = skimInfo[2]
                        skimsDict["processingVersion"] = skimInfo[3]
                        skimsDict["config"] = skimInfo[4]
                        skimsDict["globalTag"] = skimInfo[5]
            elif streamInfo[1] == "Express":
                streamDict["dataTiers"] = []
                expressConfigs = ListRunConfig.listExpressConfigForStream(self.t0astDBConn, runNumber, streamInfo[0])
                streamDict["globalTag"] = expressConfigs[0][1]
                streamDict["processingVersion"] = expressConfigs[0][2]
                streamDict["processingConfig"] = expressConfigs[0][3]
                streamDict["mergePackConfig"] = expressConfigs[0][4]
                
                for expressConfig in expressConfigs:
                    streamDict["dataTiers"].append(expressConfig[0])
            else:
                logging.error("Unknown processing style: %s" % streamInfo[1])

        couchdb = CouchServer(self.couchURL)
        if "tier0_runconf" not in couchdb.listDatabases():
            couchdb.createDatabase("tier0_runconf")

        database = couchdb.connectDatabase("tier0_runconf")
        couchConfig["_id"] = str(runNumber)
        result = database.commit(couchConfig)[0]

        if result.has_key("error"):
            logging.error("Problems adding to couch: %s" % result)
            return

        offlineConfDBHandle = open(self.offlineConfDB, 'r')
        database.addAttachment(result["id"], result["rev"], offlineConfDBHandle.read(),
                               "OfflineConfDB")
        offlineConfDBHandle.close()
        return

    def addRunConfig(self, runConfig):
        """
        _addRunConfig_

        Add a new run config, make an entry in the access times and bump the last config
        out if the cache is at capacity
        """
        self.runConfigs[runConfig.getRunNumber()] = runConfig
        self.accessTimes[runConfig.getRunNumber()] = time.time()
        logging.debug("Adding RunConfig %s @ %s" % \
                      (runConfig.getRunNumber(),
                       self.accessTimes[runConfig.getRunNumber()]))
        if len(self.runConfigs.keys()) > self.depth:
            self.dropOldestRunConfig()

    def dropOldestRunConfig(self):
        """
        _dropOldestRunConfig_

        Drop the oldest RunConfig from the cache.
        """
        accessTimeList = self.accessTimes.items()
        accessTimeList.sort(key = operator.itemgetter(1))

        if len(accessTimeList) == 0:
            return 

        oldRunNumber = accessTimeList[0][0]
        self.dropRunConfig(oldRunNumber)
        
    def dropRunConfig(self, runNumber):
        """
        _dropRunConfig_

        Drop the config for the run provided from the cache.
        """
        del self.runConfigs[runNumber]
        del self.accessTimes[runNumber]
        logging.debug("Run %s dropped" % runNumber)
        return
