#!/usr/bin/env python
"""
_StreamerInjectorComponent_

Component for creating trigger segments from streamer records

"""

__version__ = "$Revision: 1.43 $"
__revision__ = "$Id: SegmentInjectorComponent.py,v 1.43 2010/02/11 17:44:44 hufnagel Exp $"
__author__ = "Dirk Hufnagel"

import logging
import os
import subprocess
import time

from T0.State.Database.Reader import ListStreamers
from T0.State.Database.Writer import InsertStreamer
from T0.State.Database.Writer import InsertTriggerSegment
from T0.State.Database.Writer import RunLumiCloseout

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

class SegmentInjectorComponent(GenericTier0Component):
    """
    _SegmentInjectorComponent_

    Query TOAST for streamer records and create trigger segment records

    """
    def __init__(self, **args):
        GenericTier0Component.__init__(self, **args) 

        self.activePolling = True
        self.pollDelay = int(self.lookupParameter('PollInterval', 60))
        self.emulatorMode = bool(self.lookupParameter('EmulatorMode', False))
        self.streamerTimeout = int(self.lookupParameter('StreamerTimeout', 93))

        self.numberOfLoggerNodes = int(self.lookupParameter('NumberOfLoggerNodes', 4))
        self.releaseArea = self.args['ComponentDir']

        logging.info("SegmentInjector Component Started")

    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component receiving a message

        """
        logging.debug("Recieved Message: %s" % message)
        logging.debug("Payload: %s" % payload)
        
        if message == "SegmentInjector:Start":
            logging.info("Starting SegmentInjector...")
            self.activePolling = True
            return

        elif message == "SegmentInjector:Stop":
            logging.info("Stopping SegmentInjector...")
            self.activePolling = False
            return
            
        elif message == "SegmentInjector:Poll":
            self.pollStreamers()
            return            

        elif message == "SegmentInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
            return
        
        elif message == "SegmentInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
            return

        else:
            logging.info("Do not know how to deal with Message: %s" % message)
            return


    def handleNewStreamers(self):
        """
        _handleStreamers_
        
        Queries database for unsplit streamers, processes them

        """
        
        logging.debug("----Start creating trigger segments ")

        try:
            logging.info("Closing lumi sections")
            RunLumiCloseout.closeLumiSections(self.t0astDBConn, self.streamerTimeout)

            self.t0astDBConn.commit()
        except StandardError, ex:
            # close and rethrow
            self.t0astDBConn.rollback()
            self.t0astDBConn.close()
            msg = "Could not close lumi sections: %s\n" % str(ex)
            raise RuntimeError, msg

        # first check for new streamers in general, then access RunConfig
        # needed to populate T0AST with settings for run(s)
        # which is needed for the data discovery to work
        runNumbers = ListStreamers.getRunsForUnusedStreamers(self.t0astDBConn)
        for runNumber in runNumbers:
            runConfig = self.getRunConfig(runNumber)

        #
        # current time, needed for streamer age calculation
        #
        now = time.time()

        #
        # find streamer files to be split
        #
        streamersAll = ListStreamers.listSplitableStreamers(self.t0astDBConn)

        logging.debug("Number of streamers found = %d" % len(streamersAll))

        #
        # sort streamer record by run, lumi and stream
        #
        streamersDict = {}
        for streamer in streamersAll:
            run = streamer['RUN_ID']
            lumi = streamer['LUMI_ID']
            stream = streamer['STREAM_ID']
            if streamersDict.has_key(run) :
                if streamersDict[run].has_key(lumi) :
                    if streamersDict[run][lumi].has_key(stream) :
                        streamersDict[run][lumi][stream].append(streamer)
                    else :
                        streamersDict[run][lumi][stream] = [streamer]
                else :
                    streamersDict[run][lumi] = { stream : [streamer] }
            else :
                streamersDict[run] = { lumi : { stream : [streamer] } }

        #
        # loop over all streamers in run, lumi and stream order
        # and split them into trigger segments
        #
        for run in sorted(streamersDict.iterkeys()):

            # get runConfig for this run, skip other files
            # in the run if we can't get it
            #
            # if we do get it gets cached, so we do not need
            # to keep the reference
            runConfig = self.getRunConfig(run)
            if runConfig == None:
                logging.error("RunConfig error : could not retrieve RunConfig for run %s\n" % run)
                continue

            for lumi in sorted(streamersDict[run].iterkeys()):
                for stream in streamersDict[run][lumi].iterkeys():
                    self.splitStreamers(streamersDict[run][lumi][stream])

        return

    def startComponent(self):
        """
        _startComponent_

        Start up the component and define the messages that it subscribes to.
        """
        self.ms.subscribeTo("SegmentInjector:Start")
        self.ms.subscribeTo("SegmentInjector:End")
        self.ms.subscribeTo("SegmentInjector:Poll")
        self.ms.subscribeTo("SegmentInjector:StartDebug")
        self.ms.subscribeTo("SegmentInjector:EndDebug")

        self.connectT0AST()
        
        # generate first polling cycle
        self.ms.remove("SegmentInjector:Poll")
        self.ms.publishUnique("SegmentInjector:Poll", "")
        self.ms.commit()
 
        # wait for messages
        self.messageLoop()

    def pollStreamers(self):
        """
        _pollStreamers_

        starts polling cycle

        """
        try:
            if not self.activePolling:
                logging.debug("pollStreamers:Inactive")
            else:
                logging.debug("pollStreamers:Active")
                self.handleNewStreamers()

	finally:
		# generate new polling cycle
		self.ms.publishUnique('SegmentInjector:Poll', '', self.pollDelay)
		self.ms.commit()
       
	return


    def splitStreamers(self,streamers):
        """
        _splitStreamers_
        
        Takes a list of streamers as input and processes them

        """

        #
        # split streamer files into trigger segments
        #
        streamersSplit = []
        trigger_segments = []
        for streamer in streamers:

            logging.debug("Split streamer %s" % streamer['LFN'])

            # if we got this far we have a RunConfig
            runConfig = self.getRunConfig(streamer['RUN_ID'])

            repackConfig = runConfig.getRepackConfiguration(streamer['STREAM_ID'])
            if repackConfig == None:
                logging.error("RunConfig error : could not get repack config for run %s\n and stream %s" % (streamer['RUN_ID'],
                                                                                                            streamer['STREAM_ID'])
                return
            
            primaryDatasets = runConfig.getDatasetsForStream(streamer['STREAM_ID'])
            if primaryDatasets == None:
                logging.error("RunConfig error : could not get datasets for run %s\n" % streamer['RUN_ID'])
                return

            if ( not self.emulatorMode ):

                output = self.parseIndexFile(repackConfig.CMSSWVersion,streamer['INDEXPFN'])

                if ( output == None ):
                    continue

                #
                # process edmStreamerIndex output
                #
                triggerPathDetails = {}
                maxIndex = (streamer['EVENTS'] + 31) / 32
                for line in output.splitlines(False):
                    temp = line.lstrip(':').split(',')
                    intList = []
                    # consistency check on the edmStreamerIndex output
                    if ( len(temp)-1 != maxIndex):
                        logging.debug("WARNING: streamer file %s has more events than indicated (%d)"
                                      % (streamer['LFN'],streamer['EVENTS']))
                        maxIndex = len(temp)-1
                    for i in range (1, maxIndex+1):
                        intList.append(int(temp[i], 16))
                    triggerPathDetails[temp[0]] = intList

            #
            # we got this far, remember the streamer as split
            #
            streamersSplit.append(streamer)

            #
            # now loop over datasets and create a trigger segment
            # for each (provided there are events going to this
            # dataset in the streamer file)
            #
            for dataset_id in primaryDatasets:
                
                triggerpaths = runConfig.getTriggersForDataset(dataset_id)
                if triggerpaths == None:
                    logging.error("RunConfig error : could not get triggers for run %s and dataset_id %s\n" % (streamer['RUN_ID'],dataset_id))
                    return

                dataset = runConfig.getPrimaryDatasetName(dataset_id)
                if dataset == None:
                    logging.error("RunConfig error : could not get dataset name for run %s and dataset_id %s\n" % (streamer['RUN_ID'],dataset_id))
                    return

                logging.debug("Split into dataset %s" % dataset)

                if ( not self.emulatorMode ):
                    # create bitmask with selected events for dataset
                    intList = []
                    for i in range(0, maxIndex):
                        intList.append(0)
                    for triggerpath in triggerpaths:
                        if triggerpath in triggerPathDetails.keys():
                            for i in range(0, maxIndex):
                                intList[i] |= triggerPathDetails[triggerpath][i]
                        else:
                            logging.debug("Trigger paths in streamer differ from trigger paths in T0AST")
                            logging.debug("Skipping this lumi section")
                            return

                    # count bits
                    eventCount = 0
                    for i in range(0, maxIndex):
                        while intList[i]:
                            intList[i] &= intList[i] - 1
                            eventCount += 1
                else:
                    eventCount = 100

                parameters = { 'streamer_id' : streamer['STREAMER_ID'], 'lumi_id' : streamer['LUMI_ID'],
                               'run_id' : streamer['RUN_ID'], 'segment_size' : eventCount,
                               'dataset_id' : dataset_id, 'status' : 1 }
                trigger_segments.append(parameters)
                logging.debug("New Trigger Segment : Dataset = %s , StreamerID = %s, Run = %d, Lumi = %d, EventsPassed = %d"
                              % (dataset,streamer['STREAMER_ID'],streamer['RUN_ID'],streamer['LUMI_ID'],eventCount))

        #
        # insert trigger_segments all at once and also update streamer status
        #
        if ( len(trigger_segments) > 0 ) or ( len(streamersSplit) > 0 ):
            try:
                if len(trigger_segments) > 0:
                    InsertTriggerSegment.insertTriggerSegments(self.t0astDBConn,
                                                               trigger_segments)
                if len(streamersSplit) > 0:
                    streamerIDs = []
                    for streamer in streamersSplit:
                        streamerIDs.append( streamer['STREAMER_ID'] )
                    InsertStreamer.updateSplitStreamers(self.t0astDBConn,streamerIDs)

                self.t0astDBConn.commit()
            except StandardError, ex:
                # close and rethrow
                self.t0astDBConn.rollback()
                self.t0astDBConn.close()
                msg = "Failed to insert trigger segments or update streamer status: %s\n" % str(ex)
                raise RuntimeError, msg

        return


    def parseIndexFile(self,cmsswVersion,indexFile):
        """
        _parseIndexFile_
        
        Takes a list of streamers as input and processes them

        """

        #
        # setup CMSSW environment
        #
        if os.path.isfile("%s/%s/src/runtime.sh" % (self.releaseArea, cmsswVersion)):
            logging.debug("Use exisiting software area for %s" % cmsswVersion)
        else:
            logging.debug("Create software area for %s" % cmsswVersion)
            p = subprocess.Popen("/usr/local/bin/bash",stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            output, error = p.communicate("""
            cd %s
            scramv1 project CMSSW %s
            cd %s/src
            scramv1 runtime -sh > runtime.sh
            """ % (self.releaseArea,cmsswVersion,cmsswVersion))
            if ( len(error) > 0 ):
                logging.error("ERROR: could not create software area for %s\n%s" % (cmsswVersion,error))
                return None

        #
        # process index file
        #
        logging.debug("Parsing index file %s" % indexFile)
        p = subprocess.Popen("/usr/local/bin/bash",stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        output, error = p.communicate("""
        . %s/%s/src/runtime.sh
        tempfile=`mktemp`
        rfcp %s $tempfile > /dev/null
        edmStreamerIndex -i $tempfile
        rm -f $tempfile
        """ % (self.releaseArea,cmsswVersion,indexFile))
        if ( len(error) > 0 ):
            logging.error("ERROR: could not parse index file %s\n%s" % (indexFile,error))
            return None

        return output
