#!/usr/bin/env python
"""
_ListBlock_

Read operations related to the block and active_block tables in T0AST. 
"""

from T0.DataStructs.Block import Block 

def listActiveBlocksByRunStatus(dbConn, runStatusList):
    """
    _listActiveBlocksByRunStatus_
    
    Retrieve block related information from active_block.
    and return Block object
    """
    block = Block()
            
    sqlQuery = """ SELECT bl.id, bl.name, ab.dataset_path_id, bl.block_size, bl.file_count, 
                          dp.primary_dataset, ab.run_id 
                   FROM active_block ab 
                    INNER JOIN block bl ON (bl.id = ab.block_id) 
                    INNER JOIN dataset_path dp ON (dp.id = ab.dataset_path_id)
                    INNER JOIN run ON run.run_id = ab.run_id
                    INNER JOIN run_status rs ON rs.id = run.run_status
                   WHERE rs.status = :run_status  
               """
    bindVars = []
    if type(runStatusList) == str:
        runStatusList = [runStatusList]

    for runStatus in runStatusList:
        param = {'run_status': runStatus}
        bindVars.append(param)
    dbConn.execute(sqlQuery, bindVars)
    # this should only return one or None
    blockRows = dbConn.fetchall()
    listBlocks = []
    for blockRow in blockRows:
        block = Block()
        block["BLOCK_ID"] = blockRow[0]
        block["BLOCK_NAME"] = blockRow[1]
        block["DATASET_PATH_ID"] = blockRow[2]
        block["BLOCKSIZE"] = blockRow[3]
        block["FILECOUNT"] = blockRow[4] 
        block["DATASET_ID"] = blockRow[5]
        block["RUN_ID"] = blockRow[6]
        listBlocks.append(block)
    
    return listBlocks
    
def listActiveBlock(dbConn, t0astFile):    
    """
    _listActiveBlock_
    
    Retrieve block related information from active_block and
    repacked/reconstructed table.  Returns block object with updated
    parameters from t0astFile.
    """
    block = Block(t0astFile)
            
    sqlQuery = """ SELECT b.id, b.name, ab.dataset_path_id, b.block_size, b.file_count
                   FROM active_block ab INNER JOIN block b ON (b.id = ab.block_id)
                   WHERE ab.run_id = :p_1 AND ab.dataset_path_id = :p_2  
               """
             
    bindVars = {'p_1':block["RUN_ID"], 'p_2':block["DATASET_PATH_ID"]}
    
    dbConn.execute(sqlQuery, bindVars)
    
    # this should only return one or None
    blockRows = dbConn.fetchall()
        
    if len(blockRows) != 1:
        return None
    else:
        # update block information
        blockRow = blockRows[0]
        block["BLOCK_ID"] = blockRow[0]
        block["BLOCK_NAME"] = blockRow[1]
        block["DATASET_PATH_ID"] = blockRow[2]
        block["BLOCKSIZE"] = blockRow[3] 
        block["FILECOUNT"] = blockRow[4]
        
        return block     
    
    
def listAllActiveBlocks(dbConn):
    
    """
    _listActiveBlockByRun_
    
    retrieve block related information from active_block 
    and repacked/reconstructed table
    returns block object with updated parameters from t0astFile
    
    """    
    sqlQuery = """ SELECT bl.id, bl.name, ab.dataset_path_id, bl.block_size, bl.file_count, 
                          dp.primary_dataset, ab.run_id 
                   FROM active_block ab 
                    INNER JOIN block bl ON (bl.id = ab.block_id) 
                    INNER JOIN dataset_path dp ON (dp.id =ab.dataset_path_id)
               """             
    dbConn.execute(sqlQuery)
    # this should only return one or None
    blockRows = dbConn.fetchall()
    listBlocks = []
    for blockRow in blockRows:
        block = Block()
        block["BLOCK_ID"] = blockRow[0]
        block["BLOCK_NAME"] = blockRow[1]
        block["DATASET_PATH_ID"] = blockRow[2]
        block["BLOCKSIZE"] = blockRow[3]
        block["FILECOUNT"] = blockRow[4] 
        block["DATASET_ID"] = blockRow[5]
        block["RUN_ID"] = blockRow[6]
        listBlocks.append(block)
    
    return listBlocks

def listBlocksByStatus(dbConn, statusList, migrateStatus = None):
    """
    _listBlocksByStatus_
    list of blocks by status
    
    """
    
    sqlQuery = """SELECT block.id, dataset_path_id, name, run_id, 
                          dataset_path.primary_dataset, bs.status FROM block
                   INNER JOIN block_run_assoc ON (block.id = block_id)
                   INNER JOIN dataset_path
                      ON (block.dataset_path_id = dataset_path.id)
                   INNER JOIN block_status bs ON block.status = bs.id
                   WHERE bs.status = :p_1 """
    
    if type(statusList) == str:
        statusList = [statusList]
 
    if migrateStatus != None:
        sqlQuery += """ AND migrate_status = (SELECT id FROM block_migrate_status 
                        WHERE status = :mStatus) """
    
    bindVars = []
    for status in statusList:    
        params = {'p_1': status}               
        if migrateStatus != None:
            params['mStatus'] = migrateStatus
        bindVars.append(params)

    # this is hack to order by tier: TODO order by tier
    # sqlQuery += """ ORDER BY block.id"""  
    dbConn.execute(sqlQuery, bindVars)
    
    blockRows = dbConn.fetchall()
    blockList = []
    
    for blockRow in blockRows:
        block = Block()
        block["BLOCK_ID"] = blockRow[0]
        block["DATASET_PATH_ID"] = blockRow[1]
        block["BLOCK_NAME"] = blockRow[2]
        block["RUN_ID"] = blockRow[3]
        block["DATASET_ID"] = blockRow[4]
        block["STATUS"] = blockRow[5]
        blockList.append(block)
    
    return blockList           

def getBlockInfoByID(dbConn, blockID):
    """
    _getBlockInfoByID_
    get information from the block table by block_id
    
    """
    sqlQuery = """ SELECT block.id, dataset_path_id, name, block_status.status,
                          bms.status 
                   FROM block
                   INNER JOIN block_status on block.status = block_status.id
                   INNER JOIN block_migrate_status bms on block.migrate_status = bms.id
                   WHERE block.id=:p_1 """
    param = {'p_1': blockID}
    dbConn.execute(sqlQuery, param)
    
    blockRows = dbConn.fetchall()
    blockRow = blockRows[0]
    block = Block()
    block["BLOCK_ID"] = blockRow[0]
    block["DATASET_PATH_ID"] = blockRow[1]
    block["BLOCK_NAME"] = blockRow[2]
    block["STATUS"] = blockRow[3]
    block["MIGRATE_STATUS"] = blockRow[4]
    
    return block

def getBlockIDByFileID(dbConn, fileID):
    """
    _getBlockIDByFileID_
    get block_id from wmbs_file_block_assoc
    
    """
    sqlQuery = "SELECT block_id FROM wmbs_file_block_assoc WHERE file_id=:p_1"
    param = {'p_1': fileID}
    dbConn.execute(sqlQuery, param)
    result = dbConn.fetchall()
    
    if len(result) == 0:
            return None
    else:
        return result[0][0]

def listBlockInfoByStatus(dbConn, transferStatus, migrationStatus):
    """
    _listBlockInfoByStatus_

    Retrieve a list of blocks with a given transfer status and migration status.
    The result will be a list of dictionaries each with the following keys:
    BLOCK_ID, DATASET_PATH_ID, RUN_ID, STORAGE_NODE, PRIMARY_ID,
    PROCESSED_ID and TIER_ID. 
    """
    sqlQuery = """SELECT block.id, block.name, block.dataset_path_id,
                         block_run_assoc.run_id,
                         storage_node.name, dataset_path.primary_dataset,
                         dataset_path.processed_dataset, dataset_path.data_tier,
                         phedex_subscription.custodial_flag
                  FROM block
                    INNER JOIN dataset_path ON
                      block.dataset_path_id = dataset_path.id                  
                    INNER JOIN block_run_assoc ON
                      block.id = block_run_assoc.block_id
                    LEFT OUTER JOIN phedex_subscription ON
                      block_run_assoc.run_id = phedex_subscription.run_id AND
                      dataset_path.primary_dataset = phedex_subscription.primary_dataset_id AND
                      dataset_path.data_tier = phedex_subscription.data_tier_id
                    LEFT OUTER JOIN storage_node ON
                      phedex_subscription.node_id = storage_node.id                      
                   WHERE block.status =
                    (SELECT id FROM block_status WHERE status = :p_1) AND
                   block.migrate_status =
                    (SELECT id FROM block_migrate_status WHERE status = :p_2)
                   """
    bindVars = {"p_1": transferStatus, "p_2": migrationStatus}

    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    blocks = []
    for result in results:
        blockInfo = Block()
        blockInfo["BLOCK_ID"] = result[0]
        blockInfo["BLOCK_NAME"] = result[1] 
        blockInfo["DATASET_PATH_ID"] = result[2]
        blockInfo["RUN_ID"] = result[3]
        blockInfo["STORAGE_NODE"] = result[4]
        blockInfo["PRIMARY_ID"] = result[5]
        blockInfo["PROCESSED_ID"] = result[6]
        blockInfo["TIER_ID"] = result[7]
        blockInfo["CUSTODIAL"] = result[8]
        blocks.append(blockInfo)

    return blocks

def countUnskimmedBlocks(dbConn, run):
    """
    _countUnskimmedBlocks_

    Count the number of blocks for a run
    that do not have status Skimmed

    """
    sqlQuery = """SELECT COUNT(*)
                  FROM block
                  INNER JOIN block_run_assoc ON
                    block_run_assoc.block_id = block.id AND
                    block_run_assoc.run_id = :RUN_ID
                  WHERE block.status != (SELECT id FROM block_status WHERE status = 'Skimmed')
                  """

    bindVars = { "RUN_ID" : run }
    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = dbConn.fetchall()

    return results[0][0]

def listCustodialNodeForBlock(dbConn, blockID):
    """
    _listCustodialNodeForBlock_

    Retrieve the name of the custoidal site for a given block.
    """
    sqlQuery = """SELECT storage_node.name FROM storage_node
                    INNER JOIN phedex_subscription_made ON storage_node.id = phedex_subscription_made.node_id
                    INNER JOIN block ON phedex_subscription_made.dataset_path_id = block.dataset_path_id
                    INNER JOIN block_run_assoc ON block.id = block_run_assoc.block_id
                    INNER JOIN dataset_path ON block.dataset_path_id = dataset_path.id
                    INNER JOIN primary_dataset ON dataset_path.primary_dataset = primary_dataset.id
                    INNER JOIN phedex_subscription ON phedex_subscription.run_id = block_run_assoc.run_id AND
                      phedex_subscription.primary_dataset_id = primary_dataset.id
                  WHERE block.id = :blockid AND phedex_subscription.custodial_flag = 1"""
    bindVars = {"blockid": blockID}

    dbConn.execute(sqlQuery, bindVars)
    custodialNodes = []
    for result in dbConn.fetchall():
        custodialNodes.append(result[0])
        
    return custodialNodes

def isParentBlockMigrated(dbConn, blockID):
    """
    this will return True if the block doesn't have a parent.
    """
    sqlQuery = """SELECT count(*) FROM block 
                    INNER JOIN block_parentage ON block.id = input_id
                  WHERE output_id = :blockID 
                        AND migrate_status != 
                            (SELECT id FROM block_migrate_status 
                             WHERE status = 'Migrated')"""
    
    bindVars = {"blockID": blockID}

    dbConn.execute(sqlQuery, bindVars)
    if (dbConn.fetchall()[0][0] == 0):
        return True
    else:
        return False

def isParentBlockExported(dbConn, blockId):
    """
    _isParentBlockExported_

    For a given block, verify that all of it's parent blocks are exported and
    migrated.
    """
    sql = """SELECT COUNT(*) FROM block
             INNER JOIN block_parentage ON
                 block_parentage.input_id= block.id AND
                 block_parentage.output_id = :blockId
             WHERE block.migrate_status != (SELECT id FROM block_migrate_status WHERE status = 'Migrated')
             OR ( block.status != (SELECT id FROM block_status where status = 'Exported') AND
                  block.status != (SELECT id FROM block_status WHERE status = 'Skimmed') )
             """

    bindVars = {"blockId": blockId}

    dbConn.execute(sql, bindVars)
    if (dbConn.fetchall()[0][0] == 0):
        return True
    else:
        return False    

def getExpressBlockCreationTime(dbConn, blockID):
    """
    get estimated block creation time by the earlies files created
    """
    sqlQuery = """SELECT MIN(ff.insert_time) 
                  FROM wmbs_fileset_files ff
                   INNER JOIN wmbs_fileset fs ON (fs.id = ff.fileset)
                   INNER JOIN wmbs_file_block_assoc fb ON (fb.file_id = ff.fileid)
                  WHERE fb.block_id = :blockID AND fs.name = 'ExpressDBSUploadable'"""
    
    bindVars = {"blockID": blockID}
    
    dbConn.execute(sqlQuery, bindVars)
    
    result = dbConn.fetchall()
    return int(result[0][0])

def countUnExportedBlocksByRun(dbConn, runNumbers):
    """
    _countUnExportedBlocksByRun_

    Count the number of blocks associated with a particular set of runs that
    have not been exported to their custodial site.  The results are
    returned as a three level dictionary with the first level being the run
    number, the second level the primary dataset and the third level the data
    tier.  No data will be returned for run/primary/tier combinations that
    have had all their blocks exported.
    """
    sql = """SELECT block_run_assoc.run_id, primary_dataset.name,
                    data_tier.name, COUNT(*) FROM block
               INNER JOIN block_status ON
                 block.status = block_status.id
               INNER JOIN block_migrate_status ON
                 block.migrate_status = block_migrate_status.id
               INNER JOIN block_run_assoc ON
                 block.id = block_run_assoc.block_id
               INNER JOIN dataset_path ON
                 block.dataset_path_id = dataset_path.id
               INNER JOIN primary_dataset ON
                 dataset_path.primary_dataset = primary_dataset.id
               INNER JOIN data_tier ON
                 dataset_path.data_tier = data_tier.id
               WHERE block_run_assoc.run_id = :p_1 AND
                     ((block_status.status != 'Exported' AND
                     block_status.status != 'Skimmed') OR
                     block_migrate_status.status != 'Migrated')
               GROUP BY block_run_assoc.run_id, primary_dataset.name,
                        data_tier.name"""

    bindVars = []
    for runNumber in runNumbers:
        bindVars.append({"p_1": runNumber})
    dbConn.execute(sql, bindVars)
    results = dbConn.fetchall()

    runCounts = {}
    for result in results:
        if not runCounts.has_key(result[0]):
            runCounts[int(result[0])] = {}

        if not runCounts[int(result[0])].has_key(result[1]):
            runCounts[int(result[0])][result[1]] = {}

        runCounts[int(result[0])][result[1]][result[2]] = result[3]

    return runCounts
