#!/usr/bin/env python
"""
_ListDatasets_

Utilities to retrieve information from the dataset tables in T0AST.
"""

__revision__ = "$Id: ListDatasets.py,v 1.35 2010/07/22 14:44:22 hufnagel Exp $"
__version__ = "$Revision: 1.35 $"

def listDatasetPathsForCompleteRuns(dbConn):
    """
    _listDatasetPathsForCompleteRuns_

    list all bulk dataset paths for complete runs,
    where complete in this context means either
    Complete or CloseOutT1Skimming status

    """
    sql = """SELECT block_run_assoc.run_id, block.name
             FROM block
             INNER JOIN block_run_assoc ON
               block_run_assoc.block_id = block.id
             INNER JOIN run ON
               run.run_id = block_run_assoc.run_id AND
               run.run_status IN (
                 SELECT id FROM run_status
                 WHERE status IN ('CloseOutT1Skimming', 'Complete')
               )
             INNER JOIN dataset_path ON
               dataset_path.id = block.dataset_path_id
             INNER JOIN dataset_run_stream_assoc ON
               dataset_run_stream_assoc.primary_dataset_id = dataset_path.primary_dataset AND
               dataset_run_stream_assoc.run_id = block_run_assoc.run_id
             INNER JOIN run_stream_style_assoc ON
               run_stream_style_assoc.run_id = block_run_assoc.run_id AND
               run_stream_style_assoc.stream_id = dataset_run_stream_assoc.stream_id AND
               run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
             GROUP BY block_run_assoc.run_id, block.name
             """

    dbConn.execute(sql)
    results = dbConn.fetchall()

    datasetpathsByRun = {}
    for result in results:
        run = result[0]
        path = result[1].split('#')[0]
        if datasetpathsByRun.has_key(run):
            datasetpathsByRun[run].add(path)
        else:
            datasetpathsByRun[run] = set([path])

    return datasetpathsByRun


def listProcessedDatasets(dbConn):
    """
    _listProcessedDatasets_

    Retrieve a list of all the processed datasets in T0AST.  The results are
    returned in the form of a dictionary where the key is the processed dataset
    id and the value is the processed dataset name.
    """
    sqlQuery = "SELECT id, name FROM processed_dataset"

    dbConn.execute(sqlQuery)
    results = dbConn.fetchall()

    processedDatasets = {}
    for result in results:
        processedDatasets[result[0]] = result[1]

    return processedDatasets

def listDataTiers(dbConn):
    """
    _listDataTiers_

    Retrieve a list of all the data tiers in T0AST.  The results are returned
    in the form of a dictionary where the key is the data tier id and the value
    is the data tier name.
    """
    sqlQuery = "SELECT id, name FROM data_tier"

    dbConn.execute(sqlQuery)
    results = dbConn.fetchall()

    dataTiers = {}
    for result in results:
        dataTiers[result[0]] = result[1]

    return dataTiers

def listDataTiersByNames(dbConn):
    """
    _listDataTiersByNames_

    Retrieve a list of all the data tiers in T0AST.  The results are returned
    in the form of a dictionary where the key is the data tier name and the value
    is the data tier id.
    """
    sqlQuery = "SELECT name, id FROM data_tier"

    dbConn.execute(sqlQuery)
    results = dbConn.fetchall()

    dataTiers = {}
    for result in results:
        dataTiers[result[0]] = result[1]

    return dataTiers

def listRunDatasetMapping(dbConn):
    """
    _listRunDatasetMapping_

    Retrieve a list of all the runs in T0AST and the datasets that are
    associated with them.
    """
    sqlQuery = """SELECT primary_dataset_id, run_id, name from
                  dataset_run_stream_assoc INNER JOIN primary_dataset ON
                  primary_dataset.id = dataset_run_stream_assoc.primary_dataset_id"""
                
    dbConn.execute(sqlQuery)
    results = dbConn.fetchall()

    datasetInfoList = []
    for result in results:
        datasetInfo ={}
        datasetInfo["DATASET_ID"] = result[0]
        datasetInfo["RUN_ID"] = result[1]
        datasetInfo["NAME"] = result[2]
        
        datasetInfoList.append(datasetInfo)

    return datasetInfoList

def listProcessedDatasetNameFromID(dbConn, procds_id):
    """
    _listProcessedDatasetNameFromID_

    Given the ID of a processed dataset return it's name.
    """
    sqlQuery = """SELECT a.NAME from processed_dataset a
                  WHERE a.ID = :p_1"""

    bindVars = {"p_1": procds_id}
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()

    if len(result) == 1:
        return result[0][0]
    else:
        return None

def listDatasetPathInfo(dbConn, datasetpath):
    """
    _listDatasetPathInfo_

    return primary dataset id/name, processed dataset name
    and datatier name for a given datasetpath id

    """
    sqlQuery = """SELECT dataset_path.primary_dataset,
                         primary_dataset.name,
                         processed_dataset.name,
                         data_tier.name
                  FROM dataset_path
                  INNER JOIN primary_dataset ON
                    primary_dataset.id = dataset_path.primary_dataset
                  INNER JOIN processed_dataset ON
                    processed_dataset.id = dataset_path.processed_dataset
                  INNER JOIN data_tier ON
                    data_tier.id = dataset_path.data_tier
                  WHERE dataset_path.id = :DATASETPATH
                  """

    bindVars = { 'DATASETPATH' : datasetpath }
    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = dbConn.fetchall()

    return { 'primds_id' : results[0][0],
             'primds' : results[0][1],
             'procds' : results[0][2],
             'datatier' : results[0][3] }

def listDatasetNamesFromID(dbConn, datasetID):
    """
    _listDatasetNamesFromID_

    Given the ID of a dataset return the human readable names of the primary
    dataset, processed dataset and data tier.
    """
    sqlQuery = """SELECT primary_dataset.name,
                         processed_dataset.name,
                         data_tier.name
                  FROM dataset_path
                  INNER JOIN primary_dataset ON
                    primary_dataset.id = dataset_path.primary_dataset
                  INNER JOIN processed_dataset ON
                    processed_dataset.id = dataset_path.processed_dataset
                  INNER JOIN data_tier ON
                    data_tier.id = dataset_path.data_tier
                  WHERE dataset_path.id = :p_1
                  """

    bindVars = {"p_1": datasetID}
    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    result = dbConn.fetchall()

    return { 'PRIMARY' : result[0][0],
             'PROCESSED' : result[0][1],
             'TIER' : result[0][2] }

def listDatasetIDsFromID(dbConn, datasetID):
    """
    _listDatasetIDsFromID_

    Given the ID of a dataset return the IDs of the primary dataset,
    processed dataset and data tier.
    """
    sqlQuery = """SELECT primary_dataset, processed_dataset, data_tier
                  FROM dataset_path WHERE dataset_path.ID = :p_1"""

    bindVars = {"p_1": datasetID}
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()

    datasetIDs = {"PRIMARY": result[0][0], "PROCESSED": result[0][1],
                    "TIER": result[0][2]}
    return datasetIDs

def listDatasetNamesForWMBSFile(dbConn, wmbsFileID):
    """
    _listDatasetNamesForWMBSFile_

    Retrieve the names of the primary dataset, processed dataset and data tier
    for a given WMBS file.
    """
    sqlQuery = """SELECT primary_dataset.NAME, processed_dataset.NAME,
                  data_tier.NAME FROM dataset_path INNER JOIN primary_dataset ON
                  primary_dataset.ID = dataset_path.PRIMARY_DATASET INNER JOIN
                  processed_dataset ON
                  processed_dataset.ID = dataset_path.PROCESSED_DATASET INNER JOIN
                  data_tier ON data_tier.ID = dataset_path.DATA_TIER INNER JOIN
                  wmbs_file_dataset_path_assoc ON wmbs_file_dataset_path_assoc.dataset_path_id
                  = dataset_path.id WHERE wmbs_file_dataset_path_assoc.file_id = :p_1"""

    bindVars = {"p_1": wmbsFileID}
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()

    datasetNames = {"PRIMARY": result[0][0], "PROCESSED": result[0][1],
                    "TIER": result[0][2]}
    return datasetNames

def listJobGroupInfo(dbConn, wmbsJobGroupID):
    """
    _listJobGroupInfo_

    Retrieve run_id and dataset_id for files
    belonging to jobs belonging to a job group
    """
    sqlQuery = """SELECT c.run, d.dataset_path_id FROM wmbs_job a
                  INNER JOIN wmbs_job_assoc b
                  ON b.job = a.id
                  INNER JOIN wmbs_file_runlumi_map c
                  ON c.fileid = b.fileid
                  INNER JOIN wmbs_file_dataset_path_assoc d
                  ON d.file_id = b.fileid
                  WHERE a.jobgroup = :p_1
                  GROUP BY c.run, d.dataset_path_id"""

    bindVars = {"p_1": wmbsJobGroupID}
    dbConn.execute(sqlQuery, bindVars)
    rows = dbConn.fetchall()

    jobGroupInfoList = []
    for row in rows:
        jobGroupInfo = {}
        jobGroupInfo['RUN_ID'] = row[0]
        jobGroupInfo['DATASETPATH_ID'] = row[1]
        jobGroupInfoList.append(jobGroupInfo)

    return jobGroupInfoList

def listDatasetIDForWMBSFile(dbConn, wmbsFileID):
    """
    _listDatasetIDForWMBSFile_

    Retrieve the dataset_path id for a given WMBS file.
    """
    sqlQuery = """SELECT id FROM dataset_path INNER JOIN wmbs_file_dataset_path_assoc ON
                  wmbs_file_dataset_path_assoc.dataset_path_id = dataset_path.id
                  WHERE wmbs_file_dataset_path_assoc.file_id = :p_1"""

    bindVars = {"p_1": wmbsFileID}
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()
    return result[0][0]

def listProcessedDatasetID(dbConn, processedDatasetName):
    """
    _listProcessedDatasetID_

    Retrieve the ID of a processed dataset.
    """
    sqlQuery = "SELECT id FROM processed_dataset WHERE name = :p_1"
    bindVars = {"p_1": processedDatasetName}
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()

    return result[0][0]

def listDataTierID(dbConn, dataTierName):
    """
    _listDataTierID_

    Retrieve the ID of a data tier.
    """
    sqlQuery = "SELECT id FROM data_tier WHERE name = :p_1"
    bindVars = {"p_1": dataTierName}
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()

    return result[0][0]

def listDatasetPathFromIDs(dbConn, primaryDatasetID, processedDatasetID,
                           dataTierID):
    """
    _listDatasetPathFromIDs_

    Retrieve the dataset path ID given a primary dataset, processed dataset
    and data tier ID.
    """
    sqlQuery = """SELECT id FROM dataset_path WHERE primary_dataset = :p_1
                  AND processed_dataset = :p_2 AND data_tier = :p_3"""
    bindVars = {"p_1": primaryDatasetID, "p_2": processedDatasetID,
                "p_3": dataTierID}
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()

    return result[0][0]
