#!/usr/bin/env python
"""
_ListFiles_

Collection of common read method for repacked and reconstructed table
"""

__revision__ = "$Id: ListFiles.py,v 1.76 2011/01/13 11:04:18 hufnagel Exp $"
__version__ = "$Revision: 1.76 $"

import logging

from T0 import Globals


def listFileIDsByBlockID(dbConn, blockID):
    """
    _listFileIDsByBlockID_
    
    """
    sqlQuery = """SELECT file_id FROM wmbs_file_block_assoc 
                  WHERE block_id = :block_id"""
    bindVars = {"block_id": blockID}
    
    dbConn.execute(sqlQuery, bindVars)
    result = dbConn.fetchall()
    fileIDs = []
    for row in result:
        fileIDs.append(row[0])
    return fileIDs
     
def listBlockFilesForSkim(dbConn, blockID):
    """
    _listBlockFilesForSkim_

    Retrieve a list of files in a block for Tier1 skimming.  This query will
    return a list of dictionaries, each dictionary will have the following
    keys:
      EVENTS, SIZE, LFN

    """
    sqlQuery = """SELECT events, filesize, lfn FROM wmbs_file_details
                    INNER JOIN wmbs_file_block_assoc
                      ON wmbs_file_details.id = wmbs_file_block_assoc.file_id
                  WHERE wmbs_file_block_assoc.block_id = :p_1"""
    bindVars = {"p_1": blockID}
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    files = []
    for result in results:
        newFile = {}
        newFile["EVENTS"] = result[0]
        newFile["SIZE"] = result[1]
        newFile["LFN"] = result[2]
        files.append(newFile)

    return files

def listLumiInfoForFile(dbConn, lfn):
    """
    _listLumiInfoForFile_

    Retrieve a list of lumi IDs for a given LFN.

    """
    sqlQuery = """SELECT DISTINCT lumi FROM wmbs_file_runlumi_map
                    INNER JOIN wmbs_file_details
                      ON wmbs_file_details.id = wmbs_file_runlumi_map.fileid
                  WHERE wmbs_file_details.lfn = :p_1"""
    bindVars = {"p_1": lfn}
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    lumiIDs = []
    for result in results:
        lumiIDs.append(result[0])

    return lumiIDs

def listParentLFNs(dbConn, childLFN):
    """
    _listParentLFNs_

    Retrieve the LFNs of a file's parents.  This is used by the Tier1Scheduler
    as it doesn't have a good way of using the WMBS classes with more than
    one database at a time.

    This query will actually retrieve the grand parent LFNs as we don't care
    about intermediate files that get merged.

    """
    sql = """SELECT DISTINCT lfn FROM wmbs_file_details
               INNER JOIN wmbs_file_parent ON wmbs_file_details.id = wmbs_file_parent.parent
             WHERE wmbs_file_parent.child IN
               (SELECT parent FROM wmbs_file_parent
                  WHERE child = (SELECT id FROM wmbs_file_details WHERE lfn = :lfn))"""
        
    bindVars = {"lfn": childLFN}

    dbConn.execute(sql, bindVars)
    results = dbConn.fetchall()

    parentLFNs = []
    for result in results:
        parentLFNs.append(result[0])

    return parentLFNs

def countPendingFilesForMerge(dbConn, run, dataTier):
    """
    _countPendingFilesForMerge_

    Count the number of files in the Mergeable fileset that belong to a
    particular run and data tier which are not either 'complete' or 'failed'.

    Effectively all files that either are still to be merged or where
    merging has not suceeded or failed yet.

    Calls a different method id tier is RAW.

    """
    if dataTier == "RAW":

        return countPendingFilesForRawMerge(dbConn, run)

    else:

        mergeSubscription = Globals.retrieveMergeSubscription()

        sql = """SELECT COUNT(*)
                 FROM (
                   SELECT wmbs_fileset_files.fileid AS fileid
                   FROM wmbs_fileset_files
                   LEFT OUTER JOIN wmbs_sub_files_complete ON
                     wmbs_fileset_files.fileid = wmbs_sub_files_complete.fileid AND
                     wmbs_sub_files_complete.subscription = %d
                   INNER JOIN wmbs_file_parent ON
                     wmbs_file_parent.child = wmbs_fileset_files.fileid
                   INNER JOIN wmbs_job_assoc ON
                     wmbs_file_parent.parent = wmbs_job_assoc.fileid
                   INNER JOIN wmbs_job a ON
                     a.id = wmbs_job_assoc.job
                   LEFT OUTER JOIN wmbs_job b ON
                     b.id = wmbs_job_assoc.job AND
                     b.outcome = 1
                   WHERE wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = %d)
                   AND wmbs_sub_files_complete.fileid IS NULL
                   GROUP BY wmbs_fileset_files.fileid, a.jobgroup
                   HAVING COUNT(*) = COUNT(b.id)
                 ) merge_files
                 INNER JOIN (
                   SELECT wmbs_file_runlumi_map.fileid AS fileid
                   FROM wmbs_file_runlumi_map
                   WHERE wmbs_file_runlumi_map.run = :RUN_ID
                   GROUP BY wmbs_file_runlumi_map.fileid
                 ) wmbs_file_run_map ON
                   wmbs_file_run_map.fileid = merge_files.fileid
                 INNER JOIN wmbs_file_dataset_path_assoc ON
                   wmbs_file_dataset_path_assoc.file_id = merge_files.fileid
                 INNER JOIN dataset_path ON
                   dataset_path.id = wmbs_file_dataset_path_assoc.dataset_path_id
                 INNER JOIN data_tier ON
                   data_tier.id = dataset_path.data_tier
                 WHERE data_tier.name = :DATATIER
                 """ % ( mergeSubscription["id"],
                         mergeSubscription["id"])

        bindVars = { "RUN_ID": run, "DATATIER" : dataTier }
        dbConn.execute(sql, bindVars, useTransaction = False)
        results = dbConn.fetchall()

        return results[0][0]

def countPendingFilesForRawMerge(dbConn, run):
    """
    _countPendingFilesForRawMerge_

    Similar as countFilesToBeMerged, except requieres that all
    repack jobs for a given run/lumi/stream have suceeded before
    taking into account any Raw file from said run/lumi/stream.

    """
    mergeSubscription = Globals.retrieveMergeSubscription()

    sql = """SELECT COUNT(DISTINCT wmbs_fileset_files.fileid)
             FROM wmbs_fileset_files
             LEFT OUTER JOIN wmbs_sub_files_complete ON
               wmbs_sub_files_complete.fileid = wmbs_fileset_files.fileid AND
               wmbs_sub_files_complete.subscription = %d
             INNER JOIN wmbs_file_parent ON
               wmbs_file_parent.child = wmbs_fileset_files.fileid
             INNER JOIN streamer ON
               streamer.streamer_id = wmbs_file_parent.parent AND
               streamer.run_id = :RUN_ID
             LEFT OUTER JOIN lumi_section_split_active ON
               lumi_section_split_active.run_id = streamer.run_id AND
               lumi_section_split_active.lumi_id = streamer.lumi_id AND
               lumi_section_split_active.stream_id = streamer.stream_id
             WHERE wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = %d)
             AND wmbs_sub_files_complete.fileid IS NULL
             AND lumi_section_split_active.run_id IS NULL
             """ % (mergeSubscription["id"],
                    mergeSubscription["id"])

    bindVars = { "RUN_ID" : run }
    dbConn.execute(sql, bindVars, useTransaction = False)
    results = dbConn.fetchall()

    return results[0][0]

def countPendingFilesForSub(dbConn, run, subscription,
                            inputFromSplitProc = False):
    """
    _countPendingFilesForSub_

    Count the number of files still needing to be processed
    for the given subscription and run

    if inputFromSplitProc is True ignore files for which the
    split processing unit (ie. all jobs processing the same
    parent file) that created them has not been completed

    """
    if inputFromSplitProc:

        sql = """SELECT COUNT(*)
                 FROM (
                   SELECT wmbs_fileset_files.fileid AS fileid
                   FROM wmbs_fileset_files
                   LEFT OUTER JOIN wmbs_sub_files_complete ON
                     wmbs_fileset_files.fileid = wmbs_sub_files_complete.fileid AND
                     wmbs_sub_files_complete.subscription = :SUBSCRIPTION
                   LEFT OUTER JOIN wmbs_sub_files_failed ON
                     wmbs_fileset_files.fileid = wmbs_sub_files_failed.fileid AND
                     wmbs_sub_files_failed.subscription = :SUBSCRIPTION
                   INNER JOIN wmbs_file_parent ON
                     wmbs_file_parent.child = wmbs_fileset_files.fileid
                   INNER JOIN wmbs_job_assoc ON
                     wmbs_file_parent.parent = wmbs_job_assoc.fileid
                   INNER JOIN wmbs_job a ON
                     a.id = wmbs_job_assoc.job
                   LEFT OUTER JOIN wmbs_job b ON
                     b.id = wmbs_job_assoc.job AND
                     b.outcome = 1
                   WHERE wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = :SUBSCRIPTION)
                   AND wmbs_sub_files_complete.fileid IS NULL
                   AND wmbs_sub_files_failed.fileid IS NULL
                   GROUP BY wmbs_fileset_files.fileid, a.jobgroup
                   HAVING COUNT(*) = COUNT(b.id)
                 ) complete_splitunit_files
                 INNER JOIN (
                   SELECT wmbs_file_runlumi_map.fileid AS fileid
                   FROM wmbs_file_runlumi_map
                   WHERE wmbs_file_runlumi_map.run = :RUN_ID
                   GROUP BY wmbs_file_runlumi_map.fileid
                 ) wmbs_file_run_map ON
                   wmbs_file_run_map.fileid = complete_splitunit_files.fileid
                 """

    else:

        sql = """SELECT COUNT(*)
                 FROM wmbs_fileset_files
                 INNER JOIN (
                   SELECT wmbs_file_runlumi_map.fileid AS fileid
                   FROM wmbs_file_runlumi_map
                   WHERE wmbs_file_runlumi_map.run = :RUN_ID
                   GROUP BY wmbs_file_runlumi_map.fileid
                 ) wmbs_file_run_map ON
                   wmbs_file_run_map.fileid = wmbs_fileset_files.fileid
                 LEFT OUTER JOIN wmbs_sub_files_failed ON
                   wmbs_fileset_files.fileid = wmbs_sub_files_failed.fileid AND
                   wmbs_sub_files_failed.subscription = :SUBSCRIPTION
                 LEFT OUTER JOIN wmbs_sub_files_complete ON
                   wmbs_fileset_files.fileid = wmbs_sub_files_complete.fileid AND
                   wmbs_sub_files_complete.subscription = :SUBSCRIPTION
                 WHERE wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = :SUBSCRIPTION)
                 AND wmbs_sub_files_failed.fileid IS NULL
                 AND wmbs_sub_files_complete.fileid IS NULL
                 """

    bindVars = { "RUN_ID" : run, "SUBSCRIPTION" : subscription }
    dbConn.execute(sql, bindVars, useTransaction = False)
    results = dbConn.fetchall()

    return results[0][0]

def getMaxMergeLumi(dbConn, run, datasetpath):
    """
    _getMaxMergeLumi_

    return highest lumi for which merge jobs exists
    for the given run and datasetpath

    returns 0 if no jobs exist

    """
    mergeSubscription = Globals.retrieveMergeSubscription()

    sqlQuery = """SELECT MAX(wmbs_file_runlumi_map.lumi)
                  FROM wmbs_file_runlumi_map
                  INNER JOIN wmbs_file_dataset_path_assoc ON
                    wmbs_file_dataset_path_assoc.file_id = wmbs_file_runlumi_map.fileid AND
                    wmbs_file_dataset_path_assoc.dataset_path_id = :datasetpath
                  INNER JOIN wmbs_job_assoc ON
                    wmbs_job_assoc.fileid = wmbs_file_runlumi_map.fileid
                  INNER JOIN wmbs_job ON
                    wmbs_job.id = wmbs_job_assoc.job
                  INNER JOIN wmbs_jobgroup ON
                    wmbs_jobgroup.id = wmbs_job.jobgroup AND
                    wmbs_jobgroup.subscription = %d
                  WHERE wmbs_file_runlumi_map.run = :run
                  """ % mergeSubscription["id"]

    bindVars = { "run" : run, "datasetpath" : datasetpath }
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    if results[0][0] == None:
        return 0
    else:
        return results[0][0]

def getIncompleteWMBSFiles(dbConn, subscription):
    """
    _getIncompleteWMBSFiles_

    list all non complete DQMenable files

    """
    sqlQuery = """SELECT wmbs_fileset_files.fileid,
                         wmbs_file_runlumi_map.run,
                         dataset_path.primary_dataset,
                         dataset_path.processed_dataset,
                         processed_dataset.name,
                         data_tier.name
                  FROM wmbs_fileset_files
                  LEFT OUTER JOIN wmbs_sub_files_complete ON
                    wmbs_sub_files_complete.fileid = wmbs_fileset_files.fileid AND
                    wmbs_sub_files_complete.subscription = %d
                  INNER JOIN wmbs_file_runlumi_map ON
                    wmbs_file_runlumi_map.fileid = wmbs_fileset_files.fileid
                  INNER JOIN wmbs_file_dataset_path_assoc ON
                    wmbs_file_dataset_path_assoc.file_id = wmbs_fileset_files.fileid
                  INNER JOIN dataset_path ON
                    dataset_path.id = wmbs_file_dataset_path_assoc.dataset_path_id
                  INNER JOIN processed_dataset ON
                    processed_dataset.id = dataset_path.processed_dataset
                  INNER JOIN data_tier ON
                    data_tier.id = dataset_path.data_tier
                  WHERE wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = %d)
                  AND wmbs_sub_files_complete.fileid IS NULL
                  GROUP BY wmbs_fileset_files.fileid,
                           wmbs_file_runlumi_map.run,
                           dataset_path.primary_dataset,
                           dataset_path.processed_dataset,
                           processed_dataset.name,
                           data_tier.name
                  """ % (subscription,
                         subscription)

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    incompleteFiles = []
    for result in results:
        incompleteFile = {}
        incompleteFile['id'] = result[0]
        incompleteFile['run'] = result[1]
        incompleteFile['primds_id'] = result[2]
        incompleteFile['procds_id'] = result[3]
        incompleteFile['procds'] = result[4]
        incompleteFile['tier'] = result[5]
        incompleteFiles.append(incompleteFile)

    return incompleteFiles
