#!/usr/bin/env python
"""
_ListJobs_

Utilities to retrieve information from the jobs tables in T0AST.
"""

__revision__ = "$Id: ListJobs.py,v 1.41 2011/01/20 14:57:51 hufnagel Exp $"
__version__ = "$Revision: 1.41 $"

import os
import logging

from T0 import Globals


def listWMBSJobInfo(dbConn, jobID, jobType = None):
    """
    _listWMBSJobInfo_

    returns job info needed by Tier0Accounter

    """
    if jobType == "Repack" or jobType == "Express":

        sqlQuery = """SELECT wmbs_job_mask.FirstEvent,
                             wmbs_job_mask.LastEvent,
                             wmbs_job_assoc.fileid,
                             wmbs_file_details.lfn
                      FROM wmbs_job_mask
                      INNER JOIN wmbs_job_assoc ON
                        wmbs_job_assoc.job = wmbs_job_mask.job
                      INNER JOIN wmbs_file_details ON
                        wmbs_file_details.id = wmbs_job_assoc.fileid
                      WHERE wmbs_job_mask.job = :ID
                      """

    else:

        sqlQuery = """SELECT wmbs_job_mask.FirstEvent,
                             wmbs_job_mask.LastEvent,
                             wmbs_job_assoc.fileid
                      FROM wmbs_job_mask
                      INNER JOIN wmbs_job_assoc ON
                        wmbs_job_assoc.job = wmbs_job_mask.job
                      WHERE wmbs_job_mask.job = :ID
                      """

    bindVars = { 'ID' : jobID }
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    wmbsJobInfo = {}
    for result in results:

        if len(wmbsJobInfo) == 0:
            wmbsJobInfo['first_event'] = result[0]
            wmbsJobInfo['last_event'] = result[1]
            wmbsJobInfo['input_files'] = []
            if jobType == "Repack" or jobType == "Express":
                wmbsJobInfo['input_filenames'] = []

        wmbsJobInfo['input_files'].append(result[2])

        if jobType == "Repack" or jobType == "Express":
            wmbsJobInfo['input_filenames'].append(os.path.basename(result[3]))


    if len(wmbsJobInfo) == 0:
        return None
    else:
        return wmbsJobInfo


def listNewWMBSJobs(dbConn, subscription):
    """
    _listNewWMBSJobs_

    gets all not yet created jobs for a certain subscription
    also return additional information for injector component
    sort the files for each job by first lumi and first event

    """
    sqlQuery = """SELECT wmbs_job.id,
                         wmbs_job_mask.FirstEvent,
                         wmbs_job_mask.LastEvent,
                         wmbs_file_details.lfn,
                         wmbs_file_run_map.run,
                         wmbs_file_dataset_path_assoc.dataset_path_id
                  FROM wmbs_job
                  INNER JOIN wmbs_jobgroup ON
                    wmbs_jobgroup.id = wmbs_job.jobgroup AND
                    wmbs_jobgroup.subscription = :subscription
                  INNER JOIN wmbs_job_mask ON
                    wmbs_job_mask.job = wmbs_job.id
                  INNER JOIN wmbs_job_assoc ON
                    wmbs_job_assoc.job = wmbs_job.id
                  INNER JOIN wmbs_file_details ON
                    wmbs_file_details.id = wmbs_job_assoc.fileid
                  INNER JOIN (
                    SELECT wmbs_file_runlumi_map.fileid AS fileid,
                           wmbs_file_runlumi_map.run AS run,
                           MIN(wmbs_file_runlumi_map.lumi) AS first_lumi
                    FROM wmbs_file_runlumi_map
                    GROUP BY wmbs_file_runlumi_map.fileid,
                             wmbs_file_runlumi_map.run
                  ) wmbs_file_run_map ON
                    wmbs_file_run_map.fileid = wmbs_job_assoc.fileid
                  INNER JOIN wmbs_file_dataset_path_assoc ON
                    wmbs_file_dataset_path_assoc.file_id = wmbs_job_assoc.fileid
                  WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = 'new')
                  ORDER BY wmbs_file_run_map.run,
                           wmbs_file_run_map.first_lumi,
                           wmbs_file_details.first_event
                  """

    bindVars = { 'subscription' : subscription }
    dbConn.execute(sqlQuery, subscription, useTransaction = False)
    results = dbConn.fetchall()

    wmbsJobs = {}
    for result in results:
        id = result[0]
        if wmbsJobs.has_key(id):
            wmbsJob = wmbsJobs[id]
            wmbsJob['files'].append(result[3])
        else:
            wmbsJob = {}
            wmbsJob['id'] = result[0]
            wmbsJob['first_event'] = result[1]
            wmbsJob['last_event'] = result[2]
            wmbsJob['files'] = [result[3]]
            wmbsJob['run'] = result[4]
            wmbsJob['datasetpath'] = result[5]
            wmbsJobs[id] = wmbsJob

    return wmbsJobs.values()


def listNewStreamerJobs(dbConn, subscription):
    """
    _listNewStreamerJobs_

    streamer files are not associated to datasetpath, therefore
    the generic listNewWMBSJobs method does not work

    gets all not yet created jobs for a certain subscription
    also return additional information for injector component
    sort the files for each job by run, lumi and streamer id

    """
    sqlQuery = """SELECT wmbs_job.id,
                         wmbs_file_runlumi_map.run,
                         wmbs_file_runlumi_map.lumi,
                         streamer.stream_id,
                         wmbs_file_details.lfn
                  FROM wmbs_job
                  INNER JOIN wmbs_jobgroup ON
                    wmbs_jobgroup.id = wmbs_job.jobgroup AND
                    wmbs_jobgroup.subscription = :SUBSCRIPTION
                  INNER JOIN wmbs_job_assoc ON
                    wmbs_job_assoc.job = wmbs_job.id
                  INNER JOIN wmbs_file_details ON
                    wmbs_file_details.id = wmbs_job_assoc.fileid
                  INNER JOIN wmbs_file_runlumi_map ON
                    wmbs_file_runlumi_map.fileid = wmbs_job_assoc.fileid
                  INNER JOIN streamer ON
                    streamer.streamer_id = wmbs_job_assoc.fileid
                  WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = 'new')
                  ORDER BY streamer.run_id,
                           streamer.lumi_id,
                           streamer.streamer_id
                  """

    bindVars = { 'SUBSCRIPTION' : subscription }
    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = dbConn.fetchall()

    #
    # now sort results by run and stream (optimizes workflow creation)
    #
    # needs to provide mapping of job id to input file lfn
    #
    # also needs to find lumis with multiple jobs per run/stream
    # (either repack jobs with lumi splitting or express), for
    # repack we need to force unmerged output
    #
    jobDict = {}
    for result in results:

        id = result[0]
        run = result[1]
        lumi = result[2]
        stream_id = result[3]
        lfn = result[4]

        if not jobDict.has_key(run):
            jobDict[run] = {}
        if not jobDict[run].has_key(stream_id):
            jobDict[run][stream_id] = { 'lumis' : {},
                                        'files' : {} }

        # convinience variable
        lumisDict = jobDict[run][stream_id]['lumis']
        filesDict = jobDict[run][stream_id]['files']

        # keep a lumi to job id reference
        # allows to check which lumis have multiple jobs
        if not lumisDict.has_key(lumi):
            lumisDict[lumi] = set([id])
        else:
            lumisDict[lumi].add(id)

        # keep a job id to lfn reference
        if not filesDict.has_key(id):
            filesDict[id] = [lfn]
        elif lfn not in filesDict[id]:
            filesDict[id].append(lfn)

    for run in jobDict.keys():
        for stream_id in jobDict[run].keys():

            jobDict[run][stream_id]['normal'] = {}
            jobDict[run][stream_id]['split'] = {}

            # convinience variable
            lumisDict = jobDict[run][stream_id]['lumis']
            filesDict = jobDict[run][stream_id]['files']
            normalDict = jobDict[run][stream_id]['normal']
            splitDict = jobDict[run][stream_id]['split']

            # a job is either single lumi split or multiple lumis, not both
            # based on that divide jobs into normal and split lumi jobs
            for lumi in sorted(lumisDict.keys()):
                jobIds = lumisDict[lumi]
                if len(jobIds) == 1:
                    id = jobIds.pop()
                    if not normalDict.has_key(id):
                        normalDict[id] = filesDict[id]
                else:
                    for id in jobIds:
                        splitDict[id] = filesDict[id]

            lumisDict = None
            filesDict = None
            del jobDict[run][stream_id]['lumis']
            del jobDict[run][stream_id]['files']

    return jobDict
