#!/usr/bin/env python
"""
_ListRunConfig_

Utilities for retrieving RunConfig information from T0AST.
"""

__revision__ = "$Id: ListRunConfig.py,v 1.77 2010/11/05 10:05:57 gowdy Exp $"
__version__ = "$Revision: 1.77 $"

def runExists(dbConn, runNumber):
    """
    _runExists_

    Check if a run exists in T0AST.  A row is added to the run table by the
    transfer system, but the ACQ_ERA column is empty.  This is what we
    use to decide if a RunConfig needs to be inserted for a run.
    """
    sqlQuery = "SELECT ACQ_ERA FROM run WHERE RUN_ID = :p_1"
    bindVars = {"p_1": runNumber}

    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    runExistsResult = dbConn.fetchall()

    if runExistsResult == []:
        return False

    if runExistsResult[0][0] == None:
        return False

    return True

def retrieveStreamToDatasets(dbConn, runNumber):
    """
    _retrieveStreamDatasetMapping_

    Query T0AST for the mapping of streams to datasets for a given run.
    """
    sqlQuery = """SELECT stream_id, primary_dataset_id from dataset_run_stream_assoc
                  WHERE run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    # now convert into dictionary
    streamToDatasets = {}
    for row in dbConn.fetchall():
        stream = row[0]
        dataset = row[1]
        if streamToDatasets.has_key(stream):
            streamToDatasets[stream].append(dataset)
        else:
            streamToDatasets[stream] = [ dataset ]

    return streamToDatasets

def retrieveDatasetToTriggers(dbConn, runNumber):
    """
    _retrieveDatasetToTriggers_

    Query T0AST for the mapping of datasets to triggers for a given run.
    """
    sqlQuery = """SELECT trig_dataset_assoc.PRIMARY_DATASET_ID,
                  trigger_label.LABEL FROM trig_dataset_assoc INNER JOIN
                  trigger_label ON trig_dataset_assoc.TRIG_ID = trigger_label.ID
                  WHERE trig_dataset_assoc.RUN_ID = :p_1 GROUP BY
                  trig_dataset_assoc.PRIMARY_DATASET_ID, trigger_label.LABEL"""

    bindVars = {"p_1" : runNumber}
    dbConn.execute(sqlQuery, bindVars)

    # rows with 1st element the dataset id, 2nd the trigger label
    # now convert into dictionary

    datasetToTriggers = {}
    for row in dbConn.fetchall():
        if datasetToTriggers.has_key(row[0]):
            datasetToTriggers[row[0]].append(row[1])
        else:
            datasetToTriggers[row[0]] = [row[1]]

    return datasetToTriggers

def retrievePrimaryDatasets(dbConn, runNumber):
    """
    _retrievePrimaryDatasets_

    Query T0AST for dataset ids and names for this run.
    """
    sqlQuery = """SELECT a.id, a.name FROM primary_dataset a
                  INNER JOIN dataset_run_stream_assoc b
                  ON b.primary_dataset_id = a.id
                  WHERE b.run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    # rows with 1st element the dataset id
    #           2nd element the dataset name
    # now convert into dictionary
    result = {}
    for row in dbConn.fetchall():
        result[row[0]] = row[1]
    return result

def retrieveDatasetToErrorDatasetAssoc(dbConn, runNumber):
    """
    _retrieveDatasetsToErrorDatasetAssoc_

    Query T0AST for dataset ids and corresponding error dataset ids for this run.
    """
    sqlQuery = """SELECT a.parent_id, a.error_id FROM dataset_error_dataset_assoc a
                  INNER JOIN dataset_run_stream_assoc b
                  ON b.primary_dataset_id = a.parent_id
                  WHERE b.run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    # rows with 1st element the parent dataset id
    #           2nd element the error dataset id
    # convert to dictionary
    result = {}
    for row in dbConn.fetchall():
        result[row[0]] = row[1]
    return result

def retrieveStreams(dbConn, runNumber):
    """
    _retrieveStreams_

    Query T0AST for stream ids and names for this run.
    """
    sqlQuery = """SELECT a.id, a.name FROM stream a
                  INNER JOIN dataset_run_stream_assoc b
                  ON b.stream_id = a.id
                  WHERE b.run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    # rows with 1st element the stream id
    #           2nd element the stream name
    # convert to dictionary
    result = {}
    for row in dbConn.fetchall():
        result[row[0]] = row[1]
    return result

def retrieveStreamToSpecialDatasetAssoc(dbConn, runNumber):
    """
    _retrieveStreamToSpecialDatasetAssoc_

    Query T0AST for stream ids and corresponding special dataset ids for this run.
    """
    sqlQuery = """SELECT a.id, c.primary_dataset_id FROM stream a
                  INNER JOIN dataset_run_stream_assoc b
                  ON b.stream_id = a.id
                  INNER JOIN stream_special_dataset_assoc c
                  ON c.stream_id = a.id
                  WHERE b.run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    # rows with 1st element the stream id
    #           2nd element the special dataset id
    # convert to dictionary
    result = {}
    for row in dbConn.fetchall():
        result[row[0]] = row[1]
    return result

def retrieveVersionAndHLTKey(dbConn, runNumber):
    """
    _retrieveVersionAndHLTKey_

    Retrieve the Online version and HLT key for a run
    """
    sqlQuery = """SELECT b.name, a.hltkey FROM run a
                  INNER JOIN cmssw_version b
                  ON b.id = a.run_version
                  WHERE a.run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    return dbConn.fetchall()

def retrieveRunInfo(dbConn, runNumber):
    """
    _retrieveRunInfo_

    Retrieve the CMSSW version, process name, start time, acquisition era
    and processing version from the run table.
    """
    sqlQuery = """SELECT a.process, a.start_time, a.acq_era FROM run a
                  WHERE a.run_id = :p_1"""
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    return dbConn.fetchall();

def retrieveScenario(dbConn, runNumber):
    """
    _retrieveScenario_

    Retrieve scenario name from event_scenario and scenario_config table
    """
    sqlQuery = """ SELECT primary_dataset_id, scenario FROM scenario_config
                      INNER JOIN event_scenario ON scenario_id = id
                      WHERE run_id = :p_1 """

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    scenarios = {}

    for row in dbConn.fetchall():
        scenarios[row[0]] = row[1]
    return scenarios

def retrieveRepackConfig(dbConn, runNumber):
    """
    _retrieveRepackConfig_ 
    
    Retrieve data from repack_config and the repacker version from run
    Return as dict of datasets and processing versions
    """
    sqlQuery = """SELECT repack_config.stream_id,
                         repack_config.proc_version,
                         cmssw_version.name
                  FROM repack_config
                  INNER JOIN run_stream_cmssw_assoc ON
                    run_stream_cmssw_assoc.stream_id = repack_config.stream_id AND
                    run_stream_cmssw_assoc.run_id = :p_1
                  INNER JOIN cmssw_version ON
                    cmssw_version.id = run_stream_cmssw_assoc.override_version
                  WHERE repack_config.run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    repackConfigs = {}
    for row in results:

        repackConfig = {}
        repackConfig["PROC_VER"] = row[1]
        repackConfig["CMSSW"] = row[2]

        repackConfigs[row[0]] = repackConfig

    return repackConfigs
        
def retrieveRecoConfig(dbConn, runNumber):
    """
    _retrieveRecoConfig_ 
    
    Retrieve data from reco_config
    organize as dict
    
    [dataset_id][parameter] = value
    
    parameters
    ["DO_RECO"]  : boolean True, False
    ["GLOBALTAG"] : String
    ["CMSSW_VERSION"] : String
    ["CONFIG_URL"] : String
    ["PROC_VER"] : String
    """
    
    sqlQuery = """ SELECT a.primary_dataset_id, a.do_reco, a.global_tag,
                          b.name, a.config_url, a.proc_version, a.reco_split,
                          c.alca, a.write_reco, a.write_aod
                   FROM reco_config a
                   INNER JOIN cmssw_version b
                   ON cmssw_version_id = id
                   LEFT OUTER JOIN run_dataset_alca_assoc c
                   ON c.run_id = :p_1 AND c.primary_dataset_id = a.primary_dataset_id
                   WHERE a.run_id = :p_1 """

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    recoConfigs = {}
    for result in results:
        primds_id = result[0]
        if recoConfigs.has_key(primds_id):
            recoConfig = recoConfigs[primds_id]
            recoConfig["PRODUCERS"].append(result[7])
        else:
            recoConfig = {}
            recoConfig["DO_RECO"] = bool(result[1])
            recoConfig["GLOBALTAG"] = result[2]
            recoConfig["CMSSW_VERSION"] = result[3]
            recoConfig["CONFIG_URL"] = result[4]
            recoConfig["PROC_VER"] = result[5]
            recoConfig["EVENT_SPLIT"] = result[6]
            recoConfig["WRITE_RECO"] = bool(result[8])
            recoConfig["WRITE_AOD"] = bool(result[9])
            if result[7] == None:
                recoConfig["PRODUCERS"] = []
            else:
                recoConfig["PRODUCERS"] = [ result[7] ]
            recoConfigs[primds_id] = recoConfig

    return recoConfigs

def retrieveAlcaConfig(dbConn, runNumber):
    """
    _retrieveAlcaConfig_ 

    Retrieve data from alca_config
    organize as dict

    [dataset_id][parameter] = value

    parameters
    ["DO_ALCA"]  : boolean True, False 
    ["CMSSW_VERSION"] : String
    ["CONFIG_URL"] : String
    ["PROC_VER"] : String
    """

    sqlQuery = """ SELECT a.primary_dataset_id, a.do_alca, b.name,
                          a.config_url, a.proc_version, c.alca
                   FROM alca_config a
                   INNER JOIN cmssw_version b
                   ON b.id = a.cmssw_version_id
                   LEFT OUTER JOIN run_dataset_alca_assoc c
                   ON c.run_id = :p_1 AND c.primary_dataset_id = a.primary_dataset_id
                   WHERE a.run_id = :p_1 """

    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    alcaConfigs = {}
    for result in results:
        primds_id = result[0]
        if alcaConfigs.has_key(primds_id):
            alcaConfig = alcaConfigs[primds_id]
            alcaConfig["PRODUCERS"].append(result[5])
        else:
            alcaConfig = {}
            alcaConfig["DO_ALCA"] = bool(result[1])
            alcaConfig["CMSSW_VERSION"] = result[2]
            alcaConfig["CONFIG_URL"] = result[3]
            alcaConfig["PROC_VER"] = result[4]
            if result[5] == None:
                alcaConfig["PRODUCERS"] = []
            else:
                alcaConfig["PRODUCERS"] = [ result[5] ]
            alcaConfigs[primds_id] = alcaConfig

    return alcaConfigs


def retrieveCalibConfig(dbConn, runNumber):
    """
    _retrieveCalibConfig_ 

    Retrieve data from calib_config
    organize as a two level dict with first key: primary dataset id
                                     second key: processed dataset
    [primds_id][procds][parameter] = value

    parameters:
    ["CMSSW_VERSION"] : String
    ["CONFIG_URL"] : String
    ["GLOBAL_TAG"] : String
    ["SPLIT_ALGO"] : String
    ["SPLIT_PARAM"] : String
    """
    sqlQuery = """ SELECT a.primary_dataset_id, a.processed_dataset, b.name,
                   a.config_url, a.global_tag, a.split_algo, a.split_param
                   FROM calib_config a INNER JOIN cmssw_version b
                   ON b.id = a.cmssw_version_id WHERE run_id = :p_1 """
    
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    t0CalibConfigs = {}
    for row in dbConn.fetchall():
        if row[0] not in t0CalibConfigs.keys():
            t0CalibConfigs[row[0]] = {}
        if row[1] not in t0CalibConfigs[row[0]]:
            t0CalibConfigs[row[0]][row[1]]={}
            
        t0CalibConfigs[row[0]][row[1]]["CMSSW_VERSION"] = row[2]
        t0CalibConfigs[row[0]][row[1]]["CONFIG_URL"] = row[3]
        t0CalibConfigs[row[0]][row[1]]["GLOBAL_TAG"] = row [4]
        t0CalibConfigs[row[0]][row[1]]["SPLIT_ALGO"] = row [5]
        t0CalibConfigs[row[0]][row[1]]["SPLIT_PARAM"] = row [6]

    return t0CalibConfigs

        
def retrieveWMBSPublishConfig(dbConn, runNumber):
    """
    _retrieveWMBSPublishConfig_ 
    
    Retrieve data from wmbs_publish_config
    organize as dict
    
    [dataset_id][parameter] = value
    
    parameters
    ["DO_WMBS_PUBLISH"]  : boolean True, False 
    ["DATA_TIERS_TO"] : List ( of <DataTier>[,DataTier]@<location> )
    """
    
    sqlQuery = """ SELECT primary_dataset_id, do_wmbs_publish,
                          data_tiers_to
                   FROM wmbs_publish_config
                   WHERE run_id = :p_1 """ 
    
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    t0WMBSPublishConfigs = {}
    
    for row in dbConn.fetchall():
        t0WMBSPublishConfigs[row[0]] = {}
        t0WMBSPublishConfigs[row[0]]["DO_WMBS_PUBLISH"] = bool(row[1])
        dataTiersTo = []
        if row[2]:
            dataTiersTo = row[2].split()
        t0WMBSPublishConfigs[row[0]]["DATA_TIERS_TO"] = dataTiersTo
    
    return t0WMBSPublishConfigs

def retrieveDQMConfig(dbConn, runNumber):
    """
    _retrieveDQMConfig_ 
    
    Retrieve data from alca_config
    organize as dict
    
    [dataset_id][parameter] = value
    
    parameters
    ["DO_DQM"]  : boolean True, False 
    ["CMSSW_VERSION"] : String
    ["CONFIG_URL"] : String
    ["PROC_VER"] : String
    """
    sqlQuery = """ SELECT primary_dataset_id, do_dqm, name, config_url,
                   proc_version FROM dqm_config
                   INNER JOIN cmssw_version ON cmssw_version_id = id 
                   WHERE run_id = :p_1 """ 
    
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)
    t0DQMConfigs = {}
    
    for row in dbConn.fetchall():
        t0DQMConfigs[row[0]] = {}
        t0DQMConfigs[row[0]]["DO_DQM"] = bool(row[1])
        t0DQMConfigs[row[0]]["CMSSW_VERSION"] = row[2]
        t0DQMConfigs[row[0]]["CONFIG_URL"] = row[3]
        t0DQMConfigs[row[0]]["PROC_VER"] = row[4]
    
    return t0DQMConfigs       

def retrieveTier1SkimConfig(dbConn, runNumber):
    """
    _retrieveTier1SkimConfig_

    Retrieve all the Tier1 Skim configurations for a given run.  The skims are
    returned in the form of a two level dictionary where the primary dataset ID
    is the first key and the data tier ID is the second key.
    """
    sqlQuery = """SELECT primary_dataset_id, data_tier_id, cmssw_version.name,
                  two_file_read, skim_name, config_url, global_tag, proc_version
                  FROM t1skim_config INNER JOIN cmssw_version
                  ON cmssw_version.id = cmssw_version_id
                  WHERE run_id = :p_1"""
    
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    skimConfigs = {}
    for skimConfig in dbConn.fetchall():
        if skimConfig[0] not in skimConfigs.keys():
            skimConfigs[skimConfig[0]] = {}
        if skimConfig[1] not in skimConfigs[skimConfig[0]]:
            skimConfigs[skimConfig[0]][skimConfig[1]] = []

        newConfig = {}
        newConfig["CMSSW_VERSION"] = skimConfig[2]
        newConfig["TWO_FILE_READ"] = bool(skimConfig[3])
        newConfig["SKIM_NAME"] = skimConfig[4]
        newConfig["CONFIG_URL"] = skimConfig[5]
        newConfig["GLOBAL_TAG"] = skimConfig[6]
        newConfig["PROC_VER"] = skimConfig[7]

        skimConfigs[skimConfig[0]][skimConfig[1]].append(newConfig)
        
    return skimConfigs

def retrieveExpressConfig(dbConn, runNumber):
    """
    _retrieveExpressConfig_

    Retrieve all the express configurations for a given run.  The results are
    returned in the form of a dictionary that is keyed by stream id.
    """
    sqlQuery = """SELECT a.stream_id, a.processing_config_url, a.global_tag,
                         a.splitInProcessing, a.proc_version, a.alcamerge_config_url,
                         c.name, g.alca, f.scenario, i.name
                  FROM express_config a
                  INNER JOIN run_stream_tier_assoc b
                  ON ( b.run_id = :p_1 AND b.stream_id = a.stream_id )
                  INNER JOIN data_tier c
                  ON c.id = b.data_tier_id
                  INNER JOIN stream_special_dataset_assoc d
                  ON d.stream_id = a.stream_id
                  INNER JOIN scenario_config e
                  ON ( e.run_id = :p_1 AND e.primary_dataset_id = d.primary_dataset_id )
                  INNER JOIN event_scenario f
                  ON f.id = e.scenario_id
                  LEFT OUTER JOIN run_dataset_alca_assoc g
                  ON ( g.run_id = :p_1 AND g.primary_dataset_id = d.primary_dataset_id )
                  INNER JOIN run_stream_cmssw_assoc h
                  ON ( h.run_id = :p_1 AND h.stream_id = a.stream_id )
                  INNER JOIN cmssw_version i
                  ON i.id = h.override_version
                  WHERE a.run_id = :p_1"""
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    expressConfigs = {}
    for result in dbConn.fetchall():
        stream_id = result[0]
        if expressConfigs.has_key(stream_id):
            expressConfig = expressConfigs[stream_id]
            if result[6] not in expressConfig["DATATIERS"]:
                expressConfig["DATATIERS"].append(result[6])
            if result[7] != None and \
                   result[7] not in expressConfig["PRODUCERS"]:
                expressConfig["PRODUCERS"].append(result[7])
        else:
            expressConfig = {}
            expressConfig["PROC_URL"] = result[1]
            expressConfig["GLOBAL_TAG"] = result[2]
            expressConfig["SPLIT_IN_PROC"] = result[3]
            expressConfig["PROC_VER"] = result[4]
            expressConfig["ALCAMERGE_URL"] = result[5]
            expressConfig["DATATIERS"] = [ result[6] ]
            if result[7] == None:
                expressConfig["PRODUCERS"] = []
            else:
                expressConfig["PRODUCERS"] = [ result[7] ]
            expressConfig["SCENARIO"] = result[8]
            expressConfig["CMSSW"] = result[9]
            expressConfigs[stream_id] = expressConfig

    return expressConfigs

def retrieveRegistrationConfig(dbConn, runNumber):
    """
    _retrieveRegistrationConfig_

    Retrieve all the registration configurations for a given run.  The results
    are returned in the form of a dictionary that is keyed by stream id.
    """
    sqlQuery = """SELECT a.stream_id, b.name,
                         c.name, d.name FROM regist_config a
                  INNER JOIN primary_dataset b
                  ON b.id = a.primds_id
                  INNER JOIN processed_dataset c
                  ON c.id = a.procds_id
                  INNER JOIN data_tier d
                  ON d.id = a.data_tier_id
                  WHERE a.run_id = :p_1"""
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    registrationConfigs = {}
    for result in dbConn.fetchall():
        registrationConfig = {}
        registrationConfig["PRIMDS"] = result[1]
        registrationConfig["PROCDS"] = result[2]
        registrationConfig["DATATIER"] = result[3]
        registrationConfigs[result[0]] = registrationConfig

    return registrationConfigs

def retrieveConversionConfig(dbConn, runNumber):
    """
    _retrieveConversionConfig_

    Retrieve all the conversion configurations for a given run.  The results
    are returned in the form of a dictionary that is keyed by stream id.
    """
    sqlQuery = """SELECT a.stream_id, b.name, c.name, a.conv_type,
                         a.acq_era, a.proc_vers, a.proc_string, e.name FROM conv_config a
                  INNER JOIN primary_dataset b
                  ON b.id = a.primds_id
                  INNER JOIN data_tier c
                  ON c.id = a.data_tier_id
                  INNER JOIN run_stream_cmssw_assoc d
                  ON d.run_id = a.run_id AND d.stream_id = a.stream_id
                  INNER JOIN cmssw_version e
                  ON e.id = d.override_version
                  WHERE a.run_id = :p_1"""
    bindVars = {"p_1": runNumber}
    dbConn.execute(sqlQuery, bindVars)

    conversionConfigs = {}
    for result in dbConn.fetchall():
        conversionConfig = {}
        conversionConfig["PRIMDS"] = result[1]
        conversionConfig["DATATIER"] = result[2]
        conversionConfig["CONV_TYPE"] = result[3]
        conversionConfig["ACQ_ERA"] = result[4]
        conversionConfig["PROC_VERS"] = result[5]
        conversionConfig["PROC_STRING"] = result[6]
        conversionConfig["CONV_VERSION"] = result[7]
        conversionConfigs[result[0]] = conversionConfig

    return conversionConfigs

def retrieveProcessingStyles(dbConn, runNumber):
    """
    _retrieveProcessingStyles_

    Retrieve the stream ID to processing style mappins for the given run.
    """
    sql = """SELECT a.stream_id, b.name FROM run_stream_style_assoc a
             INNER JOIN processing_style b
             ON b.id = a.style_id
             WHERE run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sql, bindVars)

    processingStyle = {}
    for result in dbConn.fetchall():
        processingStyle[result[0]] = result[1] 

    return processingStyle

def listStreamInfoForRun(dbConn, runNumber):
    """
    _listStreamInfoForRun_

    List stream names, processing style and the acquisition era for a run.  This
    is used to generate the config data structure for couch.
    """
    sql = """SELECT stream.name, processing_style.name, run.acq_era,
                    run_stream_cmssw_assoc.override_version
               FROM stream
               INNER JOIN dataset_run_stream_assoc ON
                 stream.id = dataset_run_stream_assoc.stream_id
               INNER JOIN run_stream_style_assoc ON
                 dataset_run_stream_assoc.run_id = run_stream_style_assoc.run_id AND
                 dataset_run_stream_assoc.stream_id = run_stream_style_assoc.stream_id
               INNER JOIN processing_style ON
                 run_stream_style_assoc.style_id = processing_style.id
               INNER JOIN run ON
                 dataset_run_stream_assoc.run_id = run.run_id
               INNER JOIN run_stream_cmssw_assoc ON
                 run_stream_cmssw_assoc.run_id = run.run_id AND
                 run_stream_cmssw_assoc.stream_id = stream.id
             WHERE run.run_id = :p_1"""

    bindVars = {"p_1": runNumber}
    dbConn.execute(sql, bindVars)
    return dbConn.fetchall()

def listDatasetInfoForRun(dbConn, runNumber, streamName):
    """
    _listDatasetInfoForRun_

    List bulk configuration information for all datasets in a given stream for
    a given run.  This is used to generate the config data structure for couch.
    """
    sql = """SELECT primary_dataset.name, event_scenario.scenario, repack_config.proc_version,
                    repack_version.name, reco_config.global_tag, reco_config.do_reco,
                    reco_config.proc_version, reco_version.name,
                    reco_config.config_url, alca_config.do_alca, alca_config.proc_version,
                    alca_version.name, alca_config.config_url, dqm_config.do_dqm
               FROM primary_dataset
               INNER JOIN dataset_run_stream_assoc ON
                 primary_dataset.id = dataset_run_stream_assoc.primary_dataset_id
               INNER JOIN stream ON
                 dataset_run_stream_assoc.stream_id = stream.id
               INNER JOIN repack_config ON
                 dataset_run_stream_assoc.run_id = repack_config.run_id AND
                 stream.id = repack_config.stream_id
               INNER JOIN run ON
                 dataset_run_stream_assoc.run_id = run.run_id
               INNER JOIN cmssw_version repack_version ON
                 run.repack_version = repack_version.id                 
               INNER JOIN scenario_config ON
                 dataset_run_stream_assoc.run_id = scenario_config.run_id AND
                 dataset_run_stream_assoc.primary_dataset_id = dataset_run_stream_assoc.primary_dataset_id
               INNER JOIN event_scenario ON
                 scenario_config.scenario_id = event_scenario.id
               INNER JOIN reco_config ON
                 dataset_run_stream_assoc.run_id = reco_config.run_id AND
                 dataset_run_stream_assoc.primary_dataset_id = reco_config.primary_dataset_id
               INNER JOIN cmssw_version reco_version ON
                 reco_config.cmssw_version_id = reco_version.id
               INNER JOIN alca_config ON
                 dataset_run_stream_assoc.run_id = alca_config.run_id AND
                 dataset_run_stream_assoc.primary_dataset_id = alca_config.primary_dataset_id
               INNER JOIN cmssw_version alca_version ON
                 alca_config.cmssw_version_id = alca_version.id
               INNER JOIN dqm_config ON
                 dataset_run_stream_assoc.run_id = dqm_config.run_id AND
                 dataset_run_stream_assoc.primary_dataset_id = dqm_config.primary_dataset_id
             WHERE run.run_id = :p_1 and stream.name = :p_2"""

    bindVars = {"p_1": runNumber, "p_2": streamName}
    dbConn.execute(sql, bindVars)
    return dbConn.fetchall()    

def listSkimInfoForRun(dbConn, runNumber, datasetName):
    """
    _listSkimInfoForRun_

    List configuration information for all the Tier1 skimms for a particular
    dataset in a particular run.  This information is used to generate the
    config data structure for couch.
    """
    sql = """SELECT t1skim_config.skim_name, data_tier.name, cmssw_version.name,
                    t1skim_config.proc_version, t1skim_config.config_url,
                    t1skim_config.global_tag FROM t1skim_config
               INNER JOIN data_tier ON
                 t1skim_config.data_tier_id = data_tier.id
               INNER JOIN cmssw_version ON
                 t1skim_config.cmssw_version_id = cmssw_version.id
               INNER JOIN primary_dataset ON
                 t1skim_config.primary_dataset_id = primary_dataset.id
             WHERE t1skim_config.run_id = :p_1 AND primary_dataset.name = :p_2"""

    bindVars = {"p_1": runNumber, "p_2": datasetName}
    dbConn.execute(sql, bindVars)
    return dbConn.fetchall()

def listExpressConfigForStream(dbConn, runNumber, streamName):
    """
    _listExpressConfigForStream_

    Retrieve express configuration information for the given run and stream.
    This is used to generate the couch db configuration data structure.
    """
    sql = """SELECT data_tier.name, express_config.global_tag, express_config.proc_version,
                    express_config.processing_config_url, express_config.alcamerge_config_url
                    FROM data_tier
               INNER JOIN run_stream_tier_assoc ON
                 data_tier.id = run_stream_tier_assoc.data_tier_id
               INNER JOIN stream ON
                 run_stream_tier_assoc.stream_id = stream.id
               INNER JOIN express_config ON
                 run_stream_tier_assoc.run_id = express_config.run_id AND
                 run_stream_tier_assoc.stream_id = express_config.stream_id
             WHERE run_stream_tier_assoc.run_id = :p_1 AND stream.name = :p_2"""

    bindVars = {"p_1": runNumber, "p_2": streamName}
    dbConn.execute(sql, bindVars)
    return dbConn.fetchall()
