#!/usr/bin/env python
"""
_ListRuns_

Utilities to retrieve information from the run table in T0AST.
"""

__revision__ = "$Id: ListRuns.py,v 1.46 2011/01/12 21:12:52 hufnagel Exp $"
__version__ = "$Revision: 1.46 $"

import logging

from T0 import Globals


def listRunsByStatus(dbConn, status):
    """
    _listRunsByStatus_

    return all runs with the given status

    """
    sqlQuery = """SELECT run_id
                  FROM run
                  WHERE run_status = (SELECT id FROM run_status WHERE status = :STATUS)"""

    bindVars = { 'STATUS' : status }
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    runs = []
    for result in results:
        runs.append(result[0])

    return runs


def listRunsInCloseout(dbConn):
    """
    _listRunsInCloseout_

    get all runs being closed out, ie, whose
    status is not Active and not Complete

    also return status

    """
    sqlQuery = """SELECT run.run_id, run_status.status
                  FROM run
                  INNER JOIN run_status ON
                    run.run_status = run_status.id
                  WHERE run_status.status NOT IN ( 'Active', 'Complete' )"""

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    runs = {}
    for result in results:
        runs[result[0]] = result[1]

    return runs


def listRunState(dbConn, runNumbers):
    """
    _listRunState_

    Retrieve the state of a run for a given run number.

    """
    if type(runNumbers) != list:
        runNumbers = [runNumbers]

    sqlQuery = """SELECT run.run_id, run_status.status
                  FROM run
                  INNER JOIN run_status ON
                    run_status.id = run.run_status
                  WHERE run.run_id = :p_1"""

    bindVars = []
    for runNumber in runNumbers:
        bindVar = {"p_1": runNumber}
        bindVars.append(bindVar)
        
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    runStatus = {}
    for result in results:
        runStatus[result[0]] = result[1]

    return runStatus


def listExpressProcCompleteForRun(dbConn, runNumber):
    """
    _listExpressProcCompleteForRun_

    Determine whether or not a run has completed express processing.
    """
    sqlQuery = """SELECT close_time
                  FROM run
                  WHERE run_id = :RUN_ID
                  """

    bindVars = { "RUN_ID" : runNumber }
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    if results[0][0] == 0:
        logging.debug("Run status still Active for run %d, cannot express close" % runNumber)
        return False

    sqlQuery = """SELECT COUNT(*)
                  FROM streamer
                  INNER JOIN run_stream_style_assoc ON
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Express')
                  INNER JOIN run ON
                    run.run_id = streamer.run_id
                  WHERE checkForZeroState(streamer.used) = 0
                  AND streamer.run_id = :RUN_ID
                  """

    bindVars = { "RUN_ID" : runNumber }
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    if results[0][0] != 0:
        logging.debug("Unused express streamer for run %d: %d" % (runNumber, results[0][0]))
        return False

    expressSubscription = Globals.retrieveExpressSubscription()

    sqlQuery = """SELECT COUNT(*)
                  FROM wmbs_fileset_files
                  LEFT OUTER JOIN wmbs_sub_files_complete ON
                    wmbs_sub_files_complete.fileid = wmbs_fileset_files.fileid AND
                    wmbs_sub_files_complete.subscription = %d
                  LEFT OUTER JOIN wmbs_sub_files_failed ON
                    wmbs_sub_files_failed.fileid = wmbs_fileset_files.fileid AND
                    wmbs_sub_files_failed.subscription = %d
                  INNER JOIN wmbs_file_runlumi_map ON
                    wmbs_file_runlumi_map.fileid = wmbs_fileset_files.fileid AND
                    wmbs_file_runlumi_map.run = :RUN_ID
                  WHERE wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = %d)
                  AND wmbs_sub_files_complete.fileid IS NULL
                  AND wmbs_sub_files_failed.fileid IS NULL
                  """ % (expressSubscription['id'],
                         expressSubscription['id'],
                         expressSubscription['id'])

    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    if results[0][0] != 0:
        logging.debug("Unfinished express jobs for run %d" % runNumber)
        return False

    expressmergeSubscription = Globals.retrieveExpressMergeSubscription()

    sqlQuery = """SELECT COUNT(*)
                  FROM wmbs_fileset_files
                  LEFT OUTER JOIN wmbs_sub_files_complete ON
                    wmbs_sub_files_complete.fileid = wmbs_fileset_files.fileid AND
                    wmbs_sub_files_complete.subscription = %d
                  LEFT OUTER JOIN wmbs_sub_files_failed ON
                    wmbs_sub_files_failed.fileid = wmbs_fileset_files.fileid AND
                    wmbs_sub_files_failed.subscription = %d
                  INNER JOIN wmbs_file_runlumi_map ON
                    wmbs_file_runlumi_map.fileid = wmbs_fileset_files.fileid AND
                    wmbs_file_runlumi_map.run = :RUN_ID
                  WHERE wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = %d)
                  AND wmbs_sub_files_complete.fileid IS NULL
                  AND wmbs_sub_files_failed.fileid IS NULL
                  """ % (expressmergeSubscription['id'],
                         expressmergeSubscription['id'],
                         expressmergeSubscription['id'])

    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    if results[0][0] != 0:
        logging.debug("Unfinished mergepack jobs for run %d" % runNumber)
        return False

    return True
