#!/usr/bin/env python
"""
_ListStreamers_

Functions for retreiving information from the streamer table in T0AST.
"""

__version__ = "$Revision: 1.47 $"
__revision__ = "$Id: ListStreamers.py,v 1.47 2011/01/20 14:56:16 hufnagel Exp $"

import logging

from T0 import Globals


def getRunsWithoutRunConfig(dbConn):
    """
    _getRunsWithoutRunConfig_

    returns list of runs without acqusition era
    (ie. RunConfig not loaded) and unused streamers

    later might not be needed, but it's not expensive
    and a safety check against runs with no data that
    somehow ended up in T0AST

    """
    sqlQuery = """SELECT DISTINCT run.run_id
                  FROM run
                  INNER JOIN streamer ON
                    streamer.run_id = run.run_id AND
                    checkForZeroState(streamer.used) = 0
                  WHERE run.acq_era IS NULL"""

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    runs = []
    for result in results:
        runs.append(result[0])

    return runs

def getMaxRepackLumi(dbConn, run, stream):
    """
    _getMaxRepackLumi_

    return highest lumi for which repack jobs exists
    for the given run and stream

    returns 0 if no jobs exist

    """
    repackSubscription = Globals.retrieveRepackSubscription()

    sqlQuery = """SELECT MAX(streamer.lumi_id)
                  FROM streamer
                  INNER JOIN wmbs_fileset_files ON
                    wmbs_fileset_files.fileid = streamer.streamer_id AND
                    wmbs_fileset_files.fileset = (SELECT fileset FROM wmbs_subscription WHERE id = %d)
                  WHERE streamer.run_id = :RUN_ID
                  AND streamer.stream_id = ( SELECT id FROM stream WHERE name = :STREAM )
                  """ % repackSubscription["id"]

    bindVars = { "RUN_ID" : run, "STREAM" : stream }
    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = dbConn.fetchall()

    if results[0][0] == None:
        return 0
    else:
        return results[0][0]

def countRepackStreamersWithoutChild(dbConn, run):
    """
    _countRepackStreamersWithoutChild_

    count all Bulk streamers without a child

    """
    sqlQuery = """SELECT COUNT(*)
                  FROM streamer
                  INNER JOIN run_stream_style_assoc ON
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
                  LEFT OUTER JOIN wmbs_file_parent ON
                    wmbs_file_parent.parent = streamer.streamer_id
                  WHERE streamer.run_id = :RUN_ID
                  AND wmbs_file_parent.parent IS NULL
                  """

    bindVars = { "RUN_ID" : run }

    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = dbConn.fetchall()
    return results[0][0]

#
# method is only used by the ignoreStreamer script
#
def listSplitableStreamersByRun(dbConn, runNumber):
    """
    _listSplitableStreamersByRun_

    Retrieve a list of dictionaries containing rows from the streamer table
    that have the splitable column equal to one and belong to a specific run.
    """
    sqlQuery = """SELECT streamer.streamer_id,
                         wmbs_file_details.events,
                         wmbs_file_details.lfn
                  FROM streamer
                  INNER JOIN run_stream_style_assoc
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
                  INNER JOIN wmbs_file_details ON
                    wmbs_file_details.id = streamer.streamer_id
                  WHERE checkForZeroState(streamer.used) = 0
                  AND streamer.run_id = :p_1"""

    bindVars = {"p_1": runNumber}

    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    streamers = []
    for result in results:
        streamerDict = {}
        streamerDict["STREAMER_ID"] = result[0]
        streamerDict["EVENTS"] = result[1]
        streamerDict["LFN"] = result[2]
        streamers.append(streamerDict)

    return streamers
