#!/usr/bin/env python
"""
_InsertRunConfig_

Methods for inserting the configuration details for a run into the T0AST
database.
"""

__revision__ = "$Id: InsertRunConfig.py,v 1.95 2011/01/12 21:11:56 hufnagel Exp $"
__version__ = "$Revision: 1.95 $"

import logging

from WMCore.DAOFactory import DAOFactory
from T0 import Globals
from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Writer import InsertDataset


def updateRunTable(dbConn, runNumber, process, acqEra, recoTimeout, recoLockTimeout):
    """
    _updateRunTable_

    Update the run table with the process and acquisition era. We use the
    acquisition era null/non-null status to determine whether or not the
    run config for a run has been loaded into the database.
    """
    sqlQuery = """UPDATE run
                  SET process = :p_1,
                      acq_era = :p_2,
                      reco_timeout = :p_3,
                      reco_lock_timeout = :p_4
                  WHERE run_id = :p_5"""

    bindVars = { 'p_1' : process,
                 'p_2' : acqEra,
                 'p_3' : recoTimeout,
                 'p_4' : recoLockTimeout,
                 'p_5' : runNumber }
    dbConn.execute(sqlQuery, bindVars)
    return

def insertStream(dbConn, streamName):
    """
    _insertStream_

    Insert a stream into the stream table in T0AST. This will
    work even if the stream already exists inside the table.
    """
    sqlQuery = """INSERT INTO stream (ID, NAME) SELECT stream_SEQ.nextval, :p_1
                  FROM DUAL WHERE NOT EXISTS
                  (SELECT NAME FROM stream WHERE NAME = :p_1)"""
    
    bindVars = {"p_1": streamName}
    dbConn.execute(sqlQuery, bindVars)
    return

def insertTrigger(dbConn, runNumber, triggerName, datasetName):
    """
    _insertTrigger_

    Insert a trigger into the "trigger_label" table.  This will work even if
    the trigger label already exists in the table.  Rows will also be added to
    the "run_trig_assoc" and "trig_dataset_assoc" tables.
    """
    triggerInsert = """INSERT INTO trigger_label (ID, LABEL) SELECT
                       trigger_label_SEQ.nextval, :p_1 FROM DUAL
                       WHERE NOT EXISTS
                       (SELECT LABEL FROM trigger_label WHERE LABEL = :p_1)""" 
    
    bindVars = {"p_1": triggerName}
    dbConn.execute(triggerInsert, bindVars)
    
    runTrigAssoc = """INSERT INTO run_trig_assoc (RUN_ID, TRIGGER_ID)
                      SELECT :p_2, (SELECT ID FROM trigger_label WHERE
                      LABEL = :p_1) FROM DUAL WHERE NOT EXISTS (SELECT RUN_ID
                      FROM run_trig_assoc WHERE RUN_ID = :p_2 AND TRIGGER_ID =
                      (SELECT ID FROM trigger_label WHERE LABEL = :p_1))"""
    
    bindVars = {"p_1": triggerName, "p_2": runNumber}
    dbConn.execute(runTrigAssoc, bindVars)
    
    trigDSetAssoc = """INSERT INTO trig_dataset_assoc (TRIG_ID,
                       PRIMARY_DATASET_ID, RUN_ID) SELECT (SELECT ID FROM
                       trigger_label WHERE LABEL = :p_1), (SELECT ID FROM
                       primary_dataset WHERE NAME = :p_2), :p_3 FROM DUAL WHERE
                       NOT EXISTS (SELECT TRIG_ID FROM trig_dataset_assoc WHERE
                       TRIG_ID = (SELECT ID FROM trigger_label
                       WHERE LABEL = :p_1) AND PRIMARY_DATASET_ID = (SELECT ID
                       FROM primary_dataset WHERE NAME = :p_2) AND RUN_ID = :p_3)"""
    
    bindVars = {"p_1": triggerName, "p_2": datasetName, "p_3": runNumber}
    dbConn.execute(trigDSetAssoc, bindVars)
    return

def updateGlobalTagForRunDataset(dbConn, runNumber, primaryDatasetID,
                                 globalTag):
    """
    _updateGlobalTagForRunDataset_

    Update the global tag for a run and primary dataset.
    """
    sqlQuery = """UPDATE reco_config SET GLOBAL_TAG = :p_1 WHERE RUN_ID = :p_2
                  AND PRIMARY_DATASET_ID = :p_3"""
    bindVars = {"p_1": globalTag, "p_2": runNumber, "p_3": primaryDatasetID}

    dbConn.execute(sqlQuery, bindVars)
    dbConn.commit()
    return

def insertT0Config(dbConn, runNumber, version):
    """
    _insertT0Config_

    Add a row to the t0_config table that associates the run number to the
    version of the offline config.
    """
    sqlQuery = """ INSERT INTO t0_config (run_id, config_version)
                   VALUES (:RUN_ID, :CONFIG_VERSION)
               """
    bindVars = {"RUN_ID": runNumber, "CONFIG_VERSION": version}
    
    dbConn.execute(sqlQuery, bindVars)
    return

def insertCMSSWVersion(dbConn, cmsswVersion):
    """
    _insertCMSSWVersion_

    Add a CMSSW version to the cmssw_version table.  This will succeed even if
    the version already exists inside the cmssw_version table.
    """
    sqlQuery = """INSERT INTO cmssw_version (id, name) SELECT
                  cmssw_version_SEQ.nextval, :p_1 FROM DUAL WHERE NOT EXISTS
                  (SELECT id FROM cmssw_version WHERE name = :p_1)"""
    bindVar = {"p_1": cmsswVersion}

    dbConn.execute(sqlQuery, bindVar)
    return

def insertEventScenario(dbConn, eventScenario):
    """
    _insertEventScenario_
    
    Insert an event scenario so that it can be associated with a particular
    dataset and run.  This will work even if the event scenario already exists
    in the table.
    """
    sqlQuery = """INSERT INTO event_scenario (ID, SCENARIO) SELECT
                  event_scenario_SEQ.nextval, :p_1 FROM DUAL WHERE NOT EXISTS
                  (SELECT SCENARIO FROM event_scenario WHERE SCENARIO = :p_1)"""
    bindVar = {"p_1": eventScenario}
    
    dbConn.execute(sqlQuery, bindVar)
    return
    
def insertProcessingStyle(dbConn, runNumber, streamName, processingStyle):
    """
    _insertProcessingStyle_

    Associate a processing style to a stream and run.
    
    """
    sqlQuery = """INSERT INTO run_stream_style_assoc
                  (run_id, stream_id, style_id)
                  VALUES (:p_1,
                          (SELECT id FROM stream WHERE name = :p_2),
                          (SELECT id FROM processing_style WHERE name = :p_3))"""

    bindVars = {"p_1": runNumber, "p_2": streamName, "p_3": processingStyle}
    dbConn.execute(sqlQuery, bindVars)
    return

def insertStreamConfig(dbConn, runNumber, streamName, streamConfig):
    """
    _insertStreamConfig_

    Insert the configuration for a stream.

    Different code paths for different processing styles.

    Special handling of datasets for express and conversion styles

    If the processingStyle is 'Express' a primary dataset

    Stream<streamName>

    will be inserted into T0AST and associated to the stream.

    If the processingStyle is 'Convert' or 'RegisterAndConvert', the
    primary and processed datasets of the output of the conversion is
    inserted into T0AST and the primary dataset associated to the stream.

    """
    #
    # handle style specific parts
    #
    if streamConfig.ProcessingStyle == "Bulk":

        sqlQuery = """ INSERT INTO repack_config
                       (run_id, stream_id, proc_version)
                       VALUES (:RUN_ID,
                               (SELECT id FROM stream WHERE name = :STREAM),
                               :PROC_VER) """

        bindVars = { "RUN_ID" : runNumber,
                     "STREAM" : streamName,
                     "PROC_VER": streamConfig.Repack.ProcessingVersion }
    
        dbConn.execute(sqlQuery, bindVars)

    elif streamConfig.ProcessingStyle == "Express":

        datasetName = "Stream%s" % streamName

        InsertDataset.insertPrimaryDataset(dbConn, datasetName)
        InsertDataset.assocPrimaryDatasetRunStream(dbConn, datasetName,
                                                   streamName, runNumber)
        InsertDataset.assocStreamSpecialDataset(dbConn, streamName,
                                                datasetName )

        insertScenarioConfig(dbConn, runNumber, 
                             datasetName, streamConfig.Express.Scenario)
        insertCalibConfig(dbConn, runNumber, streamConfig.Express.Calibration)

        sqlQuery = """INSERT INTO run_dataset_alca_assoc
                      (run_id, primary_dataset_id, alca)
                      VALUES (:RUN_ID,
                             (SELECT id FROM primary_dataset WHERE name = :DATASET),
                             :ALCAPRODUCER)"""

        bindVarList = []
        for producer in streamConfig.Express.Producers:
            bindVarList.append({"RUN_ID": runNumber,
                                "DATASET": datasetName,
                                "ALCAPRODUCER": producer})
        if len(bindVarList) > 0:
            dbConn.execute(sqlQuery, bindVarList)

        sqlQuery = """INSERT INTO express_config
                      (run_id, stream_id, processing_config_url,
                      splitInProcessing, proc_version,
                      alcamerge_config_url, global_tag)
                      VALUES (:RUN_ID,
                              (SELECT id FROM stream WHERE name = :STREAM),
                              :PROC_URL, :SPLIT_IN_PROC, :PROC_VER,
                              :ALCAMERGE_URL, :GLOBAL_TAG)"""

        bindParams = {"RUN_ID": runNumber, "STREAM": streamName,
                      "PROC_URL": streamConfig.Express.ProcessingConfigURL,
                      "ALCAMERGE_URL": streamConfig.Express.AlcaMergeConfigURL,
                      "GLOBAL_TAG": streamConfig.Express.GlobalTag,
                      "SPLIT_IN_PROC": int(streamConfig.Express.SplitInProcessing),
                      "PROC_VER": streamConfig.Express.ProcessingVersion}
        dbConn.execute(sqlQuery, bindParams)

        sqlQuery = """INSERT INTO run_stream_tier_assoc
                      (run_id, stream_id, data_tier_id)
                      VALUES (:RUN_ID,
                              (SELECT id FROM stream WHERE name = :STREAM),
                              (SELECT id FROM data_tier WHERE name = :DATA_TIER))"""

        bindVars = []
        for data_tier in streamConfig.Express.DataTiers:
            bindVars.append({"RUN_ID": runNumber,
                             "STREAM": streamName,
                             "DATA_TIER": data_tier})
        if len(bindVars) > 0:
            dbConn.execute(sqlQuery, bindVars)

    elif streamConfig.ProcessingStyle == "Register" or \
         streamConfig.ProcessingStyle == "Convert" or \
         streamConfig.ProcessingStyle == "RegisterAndConvert":

        insertStream(dbConn, streamName)

        if streamConfig.ProcessingStyle == "Register" or \
               streamConfig.ProcessingStyle == "RegisterAndConvert":
            
            InsertDataset.insertPrimaryDataset(dbConn, streamConfig.Register.PrimaryDataset)
            InsertDataset.assocPrimaryDatasetRunStream(dbConn, streamConfig.Register.PrimaryDataset,
                                                       streamName, runNumber)
            InsertDataset.insertProcessedDataset(dbConn,
                                                 streamConfig.Register.ProcessedDataset)

            sqlQuery = """INSERT INTO regist_config
                          (run_id, stream_id, primds_id, procds_id, data_tier_id)
                          VALUES (:RUN_ID,
                                  (SELECT id FROM stream WHERE name = :STREAM),
                                  (SELECT id FROM primary_dataset WHERE name = :PRIMDS),
                                  (SELECT id FROM processed_dataset WHERE name = :PROCDS),
                                  (SELECT id FROM data_tier WHERE name = :DATA_TIER))"""

            bindParams = {"RUN_ID": runNumber, "STREAM": streamName,
                          "PRIMDS": streamConfig.Register.PrimaryDataset,
                          "PROCDS": streamConfig.Register.ProcessedDataset,
                          "DATA_TIER": streamConfig.Register.DataTier}
            dbConn.execute(sqlQuery, bindParams)

        if streamConfig.ProcessingStyle == "Convert" or \
               streamConfig.ProcessingStyle == "RegisterAndConvert":

            InsertDataset.insertPrimaryDataset(dbConn, streamConfig.Convert.PrimaryDataset)
            InsertDataset.assocPrimaryDatasetRunStream(dbConn, streamConfig.Convert.PrimaryDataset,
                                                       streamName, runNumber)
            InsertDataset.insertProcessedDataset(dbConn, streamConfig.Convert.ProcessedDataset)

            sqlQuery = """INSERT INTO conv_config
                          (run_id, stream_id, primds_id, data_tier_id, conv_type, acq_era, proc_vers, proc_string)
                          VALUES (:RUN_ID,
                                  (SELECT id FROM stream WHERE name = :STREAM),
                                  (SELECT id FROM primary_dataset WHERE name = :PRIMDS),
                                  (SELECT id FROM data_tier WHERE name = :DATA_TIER),
                                  :CONV_TYPE, :ACQ_ERA, :PROC_VERS, :PROC_STRING)"""

            bindParams = {"RUN_ID": runNumber, "STREAM": streamName,
                          "PRIMDS": streamConfig.Convert.PrimaryDataset,
                          "DATA_TIER": streamConfig.Convert.DataTier,
                          "CONV_TYPE": streamConfig.Convert.Type,
                          "ACQ_ERA": streamConfig.Convert.AcqEra,
                          "PROC_VERS": streamConfig.Convert.ProcVers,
                          "PROC_STRING": streamConfig.Convert.ProcString}
            dbConn.execute(sqlQuery, bindParams)

    #
    # take care of version override, which is generic
    # (needs to be done later as conversion and registration
    #  streams are created in style specific part)
    #

##         sqlQuery = """SELECT a.name
##                       FROM cmssw_version a
##                       INNER JOIN run_stream_cmssw_assoc b
##                       ON b.run_version = a.id
##                       WHERE b.run_id = :RUN_ID
##                       AND b.stream_id = (SELECT id FROM stream WHERE name = :STREAM)"""

##         bindVars = { "RUN_ID": runNumber,
##                      "STREAM": streamName }

##         dbConn.execute(sqlQuery, bindVars)
##         results = dbConn.fetchall()

##         runVersion = results[0][0]

##         if streamConfig.VersionOverride.has_key(runVersion):

##             overrideVersion = streamConfig.VersionOverride[runVersion]

##             insertCMSSWVersion(dbConn, overrideVersion)

##             sqlQuery = """UPDATE run_stream_cmssw_assoc
##                           SET override_version = ( SELECT id from cmssw_version WHERE name = :OVERRIDEVERSION )
##                           WHERE run_id = :RUN_ID
##                           AND stream_id = (SELECT id FROM stream WHERE name = :STREAM)"""

##             bindVars = { "RUN_ID": runNumber,
##                          "STREAM": streamName,
##                          "OVERRIDEVERSION" : overrideVersion }
##             dbConn.execute(sqlQuery, bindVars)

    #
    # HACK, have to use the run_version from run for now
    # HACK, also have to insert into run_stream_cmssw_assoc
    #
    # FIXME: properly done when RunConfig populated per stream
    #

    sqlQuery = """SELECT a.name
                  FROM cmssw_version a
                  INNER JOIN run b
                  ON b.run_version = a.id
                  WHERE b.run_id = :RUN_ID"""

    bindVars = { "RUN_ID": runNumber }
    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    runVersion = results[0][0]

    if streamConfig.VersionOverride.has_key(runVersion):
        overrideVersion = streamConfig.VersionOverride[runVersion]
    else:
        overrideVersion = runVersion

    insertCMSSWVersion(dbConn, overrideVersion)

    sqlQuery = """MERGE INTO run_stream_cmssw_assoc a
                  USING (
                    SELECT b.id as stream_id
                    FROM stream b
                    WHERE b.name = :STREAM ) c
                  ON ( a.run_id = :RUN AND
                       a.stream_id = c.stream_id )
                  WHEN MATCHED THEN
                    UPDATE SET override_version = (SELECT id from cmssw_version WHERE name = :OVERRIDEVERSION)
                  WHEN NOT MATCHED THEN
                    INSERT (run_id, stream_id, run_version, override_version)
                    VALUES (:RUN,
                            (SELECT id FROM stream WHERE name = :STREAM),
                            (SELECT id from cmssw_version WHERE name = :RUNVERSION),
                            (SELECT id from cmssw_version WHERE name = :OVERRIDEVERSION))"""

    bindVars = { "RUN" : runNumber,
                 "STREAM" : streamName,
                 "RUNVERSION" : runVersion,
                 "OVERRIDEVERSION" : overrideVersion }
    dbConn.execute(sqlQuery, bindVars)

    # finally set style for stream
    insertProcessingStyle(dbConn, runNumber, streamName,
                          streamConfig.ProcessingStyle)

    return

def insertScenarioConfig(dbConn, runNumber, datasetName, scenario):
    """
    _insertScenarioConfig_
    
    Associate an event scenario with a particular run number and dataset.  This
    will make sure that the scenario already exists inside the event_scenario
    table before making the association.
    """
    insertEventScenario(dbConn, scenario)

    sqlQuery = """ INSERT INTO scenario_config (run_id,
                                                primary_dataset_id,
                                                scenario_id)
                   VALUES (:RUN_ID,
                          (SELECT id FROM primary_dataset WHERE name = :DATASET),
                          (SELECT id FROM event_scenario WHERE scenario = :SCENARIO)) """

    bindVars = { "RUN_ID" : runNumber,
                 "DATASET" : datasetName,
                 "SCENARIO" : scenario }

    dbConn.execute(sqlQuery, bindVars)
    return

def insertRecoConfig(dbConn, runNumber, datasetConfig):
    """
    _insertRecoConfig_
    
    Set the reco config for a given run and dataset.  Currently, reco can be
    enabled/disabled, and the framework version, processing version and config
    URL can be set.  The global tag is set later on in the run using
    updateGlobalTagForRunDataset().
    """
    insertCMSSWVersion(dbConn, datasetConfig.Reco.CMSSWVersion)

    sqlQuery = """ INSERT INTO reco_config (run_id,
                                            primary_dataset_id,
                                            do_reco, 
                                            cmssw_version_id,
                                            reco_split,
                                            write_reco,
                                            write_aod,
                                            proc_version,
                                            global_tag,
                                            config_url)
                   VALUES (:RUN_ID,
                           (SELECT id FROM primary_dataset WHERE name = :DATASET_NAME),
                           :DO_RECO, 
                           (SELECT id FROM cmssw_version WHERE name = :CMSSW_VERSION),
                           :RECO_SPLIT,
                           :WRITE_RECO,
                           :WRITE_AOD,
                           :PROC_VERSION,
                           :GLOBAL_TAG,
                           :CONFIG_URL) """

    bindVars = { "RUN_ID" : runNumber,
                 "DATASET_NAME" : datasetConfig.Name,
                 "DO_RECO" : int(datasetConfig.Reco.DoReco),
                 "CMSSW_VERSION" : datasetConfig.Reco.CMSSWVersion,
                 "RECO_SPLIT" : datasetConfig.Reco.EventSplit,
                 "WRITE_RECO" : int(datasetConfig.Reco.WriteRECO),
                 "WRITE_AOD" : int(datasetConfig.Reco.WriteAOD),
                 "PROC_VERSION" : datasetConfig.Reco.ProcessingVersion,
                 "GLOBAL_TAG" : datasetConfig.Reco.GlobalTag,
                 "CONFIG_URL" : datasetConfig.Reco.ConfigURL }

    dbConn.execute(sqlQuery, bindVars)
    return

def insertAlcaConfig(dbConn, runNumber, datasetConfig):
    """
    _insertAlcaConfig_
    
    Set the alca config for a given run and dataset.  Currently, alca can be
    enabled/disabled, and the framework version, processing version and config
    URL can be set.  
    """
    insertCMSSWVersion(dbConn, datasetConfig.Alca.CMSSWVersion)
    
    sqlQuery = """ INSERT INTO alca_config (run_id,
                                            primary_dataset_id,
                                            do_alca,
                                            cmssw_version_id,
                                            proc_version,
                                            config_url)
                   VALUES (:RUN_ID,
                           (SELECT id FROM primary_dataset WHERE name = :DATASET_NAME), 
                           :DO_ALCA, 
                           (SELECT id FROM cmssw_version WHERE name = :CMSSW_VERSION),
                           :PROC_VERSION,
                           :CONFIG_URL) """

    bindVars = { "RUN_ID" : runNumber,
                 "DATASET_NAME" : datasetConfig.Name,
                 "DO_ALCA" : int(datasetConfig.Alca.DoAlca),
                 "CMSSW_VERSION": datasetConfig.Alca.CMSSWVersion,
                 "PROC_VERSION": datasetConfig.Alca.ProcessingVersion,
                 "CONFIG_URL": datasetConfig.Alca.ConfigURL }

    dbConn.execute(sqlQuery, bindVars)

    sqlQuery = """INSERT INTO run_dataset_alca_assoc
                  (run_id, primary_dataset_id, alca)
                  VALUES (:RUN_ID,
                          (SELECT id FROM primary_dataset WHERE name = :DATASET),
                          :ALCAPRODUCER)"""

    bindVarList = []
    for producer in datasetConfig.Alca.Producers:
        bindVarList.append({"RUN_ID": runNumber,
                            "DATASET": datasetConfig.Name,
                            "ALCAPRODUCER": producer})
    if len(bindVarList) > 0:
        dbConn.execute(sqlQuery, bindVarList)

    return

def insertCalibConfig(dbConn, runNumber, CalibConfigs):
    """
    _insertCalibConfig_
    Set the Calib Config for a given run. The configuration for a primary
    and processed dataset includes cmssw version, config url, global tag,
    split algorithm and split parameter.
    """
    sqlQuery = """INSERT INTO calib_config (run_id,
                                            primary_dataset_id,
                                            processed_dataset,
                                            cmssw_version_id,
                                            config_url,
                                            global_tag,
                                            split_algo,
                                            split_param)
                  VALUES (:RUN_ID,
                          (SELECT id FROM primary_dataset WHERE name = :DATASET_NAME),
                          :PROCESSED_DATASET,
                          (SELECT id FROM cmssw_version WHERE name = :CMSSW_VERSION),
                          :CONFIG_URL,
                          :GLOBAL_TAG,
                          :SPLIT_ALGO,
                          :SPLIT_PARAM) """

    bindVarList = []
    procDatasets = CalibConfigs.keys()
    for procds in procDatasets:

        Calib = CalibConfigs[procds]
        insertCMSSWVersion(dbConn, Calib.CMSSWVersion)
        bindVarList.append( { "RUN_ID" : runNumber,
                              "DATASET_NAME" : Calib.PrimaryDataset,
                              "PROCESSED_DATASET" : procds,
                              "CMSSW_VERSION" : Calib.CMSSWVersion,
                              "CONFIG_URL" : Calib.ConfigURL,
                              "GLOBAL_TAG" : Calib.GlobalTag,
                              "SPLIT_ALGO" : Calib.SplitAlgo,
                              "SPLIT_PARAM": Calib.SplitParam} )

    if len(bindVarList) > 0:
        dbConn.execute(sqlQuery, bindVarList)
    return

def insertWMBSPublishConfig(dbConn, runNumber, datasetConfig):
    """
    _insertWMBSPublishConfig_
    
    Set the WMBS Publish config for a given run and dataset.
    """
    
    sqlQuery = """ INSERT INTO wmbs_publish_config (run_id,
                                                    primary_dataset_id,
                                                    do_wmbs_publish,
                                                    data_tiers_to)
                   VALUES (:RUN_ID,
                           (SELECT id FROM primary_dataset WHERE name = :DATASET_NAME),
                           :DO_WMBS_PUBLISH, 
                           :DATA_TIERS_TO) """

    # convert the data tier strings into ids
    dataTierMap = ListDatasets.listDataTiersByNames( dbConn )
    dataTiersToString = ''
    for dataTierTo in datasetConfig.WMBSPublish.DataTiersTo:
        (dataTiers,database) = dataTierTo.split('@')
        dataTiersString = ''
        for dataTier in dataTiers.split(','):
            dataTiersString += str( dataTierMap[dataTier] ) + ','
        dataTiersToString += dataTiersString.rstrip(',') + '@' + database + ' '

    bindVars = { "RUN_ID" : runNumber,
                 "DATASET_NAME" : datasetConfig.Name,
                 "DO_WMBS_PUBLISH" : int(datasetConfig.WMBSPublish.DoWMBSPublish),
                 "DATA_TIERS_TO" : dataTiersToString.rstrip() }
    
    dbConn.execute(sqlQuery, bindVars)
    return
    
def insertDQMConfig(dbConn, runNumber, datasetConfig):
    """
    _insertDQMConfig_
    
    Set the DQM config for a given run and dataset.  Currently, DQM can be
    enabled/disabled, and the framework version, processing version and config
    URL can be set.  
    """
    insertCMSSWVersion(dbConn, datasetConfig.DQM.CMSSWVersion)
    
    sqlQuery = """ INSERT INTO dqm_config (run_id,
                                           primary_dataset_id,
                                           do_dqm,
                                           cmssw_version_id,
                                           proc_version,
                                           config_url)
                   VALUES (:RUN_ID,
                           (SELECT id FROM primary_dataset WHERE name = :DATASET_NAME),
                           :DO_DQM,
                           (SELECT id FROM cmssw_version WHERE name = :CMSSW_VERSION),
                           :PROC_VERSION,
                           :CONFIG_URL) """

    bindVars = { "RUN_ID" : runNumber,
                 "DATASET_NAME" : datasetConfig.Name,
                 "DO_DQM" : int(datasetConfig.DQM.DoDQM),
                 "CMSSW_VERSION" : datasetConfig.DQM.CMSSWVersion,
                 "PROC_VERSION" : datasetConfig.DQM.ProcessingVersion,
                 "CONFIG_URL" : datasetConfig.DQM.ConfigURL }
    
    dbConn.execute(sqlQuery, bindVars)
    return

def insertTier1Skims(dbConn, runNumber, tier1Skims):
    """
    _insertTier1Skims_
    
    Insert all the Tier1Skims that are supposed to run over a particular data
    tier and primary dataset into the t1skim_config table in T0AST.
    """
    sqlQuery = """INSERT INTO t1skim_config (run_id,
                                             primary_dataset_id,
                                             data_tier_id,
                                             cmssw_version_id,
                                             two_file_read,
                                             skim_name,
                                             proc_version,
                                             config_url,
                                             global_tag)
                  VALUES (:RUN_ID,
                          (SELECT id FROM primary_dataset WHERE name = :DATASET_NAME),
                          (SELECT id FROM data_tier WHERE name = :DATA_TIER),
                          (SELECT id FROM cmssw_version WHERE name = :CMSSW_VERSION),
                          :TWO_FILE_READ,
                          :SKIM_NAME,
                          :PROC_VERSION,
                          :CONFIG_URL,
                          :GLOBAL_TAG) """

    bindVarList = []
    for tier1Skim in tier1Skims:

        insertCMSSWVersion(dbConn, tier1Skim.CMSSWVersion)
        
        bindVarList.append( { "RUN_ID" : runNumber,
                              "DATASET_NAME" : tier1Skim.PrimaryDataset,
                              "DATA_TIER" : tier1Skim.DataTier,
                              "CMSSW_VERSION" : tier1Skim.CMSSWVersion,
                              "TWO_FILE_READ" : int(tier1Skim.TwoFileRead),
                              "SKIM_NAME" : tier1Skim.SkimName,
                              "PROC_VERSION" : tier1Skim.ProcessingVersion,
                              "CONFIG_URL" : tier1Skim.ConfigURL,
                              "GLOBAL_TAG" : tier1Skim.GlobalTag } )

    if len(bindVarList) > 0:
        dbConn.execute(sqlQuery, bindVarList)
    return

def insertStorageNode(dbConn, nodeName):
    """
    _insertStorageNode_

    Insert a PhEDEx storage node into the storage node table.  This will
    succeed even if the node already exists.  Also insert the storage node
    name as a location in WMBS.
    """
    sqlQuery = """INSERT INTO storage_node (id, name)
                  SELECT storage_node_SEQ.nextval, :p_1 FROM DUAL
                  WHERE NOT EXISTS (SELECT name FROM storage_node
                  WHERE name = :p_1)"""
    bindVars = {"p_1": nodeName}
    dbConn.execute(sqlQuery, bindVars)

    return

def insertPhEDExConfig(dbConn, runNumber, primaryDatasetName, custodialNode,
                       archivalNode, custodialPriority, custodialAutoApprove):
    """
    _insertPhEDExConfig_

    Insert the phedex subscriptions for the given primary dataset.  This
    implements the following policy:
      https://twiki.cern.ch/twiki/bin/view/CMS/T0ASTDiscussExportPolicy
    """
    sqlQuery = """INSERT INTO phedex_subscription (run_id,
                                                   primary_dataset_id,
                                                   data_tier_id,
                                                   node_id, 
                                                   custodial_flag,
                                                   request_only,
                                                   priority)
                  VALUES (:runNumber,
                          (SELECT id FROM primary_dataset WHERE name = :pDatasetName),
                          (SELECT id FROM data_tier WHERE name = :dataTierName),
                          (SELECT id FROM storage_node WHERE name = :storageNode),
                          :custodial,
                          :autoApprove,
                          :priority)""" 

    bindVars = {"runNumber": runNumber, "pDatasetName": primaryDatasetName}
    if custodialNode != None:
        insertStorageNode(dbConn, custodialNode)
        
        bindVars["storageNode"] = custodialNode
        bindVars["priority"] = custodialPriority
        bindVars["custodial"] = 1

        if custodialAutoApprove == True:
            bindVars["autoApprove"] = "n"
        else:
            bindVars["autoApprove"] = "y"

        bindVars["dataTierName"] = "RAW"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "RECO"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "DQM"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "AOD"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "ALCARECO"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "ALCAPROMPT"
        dbConn.execute(sqlQuery, bindVars)

    if archivalNode != None:
        insertStorageNode(dbConn, archivalNode)
        
        bindVars["storageNode"] = archivalNode
        bindVars["autoApprove"] = "n"
        bindVars["priority"] = "high"
        bindVars["custodial"] = 0

        bindVars["dataTierName"] = "RAW"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "RECO"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "DQM"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "AOD"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "ALCARECO"
        dbConn.execute(sqlQuery, bindVars)
        bindVars["dataTierName"] = "ALCAPROMPT"
        dbConn.execute(sqlQuery, bindVars)        
    
    return

def insertStreamDatasetTriggerMapping(dbConn, runNumber, mapping):
    """
    _insertStreamDatasetTriggerMapping_

    Write a stream to dataset to trigger mapping to T0AST

    """
    logging.debug("insertStreamDatasetTriggerMapping(): Running %s" % runNumber)

    for stream in mapping.keys():

        insertStream(dbConn, stream)

        for dataset in mapping[stream].keys():

            InsertDataset.insertPrimaryDataset(dbConn, dataset)
            InsertDataset.assocPrimaryDatasetRunStream(dbConn, dataset,
                                                       stream, runNumber)

            errorDataset = "%s-%s" % (dataset,"Error")
            InsertDataset.insertPrimaryDataset(dbConn, errorDataset)

            InsertDataset.assocPrimaryDatasetRunStream(dbConn, errorDataset,
                                                       stream, runNumber)
            InsertDataset.assocPrimaryDatasetErrorDataset(dbConn, dataset,
                                                          errorDataset)

            insertPhEDExConfig(dbConn, runNumber, errorDataset,
                               None, "T0_CH_CERN_MSS", None, None)

            for trigger in mapping[stream][dataset]:
                insertTrigger(dbConn, runNumber, trigger, dataset)

    return

def insertDatasetConfig(dbConn, runNumber, datasetConfig):
    """
    _insertDatasetConfig_

    Insert configuration information into T0AST for a given dataset and
    run number.  The datasetConfig object should have sections for Repacking,
    Reco, Alca, DQM and Skims.
    """
    insertScenarioConfig(dbConn, runNumber, datasetConfig.Name,
                         datasetConfig.Scenario)

    insertRecoConfig(dbConn, runNumber, datasetConfig)

    insertAlcaConfig(dbConn, runNumber, datasetConfig)

    insertWMBSPublishConfig(dbConn, runNumber, datasetConfig)

    insertDQMConfig(dbConn, runNumber, datasetConfig)

    insertPhEDExConfig(dbConn, runNumber, datasetConfig.Name,
                       datasetConfig.CustodialNode,
                       datasetConfig.ArchivalNode,
                       datasetConfig.CustodialPriority,
                       datasetConfig.CustodialAutoApprove)

    insertTier1Skims(dbConn, runNumber, datasetConfig.Tier1Skims)

    insertCalibConfig(dbConn, runNumber, datasetConfig.Calibration)

    return

def updatePSetHash(dbConn, tablename, run, primds_id, psethash):
    """
    _updatePSetHash_

    Update the PSet hash value in config table.
    The value shouldn't be changed once updated
    
    """
    sqlQuery = """UPDATE %s
                  SET pset_hash = :PSETHASH
                  WHERE run_id = :RUN_ID
                  AND primary_dataset_id = :PRIMDS_ID
                  AND pset_hash IS NULL
                  """ % tablename
                  
    bindVars = { "PSETHASH" : psethash,
                 "RUN_ID": run,
                 "PRIMDS_ID": primds_id}

    dbConn.execute(sqlQuery, bindVars)
    return
