#!/usr/bin/env python
"""
_RunLumiCloseout_

Methods to closeout runs and lumi sections
"""

__revision__ = "$Id: RunLumiCloseout.py,v 1.38 2011/01/13 11:50:29 hufnagel Exp $"
__version__ = "$Revision: 1.38 $"

import time
import logging

from T0 import Globals
from T0.State.Database.Writer import InsertRun


def closeRuns(t0astDBConn, timeout):
    """
    _closeRunsNew_

    old style run closing based on waiting a certain
    time after the last streamer insert for a run

    """
    currentTime = int(time.time())

    sqlQuery = """SELECT run.run_id
                  FROM run
                  INNER JOIN streamer ON
                    streamer.run_id = run.run_id
                  WHERE run.close_time = 0
                  AND run.acq_era IS NOT NULL
                  GROUP BY run.run_id
                  HAVING MAX(streamer.insert_time) < :TIME
                  """

    bindVars = { 'TIME' : currentTime - timeout }
    t0astDBConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = t0astDBConn.fetchall()

    bindVarList = []
    for result in results:
        bindVarList.append( { 'RUN_ID' : result[0],
                              'CURRENT_TIME' : currentTime } )

    if len(bindVarList) > 0:

        sqlQuery = """UPDATE run
                      SET run_status = (SELECT id FROM run_status WHERE status = 'CloseOutRepack'),
                          close_time = :CURRENT_TIME,
                          last_updated = :CURRENT_TIME
                      WHERE run_id = :RUN_ID
                      """

        t0astDBConn.execute(sqlQuery, bindVarList, useTransaction = False)

        for bindVars in bindVarList:
            run_id = bindVars['RUN_ID']
            logging.debug("Closing out run %s" % run_id)
            Globals.tier0TwitterPost("Run %s : CloseOutRepack" % run_id)

    return


def closeLumiSections(t0astDBConn, smDBConn, timeout):
    """
    _closeLumiSection_

    close lumi sections where all streamers are older than timeout

    """
    currentTime = int(time.time())

    #  - get list of open run/lumi/stream from T0AST incl. file count
    #    where the latest streamer file was inserted more than timeout ago
    #  - for that list get count of files in SM files_created table
    #  - if count match, close lumi in T0AST

    sqlQuery = """SELECT a.run_id, a.lumi_id, b.name, COUNT(*), MAX(a.insert_time)
                  FROM streamer a
                  INNER JOIN stream b
                  ON b.id = a.stream_id
                  LEFT OUTER JOIN lumi_section_closed c
                  ON c.run_id = a.run_id AND c.lumi_id = a.lumi_id AND c.stream_id = a.stream_id
                  WHERE c.run_id IS NULL
                  GROUP BY a.run_id, a.lumi_id, b.name
                  HAVING MAX(a.insert_time) < :TIMEOUT"""

    bindVar = { "TIMEOUT" : currentTime - timeout }
    t0astDBConn.execute(sqlQuery, bindVar, useTransaction = False)
    results = t0astDBConn.fetchall()

    lumiSectionsToClose = {}

    bindVarList = []
    for result in results:

        run_id = result[0]
        lumi_id = result[1]
        stream = result[2]
        count = result[3]

        if lumiSectionsToClose.has_key(run_id):
            if lumiSectionsToClose[run_id].has_key(lumi_id):
                lumiSectionsToClose[run_id][lumi_id][stream] = count
            else:
                lumiSectionsToClose[run_id][lumi_id] = { stream : count }
        else:
            lumiSectionsToClose[run_id] = { lumi_id : { stream : count } }

        bindVars = { 'RUN_ID' : result[0],
                     'LUMI_ID' : result[1],
                     'STREAM' : result[2] }

        bindVarList.append(bindVars)

    if len(bindVarList) > 0:

        sqlQuery = """SELECT a.runnumber, a.lumisection, a.stream, COUNT(*)
                      FROM CMS_STOMGR.files_created a
                      WHERE a.runnumber = :RUN_ID
                      AND a.lumisection = :LUMI_ID
                      AND a.stream = :STREAM
                      GROUP BY a.runnumber, a.lumisection, a.stream"""

        smDBConn.execute(sqlQuery, bindVarList, useTransaction = False)
        results = smDBConn.fetchall()

        bindVarList = []
        for result in results:

            run_id = result[0]
            lumi_id = result[1]
            stream = result[2]
            count = result[3]

            if lumiSectionsToClose[run_id][lumi_id][stream] == count:

                bindVars = { 'RUN_ID' : run_id,
                             'LUMI_ID' : lumi_id,
                             'STREAM' : stream,
                             'INSERT_TIME' : currentTime,
                             'CLOSE_TIME' : currentTime }

                bindVarList.append(bindVars)

                logging.info("Closing run %d, stream %s, lumi %d" % (run_id, stream, lumi_id))

        if len(bindVarList) > 0:

            sqlQuery = """INSERT INTO lumi_section_closed
                          (RUN_ID, LUMI_ID, STREAM_ID, FILECOUNT, INSERT_TIME, CLOSE_TIME)
                          SELECT :RUN_ID, :LUMI_ID, a.id, 0, :INSERT_TIME, :CLOSE_TIME
                          FROM stream a
                          WHERE a.name = :STREAM"""

            t0astDBConn.execute(sqlQuery, bindVarList, useTransaction = False)

    return

def closeRunsAndLumis(t0astDBConn, smDBConn, runTimeout = 0, closeRuns = False):
    """
    _closeRunsAndLumis_

    close runs and lumis

    """
    currentTime = int(time.time())

    # get all open or active runs in T0AST
    # if we are not closing runs, only get active runs
    if closeRuns:

        logging.info("DEBUG closing runs and lumis")

        sqlQuery = """SELECT run_id, end_time, lumicount
                      FROM run
                      WHERE close_time = 0
                      """
    else:

        logging.info("DEBUG closing lumis")

        sqlQuery = """SELECT run_id
                      FROM run
                      WHERE close_time = 0
                      """

    t0astDBConn.execute(sqlQuery, useTransaction = False)
    results = t0astDBConn.fetchall()

    activeRuns = []
    openRuns = []
    endedRuns = {}
    for result in results:

        activeRuns.append(result[0])

        # if closing runs, sort into open and ended
        if closeRuns:
            if result[1] == 0:
                openRuns.append(result[0])
            else:
                endedRuns[result[0]] = result[2]

    logging.info("DEBUG openRuns = %s" % openRuns)
    logging.info("DEBUG activeRuns = %s" % activeRuns)
    logging.info("DEBUG endedRuns = %s" % endedRuns)

    # check SM which of the open runs have ended
    if len(openRuns) > 0:

        bindVarList = []
        for run in openRuns:
            bindVarList.append( { 'RUN_ID' : run } )

        logging.info("DEBUG check if runs are closed %s" % bindVarList)

        sqlQuery = """SELECT a.runnumber, MAX(a.n_lumisections)
                      FROM CMS_STOMGR.runs a
                      WHERE a.runnumber = :RUN_ID
                      AND a.status = 0
                      GROUP BY a.runnumber
                      HAVING COUNT(*) = MAX(a.n_instances)"""

        smDBConn.execute(sqlQuery, bindVarList, useTransaction = False)
        results = smDBConn.fetchall()

        logging.info("DEBUG check if runs are closed result %s" % results)

        # then mark them as ended in T0AST
        bindVarList = []
        for result in results:

            bindVarList.append( { 'RUN_ID' : result[0],
                                  'LUMICOUNT' : result[1],
                                  'END_TIME' : currentTime } )
            endedRuns[result[0]] = result[1]

        if len(bindVarList) > 0:

            logging.info("DEBUG closing runs %s" % bindVarList)

            sqlQuery = """UPDATE run
                          SET end_time = :END_TIME,
                              lumicount = :LUMICOUNT
                          WHERE run_id = :RUN_ID"""

            t0astDBConn.execute(sqlQuery, bindVarList, useTransaction = False)


    # close stream/lumi for all active runs
    if len(activeRuns) > 0:
        for run in activeRuns:
            closeLumiSectionsForRun(t0astDBConn, smDBConn, run, currentTime)

    # now check all ended and active runs if they can be switched
    # to Closeout (needs all stream/lumi pairs to be closed)
    #
    # provide special exepction for replays where we can switch runs
    # to closeout based on last streamer age
    if len(endedRuns) > 0:

        closeoutRuns = []

        # runTimeout equals 0 is only good for processing full runs
        # like for instance live data taking or full replays
        # (full replay == all streamers from all streams)
        if runTimeout == 0:

            bindVarList = []
            for run, lumicount in endedRuns.items():
                bindVarList.append( { 'RUN_ID' : run,
                                      'LUMICOUNT' : lumicount } )

            sqlQuery = """SELECT run_id
                          FROM lumi_section_closed
                          WHERE run_id = :RUN_ID
                          AND close_time > 0
                          GROUP BY run_id, stream_id
                          HAVING COUNT(*) != MAX(lumi_id)
                          OR COUNT(*) != :LUMICOUNT
                          """

            t0astDBConn.execute(sqlQuery, bindVarList, useTransaction = False)
            results = t0astDBConn.fetchall()

            closeoutRuns = set(endedRuns.keys())
            for result in results:
                if result[0] in closeoutRuns:
                    closeoutRuns.remove(result[0])

        # runTimeout other than 0 means runs are closed out based
        # on a timeout on the newest streamer file
        else:

            bindVarList = []
            for run in endedRuns.keys():
                bindVarList.append( { 'RUN_ID' : run,
                                      'TIME' : currentTime - runTimeout } )

            sqlQuery = """SELECT run.run_id
                          FROM run
                          INNER JOIN streamer ON
                            streamer.run_id = run.run_id
                          WHERE run.run_id = :RUN_ID
                          AND run.acq_era IS NOT NULL
                          GROUP BY run.run_id
                          HAVING MAX(streamer.insert_time) < :TIME
                          """

            t0astDBConn.execute(sqlQuery, bindVarList,  useTransaction = False)
            results = t0astDBConn.fetchall()

            for result in results:
                closeoutRuns.append(result[0])

            
        if len(closeoutRuns) > 0:

            logging.info("DEBUG switch to Closeout runs %s" % closeoutRuns)

            bindVarList = []
            for run in closeoutRuns:
                bindVarList.append( { 'RUN_ID' : run,
                                      'CURRENT_TIME' : currentTime } )

            sqlQuery = """UPDATE run
                          SET run_status = (SELECT id FROM run_status WHERE status = 'CloseOutRepack'),
                              close_time = :CURRENT_TIME,
                              last_updated = :CURRENT_TIME
                          WHERE run_id = :RUN_ID
                          AND run.acq_era IS NOT NULL
                          """

            t0astDBConn.execute(sqlQuery, bindVarList,  useTransaction = False)

    return


def closeLumiSectionsForRun(t0astDBConn, smDBConn, run, currentTime):
    """
    _closeLumiSectionForRun_

    close (aka create a record in lumi_section_closed)
    for all stream/lumi combos for the given run which are closed in SM

    at the same time retrieve the filecount and store it in T0AST

    """

    logging.info("DEBUG closing lumis for run %s" % run)

    #
    # get all streams for the run
    #
    # later query will ignore run/stream with no closed lumis
    #
    sqlQuery = """SELECT stream.name
                  FROM run_stream_cmssw_assoc
                  INNER JOIN stream ON
                    stream.id = run_stream_cmssw_assoc.stream_id
                  WHERE run_stream_cmssw_assoc.run_id = :RUN_ID
                  """

    bindVars = { 'RUN_ID' : run }
    t0astDBConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = t0astDBConn.fetchall()

    streams = []
    for result in results:
        streams.append(result[0])

    # do nothing if RunConfig for run has not been done yet
    if len(streams) == 0:
        return

    #
    # get all closed stream/lumi for given run
    #
    # for each stream detect the lowest lumi at the
    # end of a continuous sequence (starting with 1)
    #
    # example: lumi 1,2,3,5,6 would find 4
    # example: lumi 2,3,4,5,6 would find 1 (as 1 does not exist)
    #
    sqlQuery = """SELECT stream.name,
                         lumi_section_closed.lumi_id
                  FROM lumi_section_closed
                  INNER JOIN stream ON
                    stream.id = lumi_section_closed.stream_id
                  WHERE lumi_section_closed.run_id = :RUN_ID"""

    bindVars = { 'RUN_ID' : run }
    t0astDBConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = t0astDBConn.fetchall()

    streamLumis = {}
    for result in results:

        stream = result[0]
        lumi_id = result[1]

        if streamLumis.has_key(stream):
            streamLumis[stream].append(lumi_id)
        else:
            streamLumis[stream] = [lumi_id]

    bindVarList = []
    for stream in streamLumis.keys():

        lastLumi = 0
        for lumi_id in sorted(streamLumis[stream]):
            if lumi_id == lastLumi + 1:
                lastLumi = lumi_id
            else:
                break

        bindVarList.append( { 'RUN_ID' : run,
                              'STREAM' : stream,
                              'LUMI_ID' : lastLumi } )

        logging.info("DEBUG check SM db for closed run %d, stream %s, lumi > %d" %  (run, stream, lastLumi))

    # pad with entries for streams that have no closed lumis
    #
    # HACK !!! REMOVE
    for stream in streams:
        if not streamLumis.has_key(stream) and stream != 'EcalCalibration':

            bindVarList.append( { 'RUN_ID' : run,
                                  'STREAM' : stream,
                                  'LUMI_ID' : 0 } )

            logging.info("DEBUG check SM db for closed run %d, stream %s, lumi > 0" %  (run, stream))

    # execute SM query, checking for closed run/stream/lumi with
    # lumis above a minimum threshold, extracting filecounts as well
    #
    sqlQuery = """SELECT a.stream, a.lumisection, SUM(a.filecount)
                  FROM CMS_STOMGR.streams a
                  WHERE a.runnumber = :RUN_ID
                  AND a.stream = :STREAM
                  AND a.lumisection > :LUMI_ID
                  GROUP BY a.stream, a.lumisection
                  HAVING COUNT(*) = ( SELECT COUNT(*) FROM CMS_STOMGR.runs b WHERE b.runnumber = :RUN_ID )
                  AND COUNT(*) = ( SELECT MAX(n_instances) FROM CMS_STOMGR.runs c WHERE c.runnumber = :RUN_ID )"""

    smDBConn.execute(sqlQuery, bindVarList, useTransaction = False)
    results = smDBConn.fetchall()

    # only insert run/stream/lumi records that do not exists yet
    # for safetly also insert the lumi, could be one with 0 files
    # or at least 0 transfered files and no lumi_section record
    bindVarList = []
    insertLumis = set([])
    for result in results:

        stream = result[0]
        lumi = result[1]
        filecount = result[2]

        if streamLumis.has_key(stream):
            if lumi in streamLumis[stream]:
                continue

        insertLumis.add(lumi)

        # if file count is 0 immediately close finally
        if filecount == 0:
            bindVarList.append( { 'RUN_ID' : run,
                                  'STREAM' : stream,
                                  'LUMI_ID' : lumi,
                                  'FILECOUNT' : filecount,
                                  'INSERT_TIME' : currentTime,
                                  'CLOSE_TIME' : currentTime } )
        else:
            bindVarList.append( { 'RUN_ID' : run,
                                  'STREAM' : stream,
                                  'LUMI_ID' : lumi,
                                  'FILECOUNT' : filecount,
                                  'INSERT_TIME' : currentTime,
                                  'CLOSE_TIME' : 0 } )

    logging.info("DEBUG insert lumis %s" % insertLumis)

    t0astDBConn.beginTransaction()
    for lumi in insertLumis:
        InsertRun.insertLumiSection(t0astDBConn, run, lumi)
    t0astDBConn.commit()

    if len(bindVarList) > 0:

        logging.info("DEBUG insert %d lumi_section_closed records" % len(bindVarList))

        sqlQuery = """INSERT INTO lumi_section_closed
                      (RUN_ID, LUMI_ID, STREAM_ID, FILECOUNT, INSERT_TIME, CLOSE_TIME)
                      SELECT :RUN_ID,
                             :LUMI_ID,
                             stream.id,
                             :FILECOUNT,
                             :INSERT_TIME,
                             :CLOSE_TIME
                      FROM stream
                      WHERE stream.name = :STREAM"""

        t0astDBConn.execute(sqlQuery, bindVarList, useTransaction = False)

    # last step check and close lumi sections in T0AST
    # if they are closed in SM and the filecount matches
    #
    # test this query to find best !
    #
    sqlQuery = """SELECT lumi_section_closed.stream_id, lumi_section_closed.lumi_id
                  FROM lumi_section_closed
                  INNER JOIN streamer ON
                    streamer.lumi_id = lumi_section_closed.lumi_id AND
                    streamer.stream_id = lumi_section_closed.stream_id AND
                    streamer.run_id = :RUN_ID
                  WHERE checkForZeroState(lumi_section_closed.close_time) = 0
                  AND lumi_section_closed.run_id = :RUN_ID
                  GROUP BY lumi_section_closed.stream_id, lumi_section_closed.lumi_id
                  HAVING COUNT(*) = MAX(lumi_section_closed.filecount)
                  """

    bindVars = { 'RUN_ID' : run }

    t0astDBConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = t0astDBConn.fetchall()

    bindVarList = []
    for result in results:

        stream_id = result[0]
        lumi_id = result[1]
        
        bindVarList.append( { 'RUN_ID' : run,
                              'STREAM_ID' : stream_id,
                              'LUMI_ID' : lumi_id,
                              'CLOSE_TIME' : currentTime } )

    if len(bindVarList) > 0:

        logging.info("DEBUG update %d lumi_section_closed records as finally closed" % len(bindVarList))

        sqlQuery = """UPDATE lumi_section_closed
                      SET close_time = :CLOSE_TIME
                      WHERE run_id = :RUN_ID
                      AND lumi_id = :LUMI_ID
                      AND stream_id = :STREAM_ID
                      """

        t0astDBConn.execute(sqlQuery, bindVarList, useTransaction = False)

    return


def checkSplitStreamerLumisForCompletion(dbConn):
    """
    _checkSplitStreamerLumisForCompletion_

    check all run/lumi/stream that are actively split processed
    remove them when:
      all the streamers are complete (processing done)

    """
    sqlQuery = """SELECT lumi_section_split_active.run_id,
                         lumi_section_split_active.lumi_id,
                         lumi_section_split_active.stream_id
                  FROM lumi_section_split_active
                  INNER JOIN streamer ON
                    streamer.run_id = lumi_section_split_active.run_id AND
                    streamer.lumi_id = lumi_section_split_active.lumi_id AND
                    streamer.stream_id = lumi_section_split_active.stream_id
                  LEFT OUTER JOIN wmbs_sub_files_complete ON
                    wmbs_sub_files_complete.fileid = streamer.streamer_id
                  GROUP BY lumi_section_split_active.run_id,
                           lumi_section_split_active.lumi_id,
                           lumi_section_split_active.stream_id
                  HAVING COUNT(*) = COUNT(wmbs_sub_files_complete.fileid)
                  """

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    bindVarList = []
    for result in results:
        bindVarList.append( { 'RUN_ID' : result[0],
                              'LUMI_ID' : result[1],
                              'STREAM_ID' : result[2] } )

    if len(bindVarList) > 0:

        sqlQuery = """DELETE FROM lumi_section_split_active
                      WHERE run_id = :RUN_ID
                      AND :lumi_id = :LUMI_ID
                      AND :stream_id = :STREAM_ID
                      """

        dbConn.execute(sqlQuery, bindVarList, useTransaction = False)

    return
