from time import *
from datetime import datetime,timedelta 
import calendar,Columns
from types import IntType

"""
 This collection of classes represent the possible input parameters.
 Some parameters need to be translated in the DAS calls so they implements the getDasReq method.
"""
 
class GeneralParameter():
  """
  Superclass for parameters
  """
  name=""
  value=None
  defaultValue=None
  message="'%s' value not found or not valid."
  DASName=""
  
  def __init__(self):
    """
    self.message is not used so far
    """
    self.value=self.defaultValue
    self.message=self.message%self.name
    if self.DASName=="": self.DASName=self.name
  
  def getDasReq(self):
    """
    Some parameters needs to be translated as they are in the DAS call (example: run_id,status,..)
    For those parameters it is possible to use this method
    """
    res=self.DASName+'='
    if self.value!=None: 
      res+=str(self.value)
    return res
  
  def checkOnDict(self,dict):
    """
    This method try to metch the parameter to a given dict (which will be the input kwargs)
    """
    pass
  
  def getFormOptions(self):
    """
    Some parameters can have a html code to set them inside a form. 
    This method is used from the plot parameters might which need to be changed by the user. 
    """
    return ""
  
  def printValue(self):
    """
    This method return some HTML formatted text (1 line) with the value of the parameter
    """
    return "<BR/> %s:"%self.name +str(self.value)
######################################################################################## INTEGER
class IntegerParameter(GeneralParameter):
  """
  Superclass for integer parameters
  """
  def checkOnDict(self,dict):
    try:
      tmp=dict[self.name]
      if len(str(tmp))>0:
        self.value=int(tmp)
    except:
      self.value=self.defaultValue
      pass
    
#################################### RUN
class RunParameter(IntegerParameter):
  name='run_id'
  DASName='run'
    
  def getFormOptions(self):
    """
    if the value is a true integer I save it to the form (hidden)
    """
    if type(self.value) is not IntType: return ""
    return """<input type="hidden" name="run_id" value="%s">"""%str(self.value)
    
  def printValue(self):
   runC=Columns.RunId()
   return "<BR/> Run id: "+runC.decorator({self.name:self.value})
 
############################## Value 
class ValueParameter(IntegerParameter):
  """
  This parameter is used only in the History page/table
  """
  name='value'
  DASName='value'
############################ Page
class PageParameter(IntegerParameter):
  name='page'
  defaultValue=1
  
  def checkOnDict(self,dict):
    try:
      tmp=dict[self.name]
      if len(str(tmp))>0:
        self.value=int(tmp)
      if (self.value<1):self.value=1
    except Exception, e:
      self.value=self.defaultValue
      
   ### page attribute is never directly translated to DAS 
  def getDasReq(self, forceName=False):
    return ""
  

####################################################################### STRINGS
class StringParameter(GeneralParameter):
  def checkOnDict(self,dict):
    try:
      tmp=dict[self.name]
      self.value=str(tmp)
    except:
      self.value=self.defaultValue
      pass
############################################## Status
class StatusParameter(StringParameter):
  name='status'
  DASName='status'
############################################# Source
class SourceParameter(StringParameter):
  ## this parameter is used only in the History page/table
  name='source'
  DASName='vtype'
  vtypes={"StreamerFiles":'Strm',
     "RepackFiles":"Repkf","RepackJobs":"Repkj","RepackMergeJobs":"Repkmj","RepackMergeFiles":"Repkmf",
     "RecoFiles":"Recof","RecoJobs":"Recoj","RecoMergeJobs":"Recomj","RecoMergeFiles":"Recomf",
     "AlcaFiles":"Alcaf","AlcaJobs":"Alcaj","AlcaMergeJobs":"Alcamj","AlcaMergeFiles":"Alcamf",
     "ExpressFiles":"Exprf","ExpressJobs":"exprj","ExpressMergeJobs":"Exprmj","ExpressMergeFiles":"Exprmf",
     "DQMJobs":'DQMJ'}
     
  def getDasReq(self):
    return self.DASName+'='+self.vtypes[self.value]
  
########################################################## PLOT PARAMETERS
############################### Tier
class TierParameter(StringParameter):
  name='tier'
  DASName='tier'
  defaultValue=""
  
  def getFormOptions(self):
    return """<input type="hidden" name="tier" value="%s">"""%self.value
  
  def printValue(self):
    if self.value=="":
      return "<BR/> Showing all Streamers"
    else: return "<BR/> Showing %s Tier"%self.value
     
############################## Autotime
class AutotimeParameter(GeneralParameter):
  name='autotime'
  defaultValue=False
  
  def checkOnDict(self,dict):
    try:
      self.value= (int(dict['autotime'])==1)
    except:
      self.value=False
    
  def getFormOptions(self):
    if self.value: return """<input type="hidden" name="autotime" value="1">"""
    return ""
    
  def printValue(self):
    if self.value:
           return "<BR/>Auto-selection of the time parameters: <span style=\"color:green\"><B> ON</B></span>"
    else: return "<BR/>Auto-selection of the time parameters: <span style=\"color:red\"><B> OFF </B></span>"
########################### Date
class DateParameter(GeneralParameter):
  name='etime'
  DASName='etime'
  minValue=datetime(2011, 1, 1, 0, 0, 0)
  startValue=None
  
  def checkOnDict(self,dict):
    try:
      fromH=int(dict['sh'])
      fromD=int(dict['sd'])
      fromM=int(dict['sm'])
      fromY=int(dict['sy'])
      d=datetime(fromY, fromM, fromD, fromH)
      self.value=d
      if (mktime(self.value.timetuple())>mktime(localtime())) or (self.value<self.minValue):
        self.message+="<BR/> Date out of range"
        raise Exception()
    except Exception, e :
        #print e
        self.message+="<BR/> Date input error: value is back to DEFAULT value."
        self.value=datetime.utcnow()
        
  def setInterval(self,start,end):
    """
    Date parameter is 'special', it can be used to represent/display an interval 
    This happens for example in the plots.
    Moreover there is the possibily that the value is set not directly from the URL but when the data are back from DAS 
    """
    self.value=end
    self.startValue=start
  
  def getFormOptions(self):
    """
    Code to set this parameter inside a HTML form
    """
    form="<BR/>with an end time at:<select name=\"sh\"> "
    if self.value.minute>0:
     self.value=self.value+timedelta(minutes=(60-self.value.minute))
    for i in range(1,24+1):
      selected=""
      if i==self.value.hour: selected="selected=\"selected\"" 
      form+="""<option value="%d" %s>%d:00</option>"""%(i,selected,i)
    form+= """</select>  of:
            <select name="sd">"""
    for i in range(1,31+1):
      selected=""
      if i==self.value.day: selected="selected=\"selected\"" 
      form+="""<option value="%d" %s>%d</option>"""%(i,selected,i)
    form+="""</select>
             <select name="sm">"""
    for i in range(1,12+1):
      selected=""
      if i==self.value.month: selected="selected=\"selected\"" 
      form+="""<option value="%d" %s>%s</option>"""%(i,selected,calendar.month_name[i])
    form+="""</select>
             <select name="sy">"""
    for i in range(2011,self.value.year+1):
      selected=""
      if i==self.value.year: selected="selected=\"selected\""
      form+="""<option value="%d" %s>%d</option>"""%(i,selected,i)
    form+="""</select>"""
    return form
          
  def printValue(self):
    try:
      res="<BR/> from <B>%d:00</B> of <B>%s %d/%d/%d</B> <BR/> to:&nbsp;&nbsp;&nbsp;&nbsp;  <B>%d:00</B> of <B>%s %d/%d/%d</B><BR/>"%(self.startValue.hour,self.startValue.strftime("%A"),self.startValue.day,self.startValue.month,self.startValue.year,self.value.hour,self.value.strftime("%A"),self.value.day,self.value.month,self.value.year)
    except:
      res=""
    return res        
    
############################################# Hours 
class HourParameter(IntegerParameter):
  name='hours'
  defaultValue=24
  unit="hour(s)"
  prefix="Show the previous:"
  def __init__(self,avRange=(1,48),default=24):
    self.avRange=avRange
    
  def checkOnDict(self,dict):
    try:
      tmp=dict[self.name]
      if len(str(tmp))>0:
        self.value=int(tmp)
      minV,maxV=self.avRange
      self.value=min(self.value,maxV)
      self.value=max(self.value,minV)
    except Exception, e:
      self.value=self.defaultValue  
      
  def getRange(self):
    minV, maxV= (self.avRange)
    if maxV<10: return range(minV,maxV+1,1) 
    return range(minV,5,1)+range(6,maxV+1,6)
    
  def getFormOptions(self):
    form=self.prefix+"""<select name=%s>"""%self.name
    for i in self.getRange():
       selected=""
       if i==self.value: selected="selected=\"selected\""
       form+="""<option value="%d" %s>%d</option>"""%(i,selected,i)
    form+="""</select> %s """%self.unit
    return form
         
  def printValue(self):
    return "<BR/>Showing : <B>%d</B> hours "%self.value     
 
######################################################### Step
class StepParameter(HourParameter):
  name='step'
  defaultValue=10
  unit="minutes"
  prefix="in intervals of"
  
  def __init__(self,avRange=(5,60),default=10):
    self.avRange=avRange
  
  def getRange(self):
    minV, maxV= (self.avRange)
    if minV==0: return range(minV,6,1)+range(10,maxV+5,5) 
    return range(minV,maxV+5,5)
  

       
  def printValue(self):
     if self.value!=None and self.value>0:
       return  " in intervals of <B>%d</B> minute(s)"%self.value
     return ""
  