#!/usr/bin/env python
"""
_Tier0AccountantComponent_

Listen for failed and successful jobs and update the information in
T0AST accordingly.
"""

__revision__ = "$Id: Tier0AccountantComponent.py,v 1.66 2011/01/06 17:14:25 hufnagel Exp $"
__version__ = "$Revision: 1.66 $"

import os
import sys
import logging
import imp
import subprocess

from ProdCommon.Database import Session
from ProdAgentDB.Config import defaultConfig as dbConfig

from ProdCommon.FwkJobRep.ReportParser import readJobReport
from ProdCommon.Core.ProdException import ProdException
from ProdAgentCore.ProdAgentException import ProdAgentException
from ProdAgent.Trigger import Trigger

from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig
from WMCore.DAOFactory import DAOFactory

from T0.GenericTier0.GenericTier0Component import GenericTier0Component
from T0.DataStructs.StreamerFile import StreamerFile
from T0.DataStructs.T0ASTFile import T0ASTFile
from T0.Tier0Accountant.PerformanceRecord import storePerformanceReport


class Tier0AccountantComponent(GenericTier0Component):
    """
    _Tier0AccountantComponent_

    Listen for failed and successful jobs and
    update the information in T0AST accordingly.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component to setup logging.
        """
        GenericTier0Component.__init__(self, **args)
        
        # get transfer system base dir
        self.transferSystemBaseDir = self.lookupParameter("TransferSystemBaseDir", "None")

        # Trigger to allow job cleanup.
        self.trigger = None

        # Data structure to hold job handlers
        self.jobHandlers = {}

        # dao factory to make WMBS calls
        self.daoFactory = None

        logging.info("Tier0Accountant Component Started")

    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "Tier0Accountant:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "Tier0Accountant:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "JobSuccess":
            self.handleSuccessfulJob(payload)
        elif event == "JobFailed":
            self.handleTransientFailure(payload)
        elif event == "GeneralJobFailure":
            logging.debug("GeneralJobFailure: %s" % payload)
            self.handleFailedJob(payload)            
        else:
            logging.error("Unrecognized event: %s" % event)

        return

    def retrieveJobHandler(self, jobType):
        """
        _retrieveJobHandler_

        Retrieve an instance of a job handler for a particular job type.  If
        no instance exists attempt to load one from disk.
        """
        if jobType in self.jobHandlers.keys():
            return self.jobHandlers[jobType]
        
        handlerBaseDir = sys.path[0] + "/JobHandlers"
        logging.debug("retrieveJobHandler(): base dir %s" % \
                      handlerBaseDir)
        (fileHandle, pathName, description) = imp.find_module(jobType,
                                                              [handlerBaseDir])
        try:
            jobHandlerModRef = imp.load_module(jobType, fileHandle, pathName,
                                               description)
        except ImportError, ex:
            logging.error("Can't find handler for job type %s: %s" % \
                          (jobType, ex))
            fileHandle.close()
            raise ImportError, str(ex)
            #return None

        fileHandle.close()
        jobHandlerClass = getattr(jobHandlerModRef, jobType)
        
        # make sure initialize runConfigCache before passing to the handler
        if self.runConfigCache == None:
            self.initRunConfigCache()
        
        if getattr(self, "stateChanger", None) == None: 
            
            DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
            self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)
        
        jobHandlerRef = jobHandlerClass(self.t0astDBConn, self.paDBConn, 
                                        self.runConfigCache, self.stateChanger,
                                        self.daoFactory)        
        self.jobHandlers[jobType] = jobHandlerRef

        return jobHandlerRef

    

    
    def loadJobReport(self, jobReportPath):
        """
        _loadJobReport_

        Given a framework job report on disk, load it and return a
        FwkJobReport instance.
        """
        # The jobReportPath may be prefixed with "file://" which needs to be
        # removed so it doesn't confuse the FwkJobReport() parser.
        jobReportPath = jobReportPath.replace("file://","")

        if not os.path.exists(jobReportPath):
            logging.error("Bad FwkJobReport Path: %s" % jobReportPath)
            return None

        if os.path.getsize(jobReportPath) == 0:
            logging.error("Empty FwkJobReport: %s" % jobReportPath)
            return None

        try:
            jobReports = readJobReport(jobReportPath)
        except Exception, msg:
            logging.error("Cannot load %s: %s" % (jobReportPath, msg))
            return None

        # The readJobReport() function will return a list of job reports,
        # but our job report files only contain one.
        if len(jobReports) == 0:
            logging.error("Bad FWJR: %s" % jobReportPath)
            return None
        jobReport = jobReports[0]

        if jobReport.jobSpecId == None:
            logging.error("No jobId in the job report.")
	    return None

        return jobReport


    def convertToStreamerFile(self, processedFile):
        newFile = StreamerFile()

        # this works because streamer files only have single run/lumi
        newFile["RunNumber"] = processedFile.runs.keys()[0]
        newFile["LumiID"] = processedFile.runs.values()[0][0]

        newFile["FileSize"] = processedFile["Size"]
        newFile["Events"] = processedFile["TotalEvents"]
        newFile["LFN"] = processedFile["LFN"]
        newFile["Exportable"] = 0
        newFile["Stream"] = processedFile["Stream"]

        return newFile

    def convertToT0ASTFile(self, runConfig, jobReport, processedFile):
        primaryDataset = processedFile.dataset[0]["PrimaryDataset"]
        datasetID = runConfig.getPrimaryDatasetID(primaryDataset)
         
        newFile = T0ASTFile()
        newFile["STREAM"] = processedFile["Stream"]
        newFile["RUN_ID"] = processedFile.runs.keys()[0]
        lumiList = processedFile.runs[newFile["RUN_ID"]]
        newFile.setRunLumi(newFile["RUN_ID"], lumiList)
        newFile["FILESIZE"] = processedFile["Size"]
        newFile["EVENTS"] = processedFile["TotalEvents"]
        newFile["LFN"] = processedFile["LFN"]
        newFile["DATASET_ID"] = datasetID
        newFile["PRIMARY_DATASET"] = primaryDataset
        newFile["PROCESSED_DATASET"] = processedFile.dataset[0]["ProcessedDataset"]
        newFile["CHECKSUMS"] = processedFile["Checksums"]
        
        newFile["DATA_TIER"] = processedFile.dataset[0]["DataTier"]
        newFile["BRANCH_HASH"] = processedFile["BranchHash"]
        #T0DO check with Dave which PsetHash is correct one
        # It seems different framework version puts the psethash in different location
        #newFile["PSET_HASH"] = processedFile.dataset[0]["PSetHash"]
        newFile["PSET_HASH"] = jobReport.psetHash
        #this is needed if PhEDEx can use SE name instead of Node name
        # commented out for now
        newFile["LOCATIONS"] = processedFile["SEName"]
        #TODO place holder for log file location
        #newFile["LOG_FILES"] = jobReport.logFiles
        # temporary code to prevent crash: 
        # if BranchHash is not set set to a default value 
        if newFile["PSET_HASH"] == None:
            newFile["PSET_HASH"] = "NO_PSET_HASH"
        
        if newFile["BRANCH_HASH"] == None:
            newFile["BRANCH_HASH"] = "NO_BRANCH_HASH"
        
        if processedFile["MergedBySize"] == "True": 
            newFile["MERGED"] = True 
            if newFile["PROCESSED_DATASET"].endswith("-merged"):
                newFile["PROCESSED_DATASET"] = newFile["PROCESSED_DATASET"][:-7]
        else:
            newFile["MERGED"] = False

        return newFile
        
        
    def parseJobReport(self, jobReport, jobInfo):
        """
        _parseJobReport_

        Pull all the new files out of a job report and return that information
        as an array of T0ASTFile data structures.
        """
        jobFiles = []

        for processedFile in jobReport.files:
            
            if processedFile.has_key("FileType"):
                if processedFile["FileType"].upper() == "STREAMER":
                    newFile = self.convertToStreamerFile(processedFile)
                elif processedFile["FileType"].upper() == "EDM":
                    runConfig = self.getRunConfig(jobInfo["RUN_NUMBER"])
                    newFile = self.convertToT0ASTFile(runConfig, jobReport, processedFile)
                    logging.debug("JobID: %s Job Type: %s File from run %s, dataset %s: %s" %
                              (jobInfo["JOB_IDENTIFIER"], jobInfo["JOB_TYPE"],
                               jobInfo["RUN_NUMBER"], newFile["PRIMARY_DATASET"], newFile))
                else:
                    logging.error("Unknown file type: %s" % processedFile["FileType"])
            else:
                # insert into dataset table if it doesen't exist
                # if the performance is problem move this when they are created
                # i.e. runConfig and AlcaSkimInjector
                # this is move to  WMBSBase handler  and RepackHandler to handle 
                # special case -merge when size is big enough and didn't go through
                # the merge job.
                runConfig = self.getRunConfig(jobInfo["RUN_NUMBER"])
                newFile = self.convertToT0ASTFile(runConfig, jobReport, processedFile)
                logging.debug("JobID: %s Job Type: %s File from run %s, dataset %s: %s" %
                              (jobInfo["JOB_IDENTIFIER"], jobInfo["JOB_TYPE"],
                               jobInfo["RUN_NUMBER"], newFile["PRIMARY_DATASET"], newFile))

            jobFiles.append(newFile)

        return jobFiles
    

    def determineJobType(self, jobSpecID):
        """
        _determineJobType_

        Using the job spec ID, determine whether the job was a Repack,
        RepackMerge or PromptReco job.  That field has the following format:
          JOBTYPE-RUNNUMBER-JOBIDENTIFIER

        This method returns a dictionary with the following keys:
          JOB_TYPE - Currently either Express, Repack, Merge, PromptReco,
                     DQMHarvest or PromptCalib.
          RUN_NUMBER - The run number
          JOB_IDENTIFIER - The job definition ID or in the case of DQM
                           harvesting jobs the dataset.
        """
        result = {}

        if jobSpecID == None:
            logging.debug("No jobSpecId, job type is UNKNOWN.")
            return None
        
        specIDParts = jobSpecID.split('-')

        if len(specIDParts) != 3:
            if specIDParts[0].startswith("Tier0MCFeeder"):
                result["JOB_TYPE"] = "MCStreamer"
                logging.debug("JobType for %s is %s." % (jobSpecID, result["JOB_TYPE"]))
                return result
            elif len(specIDParts) == 5 and specIDParts[0] == "DQMHarvest":
                result["JOB_TYPE"] = "ExpressDQMHarvest"
                result["RUN_NUMBER"] = int(specIDParts[1][3:])
                result["JOB_IDENTIFIER"] = int(specIDParts[4])
                return result 
            elif len(specIDParts) == 5 and specIDParts[0] == "ALCAPROMPTHarvest":
                result["JOB_TYPE"] = "ExpressDQMHarvest"
                result["RUN_NUMBER"] = int(specIDParts[1][3:])
                result["JOB_IDENTIFIER"] = int(specIDParts[4])
                return result 
            elif specIDParts[0] == "PromptReco" or specIDParts[0] == "AlcaSkim":
                result["JOB_TYPE"] = specIDParts[0]
                result["RUN_NUMBER"] = int(specIDParts[1][3:])
                result["JOB_IDENTIFIER"] = int(specIDParts[3])
                return result
            elif specIDParts[0].startswith("PromptCalib"):
                result["JOB_TYPE"] = "PromptCalib"
                result["RUN_NUMBER"] = int(specIDParts[1][3:])
                result["JOB_IDENTIFIER"] = int(specIDParts[3])
                return result
            else:
                logging.error(" Wrong JobspecID format, : %s , should be JobType-Run[number]-id format" % jobSpecID)
                return None

        if specIDParts[0].startswith("Merge") \
               and specIDParts[0] != "MergePack":
            result["JOB_TYPE"] = "Merge"
        else:
            result["JOB_TYPE"] = specIDParts[0]

        result["RUN_NUMBER"] = int(specIDParts[1][3:])

        result["JOB_IDENTIFIER"] = int(specIDParts[2])

        logging.debug("JobType for %s is %s." % (jobSpecID, result["JOB_TYPE"]))

        return result

    def markJobForCleanup(self, jobSpecID):
        """
        _markJobForCleanup_

        Setup the cleanup flag for a job so that the JobCleanup component
        can clean it up.
        """
        try:
            Session.set_database(dbConfig)
            Session.connect()
            Session.start_transaction()
                        
            self.trigger.setFlag("cleanup", jobSpecID,
                                 "Tier0Accountant")

            Session.commit_all()
            Session.close_all()
        except (ProdAgentException, ProdException):
            logging.error("Error marking job %s for cleanup." % jobSpecID)

        return

    def handleSuccessfulJob(self, jobReportPath):
        """
        _handleSuccessfulJob_

        Load the framework job report from a successful job and pass it
        to the handler that corresponds to the job type.  Also update the
        status of the job to "Success".
        """

        jobReport = self.loadJobReport(jobReportPath)
        if jobReport == None:
            return

        jobInfo = self.determineJobType(jobReport.jobSpecId)
        if jobInfo == None:
            self.markJobForCleanup(jobReport.jobSpecId)
            return
        
        if jobInfo["JOB_TYPE"] == "DQMHarvest":
            # not cleaned up, why ?
            return

        jobFiles = self.parseJobReport(jobReport, jobInfo)
        if jobFiles == None:
            logging.error("No files in FWKJR: %s" % jobInfo)
            self.markJobForCleanup(jobReport.jobSpecId)            
            return
        
        jobHandler = self.retrieveJobHandler(jobInfo["JOB_TYPE"])
        if jobHandler == None:
            logging.debug("Unknown job type: %s" % jobInfo["JOB_TYPE"])
            self.markJobForCleanup(jobReport.jobSpecId)            
            return
        
        # save performance data in T0AST: only for successful jobs
        #TODO: commented out for now until further testing
        #storePerformanceReport(jobReport, jobInfo)

        self.t0astDBConn.beginTransaction()
        t0astJobInfo = jobHandler.handleSuccessfulJob(jobInfo, jobFiles)
        jobHandler.changeState(jobInfo, "Success")
        self.t0astDBConn.commit()

        self.markJobForCleanup(jobReport.jobSpecId)

        if self.transferSystemBaseDir != "None" and \
               os.path.exists(self.transferSystemBaseDir) and \
               (jobInfo["JOB_TYPE"] == "Repack" or jobInfo["JOB_TYPE"] == "Express"):

            filenameParams = ""
            for filename in t0astJobInfo['input_filenames']:
                filenameParams += """-FILENAME %s """ % filename

            logging.debug("Notifying transfer system about processed streamers")
            p = subprocess.Popen("/bin/bash",stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            output, error = p.communicate("""
            export T0_BASE_DIR=%s
            export T0ROOT=${T0_BASE_DIR}/T0
            export CONFIG=${T0_BASE_DIR}/Config/TransferSystem_CERN.cfg
 
            export PERL5LIB=${T0ROOT}/perl_lib
 
            ${T0ROOT}/operations/sendRepackedStatus.pl --config $CONFIG %s
            """ % (self.transferSystemBaseDir, filenameParams))

            if len(error) > 0:
                logging.error("ERROR: Could not notify transfer system about processed streamers ")
                logging.error("ERROR: %s" % error)

        return
        
    def handleFailedJob(self, jobSpecID):
        """
        _handleFailedJob_

        Called when a job has used up it's quota of retries and failed every
        time.  The job handler's handleFailedJob() method is called and the
        job's status is updated to "Failed".
        """
        jobInfo = self.determineJobType(jobSpecID)
        if jobInfo == None:
            return
        
        # additional step for the error checking in None WMBS job
        # WMBS error checking will occur in WMBS handler plug-in 
        # this can be moved to Repack handler but stay here for now
        # in case we need if to express stream
        
        jobHandler = self.retrieveJobHandler(jobInfo["JOB_TYPE"])
        if jobHandler == None:
            logging.debug("Unknown job type: %s" % jobInfo["JOB_TYPE"])
            return
        
        self.t0astDBConn.beginTransaction()        
        jobHandler.handleFailedJob(jobInfo)
        jobHandler.changeState(jobInfo, "Failure")
        self.t0astDBConn.commit()

        self.markJobForCleanup(jobSpecID)

        return

    def handleTransientFailure(self, jobReportPath):
        """
        _handleTransientFailure_

        Called when a job fails and will be retried by the ProdAgent
        ErrorHandler.  Currently, all it does it increment the "retry_count"
        column for the job in the appropriate job def table.
        """
        jobReport = self.loadJobReport(jobReportPath)
        if jobReport == None:
            return
        
        jobInfo = self.determineJobType(jobReport.jobSpecId)        
        if jobInfo == None:
            logging.error("Could not retrieve jobInfo from FJR: %s" % \
                          jobReportPath)
            return

        errorMessage = "Job %s failed:  " % jobReport.jobSpecId

        if len(jobReport.errors) > 0:
            errorMessage += "Type:  %s" % \
                            jobReport.errors[0]["Type"]
            errorMessage += "Status: %s" % \
                            jobReport.errors[0]["ExitStatus"]
            errorMessage += "Description:  %s" % \
                            jobReport.errors[0]["Description"]        
        else:
            errorMessage += "No Error Report in FWJR."

        for logLFN in jobReport.logFiles.keys():
            errorMessage += "%s: %s" % (jobReport.logFiles[logLFN], logLFN)

        self.publishAlert("Warning", errorMessage)

        if jobInfo["JOB_TYPE"] == "DQMHarvest":
            return
        
        return
    
    
    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST, create an instance
        of the trigger class and enter the message loop.
        """
        self.ms.subscribeTo("Tier0Accountant:StartDebug")
        self.ms.subscribeTo("Tier0Accountant:EndDebug")
        self.ms.subscribeTo("SetJobCleanupFlag")
        self.ms.subscribeTo("JobSuccess")
        self.ms.subscribeTo("JobFailed")
        self.ms.subscribeTo("GeneralJobFailure")
        self.ms.commit()

        self.connectT0AST()
        self.connectPADB()

        self.daoFactory = DAOFactory(package = "WMCore.WMBS", logger = logging,
                                     dbinterface = self.t0astDBConn.dbInterface)

        self.trigger = Trigger.Trigger(self.ms)
        
        # wait for messages
        self.messageLoop()
        
