#!/usr/bin/env python
"""
_WMBSSBaseHandler_

Provide funcitons to common WMBS Job handling.
The sub class should be provide the appropriate fileset information
"""

__revision__ = "$Id: WMBSBaseHandler.py,v 1.45 2011/01/19 21:23:47 hufnagel Exp $"
__version__ = "$Revision: 1.45 $"

import logging
import threading

from WMCore.WMBS.Job import Job as WMBSJob

from T0.Tier0Accountant.JobHandlerInterface import JobHandlerInterface
from T0.State.Database.Reader import ListJobs
from T0.State.Database.Writer import InsertJob
from T0.State.Database.Writer import InsertDataset

class WMBSBaseHandler(JobHandlerInterface):
    """
    _WMBSSplit_

    Job handler WMBS jobs
    """

    def handleSuccessfulJob(self, jobInfo, jobFiles):
        """
        _handleSuccessfulJob_

        register output files
        add output files to appropriate filesets (for next processing)
        setup input to output file parent/child assoc
        assoc output files to their datasetpath

        """
        myThread = threading.currentThread()

        wmbsJobInfo = ListJobs.listWMBSJobInfo(self.t0astDBConn,
                                               jobInfo["JOB_IDENTIFIER"],
                                               jobInfo["JOB_TYPE"])

        if wmbsJobInfo == None:
            logging.error("ERROR: Job of type %s with id %s does not exist"
                          % (jobInfo["JOB_TYPE"], jobInfo["JOB_IDENTIFIER"]))
            return None

        datasetpathDict = {}
        datasetpathList = []
        fileDatasetpathList = []

        filesetFiles = {}
        procDatasets = set([])
        for t0astFile in jobFiles:

            # update first/last event information for each t0ast file
            t0astFile["FIRST_EVENT"] = wmbsJobInfo['first_event']
            t0astFile["LAST_EVENT"] = wmbsJobInfo['last_event']

            wmbsFile = t0astFile.convertToWMBSFile()

            fileid = wmbsFile["id"]
            primds = t0astFile["PRIMARY_DATASET"]
            procds = t0astFile["PROCESSED_DATASET"]
            datatier = t0astFile["DATA_TIER"]

            for inputFile in wmbsJobInfo['input_files']:
                heritageAction = self.daoFactory(classname = "Files.Heritage")
                heritageAction.execute(child = fileid,
                                       parent = inputFile,
                                       conn = myThread.transaction.conn,
                                       transaction = True)

            filesetListByFile = self.getFilesetListByFile(t0astFile)
            for fileset in filesetListByFile:
                if not filesetFiles.has_key(fileset):
                    filesetFiles[fileset] = []
                filesetFiles[fileset].append(wmbsFile['id'])

            procDatasets.add(procds)
            
            if not datasetpathDict.has_key(primds):
                datasetpathDict[primds] = {}
            if not datasetpathDict[primds].has_key(procds):
                datasetpathDict[primds][procds] = []
            if datatier not in datasetpathDict[primds][procds]:
                datasetpathDict[primds][procds].append(datatier)
                datasetpathList.append( { 'PRIMDS' : primds,
                                          'PROCDS' : procds,
                                          'DATATIER' : datatier } )

            fileDatasetpathList.append( { 'ID' : wmbsFile["id"],
                                          'PRIMDS' : primds,
                                          'PROCDS' : procds,
                                          'DATATIER' : datatier } )

        addFilesToFileset = self.daoFactory(classname = "Fileset.BulkAdd")

        filesetFileList = []
        for fileset, fileids in filesetFiles.items():
            for fileid in fileids:
                filesetFileList.append ( { 'fileset' : fileset,
                                           'fileid' : fileid } )

        if len(filesetFileList) > 0:
            addFilesToFileset.execute(binds = filesetFileList,
                                      conn = myThread.transaction.conn,
                                      transaction = True)

        for procDataset in procDatasets:
            InsertDataset.insertProcessedDataset(self.t0astDBConn,
                                                 procDataset)

        # insert datasetpath info for output files
        if len(datasetpathList) > 0:
            InsertDataset.insertDatasetByName(self.t0astDBConn,
                                              datasetpathList)

        # associate output WMBS files with datasetpath info
        if len(fileDatasetpathList) > 0:
            InsertDataset.assocWMBSFileDatasetByNames(self.t0astDBConn,
                                                      fileDatasetpathList)
        
        InsertJob.updateWMBSJobOutcome(self.t0astDBConn,
                                       jobInfo["JOB_IDENTIFIER"])

        completeInput = self.daoFactory(classname = "Jobs.CompleteInput")
        completeInput.execute(jobInfo["JOB_IDENTIFIER"],
                              conn = myThread.transaction.conn,
                              transaction = True)

        return wmbsJobInfo
    
    def handleFailedJob(self, jobInfo):
        """
        _handleFailedJob_

        handle common task for failed wmbs job

        """
        myThread = threading.currentThread()
        wmbsJobInfo = ListJobs.listWMBSJobInfo(self.t0astDBConn,
                                               jobInfo["JOB_IDENTIFIER"])

        if wmbsJobInfo == None:
            logging.error("ERROR: Job of type %s with id %s does not exist"
                          % (jobInfo["JOB_TYPE"], jobInfo["JOB_IDENTIFIER"]))
            return

        failInput = self.daoFactory(classname = "Jobs.FailInput")
        failInput.execute(jobInfo["JOB_IDENTIFIER"],
                          conn = myThread.transaction.conn,
                          transaction = True)

        return
    
    def changeState(self, jobInfo, state):

        wmbsJob = WMBSJob(id = jobInfo["JOB_IDENTIFIER"])
        
        # make state change all at once it is one transaction anyway.
        # there are many transactions since there is restriction of  
        # sequence of transaction
        #TODO: we could make seperate transaction but there is no mechanism to
        #recover between the status currently.
        if state == "Success":
            self.stateChanger.propagate(wmbsJob, "executing", "created")
            self.stateChanger.propagate(wmbsJob, "complete", "executing")
            self.stateChanger.propagate(wmbsJob, "success", "complete")
            self.stateChanger.propagate(wmbsJob, "cleanout", "success")
        elif state == "Failure":
            self.stateChanger.propagate(wmbsJob, "executing", "created")
            self.stateChanger.propagate(wmbsJob, "jobfailed", "executing")
            self.stateChanger.propagate(wmbsJob, "exhausted", "jobfailed")
            self.stateChanger.propagate(wmbsJob, "cleanout", "exhausted")
        return

    def getFilesetListByFile(self, t0astFile):
        """
        _getFilesetListByFile_ 
        
        retrieve the fileset list for next processing steps

        default is none, need to override in child

        """
        return []
