#!/usr/bin/env python
"""
_MergeWFCache

Workflow cache for Merge workflows.
"""

__revision__ = "$Id: MergeWFCache.py,v 1.6 2010/02/11 20:37:45 sfoulkes Exp $"
__version__ = "$Revision: 1.6 $"

import logging
import os

from T0.State.Database.Reader import ListDatasets

from T0.WorkflowFactory.MergeWorkflow import MergeWorkflow
from T0.WorkflowFactory.WorkflowCacheInterface import WorkflowCacheInterface

class MergeWFCache(WorkflowCacheInterface):
    """
    _MergeWFCache_

    Generate and store Merge workflows.
    """
    def createWorkflow(self, runConfig, params):
        """
        _createWorkflow_

        Create a workflow for a particular run and dataset path.  The following
        parameters are supported:
          INPUT_DATASET_ID - Must be passed in, this is the dataset path ID that
            this workflow will merge. 
          CMSSW_VERSION - Framework version to use when merging.  If this is not
            specified the workflow creation code will use the framework version
            that produced the data to merge it.
          PROCESSING_VERSION - Processing version used to create the output LFN.
            If this is not passed in the workflow creation code will use the
            same processing version that was used to create the data.
          SITE_NAME - The name of the site that the workflow is to be run at.
          MAX_MERGE_EVENTS - The maximum number of events to insert into a merge
            job.
          MIN_MERGE_SIZE - The minimum size of a merged file.
          MAX_MERGE_SIZE - The maximum size of a merged file.
          MAX_MERGE_FILES - The maximum number of files in a merge job.
          WF_CLASS - The workflow class, this is used to construct job names.
        """
        runNumber = runConfig.getRunNumber()
        datasetNames = ListDatasets.listDatasetNamesFromID(self.t0astDBConn,
                                                           params["INPUT_DATASET_ID"])

        if "CMSSW_VERSION" in params.keys():
            cmsswVer = params["CMSSW_VERSION"]
        else:
            if datasetNames["TIER"] == "RAW":
                streamID = runConfig.getStreamIDfromDatasetPathID(params["INPUT_DATASET_ID"])
                repackConfig = runConfig.getRepackConfiguration(streamID)
                cmsswVer = repackConfig.CMSSWVersion
                processingVersion = repackConfig.ProcessingVersion
            elif datasetNames["TIER"] == "RECO":
                recoConfig = runConfig.getRecoConfiguration(params["INPUT_DATASET_ID"])
                cmsswVer = recoConfig.CMSSWVersion
                processingVersion = recoConfig.ProcessingVersion
            elif datasetNames["TIER"] == "ALCARECO":
                alcaConfig = runConfig.getAlcaConfiguration(params["INPUT_DATASET_ID"])
                cmsswVer = alcaConfig.CMSSWVersion
                processingVersion = alcaConfig.ProcessingVersion            
            else:
                cmsswVer = None

        if cmsswVer == None or cmsswVer == "Undefined":
            logging.error("No framework version for run %s" % runNumber)
            return None            
            cmssw

        outputInfo = {"dataset": datasetNames["PRIMARY"], "processedDataset": datasetNames["PROCESSED"],
                      "dataTier": datasetNames["TIER"], "acquisitionEra": runConfig.getAcquisitionEra(),
                      "processingVersion": params["PROCESSING_VERSION"]}

        newWorkflow = MergeWorkflow(runNumber, cmsswVer, self["cmsPath"], self["scramArch"], outputInfo)
        newWorkflow.setWorkflowSites(params["SITE_NAME"])

        newWorkflow.setJobSplitParams({"min_merge_size": params["MIN_MERGE_SIZE"],
                                       "max_merge_size": params["MAX_MERGE_SIZE"],
                                       "max_merge_events": params["MAX_MERGE_EVENTS"],
                                       "max_merge_files": params["MAX_MERGE_FILES"]})
        newWorkflow.setWorkflowClass(params["WF_CLASS"])
        newWorkflow.setWorkflowDescription("%s-%s" % (datasetNames["PRIMARY"],
                                                      datasetNames["TIER"]))

        workflowName = "Merge-Run%s-%s-%s-%s" % (runNumber, datasetNames["TIER"], datasetNames["PRIMARY"], datasetNames["PROCESSED"])
        workflowSpec = newWorkflow.makeWorkflowSpec(name = workflowName, enableLazyDownload = True)
        self.saveWorkflow(workflowSpec)        
        return workflowSpec

    def generateKey(self, runConfig, params):
        """
        _generateKey_

        Generate a unique key for Merge workflows.  Merge workflows can be
        uniquely identified by run number and dataset path ID.  They key
        will take the following form: RunNumber-DatasetPathID.
        """
        key = "%s-%s" % (runConfig.getRunNumber(), params["INPUT_DATASET_ID"])
        return key
