#!/usr/bin/env python
"""
_Tier0MergerComponent_

Search for mergable files in T0AST that can be gathered together
and merged into larger files.  Create and submit merge jobs to
the ProdAgent JobQueue.
"""

__revision__ = "$Id: Tier0MergerComponent.py,v 1.63 2011/01/13 11:02:33 hufnagel Exp $"
__version__ = "$Revision: 1.63 $"

import os
import sys
import shutil
import logging

from JobQueue import JobQueueDB
from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0 import Globals

from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListJobs
from T0.State.Database.Writer import InsertJob

from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig
from WMCore.WMBS.Job import Job

from T0.JobFactory.MergeJob import MergeJob
from T0.WorkflowFactory.MergeWorkflow import MergeWorkflow


class Tier0MergerComponent(GenericTier0Component):
    """
    _Tier0MergerComponent_

    Search for mergable files in T0AST that can be gathered together
    and merged into larger files.  Create and submit merge jobs to
    the ProdAgent JobQueue.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)

        # Data structure to hold WorkflowSpecs.  This is a two level
        # dictionary where the key to the first level is the run number,
        # and the second level is dataset id.
        self.workflows = {}

        # cache datasetpath information to avoid queries
        self.datasetpaths = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2

        # Check the ProdAgent configuration for the following parameters:
        #   minMergeSize - Min size in bytes of a merged file
        #   maxMergeSize - Max size in bytes of a merged file
        #   maxInputFiles - Max number of input files to merge
        #   maxEdmSize - Max allowed output file size
        #                ( assume sum(inputs) = output )
        #   maxOverMergeSize - Max allowed overmerge size
        #                      (if file pushes us from <min to >max)
        #   maxMergeEvents - Maximum number of events in a file
        #   t0astPollInterval - How often T0AST is checked for new files
        #   scramArch - SCRAM Architecture
        #   cmsPath - Path to CMS software
        #   jobCacheDir - Where jobSpecs and workflows are stored
        self.minMergeSize = int(self.lookupParameter("minMergeSize", 2.1 * 1024 * 1024 * 1024))
        self.maxMergeSize = int(self.lookupParameter("maxMergeSize", 4.0 * 1024 * 1024 * 1024))
        self.maxInputFiles = int(self.lookupParameter("maxInputFiles", 1000))
        self.maxEdmSize = int(self.lookupParameter("maxEdmSize", 10 * 1024 * 1024 * 1024))
        self.maxOverMergeSize = int(self.lookupParameter("maxOverMergeSize", self.maxEdmSize))
        self.maxMergeEvents = int(self.lookupParameter("maxMergeEvents", 500000))

        # catch configuration errors
        if self.maxOverMergeSize > self.maxEdmSize:
            self.maxOverMergeSize = self.maxEdmSize

        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")
        self.pollInterval = self.lookupParameter("PollInterval", "00:01:00")

        # place for all workflows, jobspecs, configs, etc
        self.jobCacheDir = self.lookupParameter("JobCacheDir", self.args["ComponentDir"])

        # WMCore splitter factory for creating Merge jobs
        self.splitterFactory = SplitterFactory(package = "T0.JobSplitting")

        # Check if we should use error datasets
        if self.lookupParameter("UseErrorDataset", "On") == "Off":
            self.useErrorDataset = False
        else:
            self.useErrorDataset = True

        # Check if strict lumi is enabled
        if self.lookupParameter("StrictLumiOrder", "Off") == "On":
            self.strictLumiOrder = True
        else:
            self.strictLumiOrder = False

        self.mergeSubscription = None

        # See if lazy downloads are enabled
        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        return
        
    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if message == "Tier0Merger:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "Tier0Merger:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "Tier0Merger:PollForJobs":
            self.pollForJobs()
        elif message == "CleanUp":
            self.cleanupRun(payload)
        else:
            logging.error("Unknown message: %s, payload: %s" % (message, payload))

        return

    def retrieveWorkflow(self, run, datasetpath,
                         primds, procds, datatier,
                         sendToErrorDataset):
        """
        _retrieveWorkflow_

        Retrieve a workflow from the cache that the Tier0Merger maintains.
        If the workflow does not exist a new one will be created. If workflow
        creation fails for some reason None will be returned.

        """
        if self.workflows.has_key(run):
            if self.workflows[run].has_key(datasetpath):
                if self.workflows[run][datasetpath].has_key(sendToErrorDataset):
                    return self.workflows[run][datasetpath][sendToErrorDataset]
            else:
                self.workflows[run][datasetpath] = {}
        else:
            self.workflows[run] = { datasetpath : {} }
                    
        # that we made it here means we have to create a new workflow
        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %s" % run)
            return None

        # override primary dataset if sendToErrorDataset and routing to error dataset active
        if sendToErrorDataset and self.useErrorDataset:
            primds = runConfig.getErrorDatasetName(primds)
            if primds == None:
                logging.error("Could not get error dataset for dataset %s" % primds)
                return None

        workflow = self.createWorkflow(run, datasetpath,
                                       primds, procds, datatier)
        self.workflows[run][datasetpath][sendToErrorDataset] = workflow

        return workflow

    def createWorkflow(self, run, datasetpath,
                       primds, procds, datatier):
        """
        _createWorkflow_

        Creates a merge workflow for a specific run and dataset 
        combination.  If workflow creation fails None is returned.

        A NewWorkflow message is generated for each new workflow
        with the path to the workflow being passed as the payload.

        """
        logging.debug("createWorkflow(): Running run %s, datasetpath %s" % \
                      (run, datasetpath))

        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %d" % run)
            return None

        cmsswVer = None
        if datatier == "RAW":
            streamID = runConfig.getStreamIDfromDatasetPathID(datasetpath)
            repackConfig = runConfig.getRepackConfiguration(streamID)
            cmsswVer = repackConfig.CMSSWVersion
            processingVersion = repackConfig.ProcessingVersion
        elif datatier == "RECO" or datatier == "AOD" or datatier == "DQM" or datatier == "ALCAPROMPT":
            recoConfig = runConfig.getRecoConfiguration(datasetpath)
            cmsswVer = recoConfig.CMSSWVersion
            processingVersion = recoConfig.ProcessingVersion
        elif datatier == "ALCARECO":
            alcaConfig = runConfig.getAlcaConfiguration(datasetpath)
            cmsswVer = alcaConfig.CMSSWVersion
            processingVersion = alcaConfig.ProcessingVersion

        if cmsswVer == None:
            logging.error("No framework version for run %s" % run)
            return None
        elif cmsswVer == "Undefined":
            logging.error("No framework version for this dataset: %s" % datasetpath)
            return None

        outModuleInfoList = [ { "dataTier" : datatier,
                                "dataset" : primds,
                                "processedDataset" : procds,
                                "acquisitionEra" : runConfig.getAcquisitionEra(),
                                "processingVersion" : processingVersion } ]

        workflowName = "Merge-Run%s-%s-%s-%s" % (run, datatier, primds, procds)

        workflowDir = os.path.join(self.jobCacheDir, "workflows", "Run%s" % run)
        workflowSpecFile = os.path.join(workflowDir, "%s-workflow.xml" % workflowName)

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No Workflow for run, dataset: %s, %s" % (run, primds))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowSpecFile)

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return workflowSpec

        workflow = MergeWorkflow(run,
                                 cmsswVer,
                                 self.cmsPath,
                                 self.scramArch,
                                 *outModuleInfoList)

        workflowSpec = workflow.makeWorkflowSpec(workflowName,
                                                 enableLazyDownload = self.useLazyDownload,
                                                 production = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)
        
        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)

        workflowSpec.save(workflowSpecFile)

        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)

    def createJobSpecs(self, run, datasetpath, jobs, sendToErrorDataset):
        """
        _createJobSpecs_

        Loop over all jobs and process them (they are all for same run/datasetpath)

        Generate job specs for every job

        """
        logging.debug("createJobSpecs(): Running...")
        
        runConfig = self.getRunConfig(run)
        if runConfig == None:
            logging.error("Could not retrieve run config for run %s" % run)
            return {}

        if not self.datasetpaths.has_key(datasetpath):
            self.datasetpaths[datasetpath] = ListDatasets.listDatasetPathInfo(self.t0astDBConn,
                                                                              datasetpath)

        primds = self.datasetpaths[datasetpath]['primds']
        procds = self.datasetpaths[datasetpath]['procds']
        datatier = self.datasetpaths[datasetpath]['datatier']
        
        workflowSpec = self.retrieveWorkflow(run, datasetpath,
                                             primds, procds, datatier,
                                             sendToErrorDataset)
        if workflowSpec == None:
            logging.error("No workflow for run %s, datasetpath %s" % (run,
                                                                      datasetpath))
            return {}

        jobSpecBaseDir = os.path.join(self.jobCacheDir, "jobspecs", "Run%d" % run)

        jobFactory = MergeJob(workflowSpec)

        jobSpecDict = {}
        for job in jobs:

            jobSpecDict[job['id']] = jobFactory(job['id'],
                                                datatier,
                                                job['files'],
                                                jobSpecBaseDir)

        return jobSpecDict


    def pollForJobs(self):
        """
        _pollForJobs_

        Poll T0AST for new Merge jobs

        Call job creation code, then queue them in JobQueue
        and update name and status in WMBS

        """
        logging.debug("pollForJobs(): Running...")

        self.t0astDBConn.beginTransaction()

        wmbsJobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                              subscription = self.mergeSubscription)

        wmbsJobFactory(minMergeSize = self.minMergeSize,
                       maxMergeSize = self.maxMergeSize,
                       maxOverMergeSize = self.maxOverMergeSize,
                       maxInputFiles = self.maxInputFiles,
                       maxEdmSize = self.maxEdmSize,
                       maxMergeEvents = self.maxMergeEvents,
                       strictLumiOrder = self.strictLumiOrder,
                       t0astDBConn = self.t0astDBConn)

        self.t0astDBConn.commit()

        jobs = ListJobs.listNewWMBSJobs(self.t0astDBConn, self.mergeSubscription['id'])

        # sort jobs by run and datasetpath
        jobDict = {}
        for job in jobs:
            run = job['run']
            datasetpath = job['datasetpath']

            # intialize dictionary tree
            if not jobDict.has_key(run):
                jobDict[run] = {}
            if not jobDict[run].has_key(datasetpath):
                jobDict[run][datasetpath] = []

            jobDict[run][datasetpath].append(job)

        # process jobs by run and datasetpath, then sort into jobs that write
        # output normally and jobs that write to error dataset based on
        # whether first_event is -1
        jobSpecDict = {}
        for run in jobDict.keys():
            for datasetpath, jobs in jobDict[run].items():

                errorJobs = []
                normalJobs = []
                for job in jobs:
                    if self.useErrorDataset and job['first_event'] == -1:
                        errorJobs.append(job)
                    else:
                        normalJobs.append(job)

                if len(normalJobs) > 0:
                    jobSpecDict.update( self.createJobSpecs(run, datasetpath, normalJobs,
                                                            sendToErrorDataset = False) )

                if len(errorJobs) > 0:
                    jobSpecDict.update( self.createJobSpecs(run, datasetpath, errorJobs,
                                                            sendToErrorDataset = True) )

        if len(jobSpecDict) > 0:

            # aggregate information for efficient database operations
            wmbsJobs = []
            jobNames = {}
            for jobId, jobSpec in jobSpecDict.items():
                wmbsJobs.append(Job(id = jobId))
                jobNames[jobId] = jobSpec.parameters["JobName"]

            # now do actual PA and T0AST DB modifications
            try:
                self.paDBConn.beginTransaction()

                for jobSpec in jobSpecDict.values():

                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters["JobSpecFile"],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority,
                                             siteName = None,
                                             status = "new")

                self.t0astDBConn.beginTransaction()

                InsertJob.updateWMBSJobNames(self.t0astDBConn, jobNames)
                self.stateChanger.propagate(wmbsJobs, 'created', 'new')

                self.paDBConn.commit()
                self.t0astDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(jobSpecDict))
            except StandardError, ex:
                self.t0astDBConn.rollback()
                self.paDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        self.ms.publishUnique("Tier0Merger:PollForJobs", "", self.pollInterval)
        self.ms.commit()
        return

    def cleanupRun(self, payload):
        """
        _cleanupRun_

        Once a run has completed push out any files that have not yet
        been merged and perform the following cleanup:
          - Clear out the workflow template for the run
          - Clear all workflows and jobspec files off disk
        """
        try:
            run = int(payload)
        except ValueError:
            logging.error("Run number is not valid: %s" % payload)
            return

        logging.debug("Cleaning up run %s" % run)

        if run in self.workflows.keys():
            del self.workflows[run]

        workflowDir = self.jobCacheDir + "/workflows/Run%s" % run
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % run

        shutil.rmtree(workflowDir, ignore_errors=True)
        shutil.rmtree(jobSpecDir, ignore_errors=True)
        return

    def startComponent(self):
        """
        _startComponent_

        Fire up the message service and define the list of
        messages that this component subscribes to
        """
        self.ms.subscribeTo("Tier0Merger:StartDebug")
        self.ms.subscribeTo("Tier0Merger:EndDebug")
        self.ms.subscribeTo("Tier0Merger:PollForJobs")

        self.ms.subscribeTo("CleanUp")

        self.connectT0AST()
        self.connectPADB()

        # Load up the merge subscription for easy access.
        self.mergeSubscription = Globals.retrieveMergeSubscription()

        self.ms.publishUnique("Tier0Merger:PollForJobs", "", self.pollInterval)
        self.ms.commit()

        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)
        
        # Wait for messages.
        self.messageLoop()
