#!/usr/bin/env python
"""
_Tier1AccountantComponent_

"""

__revision__ = "$Id: Tier1AccountantComponent.py,v 1.32 2009/11/09 16:55:00 sfoulkes Exp $"
__version__ = "$Revision: 1.32 $"

import os
import sys
import logging
import threading

from ProdCommon.Database import Session
from ProdAgentDB.Config import defaultConfig as dbConfig

from ProdCommon.Core.ProdException import ProdException
from ProdAgentCore.ProdAgentException import ProdAgentException
from ProdAgentCore.Configuration import loadProdAgentConfiguration

from WMCore.DAOFactory import DAOFactory
from WMCore.Database import DBFactory
from WMCore.Database import DBFormatter
from WMCore.Database import Transaction

from WMCore.FwkJobReport.ReportParser import readJobReport

from WMCore.DataStructs.Run import Run
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Workflow import Workflow

from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig

from WMComponent.DBSBuffer.Database.Interface.DBSBufferFile import DBSBufferFile

from T0.GenericTier0.GenericTier0Component import GenericTier0Component
from ProdAgent.Trigger import Trigger

class Tier1AccountantComponent(GenericTier0Component):
    """
    _Tier1AccountantComponent_

    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component to setup logging.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("Tier1Accountant Component Started")

        # Trigger to allow job cleanup.
        self.trigger = None

    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "Tier1Accountant:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "Tier1Accountant:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "Tier1Accountant:CleanoutPoll":
            self.pollForCleanout()
        elif event == "JobSuccess":
            self.handleSuccessfulJob(payload)
        elif event == "JobFailed":
            self.handleTransientFailure(payload)
        elif event == "GeneralJobFailure":
            logging.debug("GeneralJobFailure: %s" % payload)
            self.handleFailedJob(payload)            
        else:
            logging.error("Unrecognized event: %s" % event)

        return

    def loadJobReport(self, jobReportPath):
        """
        _loadJobReport_

        Given a framework job report on disk, load it and return a
        FwkJobReport instance.
        """
        logging.debug("accountant: begin loadJobReport()")
        # The jobReportPath may be prefixed with "file://" which needs to be
        # removed so it doesn't confuse the FwkJobReport() parser.
        jobReportPath = jobReportPath.replace("file://","")

        if not os.path.exists(jobReportPath):
            logging.error("Bad FwkJobReport Path: %s" % jobReportPath)
            return None

        if os.path.getsize(jobReportPath) == 0:
            logging.error("Empty FwkJobReport: %s" % jobReportPath)
            return None

        try:
            jobReports = readJobReport(jobReportPath)
        except Exception, msg:
            logging.error("Cannot load %s: %s" % (jobReportPath, msg))
            return None

        # The readJobReport() function will return a list of job reports,
        # but our job report files only contain one.
        if len(jobReports) == 0:
            logging.error("Bad FWJR: %s" % jobReportPath)
            return None
        jobReport = jobReports[0]

        if jobReport.jobSpecId == None:
            logging.error("No jobId in the job report.")
            return None

        logging.debug("accountant: end loadJobReport()")

        return jobReport

    def connectTier1DB(self):
        """
        _connectTier1DB_

        Connect to the Tier1 WMBS database.  This will pull the database
        configuration out of the "LocalWMBSDB" section of the ProdAgent config.
        """
        logging.debug("accountant: begin connectTier1DB()")
        
        paConfig = loadProdAgentConfiguration()
        dbConfig = paConfig.getConfig("LocalWMBSDB")

        options = {}

        if dbConfig.has_key("unix_socket"):
            options["unix_socket"] = dbConfig["unix_socket"]
            
        dbFactory = DBFactory.DBFactory(logging, dbConfig["database"], options)
        dbInterface = dbFactory.connect()
        dbFormatter = DBFormatter.DBFormatter(logging, dbInterface)

        myThread = threading.currentThread()            
        myThread.dbi = dbInterface
        myThread.logger = logging
        myThread.transaction = Transaction.Transaction(myThread.dbi)        
        myThread.dialect = dbConfig["dialect"]

        self.daoFactory = DAOFactory(package = "WMCore.WMBS", logger = logging,
                                     dbinterface = myThread.dbi)
        self.setFWJRPathAction = self.daoFactory(classname = "Jobs.SetFWJRPath")

        logging.debug("accountant: end connectTier1DB()")
        return

    def determineJobID(self, jobSpecID):
        """
        _determineJobID_

        Using the job spec ID, determine the job ID.  The job ID is the last
        parameter to the job spec ID string.
        """
        if jobSpecID == None:
            logging.debug("No jobSpecId, job type is UNKNOWN.")
            return None
        
        specIDParts = jobSpecID.split('-')
        return int(specIDParts[-1])

    def markJobForCleanup(self, jobSpecID):
        """
        _markJobForCleanup_

        Setup the cleanup flag for a job so that the JobCleanup component
        can clean it up.
        """
        logging.debug("accountant: begin markJobForCleanup()")
        
        try:
            Session.set_database(dbConfig)
            Session.connect()
            Session.start_transaction()
                        
            self.trigger.setFlag("cleanup", jobSpecID,
                                 "Tier1Accountant")

            Session.commit_all()
            Session.close_all()
        except (ProdAgentException, ProdException):
            logging.error("Error marking job %s for cleanup." % jobSpecID)

        logging.debug("accountant: end markJobForCleanup()")

        return

    def handleSuccessfulJob(self, jobReportPath):
        """
        _handleSuccessfulJob_

        """
        jobReport = self.loadJobReport(jobReportPath)
        if jobReport == None:
            return
        
        jobID = self.determineJobID(jobReport.jobSpecId)
        if jobID == None:
            self.markJobForCleanup(jobReport.jobSpecId)
            return

        myThread = threading.currentThread()
        myThread.transaction.begin()

        wmbsJob = Job(id = jobID)
        self.setFWJRPathAction.execute(jobID = jobID, fwjrPath = jobReportPath,
                                       conn = myThread.transaction.conn,
                                       transaction = True)
        self.stateChanger.propagate([wmbsJob], "complete", "executing")
        myThread.transaction.commit()
        return

    def handleTransientFailure(self, jobReportPath):
        """
        _handleTransientFailure_

        """
        logging.debug("HANDLE TRANSIENT %s" % jobReportPath)
        jobReport = self.loadJobReport(jobReportPath)
        if jobReport == None:
            return
        
        jobID = self.determineJobID(jobReport.jobSpecId)
        if jobID == None:
            self.markJobForCleanup(jobReport.jobSpecId)
            return

        myThread = threading.currentThread()
        myThread.transaction.begin()

        wmbsJob = Job(id = jobID)
        self.setFWJRPathAction.execute(jobID = jobID, fwjrPath = jobReportPath,
                                       conn = myThread.transaction.conn,
                                       transaction = True)        
        self.stateChanger.propagate([wmbsJob], "complete", "executing")
        myThread.transaction.commit()
        return
        
    def handleFailedJob(self, jobSpecID):
        """
        _handleFailedJob_

        Called when a job has used up it's quota of retries and failed every
        time.  The job handler's handleFailedJob() method is called and the
        job's status is updated to "Failed".
        """
        jobID = self.determineJobID(jobSpecID)
        if jobID == None:
            return

        myThread = threading.currentThread()
        myThread.transaction.begin()        
        
        #wmbsJob = Job(id = jobID)
        #self.stateChanger.propagate([wmbsJob], "complete", "executing")
        
        myThread.transaction.commit()
        return

    def pollForCleanout(self):
        """
        _pollForCleanout_

        """
        logging.debug("pollForCleanout: Running...")
        myThread = threading.currentThread()
        myThread.transaction.begin()
        getJobsAction = self.daoFactory(classname = "Jobs.GetFWJRByState")
        jobIDs = getJobsAction.execute(state = "success", conn = myThread.transaction.conn,
                                       transaction = True)

        for jobID in jobIDs:
            wmbsJob = Job(id = jobID["id"])
            wmbsJob.load()
            logging.debug("Cleaning out %s" % wmbsJob["name"])            
            self.markJobForCleanup(wmbsJob["name"])
            self.stateChanger.propagate([wmbsJob], "cleanout", "success")            

        myThread.transaction.commit()
        self.ms.publishUnique("Tier1Accountant:CleanoutPoll", "", "00:01:00")
        self.ms.commit()
        return
    
    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST, create an instance
        of the trigger class and enter the message loop.
        """
        logging.debug("accountant: begin component start")
        
        self.ms.subscribeTo("Tier1Accountant:StartDebug")
        self.ms.subscribeTo("Tier1Accountant:EndDebug")
        self.ms.subscribeTo("Tier1Accountant:CleanoutPoll")
        self.ms.subscribeTo("SetJobCleanupFlag")
        self.ms.subscribeTo("JobSuccess")
        self.ms.subscribeTo("JobFailed")
        self.ms.subscribeTo("GeneralJobFailure")
        self.ms.commit()

        self.ms.publishUnique("Tier1Accountant:CleanoutPoll", "", "00:01:00")
        self.ms.commit()

        self.connectTier1DB()

        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)

        self.trigger = Trigger.Trigger(self.ms)

        # wait for messages
        self.messageLoop()
