#!/usr/bin/env python
"""
_Tier1InjectorComponent_

Poll a WMBS instance for subscriptions that have unacquired files and attempt
to create jobs for them.
"""

__revision__ = "$Id: Tier1InjectorComponent.py,v 1.29 2010/07/14 13:34:53 hufnagel Exp $"
__version__ = "$Revision: 1.29 $"

import logging
import os
import shutil
import random
import time
import urllib
import threading

from JobQueue import JobQueueDB
from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec
from ProdAgentCore.Configuration import loadProdAgentConfiguration

from WMCore.DAOFactory import DAOFactory

from WMCore.Database import DBFactory
from WMCore.Database import DBFormatter
from WMCore.Database import Transaction

from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow

from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine import DefaultConfig

from T0.GenericTier0.GenericTier0Component import GenericTier0Component
from T0.JobFactory.ProcessingJob import ProcessingJob
from T0.JobFactory.HarvestingJob import HarvestingJob
from T0.JobFactory.MergeJob import MergeJob

class Tier1InjectorComponent(GenericTier0Component):
    """
    _Tier1InjectorComponent_

    Poll a WMBS instance for subscriptions that have unacquired files and
    attempt to create jobs for them.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)

        # Data structure to hold WorkflowSpecs.  This is a two level
        # dictionary where the key to the first level is the run number,
        # and the second level is dataset id.
        self.workflows = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2
        
        # Lookup some parameters from the config file:
        #   scramArch - SCRAM Architecture
        #   cmsPath - Path to CMS software
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")

        # The job cache directory is where all workflows, jobspecs and configs
        # are stored.
        self.jobCacheDir = self.lookupParameter("JobCacheDir",
                                                self.args["ComponentDir"])

        # WMCore splitter factory for splitting up jobs.
        self.splitterFactory = SplitterFactory()

        self.workflows = {}
        self.daoFactory = None

        # See if lazy downloads are enabled.  
        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        self.minSubID = 0

        logging.info("Tier1 Injector Component Started")
        return

    def connectLocalWMBS(self):
        """
        _connectLocalWMBS_

        Connect to the local WMBS instance.
        """
        paConfig = loadProdAgentConfiguration()
        dbConfig = paConfig.getConfig("LocalWMBSDB")

        options = {}

        if dbConfig.has_key("unix_socket"):
            options["unix_socket"] = dbConfig["unix_socket"]
            
        dbFactory = DBFactory.DBFactory(logging, dbConfig["database"], options)
        dbInterface = dbFactory.connect()
        dbFormatter = DBFormatter.DBFormatter(logging, dbInterface)

        myThread = threading.currentThread()            
        myThread.dbi = dbInterface
        myThread.transaction = Transaction.Transaction(dbInterface)
        myThread.logger = logging
        myThread.dialect = dbConfig["dialect"]

        self.daoFactory = DAOFactory(package = "WMCore.WMBS", logger = logging,
                                     dbinterface = myThread.dbi)
        return
        
    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "Tier1Injector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "Tier1Injector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "Tier1Injector:pollSubscriptions":
            self.pollSubscriptions()
        elif event == "Tier1Injector:pollForClosable":
            self.pollForClosable()
        else:
            logging.error("Unknown event: %s, payload: %s" % (event, payload))

        return

    def retrieveJobSplitParams(self, subscriptionID):
        """
        _retrieveJobSplitParams_

        Retrieve job splitting parameters from the workflow.  The way this is
        setup currently sucks, we have to know all the job splitting parameters
        up front.  The following are currently supported:
          files_per_job
          min_merge_size
          max_merge_size
          max_merge_events
          max_merge_files
          job_period
        """
        newWorkflow = self.retrieveWorkflow(subscriptionID)

        jobSplitParams = {}

        if newWorkflow.parameters.has_key("files_per_job"):
            jobSplitParams["files_per_job"] = int(newWorkflow.parameters["files_per_job"])
        if newWorkflow.parameters.has_key("min_merge_size"):
            jobSplitParams["min_merge_size"] = int(newWorkflow.parameters["min_merge_size"])
        if newWorkflow.parameters.has_key("max_merge_size"):
            jobSplitParams["max_merge_size"] = int(newWorkflow.parameters["max_merge_size"])
        if newWorkflow.parameters.has_key("max_merge_events"):
            jobSplitParams["max_merge_events"] = int(newWorkflow.parameters["max_merge_events"])
        if newWorkflow.parameters.has_key("max_merge_files"):
            jobSplitParams["max_merge_files"] = int(newWorkflow.parameters["max_merge_files"])            
        if newWorkflow.parameters.has_key("job_period"):
            jobSplitParams["job_period"] = int(newWorkflow.parameters["job_period"])

        return jobSplitParams
    
    def retrieveWorkflow(self, subscriptionID):
        """
        _retrieveWorkflow_

        Retrive a workflow.
        """
        if subscriptionID in self.workflows.keys():
            return self.workflows[subscriptionID]
        
        wmbsSubscription = Subscription(id = subscriptionID)
        wmbsSubscription.load()

        wmbsWorkflow = wmbsSubscription["workflow"]
        wmbsWorkflow.load()

        workflowURL = wmbsWorkflow.spec
        workflowName = wmbsWorkflow.name

        logging.debug("CRASHDEBUG: %s, %s" % (subscriptionID, workflowName))

        (wfType, runNumber, junk) = workflowName.split("-", 2)
        workflowDir = "%s/workflows/%s" % (self.jobCacheDir, runNumber)
        workflowAbsPath = "%s/%s-workflow.xml" % (workflowDir, workflowName)

        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)

        if not os.path.exists(workflowAbsPath):
            urlFileHandle = urllib.urlopen(workflowURL)
            cacheFileHandle = open(workflowAbsPath, "w")
            cacheFileHandle.write(urlFileHandle.read())
            cacheFileHandle.close()
            urlFileHandle.close()

        workflowSpec = WorkflowSpec()
        workflowSpec.load(workflowAbsPath)

        self.workflows[subscriptionID] = workflowSpec
        return workflowSpec
        
    def createAndSubmitJob(self, wmbsJobGroup, wmbsSubscription):
        """
        _createAndSubmitJob_

        Create and submit jobspecs for each job in the given job group.  The
        job type will be determined from the subscription.
        """
        newWorkflow = self.retrieveWorkflow(wmbsSubscription["id"])
        if newWorkflow == None:
            logging.error("createAndSubmitJob(): Workflow does not exist.")
            return

        if wmbsSubscription["type"] == "Processing":
            jobGenerator = ProcessingJob(newWorkflow)
        elif wmbsSubscription["type"] == "Harvesting":
            jobGenerator = HarvestingJob(newWorkflow)
        elif wmbsSubscription["type"] == "Merge":
            jobGenerator = MergeJob(newWorkflow)
        else:
            logging.error("createAndSubmitJob(): Unknown job type: %s" % wmbsSubscription["type"])
            return
        
        jobSpecDirKey = str(random.randrange(0, 1000, 1)).zfill(3)        

        self.paDBConn.beginTransaction()
        for wmbsJob in wmbsJobGroup.jobs:
            wmbsFiles = wmbsJob.getFiles()

            wmbsFiles[0].loadData()
            runNumber = list(wmbsFiles[0]["runs"])[0].run
            
            fileLFNs = []
            parentLFNs = []
            for wmbsFile in wmbsFiles:
                fileLFNs.append(wmbsFile["lfn"])

                for parentFile in wmbsFile["parents"]:
                    parentLFNs.append(parentFile["lfn"])
            
            jobName = "%s-Run%s-%s-%s" % (newWorkflow.parameters["WorkflowClass"],
                                          runNumber,
                                          newWorkflow.parameters["WorkflowDescription"],
                                          wmbsJob["id"])

            wmbsJob["name"] = jobName
            wmbsJob.save()

            if wmbsJob["mask"]["LastEvent"] != None and \
                   wmbsJob["mask"]["FirstEvent"] != None:
                maxEvents = wmbsJob["mask"]["LastEvent"] - wmbsJob["mask"]["FirstEvent"]
                skipEvents = wmbsJob["mask"]["FirstEvent"]
            else:
                maxEvents = -1
                skipEvents = 0

            jobSpec = jobGenerator(jobName, fileLFNs, parentLFNs, maxEvents, skipEvents)
            jobSpec.parameters["RunNumber"] = runNumber
            jobSpecFileName = jobSpec.parameters["JobName"] + "-jobspec.xml"
            jobSpecDir = os.path.join(self.jobCacheDir, "jobspecs", 
                                      "Run%d" % runNumber, jobSpecDirKey)
            if not os.path.isdir(jobSpecDir):
                os.makedirs(jobSpecDir)
            
            jobSpecLocation = os.path.join(jobSpecDir, jobSpecFileName)
            jobSpec.save(jobSpecLocation)

            JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                     self.paDBConn.dbTransaction.conn,
                                     jobSpec.parameters["JobName"],
                                     jobSpecLocation,
                                     wmbsSubscription["type"],
                                     jobSpec.payload.workflow,
                                     self.jobQueuePriority,
                                     newWorkflow.parameters["OnlySites"], "new")

            self.stateChanger.propagate([wmbsJob], "created", "new")
            self.stateChanger.propagate([wmbsJob], "executing", "created")

        self.paDBConn.commit()
        return

    def pollSubscriptions(self):
        """
        _pollSubscriptions_

        Retrieve a list of all subscriptions from WMBS and attempt to run the
        job splitting code on them.  Create and submit any jobs that are
        created.
        """
        subscriptionList = self.daoFactory(classname = "Subscriptions.ListIncomplete")
        subscriptions = subscriptionList.execute(minSub = self.minSubID)

        for subscription in subscriptions:
            myThread = threading.currentThread()
            myThread.transaction.begin()
                                    
            wmbsSubscription = Subscription(id = subscription)
            wmbsSubscription.load()

            wmbsJobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                                  subscription = wmbsSubscription)

            splitParams = self.retrieveJobSplitParams(subscription)
            wmbsJobGroups = wmbsJobFactory(**splitParams)
            for wmbsJobGroup in wmbsJobGroups:
                self.createAndSubmitJob(wmbsJobGroup, wmbsSubscription)

            myThread.transaction.commit()

        abandonedGroupsAction = self.daoFactory(classname = "JobGroup.GetGroupsByJobState")
        abandonedGroups = abandonedGroupsAction.execute(jobState = "new", minSub = self.minSubID)

        for abandonedGroup in abandonedGroups:
            myThread = threading.currentThread()
            myThread.transaction.begin()

            wmbsJobGroup = JobGroup(id = abandonedGroup)
            wmbsJobGroup.loadData()
            self.createAndSubmitJob(wmbsJobGroup, wmbsJobGroup.subscription)

            myThread.transaction.commit()

        self.ms.publishUnique("Tier1Injector:pollSubscriptions",
                              "", "00:01:00")
        self.ms.commit()
        return

    def pollForClosable(self):
        """
        _pollForClosable_

        Search WMBS for filesets that can be closed and mark them as closed.
        """
        myThread = threading.currentThread()
        myThread.transaction.begin()

        closableFilesetDAO = self.daoFactory(classname = "Fileset.ListClosable")
        closableFilesets = closableFilesetDAO.execute()

        for closableFileset in closableFilesets:
            openFileset = Fileset(id = closableFileset)
            openFileset.load()

            logging.debug("Closing fileset %s" % openFileset.name)
            openFileset.markOpen(False)

        myThread.transaction.commit()
        
        self.ms.publishUnique("Tier1Injector:pollForClosable", 
                              "", "00:01:00")        
        self.ms.commit()
        return

    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("Tier1Injector:StartDebug")
        self.ms.subscribeTo("Tier1Injector:EndDebug")
        self.ms.subscribeTo("Tier1Injector:pollSubscriptions")
        self.ms.subscribeTo("Tier1Injector:pollForClosable")        

        self.connectLocalWMBS()

        DefaultConfig.config.JobStateMachine.couchurl = self.couchURL
        self.stateChanger = ChangeState(DefaultConfig.config, self.couchDB)

        self.connectPADB()

        self.ms.publishUnique("Tier1Injector:pollSubscriptions", 
                              "", "00:01:00")
        self.ms.publishUnique("Tier1Injector:pollForClosable", 
                              "", "00:01:00")        
        self.ms.commit()

        if self.lookupParameter("Tier0Mode", "Off") == "On":
            from T0 import Globals
            self.minSubID = Globals.getMaxSubscriptionID() + 1
            logging.debug("Minimum subscription ID: %s" % self.minSubID)

        # Poll for messages...
        self.messageLoop()
