#!/usr/bin/env python
"""
_SkimWFCache_

Cache for skimming workflows.
"""

__revision__ = "$Id: SkimWFCache.py,v 1.8 2010/02/11 20:37:03 sfoulkes Exp $"
__version__ = "$Revision: 1.8 $"

import logging
import os
import sys
import threading
import time

from T0.State.Database.Reader import ListDatasets

from T0.WorkflowFactory.SkimmingWorkflow import SkimmingWorkflow
from T0.WorkflowFactory.WorkflowCacheInterface import WorkflowCacheInterface

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec

class SkimWFCache(WorkflowCacheInterface):
    """
    _SkimWFCache_

    Generate and store skimming workflows.
    """
    def createWorkflow(self, runConfig, params):
        """
        _createWorkflow_

        Create a workflow for a particular run, input dataset and skimming
        config.  This expects the following parameters to be set:
          CMSSW_VERSION - CMSSW version string representing the framework
                          version to be used with the job.                
          PROCESSION_VERSION - Processing version, will be used in the processed
                               dataset name and output LFN.
          INPUT_DATASET_ID - Input dataset path id, will be used to determine
                             input dataset names.
          CONFIG_FILE - Python framework config for this skim.
          GLOBAL_TAG - Global tag for the skim.
          LAZY_DOWNLOAD - Bool, determines whether or not lazy downloads will be
                          enabled for this job.
          SKIM_NAME - The name of the skim.
          SITE_NAME - The site to run the skim at.
          FILER_PER_JOB - The maximum number of files in a single skim job.
          WF_CLASS - The workflow class, used to construct job names.
        """
        runNumber = runConfig.getRunNumber()
        
        inputDatasetNames = ListDatasets.listDatasetNamesFromID(self.t0astDBConn,
                                                                params["INPUT_DATASET_ID"])
        acquisitionEra = runConfig.getAcquisitionEra()

        pid = os.fork()
        if pid != 0:
            os.wait()

            workflowName = "Skim-Run%s-%s-%s-%s-%s" % (runNumber, params["SKIM_NAME"],
                                                       inputDatasetNames["PRIMARY"],
                                                       inputDatasetNames["PROCESSED"],
                                                       inputDatasetNames["TIER"])
            workflowPath = os.path.join(self.workflowCacheDir, "Run%d" % runNumber,
                                        "%s-workflow.xml" % workflowName)

            if not os.path.exists(workflowPath):
                logging.debug("No workflow for run, dataset, skim: %s, %s, %s" % (runNumber,
                                                                                  inputDatasetNames["PRIMARY"],
                                                                                  params["SKIM_NAME"]))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowPath)
            return workflowSpec

        newWorkflow = SkimmingWorkflow(runNumber, params["CMSSW_VERSION"],
                                       self["cmsPath"], self["scramArch"])
        newWorkflow.setParentDataset(inputDatasetNames["PRIMARY"],
                                     inputDatasetNames["PROCESSED"],
                                     inputDatasetNames["TIER"])

        newWorkflow.setSkimName(params["SKIM_NAME"])
        newWorkflow.setAcquisitionEra(acquisitionEra)
        newWorkflow.setProcessingVersion(params["PROCESSING_VERSION"])
        newWorkflow.setConfigFile(params["CONFIG_FILE"])
        newWorkflow.setGlobalTag(params["GLOBAL_TAG"])
        newWorkflow.setLazyDownload(params["LAZY_DOWNLOAD"])
        newWorkflow.setWorkflowSites(params["SITE_NAME"])
        newWorkflow.setJobSplitParams({"files_per_job": params["FILES_PER_JOB"]})
        newWorkflow.setWorkflowClass(params["WF_CLASS"])
        newWorkflow.setWorkflowDescription("%s-%s" % (params["SKIM_NAME"],
                                                      inputDatasetNames["PRIMARY"]))

        try:
            workflowSpec = newWorkflow.makeWorkflow()
            self.saveWorkflow(workflowSpec)
        finally:
            os._exit(0)

    def generateKey(self, runConfig, params):
        """
        _generateKey_

        Generate a unique key for Skimming workflows.  Skimming workflows
        can be uniquely identified by run number, input dataset and their config
        file.  The key will take the following form:
          RunNumber-InputDatasetPathID-ConfigFileName.
        """
        key = "%s-%s-%s" % (runConfig.getRunNumber(), params["INPUT_DATASET_ID"],
                            params["CONFIG_FILE"])
        return key
