#!/usr/bin/env python
"""
_Tier1SchedulerComponent_

Poll T0AST for blocks that have been migrated to DBS and then generate skimming
jobs to run at the Tier1s for any skims that have been configured.
"""

__revision__ = "$Id: Tier1SchedulerComponent.py,v 1.34 2010/04/15 18:42:08 sfoulkes Exp $"
__version__ = "$Revision: 1.34 $"

import logging
import threading
import sys

from T0.State.Database.Reader import ListBlock
from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListFiles
from T0.State.Database.Reader import ListRuns

from T0.State.Database.Writer import InsertBlock
from T0.State.Database.Writer import InsertDataset

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow

from WMCore.DataStructs.Run import Run
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.PhEDEx.PhEDEx import PhEDEx
from WMCore.Services.UUID import makeUUID

from WMCore.Database import DBFactory
from WMCore.Database import DBFormatter
from WMCore.Database import Transaction

from T0.Tier0Merger.MergeWFCache import MergeWFCache
from T0.Tier1Scheduler.SkimWFCache import SkimWFCache

from ProdAgentCore.Configuration import loadProdAgentConfiguration
from MessageService.MessageService import MessageService
from T0.GenericTier0.GenericTier0Component import GenericTier0Component

class Tier1SchedulerComponent(GenericTier0Component):
    """
    _Tier1SchedulerComponent_

    Poll T0AST for blocks that have been migrated to DBS and then generate
    skimming jobs to run at the Tier1s for any skims that have been configured.
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("Tier1Scheduler Component Started")

        self.blockPollInterval = self.lookupParameter("blockPollInterval",
                                                      "00:01:00")

        # Absolute path to the location where workflows should be placed.  This
        # should be the place the webserver looks when someone or something
        # accesses the workflow cache URL.
        self.workflowCacheDir = self.lookupParameter("WorkflowCacheDir", None)

        # URL to the workflow cache.  This should correspond to the workflow
        # cache directory.
        self.workflowCacheURL = self.lookupParameter("WorkflowCacheURL", None)

        # Scram arch and path to cmssw needed to generate workflows.
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")

        # Containers for merge and processing workflows.
        self.mergeWFCache = None
        self.skimWFCache = None

        # Job splitting parameters
        self.filesPerJob = self.lookupParameter("filesPerJob", 1)
        self.minMergeSize = self.lookupParameter("minMergeSize", 2000000000)
        self.maxMergeEvents = self.lookupParameter("maxMergeEvents", 100000)
        self.maxMergeSize = self.lookupParameter("maxMergeSize", 4100000000)
        self.maxMergeFiles = self.lookupParameter("maxMergeFiles", 10)

        # DAO factory used for adding locations to the Tier1 WMBS.
        self.daoFactory = None

        phedex = PhEDEx({"endpoint": self.lookupParameter("phedexURL", None)}, "json")
        self.nodeMap = phedex.getNodeMap()
        self.reconnect = False
        return

    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        try:
            if event == "Tier1Scheduler:StartDebug":
                logging.getLogger().setLevel(logging.DEBUG)
            elif event == "Tier1Scheduler:EndDebug":
                logging.getLogger().setLevel(logging.INFO)
            elif event == "Tier1Scheduler:poll":
                self.pollForTransferedBlocks()
            elif event == "Tier1Scheduler:pollForRunComplete":
                self.pollForRunComplete()
            else:
                logging.debug("Unknown event: %s" % event)

            self.reconnect = False
        except Exception, ex:
            if self.reconnect == True:
                raise
            if "orig" in dir(ex):
                if ex.orig[0] == 2006 and ex.orig[1] == "MySQL server has gone away":
                    logging.debug("Tier1 database connection died, restarting...")
                    self.reconnect = True

                    if self.t0astDBConn:
                        self.t0astDBConn.rollback()
                        
                    self.connectTier1DB()

            self.ms.publishUnique(event, "", "00:01:00")
            self.ms.commit()
            
        return

    def connectTier1DB(self):
        """
        _connectTier1DB_

        Connect to the Tier1 WMBS database.  This will pull the database
        configuration out of the "LocalWMBSDB" section of the ProdAgent config.
        """
        paConfig = loadProdAgentConfiguration()
        dbConfig = paConfig.getConfig("LocalWMBSDB")

        options = {}

        if dbConfig.has_key("unix_socket"):
            options["unix_socket"] = dbConfig["unix_socket"]
            
        dbFactory = DBFactory.DBFactory(logging, dbConfig["database"], options)
        self.dbInterface = dbFactory.connect()
        self.dbFormatter = DBFormatter.DBFormatter(logging, self.dbInterface)

        myThread = threading.currentThread()            
        myThread.dbi = self.dbInterface
        myThread.logger = logging
        myThread.transaction = Transaction.Transaction(myThread.dbi)
        myThread.dialect = dbConfig["dialect"]

        self.daoFactory = DAOFactory(package = "WMCore.WMBS", logger = logging,
                                     dbinterface = myThread.dbi)
        return

    def inputFilesetName(self, blockInfo):
        """
        _inputFilesetName_

        Generate a name for the fileset that will be used to hold files from
        a freshly transfered block.  The name will take the following form:
          T0-RunNNN-PrimaryDatasetName-ProcessedDatasetName-DataTierName
        """
        filesetName = "T0-Run%s-%s-%s-%s" % (blockInfo["RUN_ID"],
                                             blockInfo.getPrimaryDatasetName(),
                                             blockInfo.getProcessedDatasetName(),
                                             blockInfo.getDataTier())
        return filesetName

    def mergeFilesetName(self, runNumber, primaryDatasetName,
                         processedDatasetName, dataTierName):
        """
        _mergeFilesetName_

        Generate a name for the fileset that will be used as the input for a
        merge workflow.  The name will take the following form:
          Merge-RunNNN-PrimaryDatasetName-ProcessedDatasetName-DataTierName
        """
        mergeFilesetName = "Merge-Run%s-%s-%s-%s" % (runNumber,
                                                     primaryDatasetName,
                                                     processedDatasetName,
                                                     dataTierName)
        return mergeFilesetName

    def createOutputDataset(self, primaryDatasetID, processedDatasetName,
                            dataTierName):
        """
        _createOutputDataset_

        Create a dataset into T0AST for the output of a skim job.  The output
        of a skim job will have the same primary and processed dataset as the
        input but will have a different processed dataset.  This method will add
        the new processed dataset and then create a new dataset path for it.
        The new dataset path ID will be returned.
        """
        InsertDataset.insertProcessedDataset(self.t0astDBConn,
                                             processedDatasetName)
        InsertDataset.insertDataTier(self.t0astDBConn,
                                     dataTierName)
        processedDatasetID = ListDatasets.listProcessedDatasetID(self.t0astDBConn,
                                                                 processedDatasetName)
        dataTierID = ListDatasets.listDataTierID(self.t0astDBConn,
                                                 dataTierName)
        InsertDataset.insertDataset(self.t0astDBConn, primaryDatasetID,
                                    processedDatasetID, dataTierID)
        datasetPathID = ListDatasets.listDatasetPathFromIDs(self.t0astDBConn,
                                                            primaryDatasetID,
                                                            processedDatasetID,
                                                            dataTierID)

        return datasetPathID

    def insertFileParentsIntoTier1WMBS(self, childFile):
        """
        _insertFileParentsIntoTier1WMBS_

        For a given file, insert it's parents into the Tier1 WMBS.  This is used
        for workflows that do two file reads.  We don't need to bother inserting
        any meta data because we only need the parent LFN to create the
        jobspec.
        """        
        parentLFNs = ListFiles.listParentLFNs(self.t0astDBConn, childFile["lfn"])

        for parentLFN in parentLFNs:
            parentFile = File(lfn = parentLFN)
            parentFile.create()

            childFile.addParent(parentLFN)

        return

    def insertFilesIntoTier1WMBS(self, blockInfo, blockLocation,
                                 insertParents = False):
        """
        _insertFilesIntoTier1WMBS_

        Insert all the files from the given block into the Tier1 WMBS.  Insert
        the newly added files with the given fileset.  Also add run and lumi
        information for each file that was added.
        """
        inputFilesetName = self.inputFilesetName(blockInfo)

        locationNew = self.daoFactory(classname = "Locations.New")
        locationNew.execute(siteName = blockLocation)

        newFileset = Fileset(name = inputFilesetName, is_open = True)
        newFileset.create()

        blockFiles = ListFiles.listBlockFilesForSkim(self.t0astDBConn,
                                                     blockInfo["BLOCK_ID"])

        logging.debug("insertFilesIntoTier1WMBS(): block %s" % blockInfo["BLOCK_ID"])

        for file in blockFiles:
            logging.debug("insertFilesIntoTier1WMBS(): file %s" % file["LFN"])
            newFile = File(lfn = file["LFN"], size = file["SIZE"],
                           events = file["EVENTS"])

            lumiNumbers = ListFiles.listLumiInfoForFile(self.t0astDBConn,
                                                        file["LFN"])
            runInfo = Run(runNumber = blockInfo["RUN_ID"])
            runInfo.extend(lumiNumbers)

            newFile.addRun(runInfo)
            newFile.setLocation(blockLocation, immediateSave = False)
            newFile.create()
            if insertParents:
                self.insertFileParentsIntoTier1WMBS(newFile)
            
            newFileset.addFile(newFile)

        newFileset.commit()
        return

    def createMergeWorkflowForDataset(self, runNumber, datasetPathID,
                                      inputFilesetName, cmsswVersion,
                                      processingVersion, blockLocation):
        """
        _createMergeWorkflowForDataset_

        Generate a merge workflow for the given run number and dataset path
        ID.  Create a workflow, input fileset and subscription for merging in
        the Tier1 WMBS instance.  All of the output of the merge workflows
        will be logged into DBS.
        """
        logging.debug("MERGE FILESET: %s" % inputFilesetName)
        inputFileset = Fileset(name = inputFilesetName, is_open = True)
        inputFileset.create()
        outputFileset = Fileset(name = makeUUID(), is_open = True)
        outputFileset.create()
        
        runConfig = self.getRunConfig(runNumber)
        mergeWorkflowSpec = self.mergeWFCache.retrieveWorkflow(runConfig,
                                                               WF_CLASS = "T1Merge",
                                                               MIN_MERGE_SIZE = self.minMergeSize,
                                                               MAX_MERGE_EVENTS = self.maxMergeEvents,
                                                               MAX_MERGE_SIZE = self.maxMergeSize,
                                                               MAX_MERGE_FILES = self.maxMergeFiles,
                                                               CMSSW_VERSION = cmsswVersion,
                                                               INPUT_DATASET_ID = datasetPathID,
                                                               PROCESSING_VERSION = processingVersion,
                                                               SITE_NAME = blockLocation)
        workflowURL = "%s/Run%s/%s-workflow.xml" % \
                      (self.workflowCacheURL, runNumber,
                       mergeWorkflowSpec.workflowName())
        
        wmbsMergeWorkflow = Workflow(spec = workflowURL, name = inputFilesetName,
                                     owner = "Tier0", task = "BOGUS")
        wmbsMergeWorkflow.create()
        wmbsMergeWorkflow.addOutput("Merged", outputFileset)

        mergeSubscription = Subscription(fileset = inputFileset,
                                         workflow = wmbsMergeWorkflow,
                                         split_algo = "WMBSMergeBySize",
                                         type = "Merge")
        mergeSubscription.create()
        return inputFileset

    def createWorkflowsForBlock(self, acquisitionEra, skimConfig, blockInfo,
                                blockLocation):
        """
        _createWorkflowsForBlock_

        Create the skimming and merge workflows for a given block and a given
        skim configuration.  This method will generate the processed dataset
        name for the skim so that a dataset path for the output of the skim can
        be added to T0AST.  It will then create a merge workflow and a
        processing workflow such that the output of the processing workflow is
        injected into the merge workflow and the output of the merge workflow
        is added to DBS.
        """
        runConfig = self.getRunConfig(blockInfo["RUN_ID"])
        configFile = runConfig.retrieveConfigFromURL(skimConfig.ConfigURL)
        skimWorkflowSpec = self.skimWFCache.retrieveWorkflow(runConfig,
                              WF_CLASS = "T0Skim",
                              CMSSW_VERSION = skimConfig.CMSSWVersion,
                              PROCESSING_VERSION = skimConfig.ProcessingVersion,
                              INPUT_DATASET_ID = blockInfo["DATASET_PATH_ID"],
                              CONFIG_FILE = configFile,
                              GLOBAL_TAG = skimConfig.GlobalTag,
                              LAZY_DOWNLOAD = True,
                              SKIM_NAME = skimConfig.SkimName,
                              SITE_NAME = blockLocation,
                              FILES_PER_JOB = self.filesPerJob)

        inputFilesetName = self.inputFilesetName(blockInfo)
        inputFileset = Fileset(name = inputFilesetName, is_open = True)
        inputFileset.create()

        workflowURL = "%s/Run%s/%s-workflow.xml" % \
                      (self.workflowCacheURL, blockInfo["RUN_ID"],
                       skimWorkflowSpec.workflowName())
        wmbsSkimWorkflow = Workflow(spec = workflowURL,
                                    name = skimWorkflowSpec.workflowName(),
                                    owner = "Tier0", task = "BOGUS")
        wmbsSkimWorkflow.create()

        primaryDatasetName = blockInfo.getPrimaryDatasetName()

        logging.debug("OUTMOD KEYS: %s" % skimWorkflowSpec.payload.cfgInterface.outputModules.keys())
        for outputModuleName in skimWorkflowSpec.payload.cfgInterface.outputModules.keys():
            outputModule = skimWorkflowSpec.payload.cfgInterface.outputModules[outputModuleName]
            processedDatasetName = outputModule["processedDataset"]
            dataTierName = outputModule["dataTier"]

            logging.debug("OUTPUT DTIER: %s" % outputModule)
            outputDatasetPathID = self.createOutputDataset(blockInfo["PRIMARY_ID"],
                                                           processedDatasetName,
                                                           dataTierName)

            mergeFilesetName = self.mergeFilesetName(blockInfo["RUN_ID"],
                                                     primaryDatasetName,
                                                     processedDatasetName,
                                                     dataTierName)

            mergeFileset = self.createMergeWorkflowForDataset(blockInfo["RUN_ID"],
                                                              outputDatasetPathID,
                                                              mergeFilesetName,
                                                              skimConfig.CMSSWVersion,
                                                              skimConfig.ProcessingVersion,
                                                              blockLocation)

            wmbsSkimWorkflow.addOutput(outputModuleName, mergeFileset,
                                       outputModule.get("parentOutputModule", None))

        if skimConfig.TwoFileRead:
            splitAlgo = "TwoFileBased"
        else:
            splitAlgo = "FileBased"
            
        skimSubscription = Subscription(fileset = inputFileset,
                                        workflow = wmbsSkimWorkflow,
                                        split_algo = splitAlgo,
                                        type = "Processing")
        skimSubscription.create()
        return

    def pollForTransferedBlocks(self):
        """
        _pollForTransferedBlocks_

        Poll T0AST for any blocks that have been migrated to DBS and generate
        skims for them.  Mark the blocks as "Skimmed" once any skims have been
        injected into the Tier1 WMBS.
        """
        logging.debug("pollForTransferedBlocks(): Running...")
        
        skimmableBlocks = ListBlock.listBlockInfoByStatus(self.t0astDBConn,
                                                          "Exported", "Migrated")

        logging.debug("pollForTransferedBlocks(): Found %s blocks." % len(skimmableBlocks))

        for skimmableBlock in skimmableBlocks:
            logging.debug("pollForTransferedBlocks(): Skimmable: %s" % skimmableBlock["BLOCK_ID"])
            runConfig = self.getRunConfig(int(skimmableBlock["RUN_ID"]))
            
            skims = runConfig.getSkimConfiguration(skimmableBlock["PRIMARY_ID"],
                                                   skimmableBlock["TIER_ID"])

            if skims == None:
                InsertBlock.updateBlockStatusByID(self.t0astDBConn,
                                                  skimmableBlock, "Skimmed")
                self.t0astDBConn.commit()
                continue

            insertParents = False
            for skimConfig in skims:
                if skimConfig.TwoFileRead:
                    insertParents = True
                    break

            if insertParents:
                if not ListBlock.isParentBlockExported(self.t0astDBConn, skimmableBlock["BLOCK_ID"]):
                    logging.debug("Block %s has unexported parents." % skimmableBlock["BLOCK_ID"])
                    continue

            blockLocation = skimmableBlock["STORAGE_NODE"]
            if skimmableBlock["CUSTODIAL"] != 1:
                continue

            myThread = threading.currentThread()
            myThread.transaction.begin()

            for phedexNode in self.nodeMap["phedex"]["node"]:
                if phedexNode["name"] == blockLocation:
                    blockSEName = str(phedexNode["se"])

            logging.debug("LOCATION: %s, %s" % (blockLocation, blockSEName))
            self.insertFilesIntoTier1WMBS(skimmableBlock, blockSEName,
                                          insertParents)

            for skimConfig in skims:
                try:
                    self.createWorkflowsForBlock(runConfig.getAcquisitionEra(),
                                                 skimConfig, skimmableBlock,
                                                 blockSEName)
                except Exception, ex:
                    logging.debug("Error making workflows: %s" % str(ex))
                    self.t0astDBConn.rollback()
                    myThread.transaction.rollback()
                    break
            else:
                InsertBlock.updateBlockStatusByID(self.t0astDBConn, skimmableBlock,
                                                  "Skimmed")
                self.t0astDBConn.commit()
                myThread.transaction.commit()

        self.ms.publishUnique("Tier1Scheduler:poll", "", self.blockPollInterval)
        return

    def pollForRunComplete(self):
        """
        _pollForRunComplete_

        Query the Tier1 WMBS for any open filesets that start with "T0".  These
        filesets contain files that are injected from the Tier0 into the Tier1
        processing system.  Given the list of open T0 filesets determine what
        run they belong to and mark them as closed if the T0 is done processing
        data for that run.
        """
        myThread = threading.currentThread()
        myThread.transaction.begin()

        openFilesetDAO = self.daoFactory(classname = "Fileset.ListOpen")
        openFilesetNames = openFilesetDAO.execute()

        openInputFilesets = []
        openRuns = []
        for openFilesetName in openFilesetNames:
            (type, run, dataset) = openFilesetName.split("-", 2)
            if type == "T0":
                openInputFilesets.append(openFilesetName)
                if run[3:] not in openRuns:
                    openRuns.append(run[3:])

        if len(openRuns) == 0:
            self.ms.publishUnique("Tier1Scheduler:pollForRunComplete",
                                  "", "00:01:00")        
            return
        
        runStatuses = ListRuns.listRunState(self.t0astDBConn, openRuns)
        blockCount = ListBlock.countUnExportedBlocksByRun(self.t0astDBConn,
                                                         openRuns)

        for openInputFileset in openInputFilesets:
            (type, run, dataset) = openInputFileset.split("-", 2)
            datasetParts = dataset.split("-")
            primary = datasetParts[0]
            tier = datasetParts[-1]
            
            runNumber = int(run[3:])
            runStatus = runStatuses[runNumber]

            if runStatus == "CloseOutExport":
                logging.debug("Checking %s for closeout..." % openInputFileset)
                if blockCount.has_key(runNumber):
                    if blockCount[runNumber].has_key(primary):
                        if blockCount[runNumber][primary].has_key(tier):
                            logging.debug("  %s has %s open blocks." % (openInputFileset,
                                                                        blockCount[runNumber][primary][tier]))
                            continue
            elif runStatus != "Complete":
                    continue

            logging.debug("Closing fileset: %s" % openInputFileset)
            wmbsFileset = Fileset(name = openInputFileset)
            wmbsFileset.markOpen(False)

        myThread.transaction.commit()

        self.ms.publishUnique("Tier1Scheduler:pollForRunComplete",
                              "", "00:01:00")

        return
    
    def startComponent(self):
        """
        _startComponent_

        Fire up the message service and define the list of
        messages that this component subscribes to
        """
        self.ms.subscribeTo("Tier1Scheduler:StartDebug")
        self.ms.subscribeTo("Tier1Scheduler:EndDebug")
        self.ms.subscribeTo("Tier1Scheduler:poll")
        self.ms.subscribeTo("Tier1Scheduler:pollForRunComplete")
        self.ms.commit()

        self.connectT0AST(manageGlobal = False)
        self.connectTier1DB()

        self.mergeWFCache = MergeWFCache(self.ms, self.t0astDBConn,
                                         self.workflowCacheDir,
                                         scramArch = self.scramArch,
                                         cmsPath = self.cmsPath,
                                         useLazyDownload = False)
        self.skimWFCache = SkimWFCache(self.ms, self.t0astDBConn,
                                       self.workflowCacheDir,
                                       scramArch = self.scramArch,
                                       cmsPath = self.cmsPath,
                                       useLazyDownload = False)        

        self.ms.publishUnique("Tier1Scheduler:poll",
                              "", "00:01:00")
        self.ms.publishUnique("Tier1Scheduler:pollForRunComplete",
                              "", "00:01:00")
        self.ms.commit()

        # Wait for messages.
        self.messageLoop()
