#!/usr/bin/env python
"""
_WMBSPublishComponent_

Poll T0AST for files in datasets and export to another WMBS instance
"""

__revision__ = "$Id: WMBSPublishComponent.py,v 1.4 2009/07/01 08:41:42 gowdy Exp $"
__version__ = "$Revision: 1.4 $"

import logging
from sets import Set

from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListFiles
from T0.State.Database.Reader import ListRuns

from T0.State.Database.Writer import InsertDataset

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow

from WMCore.DAOFactory import DAOFactory

from WMCore.Database import DBFactory
from WMCore.Database import DBFormatter
from WMCore.Database import Transaction

from ProdAgentCore.Configuration import loadProdAgentConfiguration
from MessageService.MessageService import MessageService
from T0.GenericTier0.GenericTier0Component import GenericTier0Component

class WMBSPublishComponent(GenericTier0Component):
    """
    _WMBSPublishComponent_

    Poll T0AST for files and export them to another DBS instance
    """
    def __init__(self, **args):
        """
        ___init___

        Call the GenericTier0Component constructor to setup logging
        as well as the connection to T0AST.  Also, verify that the
        required parameters are in the ProdAgent configuration file.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("WMBSPublish Component Started")

        # DAO factories used for WMBS expots
        self.daoFactories = {}

        # Load up the subscription for easy access.
        self.subscription = None

        return

    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        # always have a debug on/off switch
        if event == "WMBSPublish:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "WMBSPublish:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "WMBSPublish:poll":
            self.poll()
        else:
            logging.debug("Unknown event: %s" % event)
            
        return


    def poll(self):
        """
        _poll_

        Check each file again publish list for its run to see if it should
        be published
        """
        self.t0astDBConn.beginTransaction()

        files = self.subscription.availableFiles()
        logging.debug("AvailableFiles: %d" % len( files ) )
        for file in files:
            file.loadData()

            runNumber = list(file["runs"])[0].run
            id = file["id"]
            datasetID = ListDatasets.listDatasetIDForWMBSFile(self.t0astDBConn,
                                                              id )

            runConfig = self.getRunConfig( runNumber )
            if runConfig == None:
                logging.error("Could not retrieve run config for run %s" % \
                               runNumber)
                continue
            wmbsPublishConfig = runConfig.getWMBSPublishConfiguration( datasetID )
            if wmbsPublishConfig == None:
                logging.error("No wmbsPublish config for run %s, dataset %s" %\
                                (runNumber, datasetID) )
                self.subscription.completeFiles(file)
                continue
            elif wmbsPublishConfig.DoWMBSPublish == False:
                logging.debug("WMBSPublish disabled for dataset %s, file %s" %\
                              (datasetID, id))
                self.subscription.completeFiles(file)
                continue

            fileDataTier = ListDatasets.listDatasetIDsFromID( self.t0astDBConn,
                                                              datasetID )['TIER']
            exportFileTo = []
            for dataTierTo in wmbsPublishConfig.DataTiersTo:
                (dataTiersToMatch,databases) = dataTierTo.split('@')
                dataTiersToMatchList = dataTiersToMatch.split(',')
                dataTierIDs = []
                for dataTier in dataTiersToMatchList:
                    dataTierIDs.append( int( dataTier ) )
                if fileDataTier in dataTierIDs:
                    exportFileTo.extend( databases.split(',') )

            if len( exportFileTo ) > 0:
                self.subscription.acquireFiles(file)
                for database in exportFileTo:
                    self.exportFileTo( file, database )
            else:
                logging.debug("No WMBSPublish for dataset %s, file %s, tier %s" %\
                              (datasetID, id, fileDataTier ))

            self.subscription.completeFiles(file)

        self.t0astDBConn.commit()
        self.ms.publishUnique("WMBSPublish:poll",
                              "", "00:01:00")
        self.ms.commit()
        return

    def exportFileTo( self, file, database ):
        """
        _exportFileTo_

        Actually export the file to this other database
        """
        if database not in self.daoFactories:
            self.connectDB( database )
        db = self.daoFactories[database]


        # create file in new database and get it back again
        newFileCreator = db( classname = "Files.Add" )
        fileTuple = file.getInfo()
        logging.debug( "About to export: %s, local id: %d" % \
                       ( fileTuple[0], fileTuple[1] ) )
        newFileCreator.execute( [ fileTuple ] )
        #newFileRetriever = db( classname = "Files.GetByLFN" )
        #newFile = File()
        #newFile.update( newFileRetriever.execute( fileTuple[0] ) )

        # add run and lumi section information
        runs = Set( file.getRuns() )
        newRunLumi = db( classname = "Files.AddRunLumi" )
        logging.debug( "About to add run/lumisec information." )
        newRunLumi.execute( fileTuple[0], runs )

        return

    def connectDB(self, dbName):
        """
        _connectDB_

        @dbName Name for database connection parameters in ProdAgentConfig

        Connect to a WMBS database.  This will pull the database
        configuration out of the section of the ProdAgent config
        with same value as dbName.
        """
        paConfig = loadProdAgentConfiguration()
        dbConfig = paConfig.getConfig(dbName)

        options = {}

        if dbConfig.has_key("unix_socket"):
            options["unix_socket"] = dbConfig["unix_socket"]
            
        if dbConfig.has_key("dialect"):
            options["dialect"] = dbConfig["dialect"]
            
        if dbConfig.has_key("dbName"):
            options["database"] = dbConfig["dbName"]
            
        dbFactory = DBFactory.DBFactory(logging, options=options)
        dbInterface = dbFactory.connect()
        dbFormatter = DBFormatter.DBFormatter(logging, dbInterface)

        self.daoFactories[dbName] = DAOFactory( package = "WMCore.WMBS",
                                                logger = logging,
                                                dbinterface = dbInterface )
        return

    def startComponent(self):
        """
        _startComponent_

        Fire up the message service and define the list of
        messages that this component subscribes to
        """
        self.ms.subscribeTo("WMBSPublish:StartDebug")
        self.ms.subscribeTo("WMBSPublish:EndDebug")
        self.ms.subscribeTo("WMBSPublish:poll")
        self.ms.commit()

        self.connectT0AST()

        # Load up the alca subscription for easy access.
        pubFileset = Fileset(name = "WMBSPublishable")
        pubFileset.load()
        pubWorkflow = Workflow(name = "FileWMBSPublish")
        pubWorkflow.load()
        self.subscription = Subscription( fileset = pubFileset,
                                          workflow = pubWorkflow )
        self.subscription.load()

        self.ms.publishUnique("WMBSPublish:poll",
                              "", "00:00:15")
        self.ms.commit()

        # Wait for messages.
        self.messageLoop()
