#!/usr/bin/env python
"""
_AlcaSkimWorkflow_

Workflow Factory for AlcaSkimming processing workflows

"""

__revision__ = "$Id: AlcaSkimWorkflow.py,v 1.21 2010/04/16 17:20:33 hufnagel Exp $"
__version__ = "$Revision: 1.21 $"

import os
import time
import logging

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec
import ProdCommon.MCPayloads.WorkflowTools as WorkflowTools

from T0.WorkflowFactory.FactoryInterface import FactoryInterface


class AlcaSkimWorkflow(FactoryInterface):
    """
    _AlcaSkimWorkflow_

    Util to build workflows for alcaskim processing jobs
    
    """
    def __init__(self, runNumber, version, cmsPath, scramArch, *outModuleInfo):
        FactoryInterface.__init__(self, version, cmsPath, scramArch)
        self.run = runNumber
        self.outputModules = list(outModuleInfo)


    def buildConfiguration(self, **params):
        """
        _buildConfiguration_

        mostly just a method to take the passed in information

        """
        scenario = params.get('scenario', None)
        configFile = params.get('configFile', None)
        alcaProducers = params.get('alcaProducers', [])
        globalTag = params.get('globalTag', None)
        enableLazyDownload = params.get('enableLazyDownload', True)

        if globalTag == None:
            logging.error("ERROR: global tag is mandatory for workflow creation!")
            return None

        outputModuleDetails = {}

        for moduleInfo in self.outputModules:

            moduleName = "write_%s_%s" % (moduleInfo["dataset"],
                                          moduleInfo["dataTier"])

            outputModuleDetails[moduleName] = {
                "primaryDataset" : moduleInfo["dataset"],
                "dataTier" : moduleInfo["dataTier"],
                "acquisitionEra" : moduleInfo["acquisitionEra"],
                "processingVersion" : moduleInfo["processingVersion"],
                }

        if configFile == None:

            cfgInterface = self.createConfiguration(scenario = scenario,
                                                    config = "alcaSkim",
                                                    skims = alcaProducers,
                                                    globalTag = globalTag,
                                                    enableLazyDownload = enableLazyDownload,
                                                    outputModuleTemplate = outputModuleDetails.values()[0],
                                                    setEventContentInOutput = True)

        else:

            cfgInterface = self.createConfiguration(configFile = configFile,
                                                    globalTag = globalTag,
                                                    enableLazyDownload = enableLazyDownload,
                                                    outputModuleTemplate = outputModuleDetails.values()[0])

        return cfgInterface


    def makeWorkflowSpec(self, name, **params):
        """
        _makeWorkflowSpec_

        Create a workflow spec instance

        """
        production = params.get('production', False)

        self.workflow = WorkflowSpec()
        self.workflow.setWorkflowName(name)

        if production:
            self.workflow.setActivity("T0-AlcaSkim")
        else:
            self.workflow.setActivity("T0-Replay-AlcaSkim")
            
        self.workflow.setRequestCategory("data")
        self.workflow.setRequestTimestamp(int(time.time()))
        self.workflow.parameters["WorkflowType"] = "Skim"
        self.workflow.parameters["RequestLabel"] = name
        self.workflow.parameters["ProdRequestID"] = self.run
        self.workflow.parameters["RunNumber"] = self.run
        self.workflow.parameters["WorkflowRunNumber"] = self.run
        self.workflow.parameters["CMSSWVersion"] = self.cmssw["CMSSWVersion"] 
        self.workflow.parameters["ScramArch"] = self.cmssw["ScramArch"] 
        self.workflow.parameters["CMSPath"] = self.cmssw["CMSPath"]

        cmsRunNode = self.workflow.payload
        cmsRunNode.name = "cmsRun1"
        cmsRunNode.type = "CMSSW"
        cmsRunNode.application["Version"] = self.cmssw["CMSSWVersion"]
        cmsRunNode.application["Executable"] = "cmsRun"
        cmsRunNode.application["Project"] = "CMSSW"
        cmsRunNode.application["Architecture"] = self.cmssw["ScramArch"]

        # runtime express script
        cmsRunNode.scriptControls["PreExe"].append(
            "T0.AlcaSkimInjector.RuntimeAlcaSkim")

        # build the configuration template for the workflow
        cmsRunNode.cfgInterface = self.buildConfiguration(**params)
        if cmsRunNode.cfgInterface == None:
            return None

        # generate Dataset information for workflow from cfgInterface
        for outMod,moduleInstance in cmsRunNode.cfgInterface.outputModules.items():

            primaryName = moduleInstance["primaryDataset"] 
            processedName = moduleInstance["processedDataset"] 

            outDS = cmsRunNode.addOutputDataset(primaryName,
                                                processedName,
                                                outMod)

            outDS["DataTier"] = moduleInstance["dataTier"]
            outDS["ApplicationName"] = cmsRunNode.application["Executable"]
            outDS["ApplicationFamily"] = outMod
            outDS["PhysicsGroup"] = "Tier0"
            outDS["ApplicationFamily"] = outMod

            # generate just single LFN stub (all output is unmerged)
            # insert them into the output module and dataset info
            outDS["LFNBase"] = self.getLFN(moduleInstance, Unmerged = True)
            moduleInstance["LFNBase"] = outDS["LFNBase"]
            moduleInstance["logicalFileName"] = os.path.join(
                outDS["LFNBase"], "%s.root" % outMod
                )

        WorkflowTools.addStageOutNode(cmsRunNode, "stageOut1")
        WorkflowTools.addLogArchNode(cmsRunNode, "logArchive")
        
        return self.workflow
