#!/usr/bin/env python
"""
_FactoryInterface_


Basic interface class for all workflow factories.
Start out very lightweight API only definition and
evolve to contain more common helper utils

Basic principle is that this factory is instantiated
with a config file and release and is then
used to cut out per run, per dataset workflow specs and
provide an API to customise them as needed for tier0 ops.


"""

import os
import sys
import imp
import time
import logging
import traceback

from ProdCommon.CMSConfigTools.ConfigAPI.CMSSWAPILoader import CMSSWAPILoader
from ProdCommon.CMSConfigTools.ConfigAPI.CMSSWConfig import CMSSWConfig


def installEventContent(process, dataTier, outputModule):
    outputModule.outputCommands = getattr(process, dataTier+"EventContent").outputCommands 


class FactoryInterface:
    """
    _FactoryInterface_

    Base class for Workflow Factory APIs


    """
    def __init__(self, version, cmsPath, scramArch):
        self.cmssw = {}
        self.cmssw['CMSSWVersion'] = version
        self.cmssw['CMSPath'] = cmsPath
        self.cmssw['ScramArch'] = scramArch
        self.cmssw['version1'] = 0
        self.cmssw['version2'] = 0
        self.cmssw['version3'] = 0
        self.process = None

        cmsswVersionList = version.split("_")
        try:
            self.cmssw['version1'] = int(cmsswVersionList[1])
            self.cmssw['version2'] = int(cmsswVersionList[2])
            self.cmssw['version3'] = int(cmsswVersionList[3])
        except ValueError:
            logging.info("Warning: non standard CMSSW version: %s" % version)


    def createConfiguration(self, **params):
        """
        _createConfiguration_

        Builds a configuration, does some release specifc
        customizations, adds a source and adds output modules

        Three options to make the configuration, based on a
        config file, getting it from the framework or
        making up an empty one from scratch

        """
        loader = CMSSWAPILoader(self.cmssw['ScramArch'],
                                self.cmssw['CMSSWVersion'],
                                self.cmssw['CMSPath'])

        try:
            loader.load()
        except Exception, ex:
            logging.error("Couldn't load CMSSW libraries: %s" % ex)
            return None

        import FWCore.ParameterSet.Config as cms

        # building process
        # either from config file, from the framework(release) or from scratch
        if params.has_key('configFile'):

            if params.has_key('outputModuleDetails'):

                self.process = self.createProcessFromFile(params['configFile'],
                                                          stripOutputModules = True)

            elif params.has_key('outputModuleTemplate'):

                self.process = self.createProcessFromFile(params['configFile'],
                                                          stripOutputModules = False)

                moduleTemplate = params['outputModuleTemplate']

                # override primary dataset
                #if self.process != None and moduleTemplate.has_key('primaryDataset'):
                #    for outputModule in self.process.outputModules.values():
                #        outputModule.dataset.primaryDataset = cms.untracked.string(moduleTemplate['primaryDataset'])

            else:
                logging.error("Neither output module details or template specified")
                
        elif params.has_key('processName'):

            if params.has_key('outputModuleDetails'):

                self.process = self.createProcessFromScratch(params['processName'],
                                                             configName = params.get('configName', 'auto-config'),
                                                             configVersion = params.get("configVersion", time.strftime("%d-%b-%Y-%H:%M:%S")))
            else:
                logging.error("No output module details specified")

        elif params.has_key('scenario'):

            if params.has_key('config'):

                self.process = self.createProcessFromFramework(**params)

                if params.has_key('outputModuleTemplate'):

                    moduleTemplate = params['outputModuleTemplate']

                    # override primary dataset
                    #if self.process != None and moduleTemplate.has_key('primaryDataset'):
                    #    for outputModule in self.process.outputModules.values():
                    #        outputModule.dataset.primaryDataset = cms.untracked.string(moduleTemplate['primaryDataset'])

            else:
                logging.error("No config or output module template specified")   

        else:
            logging.error("Neither config file, framework config code or process name specified")

        # check if it worked
        if self.process == None:
            logging.error("Cannot build process, bailing out")
            loader.unload()
            return None

        # recreate source
        if params.has_key('sourceType'):
            sourceType = params['sourceType']
            self.process.source = cms.Source(sourceType, fileNames = cms.untracked.vstring())
        else:
            sourceType = "PoolSource"

        # Add Tier-0 override catalogue
        self.process.source.overrideCatalog = cms.untracked.string( 'trivialcatalog_file:' + os.path.join(os.environ["CMS_PATH"], 'SITECONF/local/Tier0/override_catalog.xml' ) + '?protocol=override')

        # configure firstFreeID (works around a bug processing 2_0_X streamer files)
        if ( self.cmssw['version1'] == 2 and self.cmssw['version2'] == 0 ) \
           and sourceType == 'NewEventStreamFileReader':

            self.process.source.firstFreeID = cms.untracked.uint32(65536)

        # configure input file transition at each event (only for streamers and at least CMSSW_3_X_Y
        # makes the maxSize parameter in output for repacker jobs functional
        if ( self.cmssw['version1'] >= 3 ) \
           and params.has_key('maxSize') \
           and sourceType == 'NewEventStreamFileReader':

            self.process.source.inputFileTransitionsEachEvent = cms.untracked.bool(True)

        # override global tag (only if placeholder already exists)
        if params.has_key('globalTag'):
            if hasattr(self.process, "GlobalTag"):
                self.process.GlobalTag.globaltag = cms.string(params['globalTag'])
                if params.has_key('globalTagConnect'):
                    self.process.GlobalTag.DBParameters.transactionId = cms.untracked.string(params['globalTagConnect'])

        # configure lazy download
        # (supported earlier than 2_1_8, but we don't use these releases anymore)
        if ( self.cmssw['version1'] > 2 ) \
           or ( self.cmssw['version1'] == 2 and self.cmssw['version2'] > 1 ) \
           or ( self.cmssw['version1'] == 2 and self.cmssw['version2'] == 1 and self.cmssw['version3'] >= 8 ):

            self.configureLazyDownload(
                params.get("enableLazyDownload", None) == True,
                params.get("enableCaching", sourceType == "PoolSource")
                )

        # configure fastCloning and noEventSort
        # (supported earlier than 2_1_8, but we don't use these releases anymore)
        fastCloning = False
        if ( ( self.cmssw['version1'] > 2 ) \
           or ( self.cmssw['version1'] == 2 and self.cmssw['version2'] > 1 ) \
           or ( self.cmssw['version1'] == 2 and self.cmssw['version2'] == 1 and self.cmssw['version3'] >= 8 ) ) \
           and sourceType == 'PoolSource':

            fastCloning = True
            if params.get("noEventSort", None) == True:
                self.process.source.noEventSort = cms.untracked.bool(True)

        # add output modules
        if params.has_key('outputModuleDetails'):

            for moduleName, moduleDetails in params['outputModuleDetails'].items():
                logging.debug("Adding output module %s to workflow" % moduleName)
                self.addOutputModule(moduleName,
                                     moduleDetails['dataTier'],
                                     primaryDataset = moduleDetails.get("primaryDataset", None),
                                     selectEvents = moduleDetails.get("SelectEvents", None),
                                     setEventContentInOutput = params.get("setEventContentInOutput", False),
                                     dropDQM = params.get("dropDQM", False),
                                     maxSize = params.get("maxSize", None))

        # apply generic modifiers to output modules
        # at the moment only fastCloning
        self.modifyOutputModules(fastCloning = fastCloning)

        # create and EndPath if needed, add all output modules
        # then add EndPath to schedule, also create new if needed
        self.addOutputModulesToEndPath()

        cfgInterface = CMSSWConfig()
        loadedConfig = cfgInterface.loadConfiguration(self.process)
        loadedConfig.validateForProduction()

        # complete the output module info in workflow
        for moduleName, outMod in cfgInterface.outputModules.items():

            # easy for output modules we added
            if params.has_key('outputModuleDetails'):
                outMod.update(params['outputModuleDetails'][moduleName])

            # if we kept the output modules from the configs it's harder
            # need to combine info from template and config
            #
            # acquisitionEra, processingVersion comes from template
            # processedDataset comes from template (if there)
            # processedDataset build from template + config (if not there)
            elif params.has_key('outputModuleTemplate'):

                template = params["outputModuleTemplate"]

                outMod['primaryDataset'] = template['primaryDataset']
                outMod['acquisitionEra'] = template['acquisitionEra']
                outMod['processingVersion'] = template['processingVersion']

                if template.has_key('processedDataset'):
                    outMod['processedDataset'] = template['processedDataset']
                else:
                    if outMod.has_key('processingString') and outMod['processingString'] != '':
                        processingString = outMod['processingString']
                    elif outMod.has_key('filterName') and outMod['filterName'] != '':
                        processingString = outMod['filterName']
                        outMod['processingString'] = processingString
                    else:
                        processingString = None

                    if processingString == None:
                        outMod['processedDataset'] = "%s-%s" % (template['acquisitionEra'],
                                                                template['processingVersion'])
                    else:
                        outMod['processedDataset'] = "%s-%s-%s" % (template['acquisitionEra'],
                                                                   processingString,
                                                                   template['processingVersion'])

        if params.has_key('configFile'):
            cfgInterface.originalCfg = file(params['configFile']).read()

        loader.unload()
        
        return cfgInterface


    def configureLazyDownload(self, enableLazyDownload, enableCaching):
        """
        _configureLazyDownload_

        """
        import FWCore.ParameterSet.Config as cms

        if enableLazyDownload:

            if not self.process.services.has_key('AdaptorConfig'):
                self.process.add_(cms.Service('AdaptorConfig'))
##                 self.process.add_(cms.Service("SiteLocalConfigService"))

            self.process.services["AdaptorConfig"].cacheHint = cms.untracked.string("lazy-download")
            self.process.services["AdaptorConfig"].readHint = cms.untracked.string("read-ahead-buffered")
##             self.process.services["SiteLocalConfigService"].overrideSourceCacheHintDir = cms.untracked.string("lazy-download")
##             self.process.services["SiteLocalConfigService"].overrideSourceReadHint = cms.untracked.string("read-ahead-buffered")

            if enableCaching:
                self.process.source.cacheSize = cms.untracked.uint32(20*1024*1024)
        else:
            if enableCaching:
                self.process.source.cacheSize = cms.untracked.uint32(20*1024*1024)

        return


    def createProcessFromFile(self, configFile, **params):
        """
        _createProcessFromFile_

        """
        import FWCore.ParameterSet.Config as cms
        
        cfgBaseName = os.path.basename(configFile).replace(".py", "")
        cfgDirName = os.path.dirname(configFile)
        modPath = imp.find_module(cfgBaseName, [cfgDirName])

        loadedConfig = imp.load_module(cfgBaseName, modPath[0],
                                       modPath[1], modPath[2])

        if params['stripOutputModules']:
            for moduleName in loadedConfig.process.outputModules.keys():
                del loadedConfig.process._Process__outputmodules[moduleName]

        loadedConfig.process.maxEvents = cms.untracked.PSet( input = cms.untracked.int32(-1) )

        return loadedConfig.process


    def createProcessFromFramework(self, **params):
        """
        _createProcessFromFramework_

        """
        try:
            #sys.path.insert(0, "/afs/cern.ch/user/h/hufnagel/w0/releases/CMSSW_3_6_1_patch4/python")

            import FWCore.ParameterSet.Config as cms

            from Configuration.DataProcessing.GetScenario import getScenario

            scenarioInst = getScenario(params['scenario'])

            if hasattr(scenarioInst, params['config']):

                if params['config'] == "promptReco":

                    process = getattr(scenarioInst, params['config'])(globalTag = "NOTSET",
                                                                      writeTiers = params['writeTiers'])

                elif params['config'] == "expressProcessing":

                    process = getattr(scenarioInst, params['config'])(globalTag = "NOTSET")

                elif params['config'] == "alcaSkim":

                    process = getattr(scenarioInst, params['config'])(skims = params['skims'])

                elif params['config'] == "dqmHarvesting":

                    process = getattr(scenarioInst, params['config'])(datasetName = params['inputDataset'],
                                                                      runNumber = self.run,
                                                                      globalTag = "NOTSET")
                    if process != None:
                        process.dqmSaver.runIsComplete = cms.untracked.bool(params['endOfRun'])

                elif params['config'] == "alcaHarvesting":

                    process = getattr(scenarioInst, params['config'])(globalTag = "NOTSET")

                else:

                    logging.error("Unknown config %s from scenario %s" % (params['config'], params['scenario']))

            if process != None:
                process.maxEvents = cms.untracked.PSet( input = cms.untracked.int32(-1) )

            else:

                logging.error("Scenario %s has no config for %s" % (params['scenario'], params['config']))

            return process

        except Exception, ex:
            msg = "Failed to get process from release for scenario "
            msg += params['scenario']
            msg += " and config "
            msg += params['config']
            msg += "\nWith Error:"
            msg += str(ex)
            logging.error(msg)
            #error_type, error_value, trbk = sys.exc_info()
            #tb_list = traceback.format_tb(trbk, 6)    
            #s = "Error: %s \nDescription: %s \nTraceback:" % (error_type.__name__, error_value)
            #for i in tb_list:
            #    s += "\n" + i
            #logging.error(s)

            return None


    def createProcessFromScratch(self, processName, **params):
        """
        _createProcessFromScratch_

        """
        import FWCore.ParameterSet.Config as cms

        process = cms.Process(processName)
        process.load("FWCore.MessageLogger.MessageLogger_cfi")

        process.maxEvents = cms.untracked.PSet( input = cms.untracked.int32(-1) )

        configName = params['configName']
        configVersion = params['configVersion']
        configAnnotation = "auto generated configuration"

        process.configurationMetadata = cms.untracked.PSet(
            name = cms.untracked.string(configName),
            version = cms.untracked.string(configVersion),
            annotation = cms.untracked.string(configAnnotation)
            )

        process.options = cms.untracked.PSet(
            Rethrow = cms.untracked.vstring("ProductNotFound","TooManyProducts","TooFewProducts"),
            wantSummary = cms.untracked.bool(False)
            )
            
        return process


    def addOutputModule(self, moduleName, dataTier, **params):
        """
        _addOutputModule_

        Add an output module and a dataset PSet to it

        """
        logging.info("Called addOutputModule with %s , %s , %s" % (moduleName, dataTier, params))

        import FWCore.ParameterSet.Config as cms

        outputModule = cms.OutputModule(
            "PoolOutputModule",
            fileName = cms.untracked.string("%s.root" % moduleName)
            )

        outputModule.dataset = cms.untracked.PSet(dataTier = cms.untracked.string(dataTier))
        #if params['primaryDataset'] !=  None:
        #    outputModule.dataset.primaryDataset = cms.untracked.string(params['primaryDataset'])

        if params['maxSize'] != None:
            outputModule.maxSize = cms.untracked.int32(params['maxSize'])

        if params['setEventContentInOutput']:
            if hasattr(self.process, dataTier+"EventContent"):
                installEventContent(self.process, dataTier, outputModule)
            elif dataTier == 'ALCAPROMPT':
                outputModule.outputCommands = cms.untracked.vstring(
                    'keep *')
            elif dataTier == 'DQM':
                outputModule.outputCommands = cms.untracked.vstring(
                    'drop *',
                    'keep *_MEtoEDMConverter_*_*')
            elif dataTier == 'FEVTHLTALL':
                installEventContent(self.process, "FEVT", outputModule)
                outputModule.outputCommands.append('keep *_*_*_HLT')
            else:
                msg = "Cannot figure out event content for data tier %s!!!\n" % dataTier
                raise RuntimeError, msg

        if params['dropDQM'] and dataTier != 'DQM':
            outputModule.outputCommands.append('drop *_MEtoEDMConverter_*_*')

        if params['selectEvents'] != None:
            outputModule.SelectEvents = cms.untracked.PSet(
                SelectEvents = cms.vstring()
                )
            for selCond in params['selectEvents']:
                outputModule.SelectEvents.SelectEvents.append(selCond)

        setattr(self.process, moduleName, outputModule)

        return


    def modifyOutputModules(self, **params):
        """
        _modifyOutputModules)

        Apply generic modifiers to output modules

        At the moment that's only fastCloning

        """
        import FWCore.ParameterSet.Config as cms

        if params.has_key('fastCloning'):
            for outputModule in self.process.outputModules.values():
                outputModule.fastCloning = cms.untracked.bool(params['fastCloning'])

        return


    def addOutputModulesToEndPath(self):
        """
        _addOutputModulesToEndPath_

        For the given process, create an EndPath, add all
        outputs to it and then add it to the Schedule
        """
        from FWCore.ParameterSet.Config import EndPath
        from FWCore.ParameterSet.Config import Schedule

        # add all output modules to outputPath
        for outputModuleName, outputModule in self.process.outputModules.items():

            # add it to the outputPath
            endPath = getattr(self.process, "outputPath", None)
            if endPath == None:
                self.process.outputPath = EndPath(outputModule)
            else:
                self.process.outputPath += outputModule

        # add outputPath to schedule if it exists
        outputPath = getattr(self.process, "outputPath", None)
        if outputPath != None:
            schedule = getattr(self.process, "schedule", None)
            if schedule == None:
                self.process.schedule = Schedule(self.process.outputPath)
            else:
                self.process.schedule.append(self.process.outputPath)

        return


    def getLFN(self, outputModule, **options):
        """
        _getLFN_

        Build a default name structure based on the parameters passed.

        """
        if outputModule.get("acquisitionEra") == None:
            msg = "Acquistion Era is not specified!!!\n"
            raise RuntimeError, msg

        #lfn = "/T0/hufnagel/repacktest/store/"
        lfn = "/store/"

        if options.get("Unmerged", False):
            lfn += "t0temp/"

        if outputModule.get("acquisitionEra").startswith('HI'):
            lfn += options.get("dataType","hidata")
        else:
            lfn += options.get("dataType","data")

        lfn += "/"
        lfn += "%s/" % outputModule.get("acquisitionEra")
        lfn += "%s/" % outputModule.get("primaryDataset")
        lfn += "%s/" % outputModule.get("dataTier")

        # FIXME: hacked for now, this should be cleaned up in design
        processingString = outputModule.get("processingString", None)
        processingVersion = outputModule.get("processingVersion")
        if processingString == None:
            tempList = outputModule.get("processedDataset").split('-')
            if len(tempList) == 3:
                processingString = tempList[1]
            elif len(tempList) == 2:
                processingString = None
            else:
                msg = "Invalid processed dataset: %s\n" % outputModule.get("processedDataset")
                raise RuntimeError, msg

        if processingString == None:
            lfn += "%s/" % processingVersion
        else:
            lfn += "%s-%s/" % (processingString, processingVersion)

        runString = str(self.run).zfill(9)
        lfn += "%s/%s/%s/" % (runString[0:3],
                              runString[3:6],
                              runString[6:9])

        return lfn
