#!/usr/bin/env python
"""
_HarvestWorkflow_

Workflow Factory for DQMHarvest workflows

"""

__revision__ = "$Id: HarvestWorkflow.py,v 1.5 2010/08/03 14:12:46 hufnagel Exp $"
__version__ = "$Revision: 1.5 $"

import os
import time
import logging

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec
import ProdCommon.MCPayloads.WorkflowTools as WorkflowTools

from T0.WorkflowFactory.FactoryInterface import FactoryInterface


class HarvestWorkflow(FactoryInterface):
    """
    _HarvestWorkflow_

    Util to build workflows for DQM harvesting jobs
    
    """
    def __init__(self, runNumber, version, cmsPath, scramArch):
        FactoryInterface.__init__(self, version, cmsPath, scramArch)
        self.run = runNumber


    def buildConfiguration(self, **params):
        """
        _buildConfiguration_

        mostly just a method to take the passed in information

        """
        scenario = params.get('scenario', None)
        harvestType = params.get('harvestType', "DQM")
        globalTag = params.get('globalTag', None)
        inputDataset = params.get('inputDataset', None)
        endOfRun = params.get('endOfRun', False)
        enableLazyDownload = params.get('enableLazyDownload', True)

        if globalTag == None:
            logging.error("ERROR: global tag is mandatory for workflow creation !")
            return None

        if harvestType == "DQM":

            if inputDataset == None:
                logging.error("ERROR: input dataset is mandatory for workflow creation !")
                return None

            cfgInterface = self.createConfiguration(scenario = scenario,
                                                    config = "dqmHarvesting",
                                                    globalTag = globalTag,
                                                    inputDataset = inputDataset,
                                                    endOfRun = endOfRun,
                                                    enableLazyDownload = enableLazyDownload)
            return cfgInterface

        elif harvestType == "ALCAPROMPT":

            cfgInterface = self.createConfiguration(scenario = scenario,
                                                    config = "alcaHarvesting",
                                                    globalTag = globalTag,
                                                    enableLazyDownload = enableLazyDownload)
            return cfgInterface

        else:

            logging.error("ERROR: harvest type %s is not know !" % harvestType )
            return None



    def makeWorkflowSpec(self, name, **params):
        """
        _makeWorkflowSpec_

        Create a workflow spec instance

        """
        production = params.get('production', False)
        harvestType = params.get('harvestType', "DQM")

        #  //
        # // Initialise basic workflow
        #//
        self.workflow = WorkflowSpec()
        self.workflow.setWorkflowName(name)

        if production:
            self.workflow.setActivity("T0-Harvesting")
        else:
            self.workflow.setActivity("T0-Replay-Harvesting")
            
        self.workflow.setRequestCategory("data")
        self.workflow.setRequestTimestamp(int(time.time()))
        self.workflow.parameters["WorkflowType"] = "Harvesting"
        self.workflow.parameters["RequestLabel"] = name
        self.workflow.parameters["ProdRequestID"] = self.run
        self.workflow.parameters["RunNumber"] = self.run
        self.workflow.parameters["WorkflowRunNumber"] = self.run
        self.workflow.parameters["CMSSWVersion"] = self.cmssw["CMSSWVersion"] 
        self.workflow.parameters["ScramArch"] = self.cmssw["ScramArch"] 
        self.workflow.parameters["CMSPath"] = self.cmssw["CMSPath"]

        cmsRunNode = self.workflow.payload
        cmsRunNode.name = "cmsRun1"
        cmsRunNode.type = "CMSSW"
        cmsRunNode.application["Version"] = self.cmssw["CMSSWVersion"]
        cmsRunNode.application["Executable"] = "cmsRun"
        cmsRunNode.application["Project"] = "CMSSW"
        cmsRunNode.application["Architecture"] = self.cmssw["ScramArch"]

        if harvestType == "DQM":

            # add input dataset (needed for dqm gui upload)
            inputDataset = params.get("inputDataset")
            prim, proc, tier = str(inputDataset).split("/")[1:]
            inputDatasetInWF = cmsRunNode.addInputDataset(prim, proc)
            inputDatasetInWF['DataTier'] = tier

            # runtime DQM harvesting script
            cmsRunNode.scriptControls["PostTask"].append(
                "JobCreator.RuntimeTools.RuntimeOfflineDQM")

        elif harvestType == "ALCAPROMPT":

            # runtime DQM harvesting script
            cmsRunNode.scriptControls["PostTask"].append(
                "T0.DQMScheduler.RuntimeAlcaHarvest")

        # build the configuration template for the workflow
        cmsRunNode.cfgInterface = self.buildConfiguration(**params)
        if cmsRunNode.cfgInterface == None:
            return None

        WorkflowTools.addLogArchNode(cmsRunNode, "logArchive")
        
        return self.workflow
