#!/usr/bin/env python
"""
_PromptCalibWorkflow_

Factory for PromptCalib job workflows
"""

__revision__ = "$Id: PromptCalibWorkflow.py,v 1.10 2010/03/25 13:55:07 naderer Exp $"
__version__ = "$Revision: 1.10 $"

import os
import time
import logging

from ProdCommon.CMSConfigTools.ConfigAPI.CMSSWConfig import CMSSWConfig
from ProdCommon.CMSConfigTools.ConfigAPI.CMSSWAPILoader import CMSSWAPILoader

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec
import ProdCommon.MCPayloads.WorkflowTools as WorkflowTools

from T0.WorkflowFactory.FactoryInterface import FactoryInterface


class PromptCalibWorkflow(FactoryInterface):
    """
    _PromptCalibWorkflow_

    Util to build workflows for PromptCalib processing jobs.
    """
    
    def __init__(self, runNumber, version, globalTag, cmsPath, scramArch, outModuleInfo):
        FactoryInterface.__init__(self, version, cmsPath, scramArch)
        self.run = runNumber
        self.outputModule = outModuleInfo
        self.globalTag = globalTag

    def buildCalibConfiguration(self, configFile, enableLazyDownload):
        """
        _buildCalibConfiguration_

        method to build calib configuration interface for prompt calib workflows
        -- load calibration config from file
        -- add configuration metadata (name, version, annotation)
        -- validate for production and return cfgInterface
        """

        loader = CMSSWAPILoader(self.cmssw["ScramArch"],
                                self.cmssw["CMSSWVersion"],
                                self.cmssw["CMSPath"])
        try:
            loader.load()
        except Exception, ex:
            logging.error("Couldn't load CMSSW libraries: %s" % ex)
            return None
        
        #use method from FactoryInterface to load the config file
        self.process = self.createProcessFromFile(configFile, stripOutputModules = False)

        import FWCore.ParameterSet.Config as cms

        # check for configuration metadata (name, version and annotation)
        # and add it if you can't find it
        if not getattr(self.process, "configurationMetadata", None):
            self.process.configurationMetadata = cms.untracked(cms.PSet()) 
            self.process.configurationMetadata.name = cms.untracked(
                cms.string("PyReleaseValidation"))
            self.process.configurationMetadata.version = cms.untracked(
                cms.string(self.outputModule["configFile_rev"]))
            self.process.configurationMetadata.annotation = cms.untracked(
                cms.string(self.outputModule["calibTask"]))

        cfgInterface = CMSSWConfig()
        loadedConfig = cfgInterface.loadConfiguration(self.process)
        loadedConfig.validateForProduction()
        
        cfgInterface.originalCfg = file(configFile).read()
        loader.unload()

        return cfgInterface


    def makeWorkflowSpec(self, name, configFile, enableLazyDownload,
                         production = False):
        """
        _makeWorkflowSpec_

        Create a workflow spec instance
        
        """
        #  //
        # // Initialise basic workflow
        #//
        self.workflow = WorkflowSpec()
        self.workflow.setWorkflowName(name)
        if production:
            self.workflow.setActivity("T0-PromptCalib")
        else:
            self.workflow.setActivity("T0-Replay-PromptCalib")
        self.workflow.setRequestCategory("data")
        self.workflow.setRequestTimestamp(int(time.time()))
        self.workflow.parameters["WorkflowType"] = "Processing"
        self.workflow.parameters["RequestLabel"] = name
        self.workflow.parameters["ProdRequestID"] = self.run
        self.workflow.parameters["RunNumber"] = self.run
        self.workflow.parameters["CMSSWVersion"] = self.cmssw["CMSSWVersion"] 
        self.workflow.parameters["ScramArch"] = self.cmssw["ScramArch"] 
        self.workflow.parameters["CMSPath"] = self.cmssw["CMSPath"]

        cmsRunNode = self.workflow.payload
        cmsRunNode.name = "cmsRun1"
        cmsRunNode.type = "CMSSW"
        cmsRunNode.application["Version"] = self.cmssw["CMSSWVersion"]
        cmsRunNode.application["Executable"] = "cmsRun"
        cmsRunNode.application["Project"] = "CMSSW"
        cmsRunNode.application["Architecture"] = self.cmssw["ScramArch"]
        
        # runtime express script
        cmsRunNode.scriptControls["PreExe"].append(
            "T0.PromptCalibInjector.RuntimePromptCalib")

        # runtime post processing: stores output at a given target location
        cmsRunNode.scriptControls["PostTask"].append(
            "T0.PromptCalibInjector.RuntimePostCalib")
        
        # build the configuration template for the beamfit workflow
        cmsRunNode.cfgInterface = self.buildCalibConfiguration(configFile,
                                                         enableLazyDownload)
        if cmsRunNode.cfgInterface == None:
            return None

        # override global tag
        cmsRunNode.cfgInterface.conditionsTag = self.globalTag

        WorkflowTools.addStageOutNode(cmsRunNode, "stageOut1")
        WorkflowTools.addLogArchNode(cmsRunNode, "logArchive")

        return self.workflow


        
