#!/usr/bin/env python
"""
_PromptRecoWorkflow_

Workflow Factory for PromptReco processing workflows

"""

__revision__ = "$Id: PromptRecoWorkflow.py,v 1.51 2011/01/14 15:07:20 hufnagel Exp $"
__version__ = "$Revision: 1.51 $"

import os
import time
import logging

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec
import ProdCommon.MCPayloads.WorkflowTools as WorkflowTools

from T0.WorkflowFactory.FactoryInterface import FactoryInterface


class PromptRecoWorkflow(FactoryInterface):
    """
    _PromptRecoWorkflow_

    Util to build workflows for reco processing jobs
    
    """
    def __init__(self, runNumber, version, cmsPath, scramArch, *outModuleInfo):
        FactoryInterface.__init__(self, version, cmsPath, scramArch)
        self.run = runNumber
        self.outputModules = list(outModuleInfo)


    def buildConfiguration(self, **params):
        """
        _buildConfiguration_

        mostly just a method to take the passed in information

        """
        scenario = params.get('scenario', None)
        configFile = params.get('configFile', None)
        alcaProducers = params.get('alcaProducers', [])
        writeRECO = params.get('writeRECO', True)
        writeAOD = params.get('writeAOD', False)
        globalTag = params.get('globalTag', None)
        enableLazyDownload = params.get('enableLazyDownload', True)

        if globalTag == None:
            logging.error("ERROR: global tag is mandatory for workflow creation!")
            return None

        writeTiers = ["DQM"]
        if writeRECO:
            writeTiers.append("RECO")
        if writeAOD:
            writeTiers.append("AOD")
        if len(alcaProducers) > 0:
            writeTiers.append("ALCARECO")

        if len(writeTiers) == 0:
            logging.error("ERROR: output tier needed for workflow creation!")
            return None

        outputModuleDetails = {}

        for moduleInfo in self.outputModules:

            if moduleInfo["dataTier"] in writeTiers:
                    
                moduleName = "write_%s" % moduleInfo["dataTier"]

                outputModuleDetails[moduleName] = {
                    "primaryDataset" : moduleInfo["dataset"],
                    "processedDataset" : moduleInfo["processedDataset"],
                    "dataTier" : moduleInfo["dataTier"],
                    "acquisitionEra" : moduleInfo["acquisitionEra"],
                    "processingVersion" : moduleInfo["processingVersion"],
                    }

        if configFile == None:

            cfgInterface = self.createConfiguration(scenario = scenario,
                                                    config = "promptReco",
                                                    skims = alcaProducers,
                                                    writeTiers = writeTiers,
                                                    globalTag = globalTag,
                                                    enableLazyDownload = enableLazyDownload,
                                                    outputModuleTemplate = outputModuleDetails.values()[0],
                                                    setEventContentInOutput = True,
                                                    dropDQM = True)

##             cfgInterface = self.createConfiguration(scenario = scenario,
##                                                     config = "promptReco",
##                                                     skims = alcaProducers,
##                                                     writeTiers = [],
##                                                     globalTag = globalTag,
##                                                     enableLazyDownload = enableLazyDownload,
##                                                     outputModuleDetails = outputModuleDetails,
##                                                     setEventContentInOutput = True,
##                                                     dropDQM = True)

        else:

            cfgInterface = self.createConfiguration(configFile = configFile,
                                                    globalTag = globalTag,
                                                    enableLazyDownload = enableLazyDownload,
                                                    outputModuleDetails = outputModuleDetails,
                                                    setEventContentInOutput = True,
                                                    dropDQM = True)

        return cfgInterface


    def makeWorkflowSpec(self, name, **params):
        """
        _makeWorkflowSpec_

        Create a workflow spec instance

        """
        production = params.get('production', False)

        #  //
        # // Initialise basic workflow
        #//
        self.workflow = WorkflowSpec()
        self.workflow.setWorkflowName(name)

        if production:
            self.workflow.setActivity("T0-PromptReco")
        else:
            self.workflow.setActivity("T0-Replay-PromptReco")
            
        self.workflow.setRequestCategory("data")
        self.workflow.setRequestTimestamp(int(time.time()))
        self.workflow.parameters["WorkflowType"] = "Processing"
        self.workflow.parameters["RequestLabel"] = name
        self.workflow.parameters["ProdRequestID"] = self.run
        self.workflow.parameters["RunNumber"] = self.run
        self.workflow.parameters["WorkflowRunNumber"] = self.run
        self.workflow.parameters["CMSSWVersion"] = self.cmssw["CMSSWVersion"] 
        self.workflow.parameters["ScramArch"] = self.cmssw["ScramArch"] 
        self.workflow.parameters["CMSPath"] = self.cmssw["CMSPath"]

        cmsRunNode = self.workflow.payload
        cmsRunNode.name = "cmsRun1"
        cmsRunNode.type = "CMSSW"
        cmsRunNode.application["Version"] = self.cmssw["CMSSWVersion"]
        cmsRunNode.application["Executable"] = "cmsRun"
        cmsRunNode.application["Project"] = "CMSSW"
        cmsRunNode.application["Architecture"] = self.cmssw["ScramArch"]

        # runtime PromptReco script
        cmsRunNode.scriptControls["PreExe"].append(
            "T0.PromptRecoInjector.RuntimePromptReco")

        # build the configuration template for the workflow
        cmsRunNode.cfgInterface = self.buildConfiguration(**params)
        if cmsRunNode.cfgInterface == None:
            return None

        # generate Dataset information for workflow from cfgInterface
        for outMod,moduleInstance in cmsRunNode.cfgInterface.outputModules.items():

            primaryName = moduleInstance["primaryDataset"] 
            processedName = moduleInstance["processedDataset"] 

            outDS = cmsRunNode.addOutputDataset(primaryName,
                                                processedName,
                                                outMod)

            outDS["DataTier"] = moduleInstance["dataTier"]
            outDS["ApplicationName"] = cmsRunNode.application["Executable"]
            outDS["ApplicationFamily"] = outMod
            outDS["PhysicsGroup"] = "Tier0"
            outDS["ApplicationFamily"] = outMod

            # generate just single LFN stub (all output is unmerged)
            # insert them into the output module and dataset info
            outDS["LFNBase"] = self.getLFN(moduleInstance, Unmerged = True)
            moduleInstance["LFNBase"] = outDS["LFNBase"]
            moduleInstance["logicalFileName"] = os.path.join(
                outDS["LFNBase"], "%s.root" % outMod
                )

        WorkflowTools.addStageOutNode(cmsRunNode, "stageOut1")
        WorkflowTools.addLogArchNode(cmsRunNode, "logArchive")
        
        return self.workflow
